def map_network_classifications_to_standard_classifications(native_classifications, site_metadata, run_details, write_to_metadata_array=True):

    '''takes multiple types of reported site classifications from multiple reporting networks and produces standardised classifications across all networks

       STANDARD CLASSIFICATION DEFINITIONS:
       ----------------------------------------------------

       AREA_CLASSIFICATION: Describes type of area a measurement station is situated in
        
       urban: All areas with a level of urban influence are defined as "urban"
              2 subdivisions to separate between urban centres and suburban areas:
       ------centre: Continuously built-up urban area meaning complete (or at least highly predominant) building-up of the street front side by buildings with at least two floors or large detached buildings with at least two floors.
                     With the exception of city parks, the built-up area is not mixed with non-urbanised areas.     
       ------suburban: Largely built-up urban area. ‘Largely built-up’ means contiguous settlement of detached buildings of any size with a building density less than for ‘continuously built-up’ area.
                       The built-up area is mixed with non-urbanised areas (e.g. agricultural, lakes, woods).
                       It must also be noted that ‘suburban’ as defined here has a different meaning than in every day English i.e. ‘an outlying part of a city or town’ suggesting that a suburban area is always associated to an urban area.
                       A suburban area can be suburban on its own without any urban part.
         
       rural: All areas, that do not fulfil the criteria for an urban area are defined as "rural" 
              3 subdivisions to indicate the distance to the nearest built-up urban area
       ------near_city: area within 10 km from the border of an urban or suburban area.
       ------regional: 10-50 km from major sources/source areas.
       ------remote: > 50 km from major sources/source areas.   

       ----------------------------------------------------
       STATION_CLASSIFICATION: Categorises the dominant type of air measured by a station

       background : Located such that its pollution levels are representative of the average exposure of the general population within the type of area under assessment. The pollution level should not be dominated by a single source type (e.g. traffic), unless that source type is typical within the area
                    under assessment. The station should usually be representative of a wider area of at least several square kilometers.

       point_source: Measurements influenced by a major stationary emissions source (i.e power plant),
                     or influenced by traffic, off-road vehicles and engines, rail, marine, or aviation sources.
                     Subdivisions of 'industrial' & 'traffic' given to allow for greater speciation given by EEA_AIRBASE & EEA_AQ_eReporting)
       --------industrial: Located in close proximity to single industrial sources or industrial areas, in a location that should represent the highest concentrations to which the population within the zone are exposed to. 
                           A wide range of industrial sources can be considered here, including: a) thermal power generation, b) district heating plants, c) refineries, d) waste incineration/treatment plants, dump sites, e) mining, including gravel, oil, natural gas, f) airports or g) ports.
       --------traffic: Located in close proximity to a road, in a location that should represent the highest concentrations to which the population are exposed to within the zone.
                        Located such that the pollution level for the specific pollutant is determined predominantly by the emissions from road traffic (i.e. not ship, railway, airplane, off-road) on distinct major roads.

       ----------------------------------------------------
       MAIN_EMISSION_SOURCE: Describes the main source of the emissions influencing air at a station.   
                              **Note - this describes the main source of emissions, but does not necessarily mean these emissions dominate the type of air measured at a station (given by STATION_CLASSIFICATION). 

       agriculture: emissions associated with agriculture

       commercial_and_residential_combustion: emissions associated with commercial and residential combustion

       extraction_of_fossil_fuels: emissions associated with the extraction of fossil fuels 

       industrial_combustion: emissions associated with industrial combustion

       natural: emissions from natural sources (e.g terpenes from trees)

       other_mobile_sources_and_machinery: emissions from all other mobile sources than traffic, and from off-road vehicles and engines 

       production_processes: emissions from processes associated with production and assembly

       power_production: emissions from processes associated with the generation of power

       road_transport: emissions associated with emissions from traffic

       solvents: emissions associated with use of solvents

       waste_treatment_and_disposal: emissions associated with waste treatment and disposal

       ----------------------------------------------------
       LAND_USE: Dominant land use in the area of the reporting station 
                 categories broadly follow IGBP land cover classification system, followed by MODIS/VIIRS land use maps
                 source: http://www.eomf.ou.edu/static/IGBP.pdf

       barren: Lands with exposed soil, sand or rocks and never have more than 10% vegetated cover during any time of the year.
       --------beach: landform alongside a body of water which consists of loose particles. The particles composing a beach are typically made from rock, such as sand, gravel, shingle, pebbles, or cobblestones.
       --------desert: a barren area of landscape where little precipitation occurs and consequently living conditions are hostile for plant and animal life. 
                       The lack of vegetation exposes the unprotected surface of the ground to the processes of denudation.
       --------rock: characterised by areas of bedrock exposure, scarps, talus, slides, volcanic material, rock glaciers, and other accumulations of rock without vegetative cover 
       --------soil: thin soil, without vegetation

       forest: Lands dominated by woody vegetation or trees, with a percent cover >60% and height exceeding 2 m. 
               Includes all evergreen needleleaf, evergreen broadleaf, deciduous needleleaf, deciduous broadleaf types

       open: Lands with herbaceous, other understory systems or woody vegetation less than 2 m tall
       --------grassland: Lands with herbaceous types of cover. Tree and shrub cover is less than 10%.
       --------savanna: Lands with herbaceous and other understory systems, and with forest canopy cover between 10% and 60%. The forest cover height exceeds 2 m.
       --------shrubland: Lands with woody vegetation less than 2 m tall and with shrub canopy cover > 10%. The shrub foliage can be either evergreen or deciduous.

       snow: Lands under snow/ice cover throughout the year.

       urban: Land covered by buildings and other man-made structures.
       --------agricultural: Lands covered with temporary crops followed by harvest and a bare soil period (e.g., single and multiple cropping systems).
                             Also includes lands used for farming and raising of livestock
                             It is important to note that also agricultural is classed under urban, this class also includes naturally occuring crops.
                             **Sites with agricultural land_use can have an AREA_CLASSIFICATION of 'rural'. 
                             **They are defined as 'urban' under LAND_USE as they would not be there without human activity, but can be located extremely far from urban areas, with no major emission sources nearby (with the exception of pesticides)
       --------blighted: A blighted area is defined as: by reason of deterioration, faulty planning, inadequate or improper facilities, deleterious land use or the existence of unsafe structures, 
                          or any combination of these factors, are detrimental to the safety, health or welfare of the community.
       --------commercial: A commercial area is real estate intended for use by for-profit businesses, such as office complexes, shopping malls, service stations and restaurants. 
                            It may be purchased outright by a developer for future projects or leased through a real estate broker.
       --------industrial: Industrial land use is the use of a piece of land for industrial purposes, eg. manufacturing.
       --------military: Land used for solely military purposes
       --------park: A large public garden or area of land used for recreation.
       --------residential: A residential area is a land used in which housing predominates, as opposed to industrial and commercial areas. Housing may vary significantly between, and through, residential areas.
                             These include single-family housing, multi-family residential, or mobile homes.
       --------transportation: All types of land use used for human transportation. This includes airports, roads, railway lines and shipping ports.
       **possibly some overlap between between 'industrial' and 'transportation' classes, as some networks may class things such as airports as industrial (e.g EEA_AIRBASE), and others not

       water: Oceans, seas, lakes, reservoirs, and rivers. Can be either fresh or salt- water bodies.

       wetland: Lands with a permanent mixture of water and herbaceous or woody vegetation. The vegetation can be present either in salt, brackish, or fresh water.


       ----------------------------------------------------
       TERRAIN: Dominant terrain in the area of the reporting station 

       coastal: A coastal area is where land meets the sea or ocean.

       complex: A region having irregular topography (not including mountains or coastal). 
                 Complex terrain can include variations in land use, such as urban, irrigated, and unirrigated.

       flat: Open terrain, open country or open ground is terrain which is mostly flat and free of obstructions such as trees and buildings. 
             Examples include farmland, grassland and specially cleared areas such as an airport.

       mountain: A mountain is a large landform that stretches above the surrounding land in a limited area, usually in the form of a peak.

       rolling: Terrain where the natural slopes consistently rise and fall across a horizontal plane 


       ----------------------------------------------------
       MEASUREMENT_SCALE: A denotation of the geographic scope of the air quality measurements made by the monitor. 
                          The implication is that the same measurement made elsewhere within the measurement scale 
                          would produce an equivalent result to that produced at the monitoring site
                          **Neighbourhood and City scales have potential to overlap in applications that concern secondarily formed or homogeneously distributed air pollutants
                    
       micro: representative for: several metres - ~100m

       middle: representative for: ~100m - 0.5km, several city blocks

       neighbourhood: representative for: 0.5km - 4km, some extended area of city that has relatively uniform land use

       city: representative for: 4km - 50km, city like dimensions. 

       regional: representative for: 50km - 100s km, rural area of reasonably homogeneous geography, without large sources
    '''

    import numpy as np

    #iterate through values in native_classifications dictionary and make sure all values are lower-cased and all white spaces are stripped.
    for key in list(native_classifications.keys()):
        native_classifications[key] = native_classifications[key].replace(" ", "").lower()

#------------------------------------------------
#NOW MAP INDIVIDUAL NATIVE NETWORK CLASSIFICATIONS TO STANDARD CLASSIFICATIONS 
#------------------------------------------------

#AERONET

#AERONET NATIVE CLASSIFICATIONS = NONE
    
    if 'AERONET' in run_details['network']:
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#BJMEMC

#BJMEMC NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = urban, suburban, traffic, rural
    
    elif 'BJMEMC' == run_details['network']:
    
        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','suburban':'urban','traffic':'urban','rural':'rural','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        if native_classifications['area_classification'] in mapping_area_classification_urban_subdivision:
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        

        #STATION CLASSIFICATION - infer from area classication
                              # - as there is a traffic classification, assume other classes are background 
        mapping_station_classification = {'urban':'background','suburban':'background','traffic':'point_source','rural':'background','':''}
    
        station_classification = mapping_station_classification[native_classifications['area_classification']]
    
        if native_classifications['area_classification'] == 'traffic':
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------

#CANADA_NAPS

#CANADA_NAPS NATIVE CLASSIFICATIONS = STATION_CLASSIFICATION, LAND_USE, MEASUREMENT_SCALE

#STATION_CLASSIFICATION = pe (General Population Exposure), rb (Regional Background), t (Transportation Source Influenced), ps (Point Source Influenced)
#LAND_USE = a (Agricultural), c (Commercial), f (Forested), i (Industrial), o (Open), p (Parks), r (Residential), w (Water)
#MEASUREMENT_SCALE = mi (Micro), md (Middle), ne (Neighbourhood), sr (Sub-Regional), r (Regional)

    elif run_details['network'] == 'CANADA_NAPS':
    
        #AREA CLASSIFICATION - not given directly, but derive it from station_classification
                              #all 'General Population Exposure', 'Transportation Source Influenced' and 'Point Source Influenced' classes defined as being within 'urban' area
                              #'Regional background' is defined as being outside 'urban' area --> 'rural' 
        mapping_area_classification = {'pe':'urban','t':'urban','ps':'urban','rb':'rural','unknown':'','na':'','ns':'','':''}
        area_classification = mapping_area_classification[native_classifications['station_classification']]
        
        #STATION CLASSIFICATION
        mapping_station_classification = {'pe':'background','rb':'background','t':'point_source-traffic','ps':'point_source-industrial','unknown':'','na':'','ns':'','':''}
        station_classification = mapping_station_classification[native_classifications['station_classification']]

        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''

        #LAND USE
        mapping_land_use = {'a':'urban','c':'urban','f':'forest','i':'urban','o':'open','p':'urban','r':'urban','w':'water','unknown':'','na':'','':''}
        mapping_land_use_urban_subdivision = {'a':'agricultural','c':'commercial','i':'industrial','p':'park','r':'residential'}
        
        land_use = mapping_land_use[native_classifications['land_use']]
        
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
            
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE 
        mapping_measurement_scale = {'mi':'micro','md':'middle','ne':'neighbourhood','ur':'city','re':'regional','unknown':'','na':'','':''}
        measurement_scale = mapping_measurement_scale[native_classifications['measurement_scale']]

#------------------------------------------------
#CAPMoN

#CAPMoN NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = rural
    
    elif run_details['network'] == 'CAPMoN':
    
        #AREA CLASSIFICATION
        if native_classifications['area_classification'] == 'rural':
            area_classification = 'rural'
        else:
            area_classification = ''    
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
    
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
            
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#CHILE SINCA

#CHILE SINCA NATIVE CLASSIFICATIONS = NONE
    
    elif 'CHILE_SINCA' == run_details['network']:
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#CNEMC

#CNEMC NATIVE CLASSIFICATIONS = NONE
    
    elif 'CNEMC' == run_details['network']:
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EANET

#EANET NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = urban, rural, remote
    
    elif run_details['network'] == 'EANET':
    
        #AREA CLASSIFICATION - Enough detail reported in data manuals to subdivide rural classes
                               #'rural' class defined as more than 20km from large emission sources --> 'regional' 
                               #'remote' class defined as more than 50km from large emission sources --> 'remote' 
                               
        mapping_area_classification = {'urban':'urban','rural':'rural','remote':'rural','unknown':'','':''}
        mapping_area_classification_rural_subdivision = {'rural':'regional','remote':'remote'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision)  
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume representativeness of 'urban' sites' 
        
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EBAS

#EMEP NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, LAND_USE

#AREA_CLASSIFICATION = mountain, polar, coastal, rural, suburban, urbanandcentercity, other
#LAND_USE = agricultural, airport, blightedarea, commercial, desert, forest, grassland,
#gravelandstone, industrial, militaryreservation, remotepark, residential, singlepointsourcearea,
#snowfield, urbanpark, wetland, other

    elif run_details['network'] == 'EBAS':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'mountain':'rural','polar':'rural','coastal':'rural','rural':'rural','suburban':'urban','urbanandcentercity':'urban','other':'','notavailable':'','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urbanandcentercity':'centre','suburban':'suburban'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision)     


        #STATION CLASSIFICATION - Derive this class from mixture of AREA CLASSIFICATION AND LAND_USE
                                #Where land use is 'singlepointsourcearea', the station is classed as 'point_source'
                                #Where land use is 'airport', the station is classed as 'point_source-industrial'
                                #Otherwise, if area_classification is 'rural', then set class as 'background'
        
        #if land use is 'singlepointsourcearea' then classify station as 'point_source'
        if (native_classifications['land_use'] == 'singlepointsourcearea'):
            station_classification = 'point_source'
         #if land use is 'airport' then classify as 'point_source-industrial'
        elif (native_classifications['land_use'] == 'airport'):
            station_classification = 'point_source-industrial'
        #otherwise if area classification is 'rural', make assumption the station is a 'background' site
        elif area_classification == 'rural':
            station_classification = 'background'
        #can't be sure of station classification - set as blank
        else:
            station_classification = ''
            
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
            
        #LAND USE - Only use 'grassland' to classify land_use as 'open', know from experience it is used to describe both 'shrublands' and 'grasslands'
        mapping_land_use = {'agricultural':'urban','airport':'urban','blightedarea':'urban','commercial':'urban','desert':'barren','forest':'forest','grassland':'open','gravelandstone':'barren','industrial':'urban','militaryreservation':'urban','remotepark':'urban','residential':'urban','singlepointsourcearea':'urban','snowfield':'snow','urbanpark':'urban','wetland':'wetland','other':'','notavailable':'','unknown':'','':''}
        mapping_land_use_barren_subdivision = {'desert':'desert','gravelandstone':'rock'}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','airport':'transportation','blightedarea':'blighted','commercial':'commercial','industrial':'industrial','militaryreservation':'military','remotepark':'park','residential':'residential','singlepointsourcearea':'','urbanpark':'park'}

        land_use = mapping_land_use[native_classifications['land_use']]
        
        if land_use == 'barren':
            land_use_barren_subdivision = mapping_land_use_barren_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            #take care not to needlessly append empty string associated with 'singlepointsourcearea', too vague to give specific land use
            if land_use_urban_subdivision != '':
                land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
    
        #TERRAIN 
        terrain = ''
        if native_classifications['area_classification'] == 'mountain':
            terrain = 'mountain'
        elif native_classifications['area_classification'] == 'coastal':
            terrain = 'coastal'
        
        #if terrain is empty string still, if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if terrain == '':    
            if ('urban' in area_classification):
                terrain = 'complex'

        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#EEA_AIRBASE

#EEA_AIRBASE NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, STATION_CLASSIFICATION, MAIN_EMISSION_SOURCE, LAND_USE, AREA_CLASSIFICATION_RURAL_SUBDIVISION, REPRESENTATIVE_RADIUS

#AREA_CLASSIFICATION = suburban, urban, rural 
#STATION_CLASSIFICATION = background, industrial, traffic
#MAIN_EMISSION_SOURCE = agriculture, "commercial, institutional and residential combustion", "extraction and distribution of fossil fuels", "industrial combustion", nature, other, "other mobile sources and machinery (specify)", "production processes", "public power, co-generation and district heating", "road transport", "use of solvents", "waste treatment and disposal"
#LAND_USE = agricultural, commercial, industrial, natural, residential
#AREA_CLASSIFICATION_RURAL_SUBDIVISION = nearcity, remote, regional
#REPRESENTATIVE_RADIUS = representative radius of measurements in kilometres

    elif run_details['network'] == 'EEA_AIRBASE':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        mapping_area_classification_rural_subdivision = {'nearcity':'near_city','regional':'regional','remote':'remote'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        if area_classification == 'rural':
            if (native_classifications['area_classification_rural_subdivision'] != '') & (native_classifications['area_classification_rural_subdivision'] != 'unknown'):
                area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification_rural_subdivision']]
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
        

        #STATION CLASSIFICATION
        mapping_station_classification = {'background':'background','industrial':'point_source','traffic':'point_source','unknown':'','':''}
    
        station_classification = mapping_station_classification[native_classifications['station_classification']]
    
        if native_classifications['station_classification'] == 'industrial':
            station_classification_point_source_subdivision = 'industrial'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision) 
        if native_classifications['station_classification'] == 'traffic':
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision) 
    
        #MAIN_EMISSION_SOURCE
        mapping_main_emission_source_classification = {'agriculture':'agriculture', 'commercial,institutionalandresidentialcombustion':'commercial_and_residential_combustion', 'extractionanddistributionoffossilfuels':'extraction_of_fossil_fuels', 'industrialcombustion':'industrial_combustion', 'nature':'natural', 'other':'', 'othermobilesourcesandmachinery(specify)':'other_mobile_sources_and_machinery', 'productionprocesses':'production_processes', 'publicpower,co-generationanddistrictheating':'power_production', 'roadtransport':'road_transport', 'useofsolvents':'solvents', 'wastetreatmentanddisposal':'waste_treatment_and_disposal', 'unknown':'', '':''}
        main_emission_source = mapping_main_emission_source_classification[native_classifications['main_emission_source']]
    
    
        #LAND USE - mixed land uses are sometimes given - process all into one standardised string separated by '/'
                   #'natural' land use is too vague to map to a land use (without given details) - do not use
        
        mapping_land_use = {'agricultural':'urban','commercial':'urban','industrial':'urban','residential':'urban','natural':'','mountain':'barren-rock','unknown':'','':''}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','commercial':'commercial','industrial':'industrial','residential':'residential'}
    
        native_land_use = native_classifications['land_use'].split('/')
    
        land_use = ''
        if native_land_use != ['']:
            for nlu in native_land_use:
                if land_use != '':
                    land_use = land_use+'/'
                iter_land_use = mapping_land_use[nlu]
                if iter_land_use == 'urban':
                    iter_land_use_urban_subdivision = mapping_land_use_urban_subdivision[nlu]
                    iter_land_use = '{}-{}'.format(iter_land_use,iter_land_use_urban_subdivision)
                land_use = land_use + iter_land_use
   
        #TERRAIN - if the land_use is 'barren-rock' this is because the raw provided land use is mountain, therefore give terrain this classification
        if land_use == 'barren-rock':
            terrain = 'mountain'
        #else if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        elif ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
          
        #MEASUREMENT SCALE - use REPRESENTATIVE_RADIUS to determine standardised measurement scale
        if (native_classifications['representative_radius'] != '') & (native_classifications['representative_radius'] != 'unknown'):
            if float(native_classifications['representative_radius']) < 0.1:
                measurement_scale = 'micro'
            elif (float(native_classifications['representative_radius']) >= 0.1) & (float(native_classifications['representative_radius']) < 0.5):
                measurement_scale = 'middle'
            elif (float(native_classifications['representative_radius']) >= 0.5) & (float(native_classifications['representative_radius']) < 4.0):
                measurement_scale = 'neighbourhood'
            elif (float(native_classifications['representative_radius']) >= 4.0) & (float(native_classifications['representative_radius']) < 50.0):
                measurement_scale = 'city'
            elif float(native_classifications['representative_radius']) >= 50.0:
                measurement_scale = 'regional'
        else:
            measurement_scale = ''

#------------------------------------------------
#EEA_AQ_eReporting E1a/E2a

#EEA_AQ_eReporting NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION (EEA_AQ_eReporting OR EEA_AIRBASE), STATION_CLASSIFICATION (EEA_AQ_eReporting OR EEA_AIRBASE), MAIN_EMISSION_SOURCE (EEA_AIRBASE), LAND_USE (EEA_AIRBASE), AREA_CLASSIFICATION_RURAL_SUBDIVISION (EEA_AIRBASE), REPRESENTATIVE_RADIUS (EEA_AIRBASE)

#AREA_CLASSIFICATION = suburban, urban, rural, rural-nearcity, rural-regional, rural-remote 
#STATION_CLASSIFICATION = background, industrial, traffic
#MAIN_EMISSION_SOURCE = agriculture, "commercial, institutional and residential combustion", "extraction and distribution of fossil fuels", "industrial combustion", nature, other, "other mobile sources and machinery (specify)", "production processes", "public power, co-generation and district heating", "road transport", "use of solvents", "waste treatment and disposal"
#LAND_USE = agricultural, commercial, industrial, natural, residential
#AREA_CLASSIFICATION_RURAL_SUBDIVISION = nearcity, remote, regional
#REPRESENTATIVE_RADIUS = representative radius of measurements in kilometres

    elif 'EEA_AQ_eReporting' in run_details['network']:

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','rural-nearcity':'rural','rural-near_city':'rural','rural-regional':'rural','rural-remote':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        mapping_area_classification_rural_subdivision = {'rural-nearcity':'near_city','rural-near_city':'near_city','rural-regional':'regional','rural-remote':'remote','rural':''}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification']]
            if area_classification_rural_subdivision != '':
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
            #if do not have rural subdivisions, then try use AIRBASE only available AREA_CLASSIFICATION_RURAL_SUBDIVISION
            #check AREA_CLASSIFICATION_RURAL_SUBDIVISION not empty string or 'unknown'
            elif (native_classifications['area_classification_rural_subdivision'] != '') & (native_classifications['area_classification_rural_subdivision'] != 'unknown'):
                mapping_area_classification_rural_subdivision = {'nearcity':'near_city','regional':'regional','remote':'remote'}
                area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification_rural_subdivision']]
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
        
        #STATION CLASSIFICATION
        mapping_station_classification = {'background':'background','industrial':'point_source','traffic':'point_source','unknown':'','':''}
    
        station_classification = mapping_station_classification[native_classifications['station_classification']]
    
        if native_classifications['station_classification'] == 'traffic':
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)
            
        if native_classifications['station_classification'] == 'industrial':
            station_classification_point_source_subdivision = 'industrial'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)  

        #MAIN_EMISSION_SOURCE
        mapping_main_emission_source_classification = {'agriculture':'agriculture', 'commercial,institutionalandresidentialcombustion':'commercial_and_residential_combustion', 'extractionanddistributionoffossilfuels':'extraction_of_fossil_fuels', 'industrialcombustion':'industrial_combustion', 'nature':'natural', 'other':'', 'othermobilesourcesandmachinery(specify)':'other_mobile_sources_and_machinery', 'productionprocesses':'production_processes', 'publicpower,co-generationanddistrictheating':'power_production', 'roadtransport':'road_transport', 'useofsolvents':'solvents', 'wastetreatmentanddisposal':'waste_treatment_and_disposal', 'unknown':'', '':''}
        main_emission_source = mapping_main_emission_source_classification[native_classifications['main_emission_source']]
    
    
        #LAND USE - mixed land uses are sometimes given - process all into one standardised string separated by '/'
                   #'natural' land use is too vague to map to a land use (without given details) - do not use
        
        mapping_land_use = {'agricultural':'urban','commercial':'urban','industrial':'urban','residential':'urban','natural':'','mountain':'barren-rock','unknown':'','':''}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','commercial':'commercial','industrial':'industrial','residential':'residential'}
    
        native_land_use = native_classifications['land_use'].split('/')
    
        land_use = ''
        if native_land_use != ['']:
            for nlu in native_land_use:
                if land_use != '':
                    land_use = land_use+'/'
                iter_land_use = mapping_land_use[nlu]
                if iter_land_use == 'urban':
                    iter_land_use_urban_subdivision = mapping_land_use_urban_subdivision[nlu]
                    iter_land_use = '{}-{}'.format(iter_land_use,iter_land_use_urban_subdivision)
                land_use = land_use + iter_land_use

        #TERRAIN - if the land_use is 'barren-rock' this is because the raw provided land use is mountain, therefore give terrain this classification
        if land_use == 'barren-rock':
            terrain = 'mountain'
        #else if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        elif ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''

        #MEASUREMENT SCALE - use REPRESENTATIVE_RADIUS to determine standardised measurement scale
        if (native_classifications['representative_radius'] != '') & (native_classifications['representative_radius'] != 'unknown'):
            if float(native_classifications['representative_radius']) < 0.1:
                measurement_scale = 'micro'
            elif (float(native_classifications['representative_radius']) >= 0.1) & (float(native_classifications['representative_radius']) < 0.5):
                measurement_scale = 'middle'
            elif (float(native_classifications['representative_radius']) >= 0.5) & (float(native_classifications['representative_radius']) < 4.0):
                measurement_scale = 'neighbourhood'
            elif (float(native_classifications['representative_radius']) >= 4.0) & (float(native_classifications['representative_radius']) < 50.0):
                measurement_scale = 'city'
            elif float(native_classifications['representative_radius']) >= 50.0:
                measurement_scale = 'regional'
        else:
            measurement_scale = ''

#------------------------------------------------
#JAPAN_NIES

#JAPAN_NIES NATIVE CLASSIFICATIONS = STATION_CLASSIFICATION, LAND_USE

#STATION_CLASSFICATION = general_environmental_atmospheric_measurement_station:environmental_standard_application_bureau, general_environmental_atmospheric_measurement_station:environmental_standard_exemption_bureau
#                        vehicle_exhaust_gas_measurement_station:roadside_station, vehicle_exhaust_gas_measurement_station:roadway_station, meteorological_3D_station, descent_and_dust_station, other
#LAND USE = residential, commercial, semi-construction, engineering, technical college, unknown, scenic, harbor, other

    elif run_details['network'] == 'JAPAN_NIES':

        #AREA CLASSIFICATION - infer some information from station classification (and land use if necessary)
        mapping_area_classification = {'general_environmental_atmospheric_measurement_station':'','general_environmental_atmospheric_measurement_station:environmental_standard_application_bureau':'','general_environmental_atmospheric_measurement_station:environmental_standard_exemption_bureau':'','vehicle_exhaust_gas_measurement_station':'urban','vehicle_exhaust_gas_measurement_station:roadside_station':'urban','vehicle_exhaust_gas_measurement_station:roadway_station':'urban','meteorological_3d_station':'','descent_and_dust_station':'rural','other':'rural','':''}
        area_classification = mapping_area_classification[native_classifications['station_classification']]
        if area_classification == '':
            if native_classifications['land_use'] in ['residential','commercial','semi-construction','engineering','technical_college','scenic','harbor']:
                area_classification = 'urban'
    
        #STATION CLASSIFICATION
        mapping_station_classification = {'general_environmental_atmospheric_measurement_station':'','general_environmental_atmospheric_measurement_station:environmental_standard_application_bureau':'background','general_environmental_atmospheric_measurement_station:environmental_standard_exemption_bureau':'point_source','vehicle_exhaust_gas_measurement_station':'point_source-traffic','vehicle_exhaust_gas_measurement_station:roadside_station':'point_source-traffic','vehicle_exhaust_gas_measurement_station:roadway_station':'point_source-traffic','meteorological_3d_station':'','descent_and_dust_station':'background','other':'background','':''}
        station_classification = mapping_station_classification[native_classifications['station_classification']] 

        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        mapping_land_use = {'residential':'urban','commercial':'urban','semi-construction':'urban','engineering':'urban','technical_college':'urban','scenic':'urban','harbor':'urban','unknown':'','other':'','':''}
        mapping_land_use_urban_subdivision = {'residential':'residential','commercial':'commercial','semi-construction':'industrial','engineering':'industrial','technical_college':'industrial','scenic':'park','harbor':'transportation'}
        land_use = mapping_land_use[native_classifications['land_use']]
        if native_classifications['land_use'] in list(mapping_land_use_urban_subdivision.keys()):
            land_use = '{}-{}'.format(land_use,mapping_land_use_urban_subdivision[native_classifications['land_use']])
        
        #TERRAIN 
        terrain = ''
        if native_classifications['station_classification'] == 'other':
            terrain = 'mountain'
        #if terrain is empty string still, if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if terrain == '':    
            if ('urban' in area_classification):
                terrain = 'complex'

        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#MEXICO_CDMX 

#MEXICO_CDMX NATIVE CLASSIFICATIONS = MEASUREMENT_SCALE

#MEASUREMENT_SCALE = 1 (<100m), 2 (100m-1km), 3 (1km - 4km), 4 (4km - 20km), 5 (>=20km)

    elif run_details['network'] == 'MEXICO_CDMX':

        #AREA CLASSIFICATION - Derive from measurement scale
                             #anything <20 km representativity assume to be urban, the rest is rural
        mapping_area_classification = {'1':'urban','2':'urban','3':'urban','4':'urban','5':'rural','':''}
        area_classification = mapping_area_classification[native_classifications['measurement_scale']]
        
        #STATION CLASSIFICATION - no category that gives this, but can derive it
                                 #where measurement_scale is 'micro', assume this point source dominated (<1km)
                                 #other larger scales (>1km) are assumed to be representative of background  
        mapping_station_classification = {'1':'point_source','2':'point_source','3':'background','4':'background','5':'background','':''}
        station_classification = mapping_station_classification[native_classifications['measurement_scale']]
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - scales boundaries do not exactly to defined standards 1:1, but close enough to allow it
        mapping_measurement_scale = {'1':'micro','2':'middle','3':'neighbourhood','4':'city','5':'regional','':''}
        measurement_scale = mapping_measurement_scale[native_classifications['measurement_scale']]

#------------------------------------------------
#MITECO

#MITECO NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION (MITECO, EEA_AQ_eReporting OR EEA_AIRBASE), STATION_CLASSIFICATION (MITECO, EEA_AQ_eReporting OR EEA_AIRBASE), MAIN_EMISSION_SOURCE (EEA_AIRBASE), LAND_USE (MITECO, EEA_AIRBASE), AREA_CLASSIFICATION_RURAL_SUBDIVISION (MITECO, EEA_AIRBASE), REPRESENTATIVE_RADIUS (EEA_AIRBASE)

#AREA_CLASSIFICATION = suburban, urban, rural, rural-nearcity, rural-regional, rural-remote 
#STATION_CLASSIFICATION = background, industrial, traffic
#MAIN_EMISSION_SOURCE = agriculture, "commercial, institutional and residential combustion", "extraction and distribution of fossil fuels", "industrial combustion", nature, other, "other mobile sources and machinery (specify)", "production processes", "public power, co-generation and district heating", "road transport", "use of solvents", "waste treatment and disposal"
#LAND_USE = agricultural, commercial, industrial, natural, residential
#AREA_CLASSIFICATION_RURAL_SUBDIVISION = nearcity, remote, regional
#REPRESENTATIVE_RADIUS = representative radius of measurements in kilometres

    elif 'MITECO' in run_details['network']:

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urban':'urban','urbana':'urban','suburban':'urban','suburbana':'urban','rural':'rural','rural-nearcity':'rural','rural-near_city':'rural','rural-regional':'rural','rural-remote':'rural','unknown':'','desconocido':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','urbana':'centre','suburban':'suburban','suburbana':'suburban'}
        mapping_area_classification_rural_subdivision = {'rural-nearcity':'near_city','rural-near_city':'near_city','rural-regional':'regional','rural-remote':'remote','rural':''}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        if area_classification == 'rural':
            area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification']]
            if area_classification_rural_subdivision != '':
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
            #if do not have rural subdivisions, then try use AIRBASE only available AREA_CLASSIFICATION_RURAL_SUBDIVISION
            #check AREA_CLASSIFICATION_RURAL_SUBDIVISION not empty string or 'unknown'
            elif (native_classifications['area_classification_rural_subdivision'] != '') & (native_classifications['area_classification_rural_subdivision'] != 'unknown') & (native_classifications['area_classification_rural_subdivision'] != 'desconocido'):
                mapping_area_classification_rural_subdivision = {'nearcity':'near_city','cercadeunaciudad':'near_city','regional':'regional','remote':'remote','remota':'remote'}
                area_classification_rural_subdivision = mapping_area_classification_rural_subdivision[native_classifications['area_classification_rural_subdivision']]
                area_classification = '{}-{}'.format(area_classification,area_classification_rural_subdivision) 
        
        #STATION CLASSIFICATION
        mapping_station_classification = {'background':'background','fondo':'background','industrial':'point_source','traffic':'point_source','trafico':'point_source','unknown':'','desconocido':'','':''}
    
        station_classification = mapping_station_classification[native_classifications['station_classification']]
    
        if native_classifications['station_classification'] in ['traffic','trafico']:
            station_classification_point_source_subdivision = 'traffic'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)
            
        if native_classifications['station_classification'] == 'industrial':
            station_classification_point_source_subdivision = 'industrial'
            station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)  

        #MAIN_EMISSION_SOURCE
        mapping_main_emission_source_classification = {'agriculture':'agriculture', 'commercial,institutionalandresidentialcombustion':'commercial_and_residential_combustion', 'extractionanddistributionoffossilfuels':'extraction_of_fossil_fuels', 'industrialcombustion':'industrial_combustion', 'nature':'natural', 'other':'', 'othermobilesourcesandmachinery(specify)':'other_mobile_sources_and_machinery', 'productionprocesses':'production_processes', 'publicpower,co-generationanddistrictheating':'power_production', 'roadtransport':'road_transport', 'useofsolvents':'solvents', 'wastetreatmentanddisposal':'waste_treatment_and_disposal', 'unknown':'', '':''}
        main_emission_source = mapping_main_emission_source_classification[native_classifications['main_emission_source']]
    
    
        #LAND USE - mixed land uses are sometimes given - process all into one standardised string separated by '/'
                   #'natural' land use is too vague to map to a land use (without given details) - do not use
        
        mapping_land_use = {'agricultural':'urban','agricola':'urban','commercial':'urban','comercial':'urban','industrial':'urban','residential':'urban','residencial':'urban','natural':'','naturaleza':'','parque':'urban','mar(cerca)':'wetland','lago(cerca)':'wetland','mountain':'barren-rock','montaña':'barren-rock','unknown':'','desconocido':'','':''}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','agricola':'agricultural','commercial':'commercial','comercial':'commercial','industrial':'industrial','parque':'park','residential':'residential','residencial':'residential'}
    
        native_land_use = native_classifications['land_use'].split('/')
    
        land_use = ''
        if native_land_use != ['']:
            for nlu in native_land_use:
                if land_use != '':
                    land_use = land_use+'/'
                iter_land_use = mapping_land_use[nlu]
                if iter_land_use == 'urban':
                    iter_land_use_urban_subdivision = mapping_land_use_urban_subdivision[nlu]
                    iter_land_use = '{}-{}'.format(iter_land_use,iter_land_use_urban_subdivision)
                land_use = land_use + iter_land_use

        #TERRAIN - if the land_use is 'barren-rock' this is because the raw provided land use is mountain, therefore give terrain this classification
        if land_use == 'barren-rock':
            terrain = 'mountain'
        #if the native land use is 'mar(cerca)' then terrain is coastal
        elif native_classifications['land_use'] == 'mar(cerca)':
            terrain = 'coastal'
        #else if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        elif 'urban' in area_classification:
            terrain = 'complex'
        else:
            terrain = ''

        #MEASUREMENT SCALE - use REPRESENTATIVE_RADIUS to determine standardised measurement scale
        if (native_classifications['representative_radius'] != '') & (native_classifications['representative_radius'] != 'unknown'):
            if float(native_classifications['representative_radius']) < 0.1:
                measurement_scale = 'micro'
            elif (float(native_classifications['representative_radius']) >= 0.1) & (float(native_classifications['representative_radius']) < 0.5):
                measurement_scale = 'middle'
            elif (float(native_classifications['representative_radius']) >= 0.5) & (float(native_classifications['representative_radius']) < 4.0):
                measurement_scale = 'neighbourhood'
            elif (float(native_classifications['representative_radius']) >= 4.0) & (float(native_classifications['representative_radius']) < 50.0):
                measurement_scale = 'city'
            elif float(native_classifications['representative_radius']) >= 50.0:
                measurement_scale = 'regional'
        else:
            measurement_scale = ''

#------------------------------------------------
#NOAA_ISD

#NATIVE CLASSIFICATIONS = NONE
    
    elif run_details['network'] == 'NOAA_ISD':
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''
    
#------------------------------------------------
#SEARCH

#SEARCH NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = urban, suburban, rural
    
    elif run_details['network'] == 'SEARCH':
    
        #AREA CLASSIFICATION
        mapping_area_classification = {'urban':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urban':'centre','suburban':'suburban'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume background nature of 'urban' sites' 
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#UK_AIR

#UK_AIR NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = roadside, ruralbackground, suburban, suburbanbackground, suburbanindustrial, 
#                      unknownindustrial, urbanbackground, urbancentre, urbanindustrial, urbantraffic, 
#                      not_available

    elif run_details['network'] == 'UK_AIR':
    
        #AREA CLASSIFICATION 
        mapping_area_classification = {'roadside':'urban', 'ruralbackground':'rural', 'suburban':'urban', 
                                       'suburbanbackground':'urban', 'suburbanindustrial':'urban', 
                                       'unknownindustrial':'urban', 'urbanbackground':'urban', 'urbancentre':'urban', 
                                       'urbanindustrial':'urban', 'urbantraffic':'urban', 'notavailable':'', '':''}
        mapping_area_classification_urban_subdivision = {'suburban':'suburban', 'suburbanbackground':'suburban',
                                                         'suburbanindustrial':'suburban', 'urbanbackground':'centre', 
                                                         'urbancentre':'centre', 'urbanindustrial':'centre', 
                                                         'urbantraffic':'centre'}
        
        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            if native_classifications['area_classification'] in mapping_area_classification_urban_subdivision:
                area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
                area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 

        #STATION CLASSIFICATION - inferred from area_classification
        mapping_station_classification = {'roadside':'point_source', 'ruralbackground':'background', 
                                          'suburban':'background', 'suburbanbackground':'background', 
                                          'suburbanindustrial':'point_source', 'unknownindustrial':'point_source', 
                                          'urbanbackground':'background', 'urbancentre':'background', 
                                          'urbanindustrial':'point_source', 'urbantraffic':'point_source', 
                                          'notavailable':'', '':''}
        mapping_station_classification_point_source_subdivision = {'roadside':'traffic', 
                                                                   'suburbanindustrial':'industrial',
                                                                   'unknownindustrial':'industrial',
                                                                   'urbanindustrial':'industrial',
                                                                   'urbantraffic':'traffic'}

        station_classification = mapping_station_classification[native_classifications['area_classification']]
    
        if station_classification == 'point_source':
            if native_classifications['area_classification'] in mapping_station_classification_point_source_subdivision:
                station_classification_point_source_subdivision = mapping_station_classification_point_source_subdivision[native_classifications['area_classification']]
                station_classification = '{}-{}'.format(station_classification,station_classification_point_source_subdivision)
            
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#US_EPA_AirNow_DOS

#US_EPA_AirNow_DOS NATIVE CLASSIFICATIONS = NONE
    
    elif run_details['network'] == 'US_EPA_AirNow_DOS':
    
        #AREA CLASSIFICATION - No information
        area_classification = ''
        
        #STATION CLASSIFICATION - No information
        station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#US_EPA_AQS

#US_EPA_AQS NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION, LAND_USE, MEASUREMENT_SCALE

#AREA_CLASSIFICATION = urbanandcentercity, suburban, rural
#LAND_USE = agricultural, blightedareas, commercial, desert, forest, industrial, militaryreservation, mobile, residential
#MEASUREMENT_SCALE = microscale, middlescale, neighborhood, urbanscale, regionalscale 

    elif run_details['network'] == 'US_EPA_AQS':

        #AREA CLASSIFICATION 
        mapping_area_classification = {'urbanandcentercity':'urban','suburban':'urban','rural':'rural','unknown':'','':''}
        mapping_area_classification_urban_subdivision = {'urbanandcentercity':'centre','suburban':'suburban'}

        area_classification = mapping_area_classification[native_classifications['area_classification']] 
    
        if area_classification == 'urban':
            area_classification_urban_subdivision = mapping_area_classification_urban_subdivision[native_classifications['area_classification']]
            area_classification = '{}-{}'.format(area_classification,area_classification_urban_subdivision) 


        #STATION CLASSIFICATION - no direct category that gives this, but can derive it
                                 #where measurement_scale is 'micro' or 'middle', assume this point source dominated (<0.5km)
                                 #other larger scales (>0.5km) are assumed to be representative of background  
    
        mapping_station_classification = {'microscale':'point_source','middlescale':'point_source','neighborhood':'background','urbanscale':'background','regionalscale':'background','unknown':'','':''}
        station_classification = mapping_station_classification[native_classifications['measurement_scale']]
    
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
    
        #LAND USE
        mapping_land_use = {'agricultural':'urban','blightedareas':'urban','commercial':'urban','desert':'barren','forest':'forest','industrial':'urban','militaryreservation':'urban','mobile':'urban','residential':'urban','unknown':'','':''}
        mapping_land_use_barren_subdivision = {'desert':'desert'}
        mapping_land_use_urban_subdivision = {'agricultural':'agricultural','blightedareas':'blighted','commercial':'commercial','industrial':'industrial','militaryreservation':'military','mobile':'transportation','residential':'residential'}

        land_use = mapping_land_use[native_classifications['land_use']]
    
        if land_use == 'barren':
            land_use_barren_subdivision = mapping_land_use_barren_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
        if land_use == 'urban':
            land_use_urban_subdivision = mapping_land_use_urban_subdivision[native_classifications['land_use']]
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
    
        #MEASUREMENT SCALE 
        mapping_measurement_scale = {'microscale':'micro','middlescale':'middle','neighborhood':'neighbourhood','urbanscale':'city','regionalscale':'regional','unknown':'','':''}
        measurement_scale = mapping_measurement_scale[native_classifications['measurement_scale']]

#------------------------------------------------
#US_EPA_CASTNET 

#US_EPA_CASTNET NATIVE CLASSIFICATIONS = LAND_USE, TERRAIN

#LAND_USE = agric, arid, coastal, coastal/marsh, desert, forest, forest/arid, forest/marsh, forest/tropical, marsh, prairie, range, tropical, urban, urban/agric
#TERRAIN = coastal, complex, flat, flat/water, mountaintop, rolling, water

    elif run_details['network'] == 'US_EPA_CASTNET':
    
        #AREA CLASSIFICATION - derive from certain LAND_USE cateogories. 
                              #Specifically the 'urban', 'urban/agric' are classed as 'urban'
                              #The rest are assumed to be 'rural'
    
        mapping_area_classification = {'agric':'rural','arid':'rural','coastal':'rural','coastal/marsh':'rural','desert':'rural','forest':'rural','forest/arid':'rural','forest/marsh':'rural','forest/tropical':'rural','marsh':'rural','prairie':'rural','range':'rural','tropical':'rural','urban':'urban','urban/agric':'urban','unknown':'','':''}
        area_classification = mapping_area_classification[native_classifications['land_use']]
    
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume representativeness of 'urban' sites' 
        
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
            
        #LAND USE - mixed land uses are sometimes given, put these mixed values into just category (unlike done with AIRBASE), as way classifications are reported lends itself to this 
                    #i.e. they are are mixture of biome and land covers, and often these merge naturally into 1 category
                    #make assumption for 'range', that lands grazed by animals are naturally occurring -->, 'open'. Can also be lands grazed by domestic animals --> 'agriculture', but this is not accounted for 
                   
        mapping_land_use = {'agric':'urban','arid':'barren','coastal':'wetland','coastal/marsh':'wetland','desert':'barren','forest':'forest','forest/arid':'forest','forest/marsh':'forest','forest/tropical':'forest','marsh':'wetland','prairie':'open','range':'open','tropical':'forest','urban':'urban','urban/agric':'urban','unknown':'','':''}
        land_use = mapping_land_use[native_classifications['land_use']]
        
        if native_classifications['land_use'] == 'desert':
            land_use_barren_subdivision = 'desert'
            land_use = '{}-{}'.format(land_use,land_use_barren_subdivision) 
            
        if native_classifications['land_use'] == 'prairie':
            land_use_open_subdivision = 'grassland'
            land_use = '{}-{}'.format(land_use,land_use_open_subdivision) 
        
        if (native_classifications['land_use'] == 'agric') or (native_classifications['land_use'] == 'urban/agric'):
            land_use_urban_subdivision = 'agricultural'
            land_use = '{}-{}'.format(land_use,land_use_urban_subdivision) 
            
        #TERRAIN - first use given 'terrain' classification
        mapping_terrain_classification = {'coastal':'coastal','complex':'complex','flat':'flat','flat/water':'flat','mountaintop':'mountain','rolling':'rolling','water':'','unknown':'','':''}
        terrain = mapping_terrain_classification[native_classifications['terrain']]
        #if terrain is empty string still, if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if terrain == '':
            if ('urban' in area_classification):
                terrain = 'complex'

        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#US_NADP_AMNet

#US NADP AMoN NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = i (Isolated), r (Rural), s (Suburban), u (Urban), p (Research/Provisional)
    
    elif run_details['network'] == 'US_NADP_AMNet':
    
        #AREA CLASSIFICATION 
        mapping_area_classification = {'i':'rural-remote','r':'rural','s':'urban-suburban','u':'urban','p':'','':''}
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume background nature of 'urban' sites' 
        if 'rural' in area_classification:
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#US_NADP_AMoN

#US NADP AMoN NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = I (Isolated), R (Rural), S (Suburban), U (Urban), P (Research/Provisional)
    
    elif run_details['network'] == 'US_NADP_AMoN':
    
        #AREA CLASSIFICATION 
        mapping_area_classification = {'i':'rural-remote','r':'rural','s':'urban-suburban','u':'urban','p':'','':''}
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
                                 #No not have enough detail to assume background nature of 'urban' sites' 
        if 'rural' in area_classification:
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - if the AREA_CLASSIFICATION is some type of "urban" class, then assume the terrain is "complex"
        if ('urban' in area_classification):
            terrain = 'complex'
        else:
            terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''

#------------------------------------------------
#WMO_WDCGG 

#WMO WDCGG NATIVE CLASSIFICATIONS = AREA_CLASSIFICATION

#AREA_CLASSIFICATION = 'airpollutionmonitoring', 'backgroundobservation', "monitoringbackgroundconditions,andtheinfluenceofairpollutionunknown"

    elif run_details['network'] == 'WMO_WDCGG':
    
        #AREA CLASSIFICATION
        mapping_area_classification = {'airpollutionmonitoring':'','backgroundobservation':'rural','monitoringbackgroundconditions,andtheinfluenceofairpollution':'rural','unknown':'','':''}
        area_classification = mapping_area_classification[native_classifications['area_classification']]
        
        #STATION CLASSIFICATION - Derive from area classification, if class is 'rural', then assume it is 'background' representative
        if area_classification == 'rural':
            station_classification = 'background'
        else:
            station_classification = ''
        
        #MAIN_EMISSION_SOURCE - No information
        main_emission_source = ''
        
        #LAND USE - No information
        land_use = ''
        
        #TERRAIN - No information
        terrain = ''
        
        #MEASUREMENT SCALE - No information
        measurement_scale = ''
        
#------------------------------------------------
    #OUTPUT STANDARDISED CLASSIFICATION INFORMATION

    #map classification variables to GHOST metadata variable names
    standard_classes = {'area_classification':area_classification,
                        'station_classification':station_classification,
                        'main_emission_source':main_emission_source,
                        'land_use':land_use,
                        'terrain':terrain,
                        'measurement_scale':measurement_scale}

    #iterate through each standard class field
    for class_name in list(standard_classes.keys()):    
        #if any of the standard classifications are empty strings, convert to NaN
        if standard_classes[class_name] == '':
            standard_classes[class_name] = np.NaN

        #add standard classifications to station metadata array if desired (write_to_metadata_array == True)
        if write_to_metadata_array:
            site_metadata[class_name]['value'].append(standard_classes[class_name])
            
    #if have written to station metadata array, return the modified array
    if write_to_metadata_array:
        return site_metadata 
    #otherwise, return standard classification dictionary
    else:
        return standard_classes

###--------------------------------------------------------------------------------------------------###
#DEFINE DATA PROVIDER DATA FLAGS TO STANDARD DATA FLAG MAPPING
###--------------------------------------------------------------------------------------------------###

#define nested dictionary which defines mapping of data provider flag codes to defined standard data flags, per network
#some definitions:
#standard_data_flag_name = standard data flag name
#standard_data_flag_code = standard data flag code
#network_decreed_validity = do the providing data network deem the data to be valid (V = Valid, I = Invalid) 
#GHOST_decreed_validity = does GHOST deem the data to be valid (V = Valid, I = Invalid)
#network_provided_flag_description = network stated description of raw network provided data flag

network_data_flag_to_standard_data_flag_mapping = {

#--------------------------------------------------------------   
#AERONET
'AERONET':{},

#--------------------------------------------------------------   
#BJMEMC
'BJMEMC':{},

#--------------------------------------------------------------
#CANADA NAPS
#--------------------------------------------------------------
'CANADA_NAPS':{
    #Basic Flags
    'V0':{'standard_data_flag_name':['Valid Data'],      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid value'},
    'M1':{'standard_data_flag_name':['Missing Data'],    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing value - no value available'},
    'H1':{'standard_data_flag_name':['Un-Flagged Data'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Historical data that have not been assessed or validated'},
    '':  {'standard_data_flag_name':['Un-Flagged Data'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Validation has not been completed'},
    #Extreme/Irregular data
    'V1':{'standard_data_flag_name':['Value Below Detection Limit'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Valid value - below detection limit'},
    'V2':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid value - extreme or unusual value assessed and considered valid'},
    'M2':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing value - invalidated'},
    #Integrated data flags
    'R': {'standard_data_flag_name':['Valid Data'],  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Routine Sample'},
    'FB':{'standard_data_flag_name':['Blank Check'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field Blank'},
    'LB':{'standard_data_flag_name':['Blank Check'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Lab Blank'},
    'MB':{'standard_data_flag_name':['Blank Check'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Method Blank'},
    'TB':{'standard_data_flag_name':['Blank Check'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Travel Blank'},
    'C': {'standard_data_flag_name':['Calibration'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Control'}
},

#--------------------------------------------------------------
#CAPMoN
#--------------------------------------------------------------
'CAPMoN':{        
    #Basic Flags
    'V0':{'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid measurement'},
    'M1':{'standard_data_flag_name':['Missing Data'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'No value was reported by the field or laboratory instrument (e.g. instrument failure, lost sample).'},
    #Estimated data
    'V7':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'The reported value is below the analytical or method or instrument detection limit and has been substituted by the detection limit or lowest measurable value as described in the EXPLANATION OF REPORTED DETECTION LIMIT VALUES.'},
    #Extreme/Irregular data
    'V1':{'standard_data_flag_name':['Value Below Detection Limit'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'The reported value is below the analytical or method or instrument detection limit. Values may be positive, zero or negative.'},
    'V2':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'The reported value is unusual (e.g., an outlier) but assessed by the Principal investigator who determined that it was valid or there was insufficient evidence for invalidation. Principal investigators may have included a textual comment to indicate possible causes for the unusual value (e.g. forest fire, volcano, dust storm).'},
    'M2':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'invalidated by Principal Investigator'}
},

#--------------------------------------------------------------
#CHILE SINCA
#--------------------------------------------------------------
'CHILE_SINCA':{
    '1':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Registros validados'},
    '2':{'standard_data_flag_name':['Preliminary Data'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Registros preliminares'},
    '3':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Registros no validados'}
},
 
#--------------------------------------------------------------   
#CNEMC
'CNEMC':{},

#--------------------------------------------------------------   
#EANET
'EANET':{},

#--------------------------------------------------------------
#EBAS
#--------------------------------------------------------------
'EBAS':{
    #Basic Flags
    '999':{'standard_data_flag_name':['Missing Data'],               'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing measurement, unspecified reason'},
    '899':{'standard_data_flag_name':['Missing Data'],               'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Measurement undefined, unspecified reason'},
    '750':{'standard_data_flag_name':['Missing Data'],               'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'H+ not measured in alkaline sample'},
    '420':{'standard_data_flag_name':['Preliminary Data'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Preliminary data'},
    '259':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Unspecified error expected'},
    '220':{'standard_data_flag_name':['Preliminary Data'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Preliminary data'},
    '000':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid Measurement'},        
    #Estimated data
    '799':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Measurement missing (unspecified reason), data element contains estimated value'},
    '798':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Measurement missing (unspecified reason), data element contains estimated value. Considered valid.'},
    '797':{'standard_data_flag_name':['Estimated Data - Value Substituted from Secondary Monitor'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data element taken from co-located instrument'},
    '784':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low precipitation, concentration estimated'},
    '782':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low precipitation, concentration estimated'},
    '781':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value below detection limit, data element contains detection limit'},
    '780':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value below detection or quantification limit, data element contains estimated or measured value. Use of flag 147 is encouraged.'},
    '771':{'standard_data_flag_name':['Estimated Data - Value Above Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value above range, data element contains upper range limit'},
    '770':{'standard_data_flag_name':['Estimated Data - Value Above Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value above range, data element contains estimated value'},
    '760':{'standard_data_flag_name':['Estimated Data - Multiple Parameters Aggregated'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Value estimated by summing up the constituents measured'},
    '655':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Estimated value created by averaging or splitting samples'},
    #Measurement issues
    '741':{'standard_data_flag_name':['Measurement Issue - Unspecified'],                       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':"Non refractory AMS concentrations. Don't include compounds that volatalises above 600 deg C"},
    '699':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Mechanical problem, unspecified reason'},
    '681':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low data capture'},
    '680':{'standard_data_flag_name':['Unmeasured Key Meteorological Parameter'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Undefined wind direction'},
    '677':{'standard_data_flag_name':['Extreme External Meteorological Conditions'],            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Icing or hoar frost in the intake'},
    '675':{'standard_data_flag_name':['Unmeasured Key Meteorological Parameter'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'no visibility data available'},
    '674':{'standard_data_flag_name':['Extreme External Meteorological Conditions'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Icing or hoar frost in the intake, considered valid'},
    '670':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Incomplete data acquisition for multi-component data sets'},
    '669':{'standard_data_flag_name':['Physical Issue With Filter'],                            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Moist or wet filter, invalid'},
    '668':{'standard_data_flag_name':['Physical Issue With Filter'],                            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moist or wet filter, valid'},
    '666':{'standard_data_flag_name':['Physical Issue With Filter'],                            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter damaged, invalid'},
    '665':{'standard_data_flag_name':['Physical Issue With Filter'],                            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter damaged, valid'},
    '664':{'standard_data_flag_name':['Invalid Flow Rate'],                                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Instrument flow(s) too far off target value, considered invalid'},
    '663':{'standard_data_flag_name':['Invalid Flow Rate'],                                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Too high sampling flow, data considered invalid'},
    '662':{'standard_data_flag_name':['Invalid Flow Rate'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Too high sampling flow, data considered valid'},
    '660':{'standard_data_flag_name':['Erroneous Sampling Operation'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unspecified sampling anomaly, considered valid'},
    '659':{'standard_data_flag_name':['Erroneous Sampling Operation'],                          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Unspecified sampling anomaly'},
    '658':{'standard_data_flag_name':['Erroneous Sampling Operation'],                          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Too small air volume'},
    '656':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Wet-only collector failure, operated as bulk collector'},
    '654':{'standard_data_flag_name':['Erroneous Sampling Operation'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sampling period longer than normal, considered representative. Observed values reported'},
    '653':{'standard_data_flag_name':['Erroneous Sampling Operation'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sampling period shorter than normal, considered representative. Observed values reported'},
    '650':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Precipitation collector failure'},
    '649':{'standard_data_flag_name':['Power Failure'],                                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Temporary power fail has affected sampler operation'},
    '644':{'standard_data_flag_name':['Operational Maintenance Check Issue'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low instrument precision and/or calibration issues'},
    '641':{'standard_data_flag_name':['Human Error'],                                           'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Aerosol filters installed incorrectly'},
    '640':{'standard_data_flag_name':['Extreme Internal Instrument Meteorological Conditions'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Instrument internal relative humidity above 40%'},
    '635':{'standard_data_flag_name':['Extreme Internal Instrument Meteorological Conditions'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Internal temperatures too far off target value, considered invalid'},
    '632':{'standard_data_flag_name':['Human Error'],                                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Lid of polyurethane foam (PUF) sampler not closed'},
    '63O':{'standard_data_flag_name':['Human Error'],                                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'POP concentration from the polyurethane foam (PUF) only'},
    '549':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Impure chemicals'},
    '541':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Gold trap passiviated by unknown compound'},
    '540':{'standard_data_flag_name':['Internal Laboratory/Instrument Contamination'],          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Spectral interference in laboratory analysis'},
    '534':{'standard_data_flag_name':['Human Error'],                                           'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Wrong coated denuder used'},
    '533':{'standard_data_flag_name':['Human Error'],                                           'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filters mixed up; incorrect analysis'},
    '532':{'standard_data_flag_name':['Operational Maintenance Check Issue'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data less accurate than normal due to high field blank value'},
    '531':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low recovery, analysis inaccurate'},
    '530':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to too low or too high recovery'},
    '521':{'standard_data_flag_name':['Extreme Sample Transport Conditions'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Bactericide was added to sample for storage under warm climate. Considered valid'},
    '498':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Gold trap inconsistency in mercury monitor'},
    '478':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to inconsistency discovered through ion balance calculations'},
    '477':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to inconsistency between measured and estimated conductivity'},
    '476':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistency discovered through ion balance calculations, but considered valid'},
    '475':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistency between measured and estimated conductivity, but considered valid'},
    '298':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Gold trap inconsistency in mercury monitor'},
    '278':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to inconsistency discovered through ion balance calculations'},
    '277':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to inconsistency between measured and estimated conductivity'},
    '276':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistency discovered through ion balance calculations, but considered valid'},
    '275':{'standard_data_flag_name':['Chemical Issue'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistency between measured and estimated conductivity, but considered valid'},
    '190':{'standard_data_flag_name':['Measurement Issue - Unspecified'],                       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Not corrected for cross-sensitivity to particle scattering'},
    '101':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Denuder capture efficiency < 75%. Valid measurement'},
    #Operational Maintenance checks
    '980':{'standard_data_flag_name':['Operational Maintenance - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing due to calibration or zero/span check'},
    '687':{'standard_data_flag_name':['Span Check'],                            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to span check. Used for Level 0.'},
    '686':{'standard_data_flag_name':['Zero Check'],                            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to zero check. Used for Level 0.'},
    '685':{'standard_data_flag_name':['Operational Maintenance - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to secondary standard gas measurement. Used for Level 0.'},
    '684':{'standard_data_flag_name':['Zero Check'],                            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to zero/span check. Used for Level 0.'},
    '683':{'standard_data_flag_name':['Calibration'],                           'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to calibration. Used for Level 0.'},
    '682':{'standard_data_flag_name':['Operational Maintenance - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to calibration or zero/span check. Used for Level 0.'},
    #Local Contamination
    '652':{'standard_data_flag_name':['Construction Contamination'],        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Construction/acitivity nearby'},
    '651':{'standard_data_flag_name':['Agricultural Contamination'],        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Agricultural activity nearby'},
    '647':{'standard_data_flag_name':['Smoke'],                             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fire/wood burning nearby'},
    '646':{'standard_data_flag_name':['Traffic Contamination'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Exceptional traffic nearby'},
    '645':{'standard_data_flag_name':['Traffic Contamination'],             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Exceptional traffic nearby'},
    '599':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unspecified contamination or local influence'},
    '593':{'standard_data_flag_name':['Industrial Contamination'],          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Industrial contamination, considered invalid'},
    '591':{'standard_data_flag_name':['Agricultural Contamination'],        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Agricultural contamination, considered invalid'},
    '578':{'standard_data_flag_name':['Sea Spray'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Large sea salt contribution (ratio between marine and excess sulphate is larger than 2.0). Used for old data only. For newer data use 451/450.'},
    '568':{'standard_data_flag_name':['Dust'],                              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Dust contamination, considered invalid'},
    '567':{'standard_data_flag_name':['Insect Contamination'],              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Insect contamination, considered invalid'},
    '566':{'standard_data_flag_name':['Bird-Dropping Contamination'],       'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Bird droppings, considered invalid'},
    '565':{'standard_data_flag_name':['Pollen/Leaf Contamination'],         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Pollen and/or leaf contamination, considered invalid'},
    '559':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unspecified contamination or local influence, but considered valid'},
    '558':{'standard_data_flag_name':['Dust'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Dust contamination, but considered valid'},
    '557':{'standard_data_flag_name':['Insect Contamination'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Insect contamination, but considered valid'},
    '556':{'standard_data_flag_name':['Bird-Dropping Contamination'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Bird droppings, but considered valid'},
    '555':{'standard_data_flag_name':['Pollen/Leaf Contamination'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Pollen and/or leaf contamination, but considered valid'},
    '460':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Contamination suspected'},
    '451':{'standard_data_flag_name':['Sea Spray'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Invalid due to large sea salt contribution'},
    '450':{'standard_data_flag_name':['Sea Spray'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Considerable sea salt contribution, but considered valid'},
    '260':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Contamination suspected'},
    '251':{'standard_data_flag_name':['Sea Spray'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Invalid due to large sea salt contribution'},
    '250':{'standard_data_flag_name':['Sea Spray'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Considerable sea salt contribution, but considered valid'},
    '189':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible local contamination indicated by wind from contaminated sector (auto)'},
    '188':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible local contamination indicated by low wind speed (auto)'},
    '187':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible local contamination indicated by occurence of new particles (auto)'},
    '186':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible local contamination indicated by single scattering albedo (auto)'},
    '185':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible local contamination indicated by wind direction or velocity'},
    #Extreme/Irregular data
    '990':{'standard_data_flag_name':['Extreme External Meteorological Conditions'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Precipitation not measured due to snow-fall. Needed for historic data, should not be needed for new data'},
    '900':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Hidden and invalidated by data originator'},
    '890':{'standard_data_flag_name':['Unmeasured Key Meteorological Parameter'],                      'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Concentration in precipitation undefined, no precipitation'},
    '783':{'standard_data_flag_name':['Unmeasured Key Meteorological Parameter'],                      'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Low precipitation, concentration unknown'},
    '740':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Probably biased gas/particle ratio'},
    '701':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Less accurate than usual, unspecified reason. (Used only with old data, for new data see groups 6 and 5)'},
    '499':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistent with another unspecified measurement'},
    '472':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Less accurate than normal due to high concentration(s)'},               
    '471':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Particulate mass concentration higher than parallell mass concentration measurement with higher cut off i.e PM1_mass > PM25_mass and PM25_mass > PM10_mass. Considered invalid.'},
    '470':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Particulate mass concentration higher than parallell mass concentration measurement with higher cut off i.e PM1_mass > PM25_mass and PM25_mass > PM10_mass'},
    '459':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Extreme value, unspecified error'},
    '458':{'standard_data_flag_name':['Value Above Acceptable Range'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Extremely high value, outside four times standard deviation in a lognormal distribution'},
    '457':{'standard_data_flag_name':['Value Below Acceptable Range'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Extremely low value, outside four times standard deviation in a lognormal distribution'},
    '456':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalidated by data originator'},
    '452':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid due to large uncertainty'},
    '440':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Reconstructed or recalculated data'},
    '299':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Inconsistent with another unspecified measurement'},
    '258':{'standard_data_flag_name':['Value Above Acceptable Range'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Extremely high value, outside four times standard deviation in a log-normal distribution'},
    '257':{'standard_data_flag_name':['Value Below Acceptable Range'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Extremely low value, outside four times standard deviation in a log-normal'},
    '256':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalidated by database co-ordinator'},
    '211':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Irregular data checked and accepted by database co-ordinator. Valid measurement'},
    '210':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Episode data checked and accepted by database co-ordinator. Valid measurement'},
    '147':{'standard_data_flag_name':['Value Below Detection Limit'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Below theoretical detection limit or formal Q/A limit, but a value has been measured and reported and is considered valid'},
    '120':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sample reanalysed with similar results. Valid measurement'},
    '111':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Irregular data checked and accepted by data originator. Valid measurement'},
    '110':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Episode data checked and accepted by data originator. Valid measurement'},        
    '103':{'standard_data_flag_name':['High Variability of Data'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'CV of replicate ALPHA samplers > 15 %;. Valid measurement'},
    '102':{'standard_data_flag_name':['High Variability of Data'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'CV of replicate diffusion tubes > 30 %. Valid measurement'},
    '100':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Checked by data originator. Valid measurement, overrides any invalid flags'},
    #Exceptional Events
    '679':{'standard_data_flag_name':['Exceptional Event - Unspecified'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unspecified meteorological condition'},
    '678':{'standard_data_flag_name':['Tropical Cyclone (Cyclone/Hurricane/Typhoon)'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Hurricane'},
    '676':{'standard_data_flag_name':['Fog'],                                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'station inside cloud (visibility < 1000 m)'},
    '657':{'standard_data_flag_name':['Precipitation - Heavy'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Precipitation collector overflow. Heavy snowfall/rain shower (squall)'},
    '648':{'standard_data_flag_name':['Snow - Unspecified Intensity'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Snow sampler'},
    '410':{'standard_data_flag_name':['Dust'],                                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sahara dust event'},
    #Aggregation/Representation issues
    '395':{'standard_data_flag_name':['Data Window Completeness < 90%'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data completeness less than 90%'},
    '394':{'standard_data_flag_name':['Data Window Completeness < 90%'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data completeness less than 90%'},
    '393':{'standard_data_flag_name':['Data Window Completeness < 75%'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data completeness less than 75%'},
    '392':{'standard_data_flag_name':['Data Window Completeness < 75%'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data completeness less than 75%'},
    '391':{'standard_data_flag_name':['Data Window Completeness < 50%'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data completeness less than 50%'},
    '390':{'standard_data_flag_name':['Data Window Completeness < 50%'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data completeness less than 50%'},
    '389':{'standard_data_flag_name':['Data Window Completeness < 66%'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data completeness less than 66%'},
    '388':{'standard_data_flag_name':['Data Window Completeness < 66%'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data completeness less than 66%'},
    '382':{'standard_data_flag_name':['>= 75% of Measurements in Window Below Detection Limit'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'More than 75% of the measurements are below detection limit'},
    '380':{'standard_data_flag_name':['>= 50% of Measurements in Window Below Detection Limit'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'More than 50% of the measurements are below detection limit'},
    '370':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'For monthly averages with samples partly in two months, the number of days in each month is used for weighing the sample'},
    #Data Formatting/Processing issues
    '249':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Apparent typing error corrected. Valid measurement'},
    '248':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Illegal flag was removed by the database co-ordinator. Lost flag information (used for historic data only).'},
    '247':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Overlapping sample interval was corrected by the database co-ordinator. Possible wrong sample time (used for historic data only)'},
    '191':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data not truncation corrected - Valid measurement'}
},

#--------------------------------------------------------------
#EEA AIRBASE
#--------------------------------------------------------------
'EEA_AIRBASE':{     
    #Basic Flags
    '-1': {'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid'},
    '1':  {'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid'},
},

#--------------------------------------------------------------
#EEA AQ eReporting
#--------------------------------------------------------------
'EEA_AQ_eReporting':{
    #Basic Flags
    '1': {'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Primary data considered valid by the data provider for its related reporting purposes. Aggregated data is valid following rules set out in Directive and Guidance.'},
    '-1':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'For primary data, data is considered to be invalid due to other circumstances or data is simply missing. For aggregated data, data is not valid because of insufficient data capture following rules set out in Directive and Guidance.'},
    #Estimated data
    '2':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data considered valid by the data provider for its related reporting purposes. However, the value has been measured below the detection limit and the value reported is the detection limit (only used for primary data).'},
    '3':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data considered valid by the data provider for its related reporting purposes. However, the value has been measured below the detection limit and the value reported is half of this detection limit (only used for primary data).'},
    #Operational Maintenance checks
    '-99':{'standard_data_flag_name':['Operational Maintenance - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Data is considered to be invalid due to the regular calibration or the normal maintenance of the instrumentation (only used for primary data).'}       
},

#--------------------------------------------------------------   
#JAPAN_NIES
'JAPAN_NIES':{
    #Monthly window representativity -- artifically generated flags
    'MONTHLY_<90':{'standard_data_flag_name':['Data Window Completeness < 90%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Monthly measurement representativity < 90%'},
    'MONTHLY_<75':{'standard_data_flag_name':['Data Window Completeness < 75%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monthly measurement representativity < 75%'},
    'MONTHLY_<66':{'standard_data_flag_name':['Data Window Completeness < 66%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monthly measurement representativity < 66%'},
    'MONTHLY_<50':{'standard_data_flag_name':['Data Window Completeness < 50%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monthly measurement representativity < 50%'},
    'MONTHLY_<25':{'standard_data_flag_name':['Data Window Completeness < 25%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monthly measurement representativity < 25%'}
},

#--------------------------------------------------------------   
#MEXICO_CDMX
'MEXICO_CDMX':{},

#--------------------------------------------------------------   
#MITECO
'MITECO':{},

#--------------------------------------------------------------   
# NOAA_ISD
'NOAA_ISD':{
    '0':{'standard_data_flag_name':['Valid Data'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed gross limits check'},
    '1':{'standard_data_flag_name':['Valid Data'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed all quality control checks'},
    '2':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Suspect'},
    '3':{'standard_data_flag_name':['Invalid Data - Unspecified'],                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Erroneous'},
    '4':{'standard_data_flag_name':['Valid Data'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed gross limits check, data originate from an NCEI data source'},
    '5':{'standard_data_flag_name':['Valid Data'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed all quality control checks, data originate from an NCEI data source'},
    '6':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Suspect, data originate from an NCEI data source'},
    '7':{'standard_data_flag_name':['Invalid Data - Unspecified'],                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Erroneous, data originate from an NCEI data source'},
    '9':{'standard_data_flag_name':['Valid Data'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed gross limits check if element is present'},
    'A':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data value flagged as suspect, but accepted as a good value'},
    'C':{'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Temperature and dew point received from Automated Weather Observing System (AWOS) are reported in whole degrees Celsius. Automated QC flags these values, but they are accepted as valid.'},
    'I':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'],    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data value not originally in data, but replaced by validator'},
    'M':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Manual changes made to value based on information provided by NWS'},
    'P':{'standard_data_flag_name':['Corrected Data Formatting/Processing Issue'],    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data value not originally flagged as suspect, but replaced by validator'},
    'R':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data value replaced with value computed by NCEI software'},
    'U':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data value replaced with edited value'},
    'CHM':{'standard_data_flag_name':['Visibility Distance Unlimited'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Visibility distance exceeds 160km'},
    'VIM':{'standard_data_flag_name':['Ceiling Height Unlimited'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ceiling height exceeds 22km'}
},

# NOAA_ISD for relative humidity (as they decided to encode same numbers to different codes)
'NOAA_ISD_CH':{
    '1':{'standard_data_flag_name':['Valid Data'],                                                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Passed all quality control checks'},
    '3':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Failed all quality control checks'},
    '9':{'standard_data_flag_name':['Missing Data'],                                                 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing'}
},

# NOAA_ISD Weather Flag codes
'NOAA_ISD_WF':{
    'au':{
        'P00W0':  {'standard_data_flag_name':['No Significant Weather'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'No Precipitation, No Other Weather Reported'},
        'P01':    {'standard_data_flag_name':['Drizzle - Unspecified Intensity'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Drizzle (DZ)'},
        'P02':    {'standard_data_flag_name':['Rain - Unspecified Intensity'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Rain (RA)'}, 
        'P03':    {'standard_data_flag_name':['Snow - Unspecified Intensity'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Snow (SN)'},
        'P04':    {'standard_data_flag_name':['Snow Grains - Unspecified Intensity'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Snow Grains (SG)'},
        'P05':    {'standard_data_flag_name':['Diamond Dust - Unspecified Intensity'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ice Crystals (IC)'},
        'P06':    {'standard_data_flag_name':['Ice Pellets - Unspecified Intensity'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ice Pellets (PL)'},
        'P07':    {'standard_data_flag_name':['Hail - Unspecified Intensity'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Hail (GR)'},
        'P08':    {'standard_data_flag_name':['Snow Pellets - Unspecified Intensity'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Small Hail and/or Snow Pellets (GS)'},
        'P09':    {'standard_data_flag_name':['Precipitation - Unspecified Intensity'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unknown Precipitation (UP)'},
        'P01I1':  {'standard_data_flag_name':['Drizzle - Light'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Drizzle'},
        'P01I2':  {'standard_data_flag_name':['Drizzle - Moderate'],                            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Drizzle'},
        'P01I3':  {'standard_data_flag_name':['Drizzle - Heavy'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Drizzle'},
        'P02I1':  {'standard_data_flag_name':['Rain - Light'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Rain'},
        'P02I2':  {'standard_data_flag_name':['Rain - Moderate'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Rain'},
        'P02I3':  {'standard_data_flag_name':['Rain - Heavy'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Rain'},
        'P03I1':  {'standard_data_flag_name':['Snow - Light'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Snow'},
        'P03I2':  {'standard_data_flag_name':['Snow - Moderate'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Snow'},
        'P03I3':  {'standard_data_flag_name':['Snow - Heavy'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Snow'}, 
        'P04I1':  {'standard_data_flag_name':['Snow Grains - Light'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Snow Grains'},
        'P04I2':  {'standard_data_flag_name':['Snow Grains - Moderate'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Snow Grains'},
        'P04I3':  {'standard_data_flag_name':['Snow Grains - Heavy'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Snow Grains'}, 
        'P05I1':  {'standard_data_flag_name':['Diamond Dust - Light'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Ice Crystals'},
        'P05I2':  {'standard_data_flag_name':['Diamond Dust - Moderate'],                       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Ice Crystals'},
        'P05I3':  {'standard_data_flag_name':['Diamond Dust - Heavy'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Ice Crystals'}, 
        'P06I1':  {'standard_data_flag_name':['Ice Pellets - Light'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Ice Pellets'},
        'P06I2':  {'standard_data_flag_name':['Ice Pellets - Moderate'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Ice Pellets'},
        'P06I3':  {'standard_data_flag_name':['Ice Pellets - Heavy'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Ice Pellets'}, 
        'P07I1':  {'standard_data_flag_name':['Hail - Light'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Hail'},
        'P07I2':  {'standard_data_flag_name':['Hail - Moderate'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Hail'},
        'P07I3':  {'standard_data_flag_name':['Hail - Heavy'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Hail'},
        'P08I1':  {'standard_data_flag_name':['Snow Pellets - Light'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Small Hail and/or Snow Pellets'},
        'P08I2':  {'standard_data_flag_name':['Snow Pellets - Moderate'],                       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Small Hail and/or Snow Pellets'},
        'P08I3':  {'standard_data_flag_name':['Snow Pellets - Heavy'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Small Hail and/or Snow Pellets'},
        'P09I1':  {'standard_data_flag_name':['Precipitation - Light'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Unknown Precipitation'},
        'P09I2':  {'standard_data_flag_name':['Precipitation - Moderate'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Unknown Precipitation'},
        'P09I3':  {'standard_data_flag_name':['Precipitation - Heavy'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Unknown Precipitation'},    
        'P01D8':  {'standard_data_flag_name':['Freezing Drizzle - Unspecified Intensity'],      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Freezing Drizzle'}, 
        'P02D6':  {'standard_data_flag_name':['Rain Shower/s - Unspecified Intensity'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Rain Shower'},
        'P02D8':  {'standard_data_flag_name':['Freezing Rain - Unspecified Intensity'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Freezing Rain'},
        'P03D4':  {'standard_data_flag_name':['Drifting Snow'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Low Drifting Snow'},
        'P03D5':  {'standard_data_flag_name':['Blowing Snow'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Blowing Snow'},
        'P03D6':  {'standard_data_flag_name':['Snow Shower/s - Unspecified Intensity'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Snow Shower'},
        'P06D6':  {'standard_data_flag_name':['Ice Pellets Shower/s - Unspecified Intensity'],  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ice Pellets Shower'},
        'P07D6':  {'standard_data_flag_name':['Hail Shower/s - Unspecified Intensity'],         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Hail Shower'},
        'P08D6':  {'standard_data_flag_name':['Snow Pellets Shower/s - Unspecified Intensity'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Small Hail and/or Snow Pellets Shower'},
        'P01I1D8':{'standard_data_flag_name':['Freezing Drizzle - Light'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Freezing Drizzle'},
        'P01I2D8':{'standard_data_flag_name':['Freezing Drizzle - Moderate'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Freezing Drizzle'},
        'P01I3D8':{'standard_data_flag_name':['Freezing Drizzle - Heavy'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Freezing Drizzle'},
        'P02I1D6':{'standard_data_flag_name':['Rain Shower/s - Light'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Rain Shower'},
        'P02I1D8':{'standard_data_flag_name':['Freezing Rain - Light'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Freezing Rain'},
        'P02I2D6':{'standard_data_flag_name':['Rain Shower/s - Moderate'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Rain Shower'},
        'P02I2D8':{'standard_data_flag_name':['Freezing Rain - Moderate'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Freezing Rain'},
        'P02I3D6':{'standard_data_flag_name':['Rain Shower/s - Heavy'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Rain Shower'},
        'P02I3D8':{'standard_data_flag_name':['Freezing Rain - Heavy'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Freezing Rain'},
        'P03I1D4':{'standard_data_flag_name':['Drifting Snow'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Low Drifing Snow'},
        'P03I1D5':{'standard_data_flag_name':['Blowing Snow'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Blowing Snow'},
        'P03I1D6':{'standard_data_flag_name':['Snow Shower/s - Light'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Snow Shower'},
        'P03I2D4':{'standard_data_flag_name':['Drifting Snow'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Low Drifing Snow'},
        'P03I2D5':{'standard_data_flag_name':['Blowing Snow'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Blowing Snow'},
        'P03I2D6':{'standard_data_flag_name':['Snow Shower/s - Moderate'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Snow Shower'},
        'P03I3D4':{'standard_data_flag_name':['Drifting Snow'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Low Drifting Snow'},
        'P03I3D5':{'standard_data_flag_name':['Blowing Snow'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Blowing Snow'},
        'P03I3D6':{'standard_data_flag_name':['Snow Shower/s - Heavy'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Snow Shower'},
        'P06I1D6':{'standard_data_flag_name':['Ice Pellets Shower/s - Light'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Ice Pellets Shower'},
        'P06I2D6':{'standard_data_flag_name':['Ice Pellets Shower/s - Moderate'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Ice Pellets Shower'},
        'P06I3D6':{'standard_data_flag_name':['Ice Pellets Shower/s - Heavy'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Ice Pellets Shower'},
        'P07I1D6':{'standard_data_flag_name':['Hail Shower/s - Light'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Hail Shower'},
        'P07I2D6':{'standard_data_flag_name':['Hail Shower/s - Moderate'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Hail Shower'},
        'P07I3D6':{'standard_data_flag_name':['Hail Shower/s - Heavy'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Hail Shower'},
        'P08I1D6':{'standard_data_flag_name':['Snow Pellets Shower/s - Light'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Light Small Hail and/or Snow Pellets Shower'},
        'P08I2D6':{'standard_data_flag_name':['Snow Pellets Shower/s - Moderate'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Moderate Small Hail and/or Snow Pellets Shower'},
        'P08I3D6':{'standard_data_flag_name':['Snow Pellets Shower/s - Heavy'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Heavy Small Hail and/or Snow Pellets Shower'},
        'D7':     {'standard_data_flag_name':['Thunderstorm'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Thunderstorm (TS)'},
        'W1':     {'standard_data_flag_name':['Dust/Sand Whirls'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Well-Developed Dust/Sand Whirls (PO)'},
        'W2':     {'standard_data_flag_name':['Squalls'],                                       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Squalls (SQ)'},
        'W3':     {'standard_data_flag_name':['Funnel Cloud/s'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Funnel Cloud, Tornado, Waterspout (FC)'},
        'W4':     {'standard_data_flag_name':['Sandstorm'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sandstorm (SS)'},
        'W5':     {'standard_data_flag_name':['Duststorm'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Duststorm (DS)'},
        'O0':     {'standard_data_flag_name':['No Atmospheric Obscuration'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'No Obscuration'},
        'O1':     {'standard_data_flag_name':['Mist'],                                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Mist (BR)'},
        'O2':     {'standard_data_flag_name':['Fog'],                                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fog (FG)'},
        'O3':     {'standard_data_flag_name':['Smoke'],                                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Smoke (FU)'},
        'O4':     {'standard_data_flag_name':['Volcanic Ash'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Volcanic Ash (VA)'},
        'O5':     {'standard_data_flag_name':['Dust'],                                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Widespread Dust (DU)'},
        'O6':     {'standard_data_flag_name':['Sand'],                                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sand (SA)'},
        'O7':     {'standard_data_flag_name':['Haze'],                                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Haze (HZ)'},
        'O8':     {'standard_data_flag_name':['Sea Spray'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Spray (PY)'},
        'O2D1':   {'standard_data_flag_name':['Ground Fog'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Shallow Fog'},
        'O2D8':   {'standard_data_flag_name':['Freezing Fog'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Freezing Fog'},
        'O5D4':   {'standard_data_flag_name':['Drifting Dust'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Low Drifting Dust'},
        'O5D5':   {'standard_data_flag_name':['Blowing Dust'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Blowing Dust'},
        'O6D4':   {'standard_data_flag_name':['Drifting Sand'],                                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Low Drifting Sand'},
        'O6D5':   {'standard_data_flag_name':['Blowing Sand'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Blowing Sand'}
    },
          
    'aw_auto_atm_condition':{
        '00':{'standard_data_flag_name':['No Significant Weather'],                               'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'No significant weather observed'},
        '01':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Clouds generally dissolving or becoming less developed'},
        '02':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'State of sky on the whole unchanged during the past hour'},
        '03':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Clouds generally forming or developing during the past hour'},
        '04':{'standard_data_flag_name':['Atmospheric Obscuration - Unknown'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Haze, smoke, or dust in suspension in the air, visibility equal to or greater than 1km'},
        '05':{'standard_data_flag_name':['Smoke'],                                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Smoke'},
        '06':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '07':{'standard_data_flag_name':['Drifting Dust'],                                        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Dust or sand raised by wind at or near the station at the time of observation, but no well-developed dust irl(s) whirls(s) or sand whirl(s), and no duststorm or sandstorm seen or, in the case of ships, blowing spray at the station'},
        '08':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '09':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '10':{'standard_data_flag_name':['Mist'],                                                 'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Mist'},
        '11':{'standard_data_flag_name':['Diamond Dust - Unspecified Intensity'],                 'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Diamond dust'},
        '12':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Distant lightning'},
        '13':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '14':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '15':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '16':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '17':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '18':{'standard_data_flag_name':['Squalls'],                                              'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Squalls'},
        '19':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '20':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Fog during the preceding hour'},
        '21':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Precipitation during the preceding hour'},
        '22':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Drizzle (not freezing) or snow grains during the preceding hour'},
        '23':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Rain (not freezing) during the preceding hour'},
        '24':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Snow during the preceding hour'},
        '25':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Freezing drizzle or freezing rain during the preceding hour'},
        '26':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Thunderstorm (with or without precipitation) during the preceding hour'},
        '27':{'standard_data_flag_name':['Atmospheric Obscuration - Unknown'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Blowing or drifting snow or sand'},
        '28':{'standard_data_flag_name':['Atmospheric Obscuration - Unknown'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Blowing or drifting snow or sand, visibility equal to or greater than 1 km'},
        '29':{'standard_data_flag_name':['Atmospheric Obscuration - Unknown'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Blowing or drifting snow or sand, visibility less than 1 km'},
        '30':{'standard_data_flag_name':['Fog'],                                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog'},
        '31':{'standard_data_flag_name':['Fog'],                                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog in patches'},
        '32':{'standard_data_flag_name':['Fog'],                                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, has become thinner during the past hour'},
        '33':{'standard_data_flag_name':['Fog'],                                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, no appreciable change during the past hour'},
        '34':{'standard_data_flag_name':['Fog'],                                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, has begun or become thicker during the past hour'},
        '35':{'standard_data_flag_name':['Fog','Rime'],                                           'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Fog, depositing rime'},
        '36':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '37':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '38':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '39':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '40':{'standard_data_flag_name':['Precipitation - Unspecified Intensity'],                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Precipitation'},
        '41':{'standard_data_flag_name':['Precipitation - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Precipitation, slight or moderate'},
        '42':{'standard_data_flag_name':['Precipitation - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Precipitation, heavy'},
        '43':{'standard_data_flag_name':['Precipitation - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Liquid precipitation, slight or moderate'},
        '44':{'standard_data_flag_name':['Precipitation - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Liquid precipitation, heavy'},
        '45':{'standard_data_flag_name':['Precipitation - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Solid precipitation, slight or moderate'},
        '46':{'standard_data_flag_name':['Precipitation - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Solid precipitation, heavy'},
        '47':{'standard_data_flag_name':['Precipitation - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Freezing precipitation, slight or moderate'},
        '48':{'standard_data_flag_name':['Precipitation - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Freezing precipitation, heavy'},
        '49':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '50':{'standard_data_flag_name':['Drizzle - Unspecified Intensity'],                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle'},
        '51':{'standard_data_flag_name':['Drizzle - Light'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, slight'},
        '52':{'standard_data_flag_name':['Drizzle - Moderate'],                                   'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, moderate'},
        '53':{'standard_data_flag_name':['Drizzle - Heavy'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, heavy'},
        '54':{'standard_data_flag_name':['Freezing Drizzle - Light'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, freezing, slight'},
        '55':{'standard_data_flag_name':['Freezing Drizzle - Moderate'],                          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, freezing, moderate'},
        '56':{'standard_data_flag_name':['Freezing Drizzle - Heavy'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, freezing, heavy'},
        '57':{'standard_data_flag_name':['Drizzle - Light','Rain - Light'],                       'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Drizzle and rain, slight'},
        '58':{'standard_data_flag_name':['Drizzle - Heavy','Rain - Heavy'],                       'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Drizzle and rain, moderate or heavy'},
        '59':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '60':{'standard_data_flag_name':['Rain - Unspecified Intensity'],                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain'},
        '61':{'standard_data_flag_name':['Rain - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, slight'},
        '62':{'standard_data_flag_name':['Rain - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, moderate'},
        '63':{'standard_data_flag_name':['Rain - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, heavy'},
        '64':{'standard_data_flag_name':['Freezing Rain - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, freezing, slight'},
        '65':{'standard_data_flag_name':['Freezing Rain - Moderate'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, freezing, moderate'},
        '66':{'standard_data_flag_name':['Freezing Rain - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, freezing, heavy'},
        '67':{'standard_data_flag_name':['Rain - Light','Snow - Light'],                          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Rain or drizzle and snow, slight'},
        '68':{'standard_data_flag_name':['Rain - Heavy','Snow - Heavy'],                          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Rain or drizzle and snow, moderate or heavy'},
        '69':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '70':{'standard_data_flag_name':['Snow - Unspecified Intensity'],                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow'},
        '71':{'standard_data_flag_name':['Snow - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow, slight'},
        '72':{'standard_data_flag_name':['Snow - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow, moderate'},
        '73':{'standard_data_flag_name':['Snow - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow, heavy'},
        '74':{'standard_data_flag_name':['Ice Pellets - Light'],                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Ice pellets, slight'},
        '75':{'standard_data_flag_name':['Ice Pellets - Moderate'],                               'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Ice pellets, moderate'},
        '76':{'standard_data_flag_name':['Ice Pellets - Heavy'],                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Ice pellets, heavy'},
        '77':{'standard_data_flag_name':['Snow Grains - Unspecified Intensity'],                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow grains'},
        '78':{'standard_data_flag_name':['Diamond Dust - Unspecified Intensity'],                 'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Ice crystals'},
        '79':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '80':{'standard_data_flag_name':['Rain Shower/s - Unspecified Intensity'],                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Showers or intermittent precipitation'},
        '81':{'standard_data_flag_name':['Rain Shower/s - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain showers or intermittent rain, slight'},
        '82':{'standard_data_flag_name':['Rain Shower/s - Moderate'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain showers or intermittent rain, moderate'},
        '83':{'standard_data_flag_name':['Rain Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain showers or intermittent rain, heavy'},
        '84':{'standard_data_flag_name':['Rain Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain showers or intermittent rain, violent'},
        '85':{'standard_data_flag_name':['Snow Shower/s - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow showers or intermittent snow, slight'},
        '86':{'standard_data_flag_name':['Snow Shower/s - Moderate'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow showers or intermittent snow, moderate'},
        '87':{'standard_data_flag_name':['Snow Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow showers or intermittent snow, heavy'},
        '88':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '89':{'standard_data_flag_name':['Hail - Unspecified Intensity'],                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Hail'},
        '90':{'standard_data_flag_name':['Thunderstorm'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Thunderstorm'},
        '91':{'standard_data_flag_name':['Thunderstorm'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Thunderstorm, slight or moderate, with no precipitation'},
        '92':{'standard_data_flag_name':['Thunderstorm','Rain Shower/s - Unspecified Intensity'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, slight or moderate, with rain showers and/or snow showers'},
        '93':{'standard_data_flag_name':['Thunderstorm','Hail - Unspecified Intensity'],          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, slight or moderate, with hail'},
        '94':{'standard_data_flag_name':['Thunderstorm'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Thunderstorm, heavy, with no precipitation'},
        '95':{'standard_data_flag_name':['Thunderstorm','Rain Shower/s - Unspecified Intensity'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, heavy, with rain showers and/or snow'},
        '96':{'standard_data_flag_name':['Thunderstorm','Hail - Unspecified Intensity'],          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, heavy, with hail'},
        '97':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '98':{'standard_data_flag_name':[''],                                                     'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description': 'Empty (Reserved) Flag Code'},
        '99':{'standard_data_flag_name':['Funnel Cloud/s'],                                       'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Tornado'}
    },

    'mw_man_atm_condition':{
        #No precipitation, fog, ice fog (except for 11 and 12), duststorm, sandstorm, drifting or blowing snow at the station at the time of observation or, except for 09 and 17, during the preceding hour.
        '00':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Cloud development not observed or not observable'},
        '01':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Clouds generally dissolving or becoming less developed'},
        '02':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'State of sky on the whole unchanged'},
        '03':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Clouds generally forming or developing'},
        '04':{'standard_data_flag_name':['Smoke'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Visibility reduced by smoke, e.g. veldt or forest fires, industrial smoke or volcanic ashes'},
        '05':{'standard_data_flag_name':['Haze'],             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Haze'},
        '06':{'standard_data_flag_name':['Dust'],             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Widespread dust in suspension in the air, not raised by wind at or near the station at the time of observation'},
        '07':{'standard_data_flag_name':['Drifting Dust'],    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Dust or sand raised by wind at or near the station at the time of observation, but no well-developed dust whirl(s) sand whirl(s), and no duststorm or sandstorm seen or, in the case of ships, blowing spray at the station'},
        '08':{'standard_data_flag_name':['Dust/Sand Whirls'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Well developed dust whirl(s) or sand whirl(s) seen at or near the station during the preceding hour or at the time of observation, but no duststorm or sandstorm'},
        '09':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Duststorm or sandstorm within sight at the time of observation, or at the station during the preceding hour'},
        '10':{'standard_data_flag_name':['Mist'],             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Mist'},
        '11':{'standard_data_flag_name':['Ground Fog'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Patches of shallow fog or ice fog at the station, whether on land or sea, not deeper than about 2 meters on land or 10 meters at sea'},
        '12':{'standard_data_flag_name':['Ground Fog'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'More or less continuous shallow fog or ice fog at the station, whether on land or sea, not deeper than about 2 meters on land or 10 meters at sea'},
        '13':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Lightning visible, no thunder heard'},
        '14':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Precipitation within sight, not reaching the ground or the surface of the sea'},
        '15':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Precipitation within sight, reaching the ground or the surface of the sea, but distant, i.e., estimated to be more than 5 km from the station'},
        '16':{'standard_data_flag_name':[''],                 'network_decreed_validity':[''],  'GHOST_decreed_validity':[''],  'network_provided_flag_description': 'Precipitation within sight, reaching the ground or the surface of the sea, near to, but not at the station'},
        '17':{'standard_data_flag_name':['Thunderstorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Thunderstorm, but no precipitation at the time of observation'},
        '18':{'standard_data_flag_name':['Squalls'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Squalls at or within sight of the station during the preceding hour or at the time of observation'},
        '19':{'standard_data_flag_name':['Funnel Cloud/s'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Funnel cloud(s) (Tornado cloud or waterspout) at or within sight of the station during the preceding hour or at the time of observation'},
        #Precipitation, fog, ice fog or thunderstorm at the station during the preceding hour, but not at the time of observation
        '20':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Drizzle (not freezing) or snow grains not falling as shower(s)'},
        '21':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Rain (not freezing) not falling as shower(s)'},
        '22':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Snow not falling as shower(s)'},
        '23':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Rain and snow or ice pellets not falling as shower(s)'},
        '24':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Freezing drizzle or freezing rain not falling as shower(s)'},
        '25':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Shower(s) of rain'},
        '26':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Shower(s) of snow or of rain and snow'},
        '27':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Shower(s) of hail (Hail, small hail, snow pellets), or rain and hail'},
        '28':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Fog or ice fog'},
        '29':{'standard_data_flag_name':[''], 'network_decreed_validity':[''], 'GHOST_decreed_validity':[''], 'network_provided_flag_description': 'Thunderstorm (with or without precipitation)'},
        #Dust, sand, or blowing snow in the air, but no precipitation at the time of observation.
        '30':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Slight or moderate duststorm or sandstorm has decreased during the preceding hour'},
        '31':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Slight or moderate duststorm or sandstorm no appreciable change during the preceding hour'},
        '32':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Slight or moderate duststorm or sandstorm has begun or has increased during the preceding hour'},
        '33':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Severe duststorm or sandstorm has decreased during the preceding hour'},
        '34':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Severe duststorm or sandstorm no appreciable change during the preceding hour'},
        '35':{'standard_data_flag_name':['Duststorm'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Severe duststorm or sandstorm has begun or has increased during the preceding hour'},
        '36':{'standard_data_flag_name':['Drifting Snow'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Slight or moderate drifting snow generally low (below eye level)'},
        '37':{'standard_data_flag_name':['Drifting Snow'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Heavy drifting snow generally low (below eye level)'},
        '38':{'standard_data_flag_name':['Blowing Snow'],  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Slight or moderate blowing snow generally high (above eye level)'},
        '39':{'standard_data_flag_name':['Blowing Snow'],  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description': 'Heavy blowing snow generally high (above eye level)'},
        #Fog or ice fog at the time of observation
        '40':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog at a distance at the time of observation, but not at the station during the preceding hour, the fog or ice fog extending to a level above that of the observer'},
        '41':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog in patches'},
        '42':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky visible, has become thinner during the preceding hour'},
        '43':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky invisible, has become thinner during the preceding hour'},
        '44':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky visible, no appreciable change during the preceding hour'},
        '45':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky invisible, no appreciable change during the preceding hour'},
        '46':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky visible, has begun or has become thicker during the preceding hour'},
        '47':{'standard_data_flag_name':['Fog'],        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Fog or ice fog, sky invisible, has begun or has become thicker during the preceding hour'},
        '48':{'standard_data_flag_name':['Fog','Rime'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Fog, depositing rime, sky visible'},
        '49':{'standard_data_flag_name':['Fog','Rime'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Fog, depositing rime, sky invisible'},
        #Precipitation at the station at the time of observation – including Drizzle, Rain, Solid Precipitation, and Precipitation with current or recent Thunder
        '50':{'standard_data_flag_name':['Drizzle - Light'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, intermittent, slight at time of observation'},
        '51':{'standard_data_flag_name':['Drizzle - Light'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, continuous, slight at time of observation'},
        '52':{'standard_data_flag_name':['Drizzle - Moderate'],                                   'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, intermittent, moderate at time of observation'},
        '53':{'standard_data_flag_name':['Drizzle - Moderate'],                                   'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, continuous, moderate at time of observation'},
        '54':{'standard_data_flag_name':['Drizzle - Heavy'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, intermittent, heavy (dense) at time of observation'},
        '55':{'standard_data_flag_name':['Drizzle - Heavy'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, not freezing, continuous, heavy (dense) at time of observation'},
        '56':{'standard_data_flag_name':['Freezing Drizzle - Light'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, freezing, slight'},
        '57':{'standard_data_flag_name':['Freezing Drizzle - Heavy'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Drizzle, freezing, moderate or heavy (dense)'},
        '58':{'standard_data_flag_name':['Drizzle - Light','Rain - Light'],                       'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Drizzle and rain, slight'},
        '59':{'standard_data_flag_name':['Drizzle - Heavy','Rain - Heavy'],                       'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Drizzle and rain, moderate or heavy'},
        '60':{'standard_data_flag_name':['Rain - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, intermittent, slight at time of observation'},
        '61':{'standard_data_flag_name':['Rain - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, continuous, slight at time of observation'},
        '62':{'standard_data_flag_name':['Rain - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, intermittent, moderate at time of observation'},
        '63':{'standard_data_flag_name':['Rain - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, continuous, moderate at time of observation'},
        '64':{'standard_data_flag_name':['Rain - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, intermittent, heavy at time of observation'},
        '65':{'standard_data_flag_name':['Rain - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, not freezing, continuous, heavy at time of observation'},
        '66':{'standard_data_flag_name':['Freezing Rain - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, freezing, slight'},
        '67':{'standard_data_flag_name':['Freezing Rain - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain, freezing, moderate or heavy'},
        '68':{'standard_data_flag_name':['Rain - Light','Snow - Light'],                          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Rain or drizzle and snow, slight'},
        '69':{'standard_data_flag_name':['Rain - Heavy','Snow - Heavy'],                          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Rain or drizzle and snow, moderate or heavy'},
        '70':{'standard_data_flag_name':['Snow - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Intermittent fall of snowflakes, slight at time of observation'},
        '71':{'standard_data_flag_name':['Snow - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Continuous fall of snowflakes, slight at time of observation'},
        '72':{'standard_data_flag_name':['Snow - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Intermittent fall of snowflakes, moderate at time of observation'},
        '73':{'standard_data_flag_name':['Snow - Moderate'],                                      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Continuous fall of snowflakes, moderate at time of observation'},
        '74':{'standard_data_flag_name':['Snow - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Intermittent fall of snowflakes, heavy at time of observation'},
        '75':{'standard_data_flag_name':['Snow - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Continuous fall of snowflakes, heavy at time of observation'},
        '76':{'standard_data_flag_name':['Diamond Dust - Unspecified Intensity'],                 'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Diamond dust (with or without fog)'},
        '77':{'standard_data_flag_name':['Snow Grains - Unspecified Intensity'],                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow grains (with or without fog)'},
        '78':{'standard_data_flag_name':['Snow - Unspecified Intensity'],                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Isolated star-like snow crystals (with or without fog)'},
        '79':{'standard_data_flag_name':['Ice Pellets - Unspecified Intensity'],                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Ice pellets'},
        '80':{'standard_data_flag_name':['Rain Shower/s - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain shower(s), slight'},
        '81':{'standard_data_flag_name':['Rain Shower/s - Moderate'],                             'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain shower(s), moderate or heavy'},
        '82':{'standard_data_flag_name':['Rain Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Rain shower(s), violent'},
        '83':{'standard_data_flag_name':['Rain Shower/s - Light','Snow Shower/s - Light'],        'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Shower(s) of rain and snow mixed, slight'},
        '84':{'standard_data_flag_name':['Rain Shower/s - Heavy','Snow Shower/s - Heavy'],        'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Shower(s) of rain and snow mixed, moderate or heavy'},
        '85':{'standard_data_flag_name':['Snow Shower/s - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Show shower(s), slight'},
        '86':{'standard_data_flag_name':['Snow Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Snow shower(s), moderate or heavy'},
        '87':{'standard_data_flag_name':['Snow Pellets Shower/s - Light'],                        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Shower(s) of snow pellets or small hail, with or without rain or rain and snow mixed, slight'},
        '88':{'standard_data_flag_name':['Snow Pellets Shower/s - Heavy'],                        'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Shower(s) of snow pellets or small hail, with or without rain or rain and snow mixed, moderate or heavy'},
        '89':{'standard_data_flag_name':['Hail Shower/s - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Shower(s) of hail (hail, small hail, snow pellets), with or without rain or rain and snow mixed, not associated with thunder, slight'},
        '90':{'standard_data_flag_name':['Hail Shower/s - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Shower(s) of hail (hail, small hail, snow pellets), with or without rain or rain and snow mixed, not associated with thunder, moderate or heavy'},
        '91':{'standard_data_flag_name':['Rain - Light'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Slight rain at time of observation, thunderstorm during the preceding hour but not at time of observation'},
        '92':{'standard_data_flag_name':['Rain - Heavy'],                                         'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Moderate or heavy rain at time of observation, thunderstorm during the preceding hour but not at time of observation'},
        '93':{'standard_data_flag_name':['Precipitation - Light'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Slight snow, or rain and snow mixed or hail (Hail, small hail, snow pellets), at time of observation, thunderstorm during the preceding hour but not at time of observation'},
        '94':{'standard_data_flag_name':['Precipitation - Heavy'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description': 'Moderate or heavy snow, or rain and snow mixed or hail(Hail, small hail, snow pellets) at time of observation, thunderstorm during the preceding hour but not at time of observation'},
        '95':{'standard_data_flag_name':['Thunderstorm','Precipitation - Unspecified Intensity'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, slight or moderate, without hail (Hail, small hail, snow pellets), but with rain and/or snow at time of observation, thunderstorm at time of observation'},
        '96':{'standard_data_flag_name':['Thunderstorm','Hail - Unspecified Intensity'],          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, slight or moderate, with hail (hail, small hail, snow pellets) at time of observation, thunderstorm at time of observation'},
        '97':{'standard_data_flag_name':['Thunderstorm','Precipitation - Unspecified Intensity'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, heavy, without hail (Hail, small hail, snow pellets), but with rain and/or snow at time of observation, thunderstorm at time of observation'},
        '98':{'standard_data_flag_name':['Thunderstorm','Duststorm'],                             'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm combined with duststorm or sandstorm at time of observation, thunderstorm at time of observation'},
        '99':{'standard_data_flag_name':['Thunderstorm','Hail - Unspecified Intensity'],          'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description': 'Thunderstorm, heavy, with hail (Hail, small hail, snow pellets) at time of observation, thunderstorm at time of observation'}
    }
},

#--------------------------------------------------------------
#SEARCH
#--------------------------------------------------------------
'SEARCH':{
    'A':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Available'},
    'C1':{},
    'D':{},
    'GZ':{},
    'I':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid'},
    'N':{'standard_data_flag_name':['Missing Data'],               'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Not Available'},
    'R':{},
    'S1':{},
    'S2':{},
    'v0':{},
    'v1':{},
    'v2':{},
    'Z':{}
},

#--------------------------------------------------------------
#US EPA AirNow DOS
#--------------------------------------------------------------

'US_EPA_AirNow_DOS':{
    #Basic Flags
    'Valid':            {'standard_data_flag_name':['Valid Data'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid'},
    'Invalid':          {'standard_data_flag_name':['Invalid Data - Unspecified'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid'},
    'Missing':          {'standard_data_flag_name':['Missing Data'],                                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing'},
    'Suspect':          {'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Suspect'},
    'Insufficient Data':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Insufficient Data'}
},

#--------------------------------------------------------------
#US EPA AQS
#--------------------------------------------------------------
'US_EPA_AQS':{
    #Basic Flags
    '':{    'standard_data_flag_name':['Valid Data'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid value'},
    'T':{   'standard_data_flag_name':['Valid Data'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Multiple PM2.5 Validity Flags (Not Active)'},
    'XX':{  'standard_data_flag_name':['Preliminary Data'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Experimental Data (Legacy=9950)'},
    '9950':{'standard_data_flag_name':['Preliminary Data'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Experimental Data'},
    '8':{   'standard_data_flag_name':['Un-Flagged Data'],  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'QA/QC Unknown (Not Active)'},  
    'AF':{  'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Scheduled but not Collected (Legacy=9972)'},
    '9972':{'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Scheduled but not Collected'},
    'BG':{  'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing ozone data not likely to exceed level of standard (Legacy=9966)'},
    '9966':{'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing ozone data not likely to exceed level of standard'},
    'AM': { 'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Miscellaneous Void (Legacy=9979)'},
    '9979':{'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Miscellaneous Void'},
    'AU':{  'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monitoring Waived (Legacy=9987)'},
    '9987':{'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Monitoring Waived'},
    'AQ':{  'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Collection Error (Legacy=9983)'},
    '9983':{'standard_data_flag_name':['Missing Data'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Collection Error'},
    #Estimated data
    'ND':{'standard_data_flag_name':['Estimated Data - No Value Detected'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'No Value Detected, Zero Reported'},
    '9': {'standard_data_flag_name':['Estimated Data - Measured Negative Value'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Negative value detected - zero reported'},
    'MS':{'standard_data_flag_name':['Estimated Data - Value Below Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value reported is 1/2 MDL substituted'},
    'EH':{'standard_data_flag_name':['Estimated Data - Value Above Detection Limit'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Estimated; Exceeds Upper Range'},
    'SS':{'standard_data_flag_name':['Estimated Data - Value Substituted from Secondary Monitor'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Value substituted from secondary monitor'},
    'DN':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'DNPH peak less than NATTS TAD requirement, reported value should be considered an estimate'},
    'LJ':{'standard_data_flag_name':['Estimated Data - Unspecified'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Identification Of Analyte Is Acceptable; Reported Value Is An Estimate'},
    #Measurement issues
    '3':{   'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Field Issue'},
    '4':{   'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Lab Issue'},
    '7':{   'standard_data_flag_name':['Operational Maintenance Check Issue'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Below Lowest Calibration Level'},
    'FX':{  'standard_data_flag_name':['Physical Issue With Filter'],                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter Integrity Issue'},
    'AV':{  'standard_data_flag_name':['Power Failure'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Power Failure (Legacy=9988)'},
    '9988':{'standard_data_flag_name':['Power Failure'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Power Failure'},
    'AN':{  'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Machine Malfunction (Legacy=9980)'},
    '9980':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Machine Malfunction'},
    'AJ':{  'standard_data_flag_name':['Physical Issue With Filter'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Damage (Legacy=9976)'},
    '9976':{'standard_data_flag_name':['Physical Issue With Filter'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Damage'},
    'AK':{  'standard_data_flag_name':['Physical Issue With Filter'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Leak (Legacy=9977)'},
    '9977':{'standard_data_flag_name':['Physical Issue With Filter'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Leak'},
    'AR':{  'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Lab Error (Legacy=9984)'},
    '9984':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Lab Error'},
    'AA':{  'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Pressure out of Limits (Legacy=9967)'},
    '9967':{'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Pressure out of Limits'},
    'HT':{  'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sample pick-up hold time exceeded'},
    'AH':{  'standard_data_flag_name':['Invalid Flow Rate'],                              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Flow Rate or CV out of Limits (Legacy=9974)'},
    '9974':{'standard_data_flag_name':['Invalid Flow Rate'],                              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Flow Rate or CV out of Limits'},
    'SC':{  'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sampler Contamination (Legacy=9948)'},
    '9948':{'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sampler Contamination'},
    'QP':{  'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Pressure Sensor Questionable'},
    'QT':{  'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Temperature Sensor Questionable'},
    'W':{   'standard_data_flag_name':['Invalid Flow Rate'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Flow Rate Average out of Spec.'},
    'X':{   'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter Temperature Difference or Average out of Spec.'},
    'BK':{  'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Site computer/data logger down (Legacy=9962)'},
    '9962':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Site computer/data logger down'},
    'AB':{  'standard_data_flag_name':['No Technician'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Technician Unavailable (Legacy=9968)'},
    '9968':{'standard_data_flag_name':['No Technician'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Technician Unavailable'},
    'BB':{  'standard_data_flag_name':['No Technician'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Unable to Reach Site (Legacy=9994)'},
    '9994':{'standard_data_flag_name':['No Technician'],                                  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Unable to Reach Site'},
    'AE':{  'standard_data_flag_name':['Extreme External Meteorological Conditions'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Shelter Temperature Outside Limits (Legacy=9971)'},
    '9971':{'standard_data_flag_name':['Extreme External Meteorological Conditions'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Shelter Temperature Outside Limits'},
    'TS':{  'standard_data_flag_name':['Extreme Sample Transport Conditions'],            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Holding Time Or Transport Temperature Is Out Of Specs. (Legacy=9956)'},
    '9956':{'standard_data_flag_name':['Extreme Sample Transport Conditions'],            'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Holding Time Or Transport Temperature Is Out Of Specs.'},
    'BJ':{  'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Operator Error (Legacy=9963)'},
    '9963':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Operator Error'},
    'AG':{  'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Time out of Limits (Legacy=9973)'},
    '9973':{'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Time out of Limits'},
    'AI':{  'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Insufficient Data (cannot calculate) (Legacy=9975)'},
    '9975':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Insufficient Data (cannot calculate)'},
    'AP':{  'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Vandalism (Legacy=9982)'},
    '9982':{'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Vandalism'},
    'BH':{  'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Interference/co-elution/misidentification (Legacy=9965)'},
    '9965':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Interference/co-elution/misidentification'},
    'MC':{  'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Module End Cap Missing (Legacy=9958)'},
    '9958':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Module End Cap Missing'},
    'BI':{  'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Lost or damaged in transit (Legacy=9964)'},
    '9964':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Lost or damaged in transit'},
    'SV':{  'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Volume out of limits'},
    'DI':{  'standard_data_flag_name':['Sample Diluted for Analysis'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sample was diluted for analysis'},
    'FB':{  'standard_data_flag_name':['Operational Maintenance Check Issue'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Field Blank Value Above Acceptable Limit'},
    'LB':{  'standard_data_flag_name':['Operational Maintenance Check Issue'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Lab blank value above acceptable limit'},
    'TB':{  'standard_data_flag_name':['Operational Maintenance Check Issue'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Trip Blank Value Above Acceptable Limit'},
    'TT':{  'standard_data_flag_name':['Extreme Sample Transport Conditions'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Transport Temperaure is Out of Specs.'},
    'Y':{   'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Elapsed Sample Time out of Spec.'},
    'MX':{  'standard_data_flag_name':['Matrix Effect'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Matrix Effect'},
    '2':{   'standard_data_flag_name':['Measurement Issue - Unspecified'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Operational Deviation'},
    #Operational Maintenance checks
    'AT':{  'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Calibration (Legacy=9986)'},
    '9986':{'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Calibration'},
    'AY':{  'standard_data_flag_name':['Zero Check'],              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Q C Control Points (zero/span) (Legacy=9991)'},
    '9991':{'standard_data_flag_name':['Zero Check'],              'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Q C Control Points (zero/span)'},
    'AX':{  'standard_data_flag_name':['Precision Check'],         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Precision Check (Legacy=9990)'},
    '9990':{'standard_data_flag_name':['Precision Check'],         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Precision Check'},
    'AZ':{  'standard_data_flag_name':['Quality Control Audit'],   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Q C Audit (Legacy=9992)'},
    '9992':{'standard_data_flag_name':['Quality Control Audit'],   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Q C Audit'},
    'BA':{  'standard_data_flag_name':['Instrumental Repair'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Maintenance/Routine Repairs (Legacy=9993)'},
    '9993':{'standard_data_flag_name':['Instrumental Repair'],     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Maintenance/Routine Repairs'},
    'BC':{  'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Multi-point Calibration (Legacy=9995)'},
    '9995':{'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Multi-point Calibration'},
    'BD':{  'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Auto Calibration (Legacy=9996)'},
    '9996':{'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Auto Calibration'},
    'BF':{  'standard_data_flag_name':['Precision Check'],         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Precision/Zero/Span (Legacy=9998)'},
    '9998':{'standard_data_flag_name':['Precision Check'],         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Precision/Zero/Span'},
    'BL':{  'standard_data_flag_name':['Quality Control Audit'],   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'QA Audit (Legacy=9961)'},
    '9961':{'standard_data_flag_name':['Quality Control Audit'],   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'QA Audit'},
    'BM':{  'standard_data_flag_name':['Accuracy Check'],          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Accuracy check (Legacy=9960)'},
    '9960':{'standard_data_flag_name':['Accuracy Check'],          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Accuracy check'},
    'CS':{  'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Laboratory Calibration Standard (Legacy=9953)'},
    '9953':{'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Laboratory Calibration Standard'},
    'ST':{  'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Calibration Verification Standard (Legacy=9954)'},
    '9954':{'standard_data_flag_name':['Calibration'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Calibration Verification Standard'},
    'TC':{  'standard_data_flag_name':['Retention Time Check'],    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Component Check & Retention Time Standard (Legacy=9952)'},
    '9952':{'standard_data_flag_name':['Retention Time Check'],    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Component Check & Retention Time Standard'},
    'FI':{  'standard_data_flag_name':['Instrumental Inspection'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Inspection Flag (Legacy=9959)'},
    '9959':{'standard_data_flag_name':['Instrumental Inspection'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Filter Inspection Flag'},
    'DL':{  'standard_data_flag_name':['Detection Limits Check'],  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Detection Limit Analyses (Legacy=9949)'},
    '9949':{'standard_data_flag_name':['Detection Limits Check'],  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Detection Limit Analyses'},
    'MB':{  'standard_data_flag_name':['Blank Check'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Method Blank (Analytical) (Legacy=9955)'},
    '9955':{'standard_data_flag_name':['Blank Check'],             'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Method Blank (Analytical)'},
    'QV':{  'standard_data_flag_name':['Quality Control Audit'],   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Quality Control Multi-point Verification'},
    #Local Contamination
    'RF':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fire - Canadian '},
    'IF':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fire - Canadian'},
    'RG':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fire - Mexico/Central America '},
    'IG':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fire - Mexico/Central America'},
    'RM':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Prescribed Fire'},
    'IM':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Prescribed Fire'},
    'Q':{   'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Prescribed Burning (Not Active)'},
    'RP':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Structural Fire'},
    'IP':{  'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Structural Fire'},
    'F':{   'standard_data_flag_name':['Smoke'],                                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Structural Fire (Not Active)'},
    'J':{   'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Construction'},
    'L':{   'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Highway Construction (Not Active)'},
    'P':{   'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Roofing Operations (Not Active)'},
    'D':{   'standard_data_flag_name':['Industrial Contamination'],                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sandblasting (Not Active)'},
    'N':{   'standard_data_flag_name':['Industrial Contamination'],                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sanding/Salting of Streets (Not Active)'},
    'RI':{  'standard_data_flag_name':['Pollen/Leaf Contamination'],                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'High Pollen Count '},
    'II':{  'standard_data_flag_name':['Pollen/Leaf Contamination'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'High Pollen Count '},
    'G':{   'standard_data_flag_name':['Pollen/Leaf Contamination'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'High Pollen Count (Not Active)'},
    'RR':{  'standard_data_flag_name':['Traffic Contamination'],                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unique Traffic Disruption'},
    'IR':{  'standard_data_flag_name':['Traffic Contamination'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unique Traffic Disruption'},
    'I':{   'standard_data_flag_name':['Traffic Contamination'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unusual Traffic Congestion (Not Active)'},
    'M':{   'standard_data_flag_name':['Traffic Contamination'],                        'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Rerouting of Traffic (Not Active)'},
    'K':{   'standard_data_flag_name':['Agricultural Contamination'],                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Agricultural Tilling (Not Active)'},
    'AC':{  'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Construction/Repairs in Area (Legacy=9969)'},
    '9969':{'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Construction/Repairs in Area'},
    'NS':{  'standard_data_flag_name':['Local Contamination - Unspecified'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Influenced by nearby source'},
    'BE':{  'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Building/Site Repair (Legacy=9997)'},
    '9997':{'standard_data_flag_name':['Construction Contamination'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Building/Site Repair'},
    'CC':{  'standard_data_flag_name':['Internal Laboratory/Instrument Contamination'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Clean Canister Residue'},
     #Extreme/Irregular data
    'MD':{  'standard_data_flag_name':['Value Below Detection Limit'],                                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Value less than MDL'},
    'BR':{  'standard_data_flag_name':['Value Below Acceptable Range'],                                 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Value Below Acceptable Range'},
    'PQ':{  'standard_data_flag_name':['Value Close to Detection Limit'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Values Between PQL And MDL (Not Active)'},
    'SQ':{  'standard_data_flag_name':['Value Close to Detection Limit'],                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Values Between SQL and MDL '},
    '1':{   'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Deviation from a CFR/Critical Criteria Requirement'},
    '1V':{  'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data reviewed and validated'},
    '5':{   'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Outlier'},
    'LK':{  'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Analyte Identified; Reported Value May Be Biased High'},
    'LL':{  'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Analyte Identified; Reported Value May Be Biased Low'},
    'QX':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Does not meet QC criteria'},
    '6':{   'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'QAPP Issue'},
    'V':{   'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Validated Value'},
    '1C':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'A 1-Point QC check exceeds acceptance criteria but there is compelling evidence that the analyzer data is valid.'},
    'VB':{  'standard_data_flag_name':['Irregular Data Manually Screened and Accepted'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Value below normal; no reason to invalidate'},
    'EC':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Exceeds Critical Criteria'},
    'CL':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Surrogate Recoveries Outside Control Limits'},
    'CB':{  'standard_data_flag_name':['Reconstructed/Recalculated Data'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Values have been Blank Corrected'},
    'SX':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Does Not Meet Siting Criteria'},
    'AS':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Poor Quality Assurance Results (Legacy=9985)'},
    '9985':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Poor Quality Assurance Results'},
    'AL':{  'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Voided by Operator (Legacy=9978)'},
    '9978':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Voided by Operator'},
    'DA':{  'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Aberrant Data (Corrupt Files, Aberrant Chromatography, Spikes, Shifts) (Legacy=9951)'},
    '9951':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Aberrant Data (Corrupt Files, Aberrant Chromatography, Spikes, Shifts)'},
    'BN':{  'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Value Exceeds Media Limit (Legacy=9947)'},
    '9947':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample Value Exceeds Media Limit'},
    #Exceptional Events
    'RA':{  'standard_data_flag_name':['Dust'],                                'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'African Dust'},
    'IA':{  'standard_data_flag_name':['Dust'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'African Dust'},
    'U':{   'standard_data_flag_name':['Dust'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Sahara Dust (Not Active)'},
    'RB':{  'standard_data_flag_name':['Dust'],                                'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Asian Dust'},
    'IB':{  'standard_data_flag_name':['Dust'],                                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Asian Dust'},
    'RC':{  'standard_data_flag_name':['Chemical Spill/Industrial Accident'],  'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Chem. Spills & Indust. Accidents'},
    'IC':{  'standard_data_flag_name':['Chemical Spill/Industrial Accident'],  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Chem. Spills & Indust Accidents'},
    'H':{   'standard_data_flag_name':['Chemical Spill/Industrial Accident'],  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Chem. Spills & Indust. Accidents (Not Active)'},
    'RD':{  'standard_data_flag_name':['Cleanup After a Major Disaster'],      'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Cleanup After a Major Disaster'},
    'ID':{  'standard_data_flag_name':['Cleanup After a Major Disaster'],      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Cleanup After a Major Disaster'},
    'R':{   'standard_data_flag_name':['Cleanup After a Major Disaster'],      'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Cleanup After a Major Disaster (Not Active)'},
    'RE':{  'standard_data_flag_name':['Demolition'],                          'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Demolition'},
    'IE':{  'standard_data_flag_name':['Demolition'],                          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Demolition'},
    'RH':{  'standard_data_flag_name':['Fireworks'],                           'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Fireworks'},
    'IH':{  'standard_data_flag_name':['Fireworks'],                           'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Fireworks'},
    'RJ':{  'standard_data_flag_name':['High Winds'],                          'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'High Winds'},
    'IJ':{  'standard_data_flag_name':['High Winds'],                          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'High Winds'},
    'A':{   'standard_data_flag_name':['High Winds'],                          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'High Winds (Not Active)'},
    'RK':{  'standard_data_flag_name':['Infrequent Large Gathering'],          'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Infrequent Large Gatherings'},
    'IK':{  'standard_data_flag_name':['Infrequent Large Gathering'],          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Infrequent Large Gatherings'},
    'O':{   'standard_data_flag_name':['Infrequent Large Gathering'],          'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Infrequent Large Gatherings (Not Active)'},
    'RL':{  'standard_data_flag_name':['Exceptional Event - Unspecified'],     'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Other'},
    'IL':{  'standard_data_flag_name':['Exceptional Event - Unspecified'],     'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Other'},
    'Z':{   'standard_data_flag_name':['Exceptional Event - Unspecified'],     'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Other event (Not Active)'},
    'RN':{  'standard_data_flag_name':['Seismic Activity'],                    'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Seismic Activity'},
    'IN':{  'standard_data_flag_name':['Seismic Activity'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Seismic Activity'},
    'S':{   'standard_data_flag_name':['Seismic Activity'],                    'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Seismic Activity (Not Active)'},
    'RO':{  'standard_data_flag_name':['Stratospheric Ozone Intrusion'],       'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Stratospheric Ozone Intrusion'},
    'IO':{  'standard_data_flag_name':['Stratospheric Ozone Intrusion'],       'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Stratospheric Ozone Intrusion'},
    'B':{   'standard_data_flag_name':['Stratospheric Ozone Intrusion'],       'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Stratospheric Ozone Intrusion (Not Active)'},
    'RQ':{  'standard_data_flag_name':['Terrorist Act'],                       'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Terrorist Act'},
    'IQ':{  'standard_data_flag_name':['Terrorist Act'],                       'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Terrorist Act'},
    'RS':{  'standard_data_flag_name':['Volcanic Eruptions'],                  'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Volcanic Eruptions'},
    'IS':{  'standard_data_flag_name':['Volcanic Eruptions'],                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Volcanic Eruptions'},
    'C':{   'standard_data_flag_name':['Volcanic Eruptions'],                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Volcanic Eruption (Not Active)'},
    'RT':{  'standard_data_flag_name':['Wildfire','Smoke'],                    'network_decreed_validity':['I','I'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Wildfire-USA'},
    'IT':{  'standard_data_flag_name':['Wildfire','Smoke'],                    'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Wildfire-USA'},
    'E':{   'standard_data_flag_name':['Wildfire','Smoke'],                    'network_decreed_validity':['I','I'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Forest Fire (Not Active)'},
    'RU':{  'standard_data_flag_name':['Wildfire','Smoke'],                    'network_decreed_validity':['I','I'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Wildland Fire Use Fire-USA (Not Active)'},
    'IU':{  'standard_data_flag_name':['Wildfire','Smoke'],                    'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Wildland Fire Use Fire-USA (Not Active)'},
    'AO':{  'standard_data_flag_name':['Thunderstorm'],                        'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Bad Weather (Legacy=9981)'},
    '9981':{'standard_data_flag_name':['Thunderstorm'],                        'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Bad Weather'},
    'AD':{  'standard_data_flag_name':['Thunderstorm'],                        'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Shelter Storm Damage (Legacy=9970)'},
    '9970':{'standard_data_flag_name':['Thunderstorm'],                        'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Shelter Storm Damage'},
    'AW':{  'standard_data_flag_name':['Wildfire'],                            'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Wildlife Damage (Legacy=9989)'},
    '9989':{'standard_data_flag_name':['Wildfire'],                            'network_decreed_validity':['I'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Wildlife Damage'},
    'SA':{  'standard_data_flag_name':[''],                                    'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description':'Storm Approaching (Legacy=9957)'},
    '9957':{'standard_data_flag_name':[''],                                    'network_decreed_validity':[''],      'GHOST_decreed_validity':[''],      'network_provided_flag_description':'Storm Approaching'},
    #Daily window representativity -- manually generated flags
    'DAILY_<90':{'standard_data_flag_name':['Data Window Completeness < 90%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Daily measurement representativity < 90%'},
    'DAILY_<75':{'standard_data_flag_name':['Data Window Completeness < 75%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Daily measurement representativity < 75%'},
    'DAILY_<66':{'standard_data_flag_name':['Data Window Completeness < 66%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Daily measurement representativity < 66%'},
    'DAILY_<50':{'standard_data_flag_name':['Data Window Completeness < 50%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Daily measurement representativity < 50%'},
    'DAILY_<25':{'standard_data_flag_name':['Data Window Completeness < 25%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Daily measurement representativity < 25%'}
},

#--------------------------------------------------------------
#UK AIR
#--------------------------------------------------------------
'UK_AIR':{
    'R': {'standard_data_flag_name':['Valid Data'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ratified'},
    'V': {'standard_data_flag_name':['Valid Data'],                           'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Verified'},
    'P': {'standard_data_flag_name':['Preliminary Data'],                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Provisional'},
    'P*':{'standard_data_flag_name':['Un-Flagged Data'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'As Supplied'},   
    'N': {'standard_data_flag_name':['Un-Flagged Data'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Not Verified'},
    'S': {'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Suspect'},
    '<': {'standard_data_flag_name':['Value Below Detection Limit'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Below Detection Limit'},       
    '>': {'standard_data_flag_name':['Value Above Detection Limit'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Above Detection Limit'},    
    '':  {'standard_data_flag_name':['Un-Flagged Data'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Not Flagged'}
},

#--------------------------------------------------------------
#US EPA CASTNET
#--------------------------------------------------------------
'US_EPA_CASTNET':{
    'O3':{
        #Basic flags
        '': {'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data'}, 
        'M':{'standard_data_flag_name':['Missing Data'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field data missing'},
        #Measurement issues
        'F':{'standard_data_flag_name':['Power Failure'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument power failure'},
        'H':{'standard_data_flag_name':['Measurement Issue - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid reading - related to instrument malfunction'},
        'K':{'standard_data_flag_name':['Invalid Flow Rate'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Non-nominal flow value identified during automated screening'},
        #Operational Maintenance checks
        'C':{'standard_data_flag_name':['Calibration'],           'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data collected during field calibration of instrument'},
        'T':{'standard_data_flag_name':['Quality Control Audit'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Third Party Audit'},
        #Extreme/Irregular data
        'P':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Potential problem identified during automated screening requiring further investigation'},
        'R':{'standard_data_flag_name':['Irregular Data Manually Screened and Rejected'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Indicates that the QC check is valid but that associated ambient measurements are not valid and the check should not be submitted as a 1-point QC check to AQS.'},
        '^':{'standard_data_flag_name':['Valid Data'],                                                   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Alarm high -- ozone > 120'},
        '_':{'standard_data_flag_name':['Negative Value'],                                               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field reading less than zero'},
        'S':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],                              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Field instrument out of DQO at time of 2nd calibration'},
        'I':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                         'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid reading - Suspect data'},
        'J':{'standard_data_flag_name':['Data Does Not Meet Internal Network Quality Control Criteria'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid reading - related to critical criteria failures'},
        'Y':{'standard_data_flag_name':['Span Check'],                                                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'QC control points (zero/span)'}, 
        #Exceptional Events
        'E':{'standard_data_flag_name':['Exceptional Event - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Exceptional event designation requested by state agency'},
        #Aggregation/Representation issues
        '<':{'standard_data_flag_name':['Data Window Completeness < 90%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Missing < 25% of hourly sampling period'},
        'B':{'standard_data_flag_name':['Data Window Completeness < 25%'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument channel down for > 75% of hourly sampling period'},
        'D':{'standard_data_flag_name':['Data Window Completeness < 75%'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument channel down for 25% to 75% of hourly sampling period'}
    },
    'hourly_gas':{
        #Basic flags
        '': {'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data'},
        'I':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid reading'},    
        'M':{'standard_data_flag_name':['Missing Data'],               'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field data missing'},  
        #Measurement issues
        'F':{'standard_data_flag_name':['Power Failure'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument power failure'},    
        'N':{'standard_data_flag_name':['Erroneous Sampling Operation'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sampling time inadequate to calculate hourly average'},    
        #Operational Maintenance checks
        'C':{'standard_data_flag_name':['Calibration'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data due to field calibration of instrument'},    
        #Extreme/Irregular data
        'S':{'standard_data_flag_name':['Reconstructed/Recalculated Data'],                              'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Field instrument out of DQO at time of 2nd calibration'},    
        'U':{'standard_data_flag_name':['Value Below Detection Limit'],                                  'network_decreed_validity':['V'],     'GHOST_decreed_validity':['I'],     'network_provided_flag_description':'Concentration below reporting limit for analyzer'},    
        'Z':{'standard_data_flag_name':['Value Below Detection Limit','Data Window Completeness < 90%'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['I','V'], 'network_provided_flag_description':'< and U apply'},    
        #Local Contamination
        'A':{'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Anomalous value resulting from local anthropogenic activity'},    
        #Aggregation/Representation issues
        '<':{'standard_data_flag_name':['Data Window Completeness < 90%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Missing < 25% of hourly sampling period'},
        'B':{'standard_data_flag_name':['Data Window Completeness < 75%'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument channel down for >= 25% of hourly sampling period'}, 
        'D':{'standard_data_flag_name':['Data Window Completeness < 75%'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Field instrument channel down for 25% to 75% of hourly sampling period'} 
    },
    'drychem':{
        #Basic flags
        '':  {'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data'}, 
        '12':{'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack on tower for less than 6 or greater than 8 days'},
        '17':{'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter given to EPA for analysis'},
        '20':{'standard_data_flag_name':['Valid Data'],   'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter used for special study'},
        '22':{'standard_data_flag_name':['Missing Data'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Site closed'},        
        #Estimated data
        'U': {'standard_data_flag_name':['Estimated Data - No Value Detected'],                     'network_decreed_validity':['V'],     'GHOST_decreed_validity':['I'],     'network_provided_flag_description':'Undetected - value listed is the reporting limit corrected by flow volume'},
        '16':{'standard_data_flag_name':['Estimated Data - Unspecified'],                           'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'On/off dates and times are assumed'},
        '!': {'standard_data_flag_name':['Estimated Data - No Value Detected','Chemical Issue'],    'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['I','V'], 'network_provided_flag_description':'X and U apply'},
        '#': {'standard_data_flag_name':['Estimated Data - No Value Detected','Invalid Flow Rate'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['I','V'], 'network_provided_flag_description':'Both U and L flags apply'},
        '$': {'standard_data_flag_name':['Estimated Data - No Value Detected','Invalid Flow Rate'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['I','V'], 'network_provided_flag_description':'L and U apply'},
        #Measurement issues
        'I': {'standard_data_flag_name':['Invalid Flow Rate'],                          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid chemistry data and/or less than 75% valid flow data'},
        'L': {'standard_data_flag_name':['Invalid Flow Rate'],                          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Less than 90% but greater than or equal to 75% valid flow data'},
        'M': {'standard_data_flag_name':['Invalid Flow Rate'],                          'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Missing or completely invalid flow data'},
        'N': {'standard_data_flag_name':['Sample Not Analysed'],                        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample not analyzed'},
        'X': {'standard_data_flag_name':['Chemical Issue'],                             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Data suspect due to impregnation solution'},
        '2': {'standard_data_flag_name':['Physical Issue With Filter'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Torn;hole;ripped filter noted during unpacking'},
        '3': {'standard_data_flag_name':['Physical Issue With Filter'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Excessively wet filter noted upon unpacking'},
        '4': {'standard_data_flag_name':['Physical Issue With Filter'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Excessively dirty filter noted upon unpacking'},
        '5': {'standard_data_flag_name':['Physical Issue With Filter'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack loose upon arrival, possible leakage during sample period'},
        '6': {'standard_data_flag_name':['Mechanical Issue/Non-Operational Equipment'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Apparent solenoid problem'},
        '7': {'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack end caps cracked/missing upon receipt'},
        '8': {'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Outside of filter pack excessively dirty upon receipt'},
        '9': {'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Support screen raised up, noted during unpacking'},
        '11':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Laboratory accident'},
        '13':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack not run in field'},
        '14':{'standard_data_flag_name':['Measurement Issue - Unspecified'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unusual odor noticed during unpacking'},
        '15':{'standard_data_flag_name':['Physical Issue With Filter'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Low extraction efficiency off filter'},
        '18':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Field accident or error'},
        '19':{'standard_data_flag_name':['Measurement Issue - Unspecified'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Field equipment problem'},
        '23':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sample report form not received with filter pack'},
        '24':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Chain of custody incomplete or incorrect'},
        '25':{'standard_data_flag_name':['Human Error'],                                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack run out of sequence'},
        '28':{'standard_data_flag_name':['Chemical Issue'],                             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'SO2 concentration adjusted due to positive bias introduced by faulty cellulose filter impregnation solution'},
        #Extreme/Irregular data
        'R': {'standard_data_flag_name':['Reconstructed/Recalculated Data'],                     'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Re-run value used in calculation'},
        'S': {'standard_data_flag_name':['Reconstructed/Recalculated Data','Invalid Flow Rate'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['V','V'], 'network_provided_flag_description':'Both L and R flags apply'},
        '26':{'standard_data_flag_name':['Extreme/Irregular Data - Unspecified'],                'network_decreed_validity':['V'],     'GHOST_decreed_validity':['V'],     'network_provided_flag_description':'Suspect value, but reason not recorded'},
        #Local Contamination
        'A': {'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Anomalous value resulting from local anthropogenic activity'},
        '&': {'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Ambient concentration data measured/collected coincident with data flagged A'},
        '1': {'standard_data_flag_name':['Local Contamination - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unidentified debris/particles on filter'},
        '10':{'standard_data_flag_name':['Insect Contamination'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Insect inside filter pack; noted during unpacking'},
        '21':{'standard_data_flag_name':['Smoke'],                             'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Forest fire/agricultural activity in area'},
        #Aggregation/Representation issues
        '<':{'standard_data_flag_name':['Data Window Completeness < 90%'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Missing < 25% of hourly sampling period'}
    },
    'vischem':{
        #Basic flags
        '': {'standard_data_flag_name':['Valid Data'],                  'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data'},
        'I':{'standard_data_flag_name':['Invalid Data - Unspecified'],  'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data'},
        #Measurement issues
        'N': {'standard_data_flag_name':['Sample Not Analysed'],        'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Sample not analyzed'},
        'S': {'standard_data_flag_name':['Invalid Flow Rate'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Suspect value due to incorrect flow rate through cyclone'},
        '2': {'standard_data_flag_name':['Physical Issue With Filter'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter torn, ripped, hole noted upon unpacking'},
        '3': {'standard_data_flag_name':['Physical Issue With Filter'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Excessively wet filter noted upon unpacking'},
        '4': {'standard_data_flag_name':['Physical Issue With Filter'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Excessively dirty on outside of filter/denuder sample'},
        '5': {'standard_data_flag_name':['Physical Issue With Filter'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack, denuder connections loose upon unpacking'},
        '6': {'standard_data_flag_name':['Physical Issue With Filter'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Denuder broken upon receipt'},
        '7': {'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Shipping endcaps missing/damaged upon receipt'},
        '8': {'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Petri dish cracked/opened upon unpacking'},
        '9': {'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter/filter pack dropped in field'},
        '10':{'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter/filter pack dropped in laboratory'},
        '11':{'standard_data_flag_name':['Invalid Flow Rate'],          'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Possible flow problem'},
        '12':{'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Run data incorrect, not scheduled'},
        '13':{'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Filter pack not run in field'},
        '20':{'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Visibility Site Report Form not received'},
        '21':{'standard_data_flag_name':['Human Error'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Chain of custody incomplete or incorrect'},
        #Extreme/Irregular data
        'U':{'standard_data_flag_name':['Value Below Detection Limit'],                     'network_decreed_validity':['V'],     'GHOST_decreed_validity':['I'],     'network_provided_flag_description':'Value is less than detection limit'},
        '#':{'standard_data_flag_name':['Value Below Detection Limit','Invalid Flow Rate'], 'network_decreed_validity':['V','V'], 'GHOST_decreed_validity':['I','V'], 'network_provided_flag_description':'Both U and S flags apply'},
        #Local Contamination
        '1': {'standard_data_flag_name':['Insect Contamination'],       'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Unidentified debris or insects on filter'},
        '17':{'standard_data_flag_name':['Smoke'],                      'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Forest fire/heavy smoke during sampling'},
        '18':{'standard_data_flag_name':['Construction Contamination'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Construction/agricultural activity during sampling'},
        '19':{'standard_data_flag_name':['Thunderstorm'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Severe storms in area during sampling'}
    }
},

#--------------------------------------------------------------
#US NADP AMNet
#--------------------------------------------------------------
'US_NADP_AMNet':{
    #A Codes = Fully qualified with no problems
    'A':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fully qualified with no problems'}, 
    #B codes = Valid data with minor problems
    'B':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data with minor issues'}, 
    #C codes = Invalid data
    'C':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data'}
},

#--------------------------------------------------------------
#US NADP AMoN
#--------------------------------------------------------------
'US_NADP_AMoN':{
    #A codes = Fully qualified with no problems
    'A':{'standard_data_flag_name':['Valid Data'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Fully qualified with no problems'}, 
    #B codes = Valid data with minor problems        
    'B':{'standard_data_flag_name':['Valid Data'],                                     'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data with minor issues'}, 
    'c':{'standard_data_flag_name':['Local Contamination - Unspecified'],              'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Local source of ammonia (within ~500m), based on site operator comments'}, 
    'd':{'standard_data_flag_name':['Value Below Detection Limit'],                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Below laboratory detection limit'}, 
    'e':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Long sample time (> 360 hours)'}, 
    'h':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Sample handling issue in field, shipping, or laboratory'}, 
    'm':{'standard_data_flag_name':['Data Formatting/Processing Issue'],               'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Missing data from field operator'}, 
    'q':{'standard_data_flag_name':['Measurement Issue - Unspecified'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Minor quality control issue'}, 
    't':{'standard_data_flag_name':['Extreme Sample Transport Conditions'],            'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Elevated travel blank concentration'},
    'y':{'standard_data_flag_name':['Measurement Issue - Unspecified'],                'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Delayed sample processing (>16 days from date off to analysis date)'}, 
    #C codes = Invalid data                                    
    'C':{'standard_data_flag_name':['Invalid Data - Unspecified'],                     'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data'}, 
    'a':{'standard_data_flag_name':['Missing Data'],                                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Laboratory analytical data missing'}, 
    'f':{'standard_data_flag_name':['Erroneous Sampling Operation'],                   'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Major field sampling issue (gross contamination, vandalism, etc.)'}, 
    'l':{'standard_data_flag_name':['Measurement Issue - Unspecified'],                'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Major laboratory issue'}, 
    'n':{'standard_data_flag_name':['Human Error'],                                    'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'No sampler deployed'}, 
    #Miscellaneous codes
    's':{'standard_data_flag_name':['Aggregation/Representation Issue - Unspecified'], 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Short sample time (< 312 hours)'} 
},

#--------------------------------------------------------------
#WMO_WDCGG
#--------------------------------------------------------------
'WMO_WDCGG':{
    '1':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid (background) data considered "background'}, 
    '2':{'standard_data_flag_name':['Valid Data'],                 'network_decreed_validity':['V'], 'GHOST_decreed_validity':['V'], 'network_provided_flag_description':'Valid data considered valid other than "background'}, 
    '3':{'standard_data_flag_name':['Invalid Data - Unspecified'], 'network_decreed_validity':['I'], 'GHOST_decreed_validity':['I'], 'network_provided_flag_description':'Invalid data considered unfit for use (questionable/erroneous/absent data)'} 
}}

###--------------------------------------------------------------------------------------------------###
#DEFINE DATA PROVIDER MEASUREMENT INFORMATION TO STANDARD MEASUREMENT INFORMATION MAPPING
###--------------------------------------------------------------------------------------------------###

#define mapping of data provider measurement information to standard measurement information, per network, per species
#two types of dictionaries are defined:
# -- mapping --> dictionary mapping all network provided measurement information to standardised measurement information
# -- standard_measurement_information --> dictionary of pre-standardised measurement information

#any information provided by the networks which improves upon the default standardised measurement information is defined

def network_measurement_information_to_standard_measurement_information_mapping(raw_measurement_info, run_details, parameter_details, metadata_stamp_dt=''):

    from time_functions import fractional_minutes_from_reference_to_datetime

#--------------------------------------------------------------
#AERONET
#--------------------------------------------------------------
    if 'AERONET' in run_details['network']:
        #convert fractional minutes to datetime
        metadata_stamp_dt = fractional_minutes_from_reference_to_datetime(metadata_stamp_dt)
        #direct measurement mode?
        #oneill
        if raw_measurement_info == 'oneill':
            if metadata_stamp_dt.year < 2004:
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-1'}}
            elif(metadata_stamp_dt.year >= 2004) & (metadata_stamp_dt.year < 2014): 
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-N'}}
            elif metadata_stamp_dt.year >= 2014:  
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-T'}}
        #directsun
        elif raw_measurement_info == 'directsun':
            if metadata_stamp_dt.year < 2004:
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-1'}}
            elif(metadata_stamp_dt.year >= 2004) & (metadata_stamp_dt.year < 2014): 
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-N'}}
            elif metadata_stamp_dt.year >= 2014:  
                standard_measurement_information = {'measurement':{'methodology':'photometry - direct', 'instrument':'Cimel CE318-T'}}
        #sky measurement mode?
        elif raw_measurement_info == 'almucantar':
            if metadata_stamp_dt.year < 2004:
                standard_measurement_information = {'measurement':{'methodology':'photometry - sky', 'instrument':'Cimel CE318-1'}}
            elif(metadata_stamp_dt.year >= 2004) & (metadata_stamp_dt.year < 2014): 
                standard_measurement_information = {'measurement':{'methodology':'photometry - sky', 'instrument':'Cimel CE318-N'}}
            elif metadata_stamp_dt.year >= 2014:  
                standard_measurement_information = {'measurement':{'methodology':'photometry - sky', 'instrument':'Cimel CE318-T'}} 

#--------------------------------------------------------------
#BJMEMC
#--------------------------------------------------------------
    elif run_details['network'] == 'BJMEMC':

        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'process_details':'Make assumption method is ultraviolet photometry', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'NO2':
            standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'process_details':'Make assumption method is chemiluminescence (internal molybdenum converter)', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'SO2':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence', 'process_details':'Make assumption method is ultraviolet fluorescence', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'process_details':'Make assumption method is NDIR (GFC)', 'measuring_assumption':True}}
        elif run_details['parameter'] in ['PM10','PM2.5']:
             standard_measurement_information = {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry', 'process_details':'Make assumption method is TEOM', 'measuring_assumption':True}}
            
#--------------------------------------------------------------
#CANADA NAPS
#--------------------------------------------------------------
    elif run_details['network'] == 'CANADA_NAPS':
        
        #continuous measurements
        if raw_measurement_info[:4] == 'con_':
            raw_measurement_info = raw_measurement_info[4:].strip('_')
            if run_details['parameter'] == 'O3':
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
            elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
                standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}}
            elif run_details['parameter'] == 'CO':
                standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}
            elif run_details['parameter'] == 'SO2':
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence'}}
            elif run_details['parameter'] == 'PM10':
                standard_measurement_information = {'measurement':{'methodology':'beta-attenuation', 'process_details':'Make assumption method is beta-attenuation', 'measuring_assumption':True}}   
            elif run_details['parameter'] == 'PM2.5':
                mapping = {
                'code170':{'measurement':{'methodology':'beta-attenuation',                                                                              'instrument':'Met One BAM1020'}},
                'code181':{'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',    'instrument':'Thermo 1405-DF'}},
                'code184':{'measurement':{'methodology':'beta-attenuation - nephelometry',                                                               'instrument':'Thermo 5030 SHARP'}},
                'code195':{'measurement':{'methodology':'optical particle counter',                                                                      'instrument':'Grimm 180'}},
                'code239':{'measurement':{'methodology':'optical particle counter',                                                                      'instrument':'Teledyne 640x'}}, 
                'code703':{'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
                'code706':{'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
                'code760':{'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}} 
        #integrated measurements
        else:
            raw_measurement_info = raw_measurement_info.strip('_')
            if run_details['parameter'] == 'SO2':
                mapping = {
                'ion_chromatography':                                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',   'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                'met_one_supersass-plus_sequential_speciation_sampler_cartc_medna2co3_-_coated_denuder_(to_collect_gaseous_hno3,_so2)_prepwater_extraction_ion_chromatograph':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Met One SUPER SASS'},                                        'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}},
                'partisol_2300_speciation_cartc_medna2co3_-_coated_denuder_(to_collect_gaseous_hno3,_so2)_prepwater_extraction_ion_chromatograph':                            {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                      'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}}}
            elif run_details['parameter'] == 'NH3':
                mapping = {
                'ion_chromatography':                                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',   'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                'met_one_supersass-plus_sequential_speciation_sampler_cartc_medcitric_acid-coated_denuder_(to_collect_gaseous_nh3)_prepwater_extraction_ion_chromatograph':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Met One SUPER SASS'},                                        'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}},   
                'partisol_2300_speciation_cartc_medcitric_acid-coated_denuder_(to_collect_gaseous_nh3)_prepwater_extraction_ion_chromatograph':                            {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                      'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}}} 
            elif run_details['parameter'] == 'HNO3':
                mapping = {
                'ion_chromatography':                                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',   'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},   
                'met_one_supersass-plus_sequential_speciation_sampler_cartc_medna2co3_-_coated_denuder_(to_collect_gaseous_hno3,_so2)_prepwater_extraction_ion_chromatograph':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Met One SUPER SASS'},                                        'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}},   
                'partisol_2300_speciation_cartc_medna2co3_-_coated_denuder_(to_collect_gaseous_hno3,_so2)_prepwater_extraction_ion_chromatograph':                            {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                      'preparation_1':{'sample_preparation_type':'denuder'},                                 'measurement':{'methodology':'ion chromatography'}}} 
            elif run_details['parameter'] in ['ETHENE','ETHANE']:
                mapping = {
                '910_canister_sampler_(xonteck)_medsumma_canister_prep_gas_chromatograph_with_mass_spectrometry_detection_(all_analytes_except_ethane,_ethylene_and_acetylene);gas_chromatograph_with_flame_ionization_detection_(ethane,_ethylen':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 910'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},            
                '910a_canister_sampler_(xontech,_inc)_medsumma_canister_prepnone_gas_chromatograph_with_flame_ionization_detection':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 910'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},        
                'lvs_summacanister_gcfid':                                                                                                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                             'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}}}
            elif run_details['parameter'] in ['CH4','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE']:
                mapping = {
                '910_canister_sampler_(xonteck)_medsumma_canister_prep_gas_chromatograph_with_mass_spectrometry_detection_(all_analytes_except_ethane,_ethylene_and_acetylene);gas_chromatograph_with_flame_ionization_detection_(ethane,_ethylen':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 910'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},            
                '910a_canister_sampler_(xontech,_inc)_medsumma_canister_prepnone_gas_chromatograph_with_mass_spectrometry_detection':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 910'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},            
                '910a_canister_sampler_(xontech,_inc)_medsumma_canister_prepnone_gas_chromatograph_with_flame_ionization_detection':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 910'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
                'lvs_summacanister_gcfid':                                                                                                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                             'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}}}
            elif run_details['parameter'] == 'HCHO':
                mapping = {
                '926_two_channel_carbonyl_sampler_(xontech,_inc)_med2,4-dinitrophenylhydrazine_(dnph)_coated_silica_gel_cartridge_preporganic_solvent_extract_high_performance_liquid_chromatograph':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 926'}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}},
                '926_two_channel_carbonyl_sampler_(xonteck)_meddnph_hplc':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Xonteck 926'}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}}}
            elif run_details['parameter'] in ['PM10','PM2.5']:
                mapping = {
                'dichotomous_manual_air_sampler_(2000-d_partisol_thermo)_code177':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                  'measurement':{'methodology':'gravimetry'}}, 
                'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo);pm2.5_manual_air_sampler_(te-wilbur_tisch)_code177;219':                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                 'measurement':{'methodology':'gravimetry'}},
                'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':{'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000i'},                                                                                                                                 'measurement':{'methodology':'gravimetry'}},            
                'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_code177':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000i'},                                                                                                                                 'measurement':{'methodology':'gravimetry'}},               
                'dichotomous_sequential_air_sampler_(2025-d_thermo);dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_code179;177':                           {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                 'measurement':{'methodology':'gravimetry'}},
                'dichotomous_sequential_air_sampler_(2025-d_thermo);pm2.5_manual_air_sampler_(te-wilbur_tisch)_code179;219':                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                 'measurement':{'methodology':'gravimetry'}},        
                'dichotomous_sequential_air_sampler_(2025-d_thermo)_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                  'measurement':{'methodology':'gravimetry'}},     
                'gravimetry':                                                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',  'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
                'partisolpm2.5_gravimetry':                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                 'measurement':{'methodology':'gravimetry'}},
                'pm2.5_manual_air_sampler_(2000_partisol_thermo)_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                  'measurement':{'methodology':'gravimetry'}}, 
                'pm2.5_manual_air_sampler_(2000_partisol_thermo)_code143':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                  'measurement':{'methodology':'gravimetry'}},       
                'pm2.5_manual_air_sampler_(te-wilbur_tisch)_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Tisch Environmental Wilbur'},                                                                                                                            'measurement':{'methodology':'gravimetry'}},         
                'pm2.5_manual_air_sampler_(te-wilbur_tisch)_code219':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Tisch Environmental Wilbur'},                                                                                                                            'measurement':{'methodology':'gravimetry'}}, 
                'pm2.5_sequential_air_sampler_(2025_thermo)_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                  'measurement':{'methodology':'gravimetry'}},              
                's-1_cartn_a_medteflon_membrane_filter_preptemperature_and_humidity_equilibration_microbalance':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                 'measurement':{'methodology':'gravimetry'}}}            
            elif parameter_details['matrix'] == 'pm':   
                if run_details['parameter'] in ['C','EC','OC']:
                    mapping = {
                    'niosh_thermal-optical_analysis':  {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040'}},
                    'improve_thermal-optical_analysis':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}},
                    'thermal-optical_analysis':        {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}}} 
                else:
                    mapping = {
                    'ion_chromatography':                       {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                    'prepdenuder_ion_chromatography':           {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                             'measurement':{'methodology':'ion chromatography'}},                                                                                                                                                      
                    'prepteflon_filter_pack_ion_chromatography':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}}}
            elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
                if run_details['parameter'] in ['PM10_C','PM10_EC','PM10_OC','PM2.5_C','PM2.5_EC','PM2.5_OC']:
                    mapping = {
                    'met_one_supersass-plus_sequential_speciation_sampler_carta,b_medquartz_filter_prepnone_tor(oc_ec)':{'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SUPER SASS'},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}}, 
                    'met_one_supersass-plus_sequential_speciation_sampler_carta_medquartz_filter_prepnone_tor(oc_ec)':  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SUPER SASS'},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}}, 
                    'partisol_2300_speciation_carta,b_medquartz_filter_prepnone_tor(oc_ec)':                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}},
                    'partisol_2300_speciation_carta_medquartz_filter_prepnone_tor(oc_ec)':                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}}}
                else:
                    mapping = {
                    'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000i'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},          
                    'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000i'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
                    'dichotomous_manual_air_sampler_(2000i-d_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepacid_digestion_icpms':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000i'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},          
                    'dichotomous_sequential_air_sampler_(2025-d_thermo)_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},                           
                    'dichotomous_sequential_air_sampler_(2025-d_thermo)_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                          {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},               
                    'dichotomous_sequential_air_sampler_(2025-d_thermo)_cartn_a_medteflon_membrane_filter_prepacid_digestion_icpms':                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},                             
                    'icpms':                                                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},                  
                    'ion_chromatography':                                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                    'met_one_supersass-plus_sequential_speciation_sampler_cartc_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                          {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Met One SUPER SASS'},                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
                    'partisol_2300_speciation_cartc_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                                                      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},            
                    'partisolpm2.5_ion_chromatography':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                    'pm2.5_manual_air_sampler_(2000_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                             {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
                    'pm2.5_manual_air_sampler_(2000_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},               
                    'pm2.5_manual_air_sampler_(2000_partisol_thermo)_cartn_a_medteflon_membrane_filter_prepacid_digestion_icpms':                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2000'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},                            
                    'pm2.5_manual_air_sampler_(te-wilbur_tisch)_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Tisch Environmental Wilbur'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},           
                    'pm2.5_manual_air_sampler_(te-wilbur_tisch)_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':              {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Tisch Environmental Wilbur'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},                           
                    'pm2.5_sequential_air_sampler_(2025_thermo)_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
                    'pm2.5_sequential_air_sampler_(2025_thermo)_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':              {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},                 
                    'pm2.5_sequential_air_sampler_(2025_thermo)_cartn_a_medteflon_membrane_filter_prepacid_digestion_icpms':                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2025'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},                
                    's-1_cartn_a_medteflon_membrane_filter_prepwater_extraction_ion_chromatograph':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
                    's-1_cartn_a_medteflon_membrane_filter_prepnone_energy_dispersive_x-ray_fluorescence_spectrometer':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
                    's-1_cartn_a_medteflon_membrane_filter_prepacid_digestion_icpms':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}}
                    

#--------------------------------------------------------------
#CAPMoN
#--------------------------------------------------------------
    elif run_details['network'] == 'CAPMoN':
        if run_details['parameter'] == 'O3':
            mapping = {             
            'dasibi':    {'measurement':{'methodology':'ultraviolet photometry',                            'reported_uncertainty':'0.6'}},
            'thermo_49': {'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49',  'reported_uncertainty':'0.6'}},
            'thermo_49c':{'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49C', 'reported_uncertainty':'0.6'}},
            'thermo_49i':{'measurement':{'methodology':'ultraviolet photometry', 'instrument':'Thermo 49i', 'reported_uncertainty':'0.6'}}}
#-------------------------------------------
        else:
            mapping = {
            'high_volume_icp-es':              {'primary_sampling':{'sampling_type':'high volume continuous'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'high_volume_icp-ms':              {'primary_sampling':{'sampling_type':'high volume continuous'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},    
            'tisch_environmental_6070v_icp-ms':{'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Tisch Environmental 6070V'}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}}
 
#--------------------------------------------------------------
#CHILE SINCA
#--------------------------------------------------------------
    elif run_details['network'] == 'CHILE_SINCA':
        if run_details['parameter'] == 'O3':
            mapping = {  
            '':                {'measurement':{'methodology':'unknown'}},
            'ecl':             {'measurement':{'methodology':'ethylene chemiluminescence'}},
            'environnement42m':{'measurement':{'methodology':'ultraviolet photometry',       'instrument':'Environnement SA O342M'}},
            'teledynet400':    {'measurement':{'methodology':'ultraviolet photometry',       'instrument':'Teledyne T400'}},
            'thermo49i':       {'measurement':{'methodology':'ultraviolet photometry',       'instrument':'Thermo 49i'}},
            'uvp':             {'measurement':{'methodology':'ultraviolet photometry'}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):        
            mapping = {
            '':                  {'measurement':{'methodology':'unknown'}},
            'cl':                {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                          'process_details':'Assume conversion using molybdenum converter.', 'measuring_assumption':True}},      
            'environnementac32m':{'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Environnement SA AC32M'}},
            'thermo42i':         {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Thermo 42i'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'CO':
            mapping = {
            '':            {'measurement':{'methodology':'unknown'}},
            'ndir':        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                   'process_details':'Assume gas-filter correlation', 'measuring_assumption':True}},        
            'ndir-gfc':    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'teledynet300':{'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'thermo48i':   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}}}
#-------------------------------------------
        elif run_details['parameter'] in ['CH4','NMHC']:
            mapping = {
            '':   {                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'fid':{'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection', 'process_details':'Method stated purely as FID, but FID cannot solely distinguish CH4, therefore assume uses GC beforehand.', 'measuring_assumption':True}},
            'gc': {'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - unknown detection'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HC':
            mapping = {
            '':   {                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'fid':{                                                                                                                                                                                       'measurement':{'methodology':'flame ionisation detection'}},
            'gc': {'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - unknown detection'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '':            {'measurement':{'methodology':'unknown'}},
            'teledynet100':{'measurement':{'methodology':'ultraviolet fluorescence',   'instrument':'Teledyne T100'}},        
            'thermo43i':   {'measurement':{'methodology':'ultraviolet fluorescence',   'instrument':'Thermo 43i'}},              
            'uvf':         {'measurement':{'methodology':'ultraviolet fluorescence'}}} 
#-------------------------------------------
        elif run_details['parameter'] in ['PM10','PM2.5']:
            mapping = {
            '':                   {                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'ba':                 {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation'}},
            'environnementmp101m':{                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Environnement SA MP101M'}}, 
            'gr':                 {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'}, 'measurement':{'methodology':'gravimetry'}},
            'gr-auto':            {                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'hivol-gr':           {'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'filter'}, 'measurement':{'methodology':'gravimetry'}},
            'metone1020':         {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Met One BAM1020'}},
            'opc':                {                                                                                                                                                        'measurement':{'methodology':'optical particle counter'}},
            'thermo1400ab':       {                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',   'instrument':'Thermo 1400ab'}},  
            'thermo5014i':        {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo 5014i'}},
            'thermofh62c-14':     {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo FH62 C14'}}}
#-------------------------------------------
        elif run_details['parameter'] in ['PM10_As','PM10_Cu','PM10_Pb']: 
            mapping = {
            '':   {                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}},
            'aas':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy', 'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}}}

#--------------------------------------------------------------
#CNEMC
#--------------------------------------------------------------
    elif run_details['network'] == 'CNEMC':

        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry', 'process_details':'Make assumption method is ultraviolet photometry', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'NO2':
            standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'process_details':'Make assumption method is chemiluminescence (internal molybdenum converter)', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'SO2':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence', 'process_details':'Make assumption method is ultraviolet fluorescence', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'process_details':'Make assumption method is NDIR (GFC)', 'measuring_assumption':True}}
        elif run_details['parameter'] in ['PM10','PM2.5']:
             standard_measurement_information = {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry', 'process_details':'Make assumption method is TEOM', 'measuring_assumption':True}}
            
#--------------------------------------------------------------
#EANET
#--------------------------------------------------------------
    elif run_details['network'] == 'EANET':
        mapping = {
        '3-stage_filter_pack':    {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'}, 'measurement':{'methodology':'ion chromatography'}}, 
        'dasibi_1006-ahj':        {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1006-AHJ'}},        
        'dasibi_1007-ahj':        {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1007-AHJ'}},
        'dkk-toa_frm377-1':       {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'DKK-TOA FPM-377'}},
        'dkk_toa_frm377-1':       {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'DKK-TOA FPM-377'}},
        'dylec_1150':             {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dylec 1150'}},
        'ecotech_serinus10':      {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Ecotech Serinus10'}},
        'ecotech_serinus40':      {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Ecotech Serinus40'}},
        'ecotech_serinus50':      {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Ecotech Serinus50'}},
        'environmental_sa_ac32m': {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Environnement SA AC32M'}},
        'environmental_sa_af22m': {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Environnement SA AF22M'}},              
        'environmental_sa_mp101m':{                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Environnement SA MP101M'}}, 
        'environmental_sa_o342m': {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O342M'}},
        'esa_mp101m':             {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Environnement SA MP101M'}},
        'esa_pm101m':             {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Environnement SA MP101M'}},
        'esa._mp101m':            {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Environnement SA MP101M'}},
        'horiba_apna-365':        {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Horiba APNA365'}},
        'horiba_apoa-360':        {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA360'}},
        'horiba_apoa-370':        {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA370'}},
        'horiba_apsa-365':        {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Horiba APSA365'}},
        'kimoto_pm-712':          {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Kimoto PM-712'}},
        'met_one_bam1020':        {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Met One BAM1020'}},
        'metone_bam_1020':        {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Met One BAM1020'}},
        'nilu_4_stages':          {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'4-Stage Filter Pack'}, 'measurement':{'methodology':'ion chromatography'}}, 
        'passive_ic':             {'primary_sampling':{'sampling_type':'passive'},               'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},   'measurement':{'methodology':'ion chromatography'}},   
        'teledyne_api_t100':      {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Teledyne T100'}},
        'teledyne_api_t200':      {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Teledyne T200'}},
        'teledyne_api_t400':      {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne T400'}},
        'thermo_42i-tl':          {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Thermo 42i-TL'}},
        'thermo_43c':             {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Thermo 43C'}},
        'thermo_5014i':           {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Thermo 5014i'}},
        'thermo-doas-2000':       {                                                                                                                                                                          'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'Thermo DOAS 2000'}},
        'thermo-fh62-c14':        {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Thermo FH62 C14'}},
        'thermo_fh62c14':         {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                  'instrument':'Thermo FH62 C14'}},
        'thermo_model_42i':       {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Thermo 42i'}},
        'thermo_model_49i':       {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Thermo 49i'}},
        'o3_at_default':          {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet photometry',                                                                      'process_details':'Make assumption method is ultraviolet photometry',                                'measuring_assumption':True}},
        'no_at_default':          {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                           'process_details':'Make assumption method is chemiluminescence (internal molybdenum converter)',     'measuring_assumption':True}}, 
        'no2_at_default':         {                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                           'process_details':'Make assumption method is chemiluminescence (internal molybdenum converter)',     'measuring_assumption':True}},
        'so2_at_default':         {                                                                                                                                                                          'measurement':{'methodology':'ultraviolet fluorescence',                                                                    'process_details':'Make assumption method is ultraviolet fluorescence',                              'measuring_assumption':True}}, 
        'pm10_at_default':        {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                                                            'process_details':'Make assumption method is beta-attenuation',                                      'measuring_assumption':True}},
        'pm2.5_at_default':       {                                                                                                                                                                          'measurement':{'methodology':'beta-attenuation',                                                                            'process_details':'Make assumption method is beta-attenuation',                                      'measuring_assumption':True}},
        'fp_default':             {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                  'measurement':{'methodology':'ion chromatography',                                                                          'process_details':'Make assumption method is ion chromatography',                                    'measuring_assumption':True}},
        'ps_default':             {'primary_sampling':{'sampling_type':'passive'},               'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},   'measurement':{'methodology':'ion chromatography',                                                                          'process_details':'Make assumption method is ion chromatography',                                    'measuring_assumption':True}}}   


#--------------------------------------------------------------
#EBAS
#--------------------------------------------------------------
    elif run_details['network'] == 'EBAS':      
        if run_details['parameter'] == 'O3':
            mapping = {
            'uv_abs':                              {'measurement':{'methodology':'ultraviolet photometry'}},
            'uv_abs,dasibi,1108':                  {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Dasibi 1108 W/GEN',        'process_details':'Make assumption measuring instrument is Dasibi 1108 W/GEN', 'measuring_assumption':True}},
            'uv_abs,dylec,1100':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Dylec 1100'}},
            'uv_abs,ebara,eg-2001f':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-2001F'}},
            'uv_abs,ebara,eg-2001ftp':             {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-2001FTP'}},
            'uv_abs,ebara,eg-3000f':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ebara EG-3000F'}},
            'uv_abs,ecotech,ml_9810':              {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Ecotech EC9810'}},
            'uv_abs,ansyco,41m':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Environnement SA O341M'}}, 
            'uv_abs,horiba,apoa-370':              {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Horiba APOA370'}},
            'uv_abs,teledyne,400a':                {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne 400A'}},
            'uv_abs,monitor_labs,9810':            {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne ML9810'}},
            'uv_abs,teledyne_monitor_labs,ml9810b':{'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne ML9810B'}},
            'uv_abs,teledyne,t400':                {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Teledyne T400'}},
            'uv_abs,thermo,49':                    {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49'}},
            'uv_abs,thermo,49c':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49C'}},
            'uv_abs,thermo,49c+49i':               {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49C'}},
            'uv_abs,thermo,49i':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49i'}},
            'uv_abs,thermo,49w':                   {'measurement':{'methodology':'ultraviolet photometry',   'instrument':'Thermo 49w'}},
            'na':                                  {'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {
            'abs_solution':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                            'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'abs_solution,imwm,imwm_abs_solution':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                            'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'abs_tube':                                                                                   {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                  'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'continuous_colorimetric':                                                                    {                                                                                                                               'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                            'measurement':{'methodology':'colorimetry',                                                                                  'sampling_type':'continuous injection'}}, 
            'diffusion_tube':                                                                             {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                  'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'caps,teledyne,t500u':                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',          'instrument':'Teledyne T500U'}},
            'chemiluminescence_molybdenum':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}},
            'chemiluminescence_photometer':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_photolytic,airqualitydesign+airqualitydesign,blc+custom':                  {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,dmt+thermo,blc+42itl':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,ecophysics,cld700alppt':                                        {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecophysics CLD700AL',                                                'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_photolytic,ecophysics,cld770alppt':                                        {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecophysics CLD770AL',                                                'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,ecophysics,cranox':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',   'instrument':'Ecophysics CraNOx'}},
            'chemiluminescence_photolytic,ecotech,9841t':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Ecotech EC9841T'}},
            'chemiluminescence_molybdenum,environnement,ac32m':                                           {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Environnement SA AC32M'}}, 
            'chemiluminescence_molybdenum,horiba,apna-360ce':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA360'}},
            'chemiluminescence_molybdenum,horiba,apna-370':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370'}}, 
            'chemiluminescence_molybdenum,horiba,apna-370ce':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370'}},
            'chemiluminescence_photolytic,horiba,apna-370':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Horiba APNA370',                                                     'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}}, 
            'chemiluminescence_molybdenum,teledyne_api,200eu':                                            {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Teledyne 200EU'}},
            'chemiluminescence_molybdenum,teledyne_api,200e':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Teledyne 200E'}},
            'chemiluminescence_photolytic,teledyne_api,200e':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Teledyne 200E',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,teledyne,t200up':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',   'instrument':'Teledyne T200UP'}},
            'chemiluminescence_molybdenum,thermo,42c':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42C'}},
            'chemiluminescence_molybdenum,thermo,42c-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42C-TL'}},
            'chemiluminescence_photometer,thermo,42c-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42C-TL',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,thermo,42c-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42C-TL',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_molybdenum,thermo,42i':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i'}},
            'chemiluminescence_photometer,thermo,42i':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i',                                                         'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_molybdenum,thermo,42i-ls':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-LS'}},
            'chemiluminescence_molybdenum,thermo,42i-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL'}},
            'chemiluminescence_photometer,thermo,42i-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_photolytic,thermo,42i-tl':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',   'instrument':'Thermo 42i-TL',                                                      'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'chemiluminescence_molybdenum,thermo,49i':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                                                                                     'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) as stated and not ultraviolet photometry as Thermo 49i instrument uses', 'measuring_assumption':True}},
            'chemiluminescence_photometer,thermo,49i':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                                                                                                                                                                     'process_details':'Make assumption the method is chemiluminescence (internal photolytic converter) as stated and not ultraviolet photometry as Thermo 49i instrument uses', 'measuring_assumption':True}},
            'doas':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'filter_1pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume method is ion chromatography after sampling.',                                                                                                    'measuring_assumption':True}},
            'filter_1pack,eccc,one_stage_inline,ic,dionex,dx-600':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                  'instrument':'Dionex DX-600'}},
            'filter_2pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume method is ion chromatography after sampling.',                                                                                                    'measuring_assumption':True}},
            'filter_3pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume method is ion chromatography after sampling.',                                                                                                    'measuring_assumption':True}}, 
            'filter_abs_solution':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                      'preparation_2':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}}, 
            'glass_sinter':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume measurement methodology is ion chromatography after sampling',                                                                                    'measuring_assumption':True}}, 
            'glass_sinter,mcz,micropns':                                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'MCZ LVS1',     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume measurement methodology is ion chromatography after sampling',                                                                                    'measuring_assumption':True}},
            'glass_sinter,nilu,fk':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU FK'},                                          'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume measurement methodology is ion chromatography after sampling',                                                                                    'measuring_assumption':True}},
            'glass_sinter,nilu,fk,spectrophotometric_griess,bran+luebbe,autoanalyzer_3':                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU FK'},                                          'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'continuous flow analysis',                            'instrument':'Bran+Luebbe Autoanalyzer 3'}},                                                                                                                                                                         
            'glass_sinter,nilu,fk,spectrophotometric,seal_analytical,aa500_autoanalyzer':                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU FK'},                                          'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'continuous flow analysis',                            'instrument':'SEAL AA500'}},                                                                                                                                                                         
            'glass_sinter,nilu,ss2000':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume measurement methodology is ion chromatography after sampling',                                                                                    'measuring_assumption':True}},
            'glass_sinter,nilu,ss2000,ic,dionex,2010i':                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                  'instrument':'Dionex 2010i'}},                                                                                                                                                                                                  
            'glass_sinter,nilu,ss2000,spectrophotometric,thermo_scientific,gallery_plus':                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'continuous flow analysis',                            'instrument':'Thermo Gallery Plus'}},                                                                                                                                                                                                
            'glass_sinter,werner-glas_&_instrument_ab,fk':                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU FK'},                                          'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                    'process_details':'Assume measurement methodology is ion chromatography after sampling',                                                                                    'measuring_assumption':True}},
            'glass_sinter,werner-glas_&_instrument_ab,fk,spectrophotometric,perkinelmer,lambda_18_uv_vis':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU FK'},                                          'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                            'measurement':{'methodology':'spectrophotometry',                                   'instrument':'Perkin Elmer 18 UV/VIS'}},
            'naj_solution':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                              'process_details':'Cannot find naj solution so just assume reagant reaction',  'preparation_assumption':True},                                                                 'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'passive_sampler':                                                                            {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                  'measurement':{'methodology':'spectrophotometry',                                                                                                                                                                                                     'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                     'measuring_assumption':True}},
            'uv_fluoresc':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'laser-induced fluorescence',                                                                                                                                                                                            'process_details':'Make assumption this technique is laser-induced fluorescence (proven for NO/NO2), and not ultraviolet fluorescence (used exclusively for SO2).',         'measuring_assumption':True}},
            'na':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'CO':
            mapping = {
            'filter_1pack':                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'}, 'measurement':{'methodology':'ion chromatography',                                                                                                                     'process_details':'Assume method is ion chromatography after sampling.',                                                                                                       'measuring_assumption':True}},
            'gc-hgo':                                           {                                                                                                                                                                                                         'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                        'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                                                                                               'measuring_assumption':True}},
            'infrared_absorption':                              {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                                            'measuring_assumption':True}},      
            'infrared_absorption,teledyne,t300':                {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne T300'}},
            'ndir':                                             {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                                            'measuring_assumption':True}},                            
            'ndir,horiba,apma-360':                             {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'ndir,horiba,apma-370':                             {                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA370'}},
            'online_crds':                                      {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'online_crds,picarro,g2301':                        {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'Picarro G2301'}},
            'online_crds,picarro,g2401':                        {                                                                                                                                                                                                         'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'Picarro G2401'}},     
            'steel_canister,restek,silcocan,crds,picarro,g2401':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                   'measurement':{'methodology':'cavity ringdown spectroscopy',                                'instrument':'Picarro G2401',      'sampling_type':'injection'}},
            'uv_fluoresc':                                      {                                                                                                                                                                                                         'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                                                                                              'process_details':'Make assumption this technique is vacuum ultraviolet resonance fluorescence (proven for CO), and not ultraviolet fluorescence (used exclusively for SO2).', 'measuring_assumption':True}},
            'vuv_fluorescence,aero-laser,al5001':               {                                                                                                                                                                                                         'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                   'instrument':'Aerolaser AL5001'}},
            'na':                                               {                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] in ['CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','M-P-XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE','NMVOC','NMHC']:
            mapping = {
            'ads_tube,tera_environnement,unspecified,gc-ms,perkinelmer,clarus_600':{'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                     'measurement':{'methodology':'gas chromatography - mass spectrometry'}},          
            'ads_tube,waters,sep-pak_xposure_sampler_wat047205':                   {'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                     'measurement':{'methodology':'high performance liquid chromatography - unknown detection',                                                                           'process_details':'Assume method is high performance liquid chromatography - unknown detection', 'measuring_assumption':True}},
            'gc-fid':                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'glass_flask,gc-fid,hewlett_packard,hp-5890_series_ii':                {'primary_sampling':{'sampling_type':'high volume continuous'},                                     'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},      
            'online_crds':                                                         {                                                                                                                                                                                       'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'online_crds,picarro,g1301':                                           {                                                                                                                                                                                       'measurement':{'methodology':'cavity ringdown spectroscopy',                               'instrument':'Picarro G1301'}},
            'online_crds,picarro,g2301':                                           {                                                                                                                                                                                       'measurement':{'methodology':'cavity ringdown spectroscopy',                               'instrument':'Picarro G2301'}},
            'online_crds,picarro,g2401':                                           {                                                                                                                                                                                       'measurement':{'methodology':'cavity ringdown spectroscopy',                               'instrument':'Picarro G2401'}},
            'online_gc':                                                           {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                     'sampling_type':'continuous injection'}}, 
            'online_gc,agage+agilent+agilent,medusa+6890n+5975b':                  {                                                                                                   'preparation_1':{'sample_preparation_type':'preconcentration'},                     'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                     'sampling_type':'continuous injection'}},
            'online_gc,agilent,5975gc_ms':                                         {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                     'sampling_type':'continuous injection'}},
            'online_gc,agilent,6890gc_fid':                                        {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                            'sampling_type':'continuous injection'}},
            'online_gc,agilent,7890a_uecd':                                        {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - electron capture detection',                                            'sampling_type':'continuous injection'}},                                                                                                
            'online_gc,markes+markes+agilent+agilent,air_server+unity2+6890+5975c':{                                                                                                   'preparation_1':{'sample_preparation_type':'preconcentration'},                     'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                     'sampling_type':'continuous injection'}},
            'online_gc,markes+markes+agilent,air_server+unity2+7890a_gc_fid':      {                                                                                                   'preparation_1':{'sample_preparation_type':'preconcentration'},                     'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                            'sampling_type':'continuous injection'}},
            'online_gc,perkinelmer+varian+varian,atd300+gc3800+saturn_ms2200':     {                                                                                                   'preparation_1':{'sample_preparation_type':'preconcentration'},                     'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                     'sampling_type':'continuous injection'}},
            'online_gc,sri,8610_gc':                                               {                                                                                                   'preparation_1':{'sample_preparation_type':'preconcentration'},                     'measurement':{'methodology':'gas chromatography - unknown detection',                                                     'sampling_type':'continuous injection'}},
            'online_gc,varian,gc-3600_fid':                                        {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                            'sampling_type':'continuous injection'}},    
            'online_gc,varian,gc-3800':                                            {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                     'sampling_type':'continuous injection'}},
            'online_gc,varian,gc-3800_fid':                                        {                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                            'sampling_type':'continuous injection'}},
            'online_ptr':                                                          {                                                                                                                                                                                       'measurement':{'methodology':'proton transfer reaction - unknown detection',                                               'sampling_type':'continuous injection'}},
            'online_ptr,ionicon_analytik,high-sensitivity_ptr-ms':                 {                                                                                                                                                                                       'measurement':{'methodology':'proton transfer reaction - mass spectrometry',                                               'sampling_type':'continuous injection'}},
            'online_ptr,ionicon_analytik,high-sensitivity_ptr-qms_500':            {                                                                                                                                                                                       'measurement':{'methodology':'proton transfer reaction - mass spectrometry',                                               'sampling_type':'continuous injection'}},
            'steel_canister':                                                      {'primary_sampling':{'sampling_type':'high volume continuous'},                                     'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                               'process_details':'Assume method is GC- unknown detection',                                      'measuring_assumption':True}},
            'steel_canister,andersen,unknown':                                     {'primary_sampling':{'sampling_type':'high volume continuous'},                                     'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                               'process_details':'Assume method is GC- unknown detection',                                      'measuring_assumption':True}},
            'steel_canister,restek,silcocan,crds,picarro,g2401':                   {'primary_sampling':{'sampling_type':'high volume continuous'},                                     'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'cavity ringdown spectroscopy',                               'instrument':'Picarro G2401',   'sampling_type':'injection'}}, 
            'steel_canister,restek+entech,silcocan+silonite':                      {'primary_sampling':{'sampling_type':'high volume continuous'},                                     'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                               'process_details':'Assume method is GC- unknown detection',                                      'measuring_assumption':True}},
            'steel_canister,riemer,autocan_2_4':                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Riemer Autocan 2/4'},  'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                               'process_details':'Assume method is GC- unknown detection',                                      'measuring_assumption':True}},
            'steel_canister,riemer,autocan_2_4,gc,perkinelmer,clarus_580':         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Riemer Autocan 2/4'},  'preparation_1':{'sample_preparation_type':'flask'},                                'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'na':                                                                  {                                                                                                                                                                                       'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] in ['VOC','HC']:
            mapping = {
            'steel_canister':{'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - unknown detection', 'process_details':'Assume method is GC- unknown detection', 'measuring_assumption':True}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HCHO':
            mapping = {
            'ads_tube':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                 'measurement':{'methodology':'high performance liquid chromatography - unknown detection', 'process_details':'Assume method is high performance liquid chromatography - unknown detection', 'measuring_assumption':True}},
            'ads_tube,waters,sep-pak_xposure_sampler_wat047205':                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                 'measurement':{'methodology':'high performance liquid chromatography - unknown detection', 'process_details':'Assume method is high performance liquid chromatography - unknown detection', 'measuring_assumption':True}},
            'ads_tube,waters,sep-pak_xposure_sampler_wat047205,hplc,thermo_scientific,ultimate_3000':{'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                 'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}}, 
            'high_performance_liq':                                                                  {'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',             'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'SO2':
            mapping = {
            'abs_solution':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                             'measurement':{'methodology':'spectrophotometry',                                                                                                          'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                'measuring_assumption':True}},
            'abs_tube':                                                                                     {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                          'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                'measuring_assumption':True}},
            'ams':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Some additional details in raw files which suggest this is Ion Chromatography',                                                     'measuring_assumption':True}}, 
            'bulk_sampler':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                       'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'denuder':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'doas':                                                                                         {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy'}}, 
            'flame_photometry':                                                                             {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'flame photometric detection'}},
            'filter_abs_solution':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                       'preparation_2':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'spectrophotometry',                                                                                                          'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                'measuring_assumption':True}},                                                                         
            'filterpack':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                  'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_1pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_3pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack,imwm,two_stage_inline':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face,ic,dionex,ics-3000':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-3000'}},
            'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-2000'}},
            'filter_3pack,ic,dionex,ics-2100':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-2100'}},
            'filter_3pack,nilu,three_stage_open_face':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},   
            'filter_3pack,nilu,three_stage_open_face,ic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,2010i':                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Dionex 2010i'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,dx-120':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Dionex DX-120'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-2000'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2100':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-2100'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-3000':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-3000'}}, 
            'filter_3pack,nilu,three_stage_open_face,ic,metrohm,881_compact_ic_pro':                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Metrohm 881 compact IC pro'}},
            'filter_3pack,riemer,spf4':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'filter_3pack,riemer,spf4,ic,dionex,ics-2000':                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                                                                                 'measurement':{'methodology':'ion chromatography',                             'instrument':'Thermo ICS-3000'}}, 
            'lvs_teflon':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                                               'measuring_assumption':True}},
            'online_ic':                                                                                    {                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'ion chromatography',                                                               'sampling_type':'continuous injection'}},
            'passive_sampler':                                                                              {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                          'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                'measuring_assumption':True}},
            'trace_level_analyzer':                                                                         {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                   'process_details':'Make assumption this technique is ultraviolet fluorescence (exclusive for SO2).',                                                   'measuring_assumption':True}},
            'tub_denuder':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                                         'process_details':'Assume ion chromatography after sampling',                                                                                          'measuring_assumption':True}},
            'uv_abs':                                                                                       {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                   'process_details':'Make assumption this technique is ultraviolet fluorescence (exclusive for SO2), and not ultraviolet photometry (for measuring O3)', 'measuring_assumption':True}},
            'uv_fluoresc':                                                                                  {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv_fluoresc,ecotech,ec9850':                                                                   {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech EC9850'}},
            'uv_fluoresc,teledyne,t100':                                                                    {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne T100'}},
            'uv_fluoresc,teledyne,t100u':                                                                   {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne T100U'}},
            'uv_fluoresc,thermo,43i':                                                                       {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i'}},
            'uv_fluoresc,thermo,43i-tl':                                                                    {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i-TLE'}},
            'uv_fluoresc,thermo,43i-tle':                                                                   {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i-TLE'}},
            'na':                                                                                           {                                                                                                                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'NH3':
            mapping = {
            'abs_solution':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                             'measurement':{'methodology':'spectrophotometry',                                                                                                                   'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},
            'abs_tube':                                                                                     {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'spectrophotometry',                                                                                                                   'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},
            'ams':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
            'ceh_denuder':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder',          'specific_technique':'CEH DELTA'},                                           'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder,riemer,den2':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder',          'specific_technique':'Riemer DEN2'},                                         'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder,riemer,den2,spectrophotometric,foss,fiastar_5000':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder',          'specific_technique':'Riemer DEN2'},                                         'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Foss FIAstar 5000'}},
            'denuder,uba,olaf':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder',          'specific_technique':'UBA Olaf'},                                            'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'doas,rivm-mil,minidoas_1.0':                                                                   {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy', 'instrument':'miniDOAS'}}, 
            'tub_denuder':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'denuder,ecn,amor':                                                                             {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'conductimetry',                                'instrument':'ECN AMOR'}},
            'filter_1pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face,spectrophotometric,bran+luebbe,autoanalyzer_3':          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Bran+Luebbe Autoanalyzer 3'}}, 
            'filter_3pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face':                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
            'filter_3pack,ic,dionex,ics-2100':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2100'}},
            'filter_3pack,riemer,spf4':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,dx-120':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                           'instrument':'Dionex DX-120'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2100':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2100'}},
            'filter_3pack,nilu,three_stage_open_face,spectrophotometric,shimadzu,uv-1201':                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'spectrophotometry',                            'instrument':'Shimadzu UV-1201'}},
            'filter_3pack,nilu,three_stage_open_face,spectrophotometric,thermo_scientific,gallery_plus':    {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Thermo Gallery Plus'}},
            'filter_3pack,riemer,spf4,spectrophotometric,foss,fiastar_5000':                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Foss FIAstar 5000'}}, 
            'high_vol_sampler':                                                                             {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},                      
            'lvs_teflon':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                                                  'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'online_ic':                                                                                    {                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                        'sampling_type':'continuous injection'}},    
            'passive_sampler':                                                                              {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'spectrophotometry',                                                                                                                   'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},
            'passive_sampler,radiello,rad168,ic,thermo_scientific,ics-3000':                                {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-3000'}},
            'passive_sampler,radiello,rad168,ic,thermo_scientific,ics-5000':                                {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-5000'}},
            'passive_sampler,radiello,rad168,spectrophotometric,thermo_scientific,aquakem_250':             {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'spectrophotometry',                            'instrument':'Thermo Aquakem 250'}},
            'passive_sampler,radiello,rad168,spectrophotometric,thermo_scientific,aquakem_600':             {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'spectrophotometry',                            'instrument':'Thermo Aquakem 600'}},
            'na':                                                                                           {                                                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HNO3':
            mapping = { 
            'abs_tube':                                                                                     {'primary_sampling':{'sampling_type':'passive',                                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                   'measurement':{'methodology':'spectrophotometry',                                                                                       'process_details':'Assume measurement methodology is spectrophotometry after sampling',                          'measuring_assumption':True}},
            'ams':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Some additional details in raw files which suggest this is Ion Chromatography',               'measuring_assumption':True}}, 
            'ceh_denuder':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder',          'specific_technique':'CEH DELTA'},                                           'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'denuder':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_1pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'1-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_2pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_2pack,nilu,two_stage_open_face,ic,dionex,ics-3000':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-3000'}},
            'filter_3pack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                      'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-2000'}}, 
            'filter_3pack,ic,dionex,ics-2100':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-2100'}},
            'filter_3pack,nilu,three_stage_open_face':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_3pack,nilu,three_stage_open_face,ce':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'capillary electrophoresis'}},
            'filter_3pack,nilu,three_stage_open_face,ce,waters,ce_1994':                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'capillary electrophoresis',   'instrument':'Waters Alliance'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,dx-120':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Dionex DX-120'}},
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-2000'}},             
            'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2100':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-2100'}},
            'filter_3pack,riemer,spf4':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                           'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'filter_3pack,riemer,spf4,ic,dionex,ics-2000':                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},                                           'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',          'instrument':'Thermo ICS-2000'}}, 
            'lvs_teflon':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                         'measuring_assumption':True}},
            'online_ic':                                                                                    {                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                            'sampling_type':'continuous injection'}}, 
            'na':                                                                                           {                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HCl':
            mapping = {
            'ams':                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                                                      'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
            'filter_3pack':                               {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,riemer,spf4':                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},         'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
            'filter_3pack,riemer,spf4,ic,dionex,ics-2000':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Riemer SPF4'},         'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                          'instrument':'Thermo ICS-2000'}},
            'online_ic':                                  {                                                                                                  'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                            'sampling_type':'continuous injection'}}, 
            'trace_level_analyzer':                       {                                                                                                                                                                                                                                         'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                             'process_details':'Some additional details in raw files which suggest this is NDIR - assume GFC',  'measuring_assumption':True}}}
#-------------------------------------------
        elif run_details['parameter'] in ['Hg-GEM','Hg-GOM','Hg-TGM']:
            mapping = {}
#-------------------------------------------
        elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
            mapping = {
            'abs_solution':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume measurement methodology is gravimetry before reagent reacttion',           'measuring_assumption':True}},
            'b-attenuation':                                            {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation'}},
            'b-attenuation,opsis,sm200':                                {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'beta_gauge_particulate_sampler':                           {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation'}},
            'beta_gauge_particulate_sampler,environnement,mp101m':      {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'beta_gauge_particulate_sampler,metone,bam1020':            {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'beta_gauge_particulate_sampler,thermo,5014i':              {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'beta_gauge_particulate_sampler,thermo,5030sharp':          {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'beta_gauge_particulate_sampler,thermo,fh62c14':            {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'beta_gauge_particulate_sampler,thermo,fh62i-r':            {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'beta_ray_absorption':                                      {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation'}},
            'dicho':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Some additional details in raw files which suggest this is a gravimetric method', 'measuring_assumption':True}},
            'filterpack':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'filter_1pack':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'filter_2pack':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'filter_3pack':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler':                                         {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                              'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,digitel,dha-80,by_weight':                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,digitel,dha-80,by_weight,sartorius,le26p':{'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry'}},                                                                                                                             
            'high_vol_sampler,digitel,da-80':                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,digitel,dha-80':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_mb':                            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/mb'},                                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,mcv,cav-a_msb':                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'high_vol_sampler,thermo,vfc-pm10':                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Thermo VFC-PM10'},                                                                                                                                                                                                                                                    'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler':                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,derenda,lvs_3.1':                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,derenda,pns_16t':                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda PNS 16T'},                                                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,fai_instruments,swam_5a_dual_channel':     {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'low_vol_sampler,leckel,lvs3':                              {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,leckel,lvs3,by_weight':                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,leckel,mvs6':                              {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,leckel,seq47_50':                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,rupprecht&patashnick,partisol2025':        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,thermo,partisol2000i':                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2000i'},                                                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'low_vol_sampler,thermo,partisol2025':                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'lvs_qbq':                                                  {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'lvs_teflon':                                               {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'impactor':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                  'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},                                                                                                                                                                                                                                                                                                                        
            'impactor,dekati,cascade':                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                  'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},                                                                                                                                                                                                                                                                                                                        
            'ivlpm':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'IVL P-Model',            'process_details':'Rough guess it is this specific instrument.', 'primary_sampling_assumption':True},                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                             'process_details':'Assume method is gravimetry after sampling.',                                     'measuring_assumption':True}},
            'optical_particle_counter':                                 {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter'}},
            'optical_particle_size_spectrometer':                       {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter'}},
            'optical_particle_size_spectrometer,grimm,180':             {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'optical_particle_size_spectrometer,grimm,190':             {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 190'}},
            'optical_particle_size_spectrometer,palas,fidas_200':       {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter'},                                                                    'instrument':'Palas Fidas 200'},
            'optical_particle_size_spectrometer,palas,fidas_200_e':     {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'optical particle counter'},                                                                    'instrument':'Palas Fidas 200E'},
            'radiometric':                                              {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'beta-attenuation'}},
            'teom':                                                     {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'teom_fdms':                                                {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom,rupprecht&patashnick,1400':                           {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom,rupprecht&patashnick,1400a':                          {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},            
            'teom,rupprecht&patashnick,1400ab':                         {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},  
            'thermo':                                                   {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                       'measuring_assumption':True}},
            'teom,thermo,1405':                                         {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},       
            'teom,thermo,1405d':                                        {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405D'}},
            'teom,thermo,1405df':                                       {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'na':                                                       {                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif parameter_details['matrix'] == 'pm':   
            if run_details['parameter'] in ['C','EC','OC']:
                mapping = {
                'filter_2pack':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol', 'measuring_assumption':True}},
                'high_vol_sampler':                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol', 'measuring_assumption':True}},                       
                'high_vol_sampler,digitel,dha-80':                                                                   {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol', 'measuring_assumption':True}},                       
                'low_vol_sampler,knf,ip20-t,thermal-optical_analysis,dri,model_2001_thermal_optical_carbon_analyzer':{'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'KNF IP20-T'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2', 'instrument':'DRI Model 2001',      'process_details':'Assume protocol is EUSAAR2',                                          'measuring_assumption':True}},
                'low_vol_sampler,knf,ip20-t,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5': {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'KNF IP20-T'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2', 'instrument':'Sunset OCEC Model 5', 'process_details':'Assume protocol is EUSAAR2',                                          'measuring_assumption':True}}}  
            elif run_details['parameter'] in ['BC']:
                mapping = {
                'filter_absorption_photometer':            {'measurement':{'methodology':'aethalometer'}},
                'filter_absorption_photometer,magee':      {'measurement':{'methodology':'aethalometer',                      'instrument':'Magee AE31',    'process_details':'Assume instrument is Magee AE31.', 'measuring_assumption':True}},
                'filter_absorption_photometer,magee,ae31': {'measurement':{'methodology':'aethalometer',                      'instrument':'Magee AE31'}},
                'filter_absorption_photometer,thermo,5012':{'measurement':{'methodology':'multi angle absorption photometer', 'instrument':'Thermo 5012'}}}
            elif 'AERO_ABS_COEFF' in run_details['parameter']:
                mapping = {
                'filter_absorption_photometer':                          {'measurement':{'methodology':'aethalometer'}},
                'filter_absorption_photometer,magee':                    {'measurement':{'methodology':'aethalometer'}},
                'filter_absorption_photometer,magee,ae8':                {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE8'}},
                'filter_absorption_photometer,magee,ae16':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE16'}},
                'filter_absorption_photometer,magee,ae31':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE31'}},
                'filter_absorption_photometer,magee,ae33':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE33'}},
                'filter_absorption_photometer,noaa_esrl_gmd,clap-3w':    {'measurement':{'methodology':'continuous light absorption photometer', 'instrument':'NOAA CLAP'}},
                'filter_absorption_photometer,noaa_esrl_gmd,clap-10':    {'measurement':{'methodology':'continuous light absorption photometer', 'instrument':'NOAA CLAP'}},
                'filter_absorption_photometer,radiance-research,psap-1w':{'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'Radiance PSAP-1W'}},
                'filter_absorption_photometer,radiance-research,psap-3w':{'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'Radiance PSAP-3W'}},
                'filter_absorption_photometer,thermo,5012':              {'measurement':{'methodology':'multi angle absorption photometer',      'instrument':'Thermo 5012'}}}
            elif ('AERO_LIGHT_BACKSCAT_COEFF' in run_details['parameter']) or ('AERO_LIGHT_SCAT_COEFF' in run_details['parameter']):
                mapping = {
                'nephelometer':                              {'measurement':{'methodology':'nephelometry'}},
                'nephelometer,ecotech,aurora3000':           {'measurement':{'methodology':'nephelometry', 'instrument':'Ecotech Aurora 3000'}},
                'nephelometer,ecotech,aurora4000':           {'measurement':{'methodology':'nephelometry', 'instrument':'Ecotech Aurora 4000'}},
                'nephelometer,ecotech,m9003':                {'measurement':{'methodology':'nephelometry', 'instrument':'Ecotech 9003'}},
                'nephelometer,meteorology_research,inc.,4-w':{'measurement':{'methodology':'nephelometry', 'instrument':'Meteorology Research Inc. 4-W'}},
                'nephelometer,ms_electron,3w-02':            {'measurement':{'methodology':'nephelometry', 'instrument':'MS Electron 3W-02'}},
                'nephelometer,optec,ngn-2':                  {'measurement':{'methodology':'nephelometry', 'instrument':'Optec NGN-2'}},
                'nephelometer,optech,ngn-2':                 {'measurement':{'methodology':'nephelometry', 'instrument':'Optec NGN-2'}},
                'nephelometer,tsi,3563':                     {'measurement':{'methodology':'nephelometry', 'instrument':'TSI 3563'}}}
            else:
                mapping = { 
                'abs_solution':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'},                                'measurement':{'methodology':'spectrophotometry',                                                                                                                 'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},
                'abs_tube':                                                                                          {'primary_sampling':{'sampling_type':'passive',                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',         'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                                                                                                     'measurement':{'methodology':'spectrophotometry',                                                                                                                 'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},
                'air_uk':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}}, 
                'ams':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Some additional details in raw files which suggest this is Ion Chromatography', 'measuring_assumption':True}}, 
                'bulk_sampler':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                               'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'ceh_denuder':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder',                  'specific_technique':'CEH DELTA'},                                                                                                                         'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'denuder':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},                                                                                                                                                      
                'filter_abs_solution':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                               'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'},                                'measurement':{'methodology':'spectrophotometry',                                                                                                                 'process_details':'Assume measurement methodology is spectrophotometry after sampling',            'measuring_assumption':True}},                                                                         
                'filter_1pack':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_1pack,eccc,one_stage_inline,ic,dionex,dx-600':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Dionex DX-600'}},
                'filter_1pack,imwm,one_stage_inline':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_1pack,imwm,one_stage_inline,f-aas,perkinelmer,aanalyst_700':                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',         'instrument':'Perkin Elmer AAnalyst 700'}},     
                'filter_2pack':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_2pack,imwm,two_stage_inline':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_2pack,nilu,two_stage_open_face':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_2pack,nilu,two_stage_open_face,ic,dionex,ics-3000':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-3000'}},
                'filter_2pack,nilu,two_stage_open_face,spectrophotometric,bran+luebbe,autoanalyzer_3':               {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Bran+Luebbe Autoanalyzer 3'}},  
                'filter_3pack':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},              
                'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face':                        {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                        'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}}, 
                'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,ic,dionex,ics-2000':     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                        'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
                'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,icp-ms,agilent,7500':    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                        'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry', 'instrument':'Agilent 7500'}},
                'filter_3pack,custom_build_-_leckel,modified_seq47_50_three_stage_open_face,icp-ms,agilent,7900':    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                        'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry', 'instrument':'Agilent 7900'}},
                'filter_3pack,ic,dionex,ics-2100':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2100'}},
                'filter_3pack,nilu,three_stage_open_face':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_3pack,nilu,three_stage_open_face,ce':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'capillary electrophoresis'}},
                'filter_3pack,nilu,three_stage_open_face,ce,waters,ce_1994':                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'capillary electrophoresis',                    'instrument':'Waters Alliance'}},
                'filter_3pack,nilu,three_stage_open_face,f-aas,perkinelmer,aanalyst_700':                            {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',         'instrument':'Perkin Elmer AAnalyst 700'}},
                'filter_3pack,nilu,three_stage_open_face,f-aas,perkinelmer,aanalyst_800':                            {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',         'instrument':'Perkin Elmer AAnalyst 800'}},
                'filter_3pack,nilu,three_stage_open_face,ic':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
                'filter_3pack,nilu,three_stage_open_face,ic,dionex,dx-120':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Dionex DX-120'}},
                'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2000':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
                'filter_3pack,nilu,three_stage_open_face,ic,dionex,2010i':                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Dionex 2010i'}},  
                'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-2100':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2100'}},
                'filter_3pack,nilu,three_stage_open_face,ic,dionex,ics-3000':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-3000'}},
                'filter_3pack,nilu,three_stage_open_face,ic,metrohm,881_compact_ic_pro':                             {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Metrohm 881 compact IC pro'}},
                'filter_3pack,nilu,three_stage_open_face,icp-ms,agilent,7500':                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry', 'instrument':'Agilent 7500'}},
                'filter_3pack,nilu,three_stage_open_face,spectrophotometric,shimadzu,uv-1201':                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                            'instrument':'Shimadzu UV-1201'}},
                'filter_3pack,nilu,three_stage_open_face,spectrophotometric,thermo_scientific,gallery_plus':         {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis',                     'instrument':'Thermo Gallery Plus'}},
                'filter_3pack,riemer,spf4':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Riemer SPF4'},                                           'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'filter_3pack,riemer,spf4,ic,dionex,ics-2000':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Riemer SPF4'},                                           'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-2000'}},
                'hg_mon':                                                                                            {                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                              'sampling_type':'continuous injection', 'process_details':'Assume method is continuous cold vapour atomic fluorescence spectroscopy',      'measuring_assumption':True}},  
                'high_vol_sampler':                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                        'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},              
                'high_vol_sampler,digitel,dh-77,icp-ms,agilent,7700x':                                               {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DH-77'},                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry', 'instrument':'Agilent 7700x'}},
                'high_vol_sampler,digitel,da-80':                                                                    {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'high_vol_sampler,digitel,dha-80':                                                                   {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'high_vol_sampler,digitel,dha-80,ic,dionex,ics-3000':                                                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-3000'}},
                'high_vol_sampler,sierra_instruments,basic_hi_vol_model_305-2000':                                   {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Sierra 305-2000'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'high_vol_sampler,sierra_instruments,basic_hi_vol_model_305-2000,ic':                                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Sierra 305-2000'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},
                'high_vol_sampler,sierra_instruments,basic_high_volume_air_sampler,ic':                              {'primary_sampling':{'sampling_type':'high volume continuous',},                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},
                'low_vol_sampler':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                         'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},    
                'low_vol_sampler,derenda,pns_16t,gc-ms':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda PNS 16T'},                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'low_vol_sampler,knf,ip20-t':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'KNF IP20-T'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'low_vol_sampler,knf,ip20-t,ic,thermo_scientific,ics-1000':                                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'KNF IP20-T'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                           'instrument':'Thermo ICS-1000'}},
                'low_vol_sampler,leckel,seq47_50':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                  'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'mini_trap':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'sorbent trapping',                                     'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'preconcentration', 'preparation_assumption':True}, 'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                                                                      'process_details':'Assume method is cold vapour atomic fluorescence spectroscopy',                 'measuring_assumption':True}}, 
                'sulphate_monitor':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'thermal reduction - ultraviolet fluorescence', 'instrument':'Thermo 5020i',                                                         'process_details':'Some details in raw files suggest instrument is Thermo 5020i',                  'measuring_assumption':True}}, 
                'tub_denuder':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                'process_details':'Assume method is ion chromatography after sampling.',                           'measuring_assumption':True}},
                'na' :                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}}} 
#-------------------------------------------                                                                                                                                                                                                                                                                                                                                                          
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
            if run_details['parameter'] in ['PM10_C','PM10_EC','PM10_OC','PM2.5_C','PM2.5_EC','PM2.5_OC','PM1_C','PM1_EC','PM1_OC']:
                mapping = {
                'denuder':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                         'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                             'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},                                                                                                                                                                                                                                                       
                'filter_1pack':                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                         'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},
                'filter_2pack':                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                         'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},    
                'high_vol_sampler':                                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                            'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},    
                'high_vol_sampler,digitel,dh-77':                                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DH-77'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},     
                'high_vol_sampler,digitel,dh-77,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                               {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DH-77'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},         
                'high_vol_sampler,digitel,da-80,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_4l_eusaar':                   {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 4'}},             
                'high_vol_sampler,digitel,da-80,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l_eusaar':                   {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5'}},                  
                'high_vol_sampler,digitel,da-80_eusaar':                                                                                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},                
                'high_vol_sampler,digitel,dha-77,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                              {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DH-77'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}}, 
                'high_vol_sampler,digitel,dha-80':                                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},      
                'high_vol_sampler,digitel,dha-80,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                              {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},          
                'high_vol_sampler,digitel,dha-80,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_4l':                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 4',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},         
                'high_vol_sampler,digitel,dha-80,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},      
                'high_vol_sampler,digitel,dha-80,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l':                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},      
                'high_vol_sampler,digitel,dha-80_eusaar':                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},      
                'high_vol_sampler,mcv,cav-a_msb':                                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},        
                'high_vol_sampler,mcv,cav-a_msb,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                               {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},      
                'high_vol_sampler,mcv,cav-a_msb_eusaar':                                                                                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},     
                'high_vol_sampler,mcv,cav-a_msb,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},      
                'high_vol_sampler,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_4l_eusaar':                                 {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                            'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 4'}},                        
                'high_vol_sampler,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                                         {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                            'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},                            
                'high_vol_sampler,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5_eusaar':                                  {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                            'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5'}},                                
                'high_vol_sampler,thermo,mfc-pm10':                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Thermo MFC-PM10'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},
                'high_vol_sampler_eusaar':                                                                                                         {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                            'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},                   
                'impactor':                                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                         'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},                                                                                                                                                  
                'impactor_eusaar':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                         'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},                                                                                                                                                                
                'low_vol_sampler':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}}, 
                'low_vol_sampler,derenda,lvs_3.1,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda LVS 3.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2 protocol',                                                 'measuring_assumption':True}},                      
                'low_vol_sampler,derenda,mvs_6.1':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda MVS 6.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},            
                'low_vol_sampler,derenda,pns_16t,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l':                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2 protocol',                                                 'measuring_assumption':True}},              
                'low_vol_sampler,digitel,dpa-14,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Digitel DPA-14'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2 protocol',                                                 'measuring_assumption':True}},              
                'low_vol_sampler,eccc_cair_lab,lvs1':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS1'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},           
                'low_vol_sampler,fai_instruments,swam_5a_dual_channel,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':         {                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'FAI Swam 5a'},                                                                                                                                     'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},
                'low_vol_sampler,leckel,lvs3':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},         
                'low_vol_sampler,leckel,lvs3,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                                  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},                     
                'low_vol_sampler,leckel,lvs3,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2 protocol',                                                 'measuring_assumption':True}},              
                'low_vol_sampler,leckel,lvs3,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5l':                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2 protocol',                                                 'measuring_assumption':True}},              
                'low_vol_sampler,leckel,seq47_50':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},          
                'low_vol_sampler,leckel,seq47_50,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},                     
                'low_vol_sampler,leckel,seq47_50,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},                     
                'low_vol_sampler,rupprecht&patashnick,partisol2025':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},                  
                'low_vol_sampler,rupprecht&patashnick,partisol2025,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer_eusaar':     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},                  
                'low_vol_sampler,rupprecht&patashnick,partisol2025,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_4l_eusaar':{'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 4'}},                             
                'low_vol_sampler,rupprecht&patashnick,partisol2025_eusaar':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},                  
                'low_vol_sampler,tcr_tecora,skypost,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Terora Skypost PM FG'},                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},                     
                'low_vol_sampler,thermal-optical_analysis':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume EUSAAR2 protocol',                                                             'measuring_assumption':True}},        
                'low_vol_sampler,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_4l':                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 4',                                             'process_details':'Assume EUSAAR2 protocol',                                                             'measuring_assumption':True}},     
                'low_vol_sampler,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume EUSAAR2 protocol',                                                             'measuring_assumption':True}},         
                'low_vol_sampler,thermo,partisol2025':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},                  
                'low_vol_sampler,thermo,partisol2025,thermal-optical_analysis,sunset_laboratory,carbon_aerosol_analyzer':                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5 and protocol is EUSAAR2', 'measuring_assumption':True}},              
                'low_vol_sampler,thermo,partisol2025,thermal-optical_analysis,sunset_laboratory,ocec_lab_instrument_model_5':                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',     'instrument':'Sunset OCEC Model 5',                                             'process_details':'Assume EUSAAR2 protocol',                                                             'measuring_assumption':True}},              
                'low_vol_sampler_eusaar':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},               
                'low_vol_sampler_niosh':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040'}},            
                'lvs_denuder_tandem_eusaar':                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                             'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},         
                'lvs_qbq_eusaar':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},
                'lvs_single':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},
                'lvs_single_twin':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},
                'lvs_single_twin_eusaar':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2'}},       
                'lvs_teflon':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                                                                     'process_details':'Assume methodolgy is thermal-optical analysis with EUSAAR2 protocol',                 'measuring_assumption':True}},
                'online_thermal-optical_analysis':                                                                                                 {                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                                           'sampling_type':'continuous injection',   'process_details':'Assume protocol is EUSAAR2',                                                          'measuring_assumption':True}},           
                'online_thermal-optical_analysis,sunset_laboratory,model-4_semi-continuous_field_analyzer_eusaar':                                 {                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',      'instrument':'Sunset OCEC Model 4',  'sampling_type':'continuous injection'}}, 
                'na':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}}} 
            elif run_details['parameter'] in ['PM10_BC','PM2.5_BC','PM1_BC']:
                mapping = {  
                'filter_absorption_photometer,magee,ae16':{                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'aethalometer', 'instrument':'Magee AE16'}}, 
                'filter_absorption_photometer,magee,ae31':{                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'aethalometer', 'instrument':'Magee AE31'}},
                'filter_absorption_photometer,magee,ae33':{                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'aethalometer', 'instrument':'Magee AE33'}},
                'high_vol_sampler':                       {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'aethalometer',                              'sampling_type':'injection', 'process_details':'Assume method is aethalometer after sampling.', 'measuring_assumption':True}}}                 
            elif 'AERO_ABS_COEFF' in run_details['parameter']:
                mapping = {
                'filter_absorption_photometer,magee,ae16':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE16'}},
                'filter_absorption_photometer,magee,ae22':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE22'}},
                'filter_absorption_photometer,magee,ae30':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE30'}},
                'filter_absorption_photometer,magee,ae31':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE31'}},
                'filter_absorption_photometer,magee,ae33':               {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE33'}},
                'filter_absorption_photometer,magee,ae42-7':             {'measurement':{'methodology':'aethalometer',                           'instrument':'Magee AE42-7'}},
                'filter_absorption_photometer,radiance-research,psap-1w':{'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'Radiance PSAP-1W'}},
                'filter_absorption_photometer,radiance-research,psap-3w':{'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'Radiance PSAP-3W'}},
                'filter_absorption_photometer,noaa_esrl_gmd,clap-10':    {'measurement':{'methodology':'continuous light absorption photometer', 'instrument':'NOAA CLAP'}},
                'filter_absorption_photometer,noaa_esrl_gmd,clap-3w':    {'measurement':{'methodology':'continuous light absorption photometer', 'instrument':'NOAA CLAP'}},
                'filter_absorption_photometer':                          {'measurement':{'methodology':'aethalometer'}},
                'filter_absorption_photometer,misu,psap-1w':             {'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'MISU PSAP-1W'}},
                'filter_absorption_photometer,itm,custom':               {'measurement':{'methodology':'particulate soot absorption photometer', 'instrument':'ITM Custom'}},
                'filter_absorption_photometer,thermo,5012':              {'measurement':{'methodology':'multi angle absorption photometer',      'instrument':'Thermo 5012'}}}
            elif ('AERO_LIGHT_BACKSCAT_COEFF' in run_details['parameter']) or ('AERO_LIGHT_SCAT_COEFF' in run_details['parameter']):
                mapping = {
                'dmps':                                       {'measurement':{'methodology':'differential mobility particle sizer'}},
                'nephelometer':                               {'measurement':{'methodology':'nephelometry'}},
                'nephelometer,ecotech,aurora1000':            {'measurement':{'methodology':'nephelometry',                           'instrument':'Ecotech Aurora 1000'}},
                'nephelometer,ecotech,aurora3000':            {'measurement':{'methodology':'nephelometry',                           'instrument':'Ecotech Aurora 3000'}},
                'nephelometer,ecotech,m9003':                 {'measurement':{'methodology':'nephelometry',                           'instrument':'Ecotech 9003'}},
                'nephelometer,meteorology_research,inc.,1550':{'measurement':{'methodology':'nephelometry',                           'instrument':'Meteorology Research Inc. 1550'}},
                'nephelometer,ms_electron,3w-02':             {'measurement':{'methodology':'nephelometry',                           'instrument':'MS Electron 3W-02'}},
                'nephelometer,radiance-research,m903':        {'measurement':{'methodology':'nephelometry',                           'instrument':'Radiance Research 903'}},
                'nephelometer,tsi,3563':                      {'measurement':{'methodology':'nephelometry',                           'instrument':'TSI 3563'}},
                'smps,tropos,tropos-smps':                    {'measurement':{'methodology':'scanning mobility particle sizer',       'instrument':'TROPOS'}},
                'smps,tropos,tropos-tsmps':                   {'measurement':{'methodology':'scanning mobility particle sizer',       'instrument':'TROPOS'}}}
            else:
                mapping = {
                'b-attenuation':                                                                              {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation'},                                                                                                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                                
                'b-attenuation,opsis,sm200':                                                                  {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                                
                'b-attenuation,opsis,sm200,gc-ms,agilent+agilent,6890n+5973':                                 {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'b-attenuation,opsis,sm200,gf-aas,varian,spectraa_880z':                                      {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                    'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',              'instrument':'Varian SpectrAA 880z'}},                                                                                                                 
                'b-attenuation,opsis,sm200,ic,dionex,ics-3000':                                               {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-3000'}},                                                                                                                
                'denuder':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                                                                      
                'filter_1pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'filter_2pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},  
                'filter_3pack':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                   
                'filterpack':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter pack'},                                                                             'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'hg_mon':                                                                                     {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                                                     'sampling_type':'continuous injection',   'process_details':'Assume method is continuous cold vapour atomic fluorescence spectroscopy',                           'measuring_assumption':True}}, 
                'high_vol_sampler':                                                                           {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},              
                'high_vol_sampler,andersen,graseby_andersen_gmw':                                             {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Graseby Andersen GMW 1200',   'process_details':'Make assumption sampling instrument is Graseby Andersen GMW 1200', 'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'high_vol_sampler,andersen,mfc-pm10':                                                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Thermo MFC-PM10'},                                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                             
                'high_vol_sampler,digitel,da-80':                                                             {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                              
                'high_vol_sampler,digitel,dh-77':                                                             {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DH-77'},                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'high_vol_sampler,digitel,dha-80':                                                            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'high_vol_sampler,digitel,dha-80,hplc':                                                       {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}},  
                'high_vol_sampler,digitel,dha-80,hplc,waters,alliance':                                       {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'high performance liquid chromatography - unknown detection'}},
                'high_vol_sampler,digitel,dha-80,gf-aas':                                                     {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}}, 
                'high_vol_sampler,digitel,dha-80,gf-aas,perkinelmer,aanalyst_600':                            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',              'instrument':'Perkin Elmer AAnalyst 600'}}, 
                'high_vol_sampler,digitel,dha-80,gf-aas,unicam,solar_989_qz':                                 {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',              'instrument':'Unicam Solar 989 QZ'}}, 
                'high_vol_sampler,digitel,dha-80,ic':                                                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},
                'high_vol_sampler,digitel,dha-80,ic,thermo_scientific,ics-1100':                              {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-1100'}},
                'high_vol_sampler,digitel,dha-80,ic,waters,ic_system':                                        {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Waters Alliance'}},
                'high_vol_sampler,digitel,dha-80,icp-aes,jarell_ash_(thermo),iris_advantage':                 {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',      'instrument':'Thermo Jarrell Ash IRIS Advantage'}}, 
                'high_vol_sampler,digitel,dha-80,icp-aes,varian,aa-240':                                      {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'flame atomic absorption spectroscopy',                         'instrument':'Varian 240 AA',                                                                 'process_details':'Assume method is F-AAS as referenced instrument uses, not ICP-AES',                                  'measuring_assumption':True}},
                'high_vol_sampler,digitel,dha-80,icp-aes,varian_(agilent),varian_vista-pro':                  {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',      'instrument':'Varian Vista-Pro'}},
                'high_vol_sampler,digitel,dha-80,icp-ms':                                                     {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},                                                                                                                                                                                                                                                                              
                'high_vol_sampler,digitel,dha-80,icp-ms,thermo,x_series_ii':                                  {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Thermo XSERIES 2'}},
                'high_vol_sampler,digitel,dha-80,spectrophotometric,shimadzu,uv-1201':                        {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                            'instrument':'Shimadzu UV-1201'}},  
                'high_vol_sampler,ic,dionex,dx-600':                                                          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                           'instrument':'Dionex DX-600'}},
                'high_vol_sampler,gc-ms':                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'high_vol_sampler,mcv,cav-a_mb':                                                              {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/mb'},                                                                                                                                            'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'high_vol_sampler,mcv,cav-a_msb':                                                             {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},          
                'high_vol_sampler,mcv,cav-a_msb,ic,waters,ic_system':                                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Waters Alliance'}},         
                'high_vol_sampler,mcv,cav-a_msb,icp-aes,jarell_ash_(thermo),iris_advantage':                  {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',      'instrument':'Thermo Jarrell Ash IRIS Advantage'}},       
                'high_vol_sampler,mcv,cav-a_msb,icp-ms,thermo,x_series_ii':                                   {'prinmary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/msb'},                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Thermo XSERIES 2'}},       
                'high_vol_sampler,thermo,mfc-pm10':                                                           {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Thermo MFC-PM10'},                                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'high_vol_sampler,ic,thermo_scientific,ics-3000':                                             {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-3000'}},
                'high_vol_sampler,ic,thermo_scientific,ics-6000':                                             {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-6000'}},
                'high_vol_sampler,tisch_environmental,te-6070dx':                                             {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Tisch Environmental 6070DX'},                                                                                                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'high_vol_sampler,tisch_environmental,te-6070dx,icp-ms,agilent,7500':                         {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Tisch Environmental 6070DX'},                                                                                                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7500'}},
                'impactor':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter'},                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                                                               
                'ivlpm':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'IVL P-Model',                 'process_details':'Rough guess it is this specific instrument.',                      'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},                                                                                                                                                          
                'low_vol_sampler':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},   
                'low_vol_sampler,apm,pms-104,ic,metrohm,881_compact_ic_pro':                                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'APM PMS-104'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Metrohm 881 compact IC pro'}},  
                'low_vol_sampler,derenda,lvs_3.1':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'low_vol_sampler,derenda,lvs_3.1,icp-ms,agilent,7500':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7500'}},  
                'low_vol_sampler,derenda,lvs_3.1,icp-ms,agilent,8900':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 8900'}},  
                'low_vol_sampler,derenda,lvs_3.1,icp-ms,thermo_(vg),icap_rq':                                 {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Thermo iCAP RQ'}},  
                'low_vol_sampler,derenda,mvs_6.1':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda MVS 6.1'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'low_vol_sampler,derenda,pns_16t,gc-ms':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda PNS 16T'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'low_vol_sampler,eccc_cair_lab,lvs1':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1'},                                                                                                                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Cannot find information about "eccc_cair_lab", assume method is ion chromatography after sampling.', 'measuring_assumption':True}},
                'low_vol_sampler,fai_instruments,swam_5a_dual_channel,ic,dionex,ics-2100':                    {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'FAI Swam 5a'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-2100'}},
                'low_vol_sampler,fai_instruments,swam_5a_dual_channel,ic,metrohm,881_compact_ic_pro':         {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'FAI Swam 5a'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                           'instrument':'Metrohm 881 compact IC pro'}},
                'low_vol_sampler,fai_instruments,swam_5a_dual_channel,ic,thermo_scientific,dionex_aquion':    {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'FAI Swam 5a'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo Dionex Aquion'}},
                'low_vol_sampler,fai_instruments,swam_5a_dual_channel,ic,thermo_scientific,ics-2100':         {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'FAI Swam 5a'},                                                                                                                                                                                                                                    'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-2100'}},
                'low_vol_sampler,icp-ms':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}, 
                'low_vol_sampler,leckel,lvs3':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'low_vol_sampler,leckel,lvs3,icp-ms,agilent,7700x':                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7700x'}},
                'low_vol_sampler,leckel,mvs6':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},    
                'low_vol_sampler,leckel,mvs6,icp-ms,agilent,7700x':                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7700x'}},
                'low_vol_sampler,leckel,seq47_50':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},  
                'low_vol_sampler,leckel,seq47_50,gc-ms':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'low_vol_sampler,leckel,seq47_50,gc-ms,agilent,5975b':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'low_vol_sampler,leckel,seq47_50,gc-ms,agilent,5977b':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
                'low_vol_sampler,leckel,seq47_50,ic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},  
                'low_vol_sampler,leckel,seq47_50,ic,dionex,dx-120':                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Dionex DX-120'}},  
                'low_vol_sampler,leckel,seq47_50,ic,metrohm,881_compact_ic_pro':                              {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Metrohm 881 compact IC pro'}},    
                'low_vol_sampler,leckel,seq47_50,ic,shimadzu,cdd-10avp':                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Shimadzu CDD-10Avp'}},             
                'low_vol_sampler,leckel,seq47_50,icp-aes,agilent_technologies,agilent_5100_dual_view_icp-oes':{'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',      'instrument':'Agilent 5100'}},              
                'low_vol_sampler,leckel,seq47_50,icp-hrms':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},  
                'low_vol_sampler,leckel,seq47_50,icp-ms,agilent,7700x':                                       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7700x'}}, 
                'low_vol_sampler,leckel,seq47_50,icp-ms,agilent,7800':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7800'}},  
                'low_vol_sampler,leckel,seq47_50,icp-ms,agilent,7900':                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Agilent 7900'}},  
                'low_vol_sampler,leckel,seq47_50,icp-ms,thermo_(vg),x_series':                                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                 'instrument':'Thermo XSERIES'}},
                'low_vol_sampler,rupprecht&patashnick,partisol2025':                                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},  
                'low_vol_sampler,rupprecht&patashnick,partisol2025,ic,dionex,ics-2000':                       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                           'instrument':'Thermo ICS-2000'}},  
                'low_vol_sampler,thermo,partisol2000i':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2000i'},                                                                                                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'low_vol_sampler,thermo,partisol2025':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},     
                'low_vol_sampler,umwettechnik_mcz_gmbh,micropns':                                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1',                                                                                                          'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'low_vol_sampler,umwettechnik_mcz_gmbh,micropns,icp-ms':                                      {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1',                                                                                                          'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',     'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}, 
                'lvs_denuder_tandem':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'denuder'},                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'lvs_qbq':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},
                'lvs_single':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},         
                'lvs_single_twin':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}},         
                'lvs_teflon':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                         'process_details':'Assume method is ion chromatography after sampling.',                                                'measuring_assumption':True}}, 
                'mini_trap':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                     'primary_sampling_assumption':True},   'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'sorbent trapping',                                     'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'preconcentration', 'preparation_assumption':True}, 'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                                                                                                               'process_details':'Assume method is cold vapour atomic fluorescence spectroscopy',                                      'measuring_assumption':True}},
                'nitrate_monitor':                                                                            {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                               'sampling_type':'continuous injection',   'process_details':'Internal file metadata suggests method is ion chromatography after sampling.',                       'measuring_assumption':True}},
                'online_ic':                                                                                  {                                                                                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter',                                                   'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                               'sampling_type':'continuous injection'}},                           
                'na' :                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}}} 
#-------------------------------------------                                                                                                                                                                                                                                                                                                                                                          
        elif parameter_details['matrix'] == 'aod':
            mapping = {}
            #moon_tracking_filter_radiometer
            #rotating_shadowband_spectral_radiometer,yes,mfr-7
            #sun_tracking_filter_radiometer
            #sun_tracking_filter_radiometer,carter-scott_design,spo1a
            #sun_tracking_filter_radiometer,middleton_solar,sp02
            #sun_tracking_filter_radiometer,middleton_solar,sp02-l
            #sun_tracking_filter_radiometer,prede,pom-02
#--------------------------------------------------------------
#EEA AIRBASE
#--------------------------------------------------------------
    elif run_details['network'] == 'EEA_AIRBASE':
        if run_details['parameter'] == 'O3':
            mapping = {
            '':                                                                                                                          {                                                                                               'measurement':{'methodology':'unknown'}},
            '2008a_ultraviolet_(uv)_photometry_automatic':                                                                               {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument 2008a, therefore assume uses instrument with ultraviolet photometry method as stated.',                     'measuring_assumption':True}},
            '342_m_ultraviolet_(uv)_photometry_automatic':                                                                               {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},                                                                           
            'amibrack_ultraviolet_(uv)_photometry_automatic':                                                                            {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument Amibrack, therefore assume uses instrument with ultraviolet photometry method as stated.',                  'measuring_assumption':True}},
            'analysis_automation_mod._427_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet photometry method as stated.',       'measuring_assumption':True}},
            'analysis_automation_ultraviolet_(uv)_photometry_automatic':                                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet photometry method as stated.',       'measuring_assumption':True}},
            'c.s.i._3.100_ultraviolet_(uv)_photometry_automatic':                                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'CSI 3100'}},
            'columbia_scientific_ic_3100_ultraviolet_(uv)_photometry_automatic':                                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'CSI 3100'}},
            'dani_environment_1003_ah_ultraviolet_(uv)_photometry_automatic':                                                            {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-AH'}},
            'dasibi_1003-pc_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-PC'}},
            'dasibi_1003-rs_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-RS'}},
            'dasibi_1003_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                             {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-AH',                                                    'process_details':'Make assumption measuring instrument is Dasibi 1003-AH',                                                                           'measuring_assumption':True}},
            'dasibi_1008-ah_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH'}},
            'dasibi_1008-pc_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-PC'}},
            'dasibi_1008-rs_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-RS'}},                                
            'dasibi_1008_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                             {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH',                                                    'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                           'measuring_assumption':True}},
            'dasibi_1108_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                             {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1108 W/GEN',                                                 'process_details':'Make assumption measuring instrument is Dasibi 1108 W/GEN',                                                                        'measuring_assumption':True}},
            'dasibi_5014_ultraviolet_(uv)_photometry_automatic':                                                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'SIR S-5014',                                                        'process_details':'Make assumption measuring instrument is SIR S-5014',                                                                               'measuring_assumption':True}},
            'dasibi_ultraviolet_(uv)_photometry_automatic':                                                                              {                                                                                               'measurement':{'methodology':'ultraviolet photometry'}},
            'doas_ultraviolet_(uv)_photometry_automatic':                                                                                {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption method is ultraviolet photometry and not DOAS',                                                                    'measuring_assumption':True}},
            'environics_300__environics_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environics 300'}},
            'environment_1003_ah_ultraviolet_(uv)_photometry_automatic':                                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-AH'}},
            'environment_1003rs_ultraviolet_(uv)_photometry_automatic':                                                                  {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-RS'}},
            'environnement_s.a._model_o331m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_automatic':                                    {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O331M'}}, 
            'environnement_s.a._model_o341m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_automatic':                                    {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_ultraviolet_(uv)_photometry_automatic':                                     {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}}, 
            'environnement_s.a._sanoa_multigas_longpath_monitoring_system_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                               'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'Environnement SA SANOA'}},
            'horiba_model_apna_350e_nox_analyser_chemiluminescence_automatic':                                                           {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA350'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                                            {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},
            'horiba_model_apoa_300_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                   {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA300'}},
            'horiba_model_apoa_350_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                   {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                   {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'iseo_argopol-sam-xair_ultraviolet_(uv)_photometry_automatic':                                                               {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument ISEO argopol sam-xair, therefore assume uses instrument with ultraviolet photometry method as stated.',     'measuring_assumption':True}},
            'mcv_48-auv_ultraviolet_(uv)_photometry_automatic':                                                                          {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'MCV 48AUV'}},
            'mlu_ultraviolet_(uv)_photometry_automatic':                                                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Airpointer-O3'}},
            'monitor_labs_model_8810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_9800_ultraviolet_(uv)_photometry_automatic':                                                             {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument ML 9800, assume method is ultraviolet photometry as stated',                                                'measuring_assumption':True}},       
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810b_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810B'}},
            'monitor_labs_model_9811_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9811'}},
            'monitor_labs_model_9812_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9812'}},
            'monitor_labs_model_9830_co_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne ML9830 instrument uses',                'measuring_assumption':True}},
            'monitor_labs_undetermined_ultraviolet_(uv)_photometry_automatic':                                                           {                                                                                               'measurement':{'methodology':'ultraviolet photometry'}},
            'monitor_labs_undetermined_uv_fluorescence_automatic':                                                                       {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                               'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'orion_ultraviolet_(uv)_photometry_automatic':                                                                               {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet photometry method as stated.',                     'measuring_assumption':True}},
            'orion_ultraviolet_absorption_not-automatic':                                                                                {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet photometry method as stated.',                     'measuring_assumption':True}},
            'philips_k50094_api_400_ultraviolet_(uv)_photometry_automatic':                                                              {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Philips K50094'}},
            'philips_k50110_00_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Philips K50110'}},
            'philips_pw977100_chemiluminescence_automatic':                                                                              {                                                                                               'measurement':{'methodology':'ethylene chemiluminescence',                                                                                                                    'process_details':'Cannot find information regarding Philips PW977100, assume ethylene chemiluminescence is method',                                  'measuring_assumption':True}},  
            's-5014_ultraviolet_(uv)_photometry_automatic':                                                                              {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'SIR S-5014'}},
            'sfi_o342m_ultraviolet_(uv)_photometry_automatic':                                                                           {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}}, 
            'teledyne_api_200a_chemiluminescent_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                     {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Teledyne 200A instrument uses',     'measuring_assumption':True}},
            'teledyne_api_400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_uv_fluorescence_automatic':                                                     {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400a_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_undertermined_chemiluminescence_automatic':                                                                    {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                             'process_details':'Make assumption uses internal molybdenum converter',                                                                               'measuring_assumption':True}},
            'thermo_49_cps_ozone_primary_standard_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49CPS'}},
            'thermo_49c_uv_fluorescence_automatic':                                                                                      {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C',                                                        'process_details':'Make assumption the method is ultraviolet photometry as reported instrument uses and not ultraviolet fluorescence',                'measuring_assumption':True}},
            'thermo_model_42_no_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                                     {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Thermo 42 instrument uses',         'measuring_assumption':True}},
            'thermo_model_43_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Thermo 43 instrument uses',  'measuring_assumption':True}},
            'thermo_model_43s_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                       {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Thermo 43S instrument uses', 'measuring_assumption':True}},
            'thermo_model_48_co_analyser_flame_photometry_automatic':                                                                    {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48'}},
            'thermo_model_48_co_analyser_ultraviolet_(uv)_photometry_automatic':                                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                        'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Thermo 48 instrument uses',                      'measuring_assumption':True}},
            'thermo_model_49_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}}, 
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo_model_49w_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49w'}},
            'too_generic_ultraviolet_(uv)_photometry_automatic':                                                                         {                                                                                               'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_uv_fluorescence_automatic':                                                                                     {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_chemiluminescence_automatic':                                                                                       {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Make assumption uses internal molybdenum converter',                                                                              'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                                          {                                                                                               'measurement':{'methodology':'gas chromatography - unknown detection',                                                               'sampling_type':'continuous injection'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                     {                                                                                               'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_gas_chromatography_(ref)_automatic':                                                                                {                                                                                               'measurement':{'methodology':'gas chromatography - unknown detection',                                                               'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                            {                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                      'sampling_type':'continuous injection'}},
            'unknown_spectrophotometry_automatic':                                                                                       {'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                             {                                                                                               'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                                 {                                                                                               'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                             {                                                                                               'measurement':{'methodology':'unknown'}},
            'unknown_uv_absorption_(ref)_not-automatic':                                                                                 {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                                                                                                         'process_details':'Make assumption method is ultraviolet photometry',                                                                                'measuring_assumption':True}},
            'unknown_uv_fluorescence_automatic':                                                                                         {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence'}}}
#------------------------------------------- 
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {   
            '':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'airpointer_chemiluminescence_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Airpointer-NOx'}},      
            'ambirack_chemiluminescence_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Cannot find instrument Ambirack, therefore assumes uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                    'measuring_assumption':True}},
            'analysis_automation_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Cannot find instrument Analysis Automation, therefore assumes uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',         'measuring_assumption':True}},
            'analysis_automation_mod._447_chemiluminescence_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Cannot find instrument Analysis Automation Mod 447, therefore assumes uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).', 'measuring_assumption':True}},
            'aspirator_neda_griess-ylosway_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},
            'bendix_chemiluminescence_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Bendix 8101-C',                                                      'process_details':'Assume instrument is Bendix 8101-C',                                                                                                                                           'measuring_assumption':True}},
            'bendix_combustion_engineering_model_8101-c_oxides_of_nitrogen_analyze_chemiluminescence_automatic':                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Bendix 8101-C'}},
            'bubbler_24_h_griess-saltzman_reaction_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry',                                                                                                                              'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                                           'measuring_assumption':True}},
            'bubbler_24_h_spectrometry_not-automatic':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'bubbler_24_h_spectrophotometry_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'columbia_scientific_industries_models_1600_chemiluminescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'CSI 1600'}},
            'contiflo_ultraviolet_(uv)_photometry_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                                                                                                         'process_details':'Could not find instrument Contiflo, assume method is ultraviolet photometry as stated',                                                                                        'measuring_assumption':True}},                    
            'da_verificare_chemiluminescence_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_not-automatic':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                        'measurement':{'methodology':'ion chromatography'}},
            'dasibi_2108_nox_analyser_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Dasibi 2108'}},                                         
            'doas_ultraviolet_(uv)_photometry_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                                                                                                   'process_details':'Asssume method is DOAS not ultraviolet photometry',                                                                                                                            'measuring_assumption':True}},
            'ec9843_chemiluminescence_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',           'instrument':'Ecotech EC9843'}},
            'eco_physics_cld_700_al_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Ecophysics CLD700AL'}},
            'environment_chemiluminescence_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'environment_zc_32m_chemiluminescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC32M'}},
            'environnement_ac_30m_chemiluminescence_not-automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_af21m_so2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet fluorescence as Environnement SA AF21M instrument uses',                                    'measuring_assumption':True}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF21M'}},    
            'environnement_s.a._model_af22m_so2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet fluorescence as Environnement SA AF22M instrument uses',                                    'measuring_assumption':True}},
            'environnement_s.a._sanoa_multigas_longpath_monitoring_system_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'Environnement SA SANOA'}},
            'environnement_voc71m_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                               {                                                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                      'sampling_type':'continuous injection'}},
            'filter_pack_griess-saltzman_reaction_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                'preparation_2':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                 'measurement':{'methodology':'spectrophotometry',                                                                                                                              'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                                           'measuring_assumption':True}},
            'filter_pack_spectrophotometry_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry'}},
            'filter_sampling_photometry_not-automatic':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                     'preparation_2':{'sample_preparation_type':'reagent reaction',                                          'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'filter_tgs-ansa_not-automatic':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                     'preparation_2':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TGS-ANSA'},                                        'measurement':{'methodology':'spectrophotometry'}},
            'glass_filter_spectrophotometry_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'glass_sinter_spectrophotometry_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'glass_tubes_spectrophotometry_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'horiba_apna_350_e_chemiluminescence_not-automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA350'}},
            'horiba_model_apha_360e_hydrocarbons_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},
            'horiba_model_apna_300_nox_analyser_chemiluminescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA300'}},     
            'horiba_model_apna_300e_nox_analyser_chemiluminescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA300'}},
            'horiba_model_apna_350_nox_analyser_chemiluminescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA350'}},   
            'horiba_model_apna_350e_nox_analyser_chemiluminescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA350'}},   
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},   
            'horiba_model_apna_360_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},  
            'horiba_model_apna_360_nox_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},  
            'horiba_model_apna_360e_nox_analyser_chemiluminescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},  
            'horiba_model_apna_370_nox_analyser_chemiluminescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA370'}},  
            'impinger_triethylonamine_unknown_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TEA'},                                                                                                                                                                                       'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_griess-saltzman_reaction_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'lap_884_chemiluminescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Cannot find instrument Lap 884, assume method is chemiluminescence as stated (assume internal molybdenum converter)',                                                          'measuring_assumption':True}},               
            "lipinski's_aspirator_agt24_griess-saltzman_reaction_not-automatic":                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},    
            "lipinski's_aspirator_griess-saltzman_reaction_not-automatic":                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},    
            "lipinski's_aspirator_spectrophotometry_not-automatic":                                                                      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'spectrophotometry'}},    
            "lipinski's_aspirator_unknown_not-automatic":                                                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'spectrophotometry',                                                                                                                              'process_details':'Asssume method is spectrophotometry',                                                                                                                                          'measuring_assumption':True}},    
            'mcv_30-ql_chemiluminescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'MCV 30QL'}},      
            'meloy_s1600_chemiluminescence_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'CSI 1600'}},
            'monitor_labs_model_8440_nox_analyser_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML8440'}},                                                         
            'monitor_labs_model_8440_nox_analyser_chromatography_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML8440'}},  
            'monitor_labs_model_8840_nox_analyser_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML8840'}},                                                  
            'monitor_labs_model_8841_nox_analyser_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML8841'}}, 
            'monitor_labs_model_8941a_nox_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML8941A'}},
            'monitor_labs_model_9841_nox_analyser_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML9841'}}, 
            'monitor_labs_model_9841a_nox_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML9841A'}}, 
            'monitor_labs_model_9841b_nox_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML9841B'}}, 
            'monitor_labs_model_9850_so2_analyser_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet flurescence as Teledyne ML9850 instrument uses',                                            'measuring_assumption':True}},
            'monitor_labs_undetermined_chemiluminescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'monitor_labs_undetermined_ultraviolet_(uv)_photometry_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'orion_chemiluminescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Cannot find instrument Orion, therefore assume uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                        'measuring_assumption':True}},
            'palmes_passive_sampler_automatic':                                                                                          {'primary_sampling':{'sampling_type':'passive'},                                                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                         'measurement':{'methodology':'spectrophotometry',                                                                                                                              'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                                                           'measuring_assumption':True}},
            'philips_k50034_api_200a_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Philips K50034'}},
            'philips_k50102_no_chemiluminescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Philips K50102'}},
            'philips_k50109_00_gas_filter_correlation_co_analyser_chemiluminescence_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence as stated and not NDIR as instrument uses (assume internal molybdenum converter)',                                                          'measuring_assumption':True}},
            'philips_k50235_00_no-nox-no2_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Philips K50235'}},
            'philips_pw9762_00_no_no2_nox_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Philips PW9762/02'}},
            'philips_42_chemiluminescence_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42'}},                     
            'photocolorimeter_spectrophotometry_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'spectrophotometry'}},
            's-5012_chemiluminescence_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'SIR S-5012'}},
            'sequential_air_sampler,_type_ss2000._nai-impregnated_glass_sinters_spectrophotometry_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                         'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry'}},
            'sequential_sampler_griess-saltzman_reaction_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                     'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},
            'sfi_ac32m_chemiluminescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC32M'}},      
            'tecan_chemiluminescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'No specific instrument, but assume method is chemiuluminescence as stated (internal molybdenum converter)',                                                                    'measuring_assumption':True}},        
            'tecan_cld_502_chemiluminescence_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Tecan CLD 502'}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet flurescence as Teledyne 100A instrument uses',                                              'measuring_assumption':True}},
            'teledyne_api_101_uv_fluorescent_so2_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet flurescence as Teledyne 101 instrument uses',                                               'measuring_assumption':True}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                                {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_beta_ray_attenuation_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200A'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200E'}},
            'teledyne_api_400_uv_photometric_o3_analyser_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet photometry as Teledyne 400 instrument uses',                                                'measuring_assumption':True}},
            'teledyne_api_undertermined_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'teledyne_api_undertermined_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry'}},
            'thermo_electron_instruments_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'thermo_environmental_instruments_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'thermo_model_14b_chemiluminescence_no-no2-nox_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 14B'}},
            'thermo_model_14b_e_chemiluminescence_no-no2-nox_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 14B/E'}},
            'thermo_model_42_no_nox_analyser_chemiluminescence_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42'}},
            'thermo_model_42_no_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42'}},
            'thermo_model_42_no_nox_analyser_uv_fluorescence_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42'}},
            'thermo_model_42c-tl_(trace_level_nox)_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42C-TL'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42C'}},
            'thermo_model_42i-tl_(trace_level_nox)_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42i-TL'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42i'}},
            'thermo_model_42s_no_nox_analyser_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42S'}},
            'thermo_model_42w_no_nox_analyser_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42W'}},
            'thermo_model_43_so2_analyser_chemiluminescence_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not ultraviolet flurescence as Thermo 43 instrument uses',                                                  'measuring_assumption':True}},                          
            'thermo_model_43a_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43A'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43C'}},
            'thermo_model_48_co_analyser_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume method is chemiluminescence (internal molybdenum converter) not NDIR as Thermo 48 instrument uses',                                                                     'measuring_assumption':True}},                          
            'thermo_model_48_co_analyser_flame_photometry_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48'}},
            'too_generic_chemiluminescence_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                      'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                                                 'measuring_assumption':True}},
            'unknown_chemiluminescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'unknown_chemiluminescence_not-automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Assume internal molybdenum converter',                                                                                                                                         'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - unknown detection',                                                               'sampling_type':'continuous injection'}},
            'unknown_colorimetry_automatic':                                                                                             {                                                                                                                                  'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'colorimetry',                                                                                          'sampling_type':'continuous injection'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_diffusive_sampler_automatic':                                                                                       {                                                                                                                                  'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection',   'process_details':'Make assumption method is spectrophotometry',                                                                                                                                  'measuring_assumption':True}},
            'unknown_griess-saltzman_reaction_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                           'measurement':{'methodology':'spectrophotometry'}},
            'unknown_spectrophotometry_automatic':                                                                                       {                                                                                                                                  'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection'}},
            'unknown_spectrophotometry_not-automatic':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                         'measurement':{'methodology':'spectrophotometry'}},
            'unknown_uv_fluorescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}}}   
#------------------------------------------- 
        elif run_details['parameter'] == 'CO':
            mapping = {
            '':                                                                                                            {                                                                                               'measurement':{'methodology':'unknown'}},
            'aal_407_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Cannot find instrument AAL 407, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation)',             'measuring_assumption':True}},
            'ambirack_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                              {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Cannot find instrument Ambirack, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation)',            'measuring_assumption':True}},
            'analysis_automation_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation)', 'measuring_assumption':True}},
            'co_analayzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                          {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'dasibi_3008_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                               {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Dasibi 3008'}},
            'environment_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'environnement_s.a._model_co10m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO10M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co12m_co_analyzer_chemiluminescence_automatic':                                      {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO12M'}},
            'environnement_s.a._model_co12m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO12M'}},
            'fuji_zrc_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                              {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Fuji ZRC'}},
            'horiba_model_apma_300_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA300'}},
            'horiba_model_apma_300e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                    {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA300'}},
            'horiba_model_apma_350_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350_co_analyser_unknown_automatic':                                                         {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                    {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA350'}},
            'horiba_model_apma_360_co_analyser_infrared_gas_filter_correlation_automatic':                                 {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360ce_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                    {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA370'}},
            'maihak_unor_6n_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                        {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Maihak Unor 6N'}},
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_non-dispersive_infrared_spectroscopy_(ndir)_automatic':{                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as instrument uses',                                      'measuring_assumption':True}},                  
            'mlu_300_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Airpointer-CO'}},
            'mlu_beta_ray_attenuation_automatic':                                                                          {                                                                                               'measurement':{'methodology':'beta-attenuation', 'instrument':'Airpointer-PMBA'}},
            'mlu_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Airpointer-CO'}},                                          
            'monitor_labs_model_8830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8831_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8831'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                   {                                                                                               'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9830b_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                  {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9841a_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                 {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not chemiluminescence as instrument uses',                                             'measuring_assumption':True}}, 
            'monitor_labs_model_9841b_nox_analyser_chemiluminescence_automatic':                                           {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne ML9841B'}},
            'monitor_labs_undetermined_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                             {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                 {                                                                                               'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'orion_beta_ray_attenuation_automatic':                                                                        {                                                                                               'measurement':{'methodology':'beta-attenuation',                                                                                                                               'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                                         'measuring_assumption':True}},
            'orion_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                 {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Cannot find instrument Orion, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',              'measuring_assumption':True}},
            'philips_48_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                            {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48',                                                          'process_details':'Assume instrument is Thermo 48',                                                                                                          'measuring_assumption':True}},
            'philips_k50093_api_300a_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                               {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Philips K50093'}},
            'philips_k50109_00_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Philips K50109'}},
            's-5006_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'SIR S-5006'}},           
            'sfi_co12m_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO12M'}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':          {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as instrument uses',                                      'measuring_assumption':True}}, 
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                 {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':       {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not chemiluminescence as instrument uses',                                             'measuring_assumption':True}}, 
            'teledyne_api_300_gas_filter_correlation_co_analyser_flame_photometry_automatic':                              {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_unknown_automatic':                                       {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300'}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation)',       'measuring_assumption':True}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_infrared_gas_filter_correlation_automatic':              {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300E'}},
            'teledyne_api_undertermined_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                            {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'thermo_electron_instruments_48c_infrared_gas_filter_correlation_automatic':                                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C'}},
            'thermo_electron_instruments_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'thermo_model_42_no_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                       {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not chemiluminescence as instrument uses',                                             'measuring_assumption':True}}, 
            'thermo_model_43_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43'}},                                       
            'thermo_model_43_so2_analyser_uv_fluorescence_automatic':                                                      {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43'}}, 
            'thermo_model_43a_so2_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                         {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as instrument uses',                                      'measuring_assumption':True}}, 
            'thermo_model_48_co_analyser_flame_photometry_automatic':                                                      {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48'}},
            'thermo_model_48_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48'}},
            'thermo_model_48c_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                          {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C'}},
            'thermo_model_48i_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                          {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48i'}},
            'thermo_model_48w_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                          {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48W'}},
            'thermo_model_49_o3_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Assume method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as instrument uses',                                        'measuring_assumption':True}}, 
            'too_generic_chemiluminescence_automatic':                                                                     {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Make assumption uses internal molybdenum converter',                                                                                      'measuring_assumption':True}},
            'too_generic_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'unknown_chemiluminescence_automatic':                                                                         {                                                                                               'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                              'process_details':'Make assumption uses internal molybdenum converter',                                                                                      'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                            {                                                                                               'measurement':{'methodology':'gas chromatography - unknown detection',                                                               'sampling_type':'continuous injection'}},
            'unknown_coulometry_automatic':                                                                                {                                                                                               'measurement':{'methodology':'coulometry'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                              {                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                      'sampling_type':'continuous injection'}},
            'unknown_infrared_absorption_not-automatic':                                                                   {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'unknown_infrared_gas_filter_correlation_automatic':                                                           {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'unknown_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                          'measuring_assumption':True}},
            'unknown_spectrophotometry_automatic':                                                                         {'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                    'sampling_type':'continuous injection'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                               {                                                                                               'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                   {                                                                                               'measurement':{'methodology':'unknown'}},
            'unknown_uv_fluorescence_automatic':                                                                           {                                                                                               'measurement':{'methodology':'ultraviolet fluorescence'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'SO2':
            mapping = {    
            '':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            '47_uv_fluorescence_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument 47, therefore assume uses instrument with ultraviolet fluorescence method as stated.',                    'measuring_assumption':True}},              
            'ambirack_uv_fluorescence_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Ambirack, therefore assume uses instrument with ultraviolet fluorescence method as stated.',              'measuring_assumption':True}},
            'analysis_automation_mod._477_uv_fluorescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet fluorescence method as stated.',   'measuring_assumption':True}},
            'analysis_automation_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet fluorescence method as stated.',   'measuring_assumption':True}},
            'ats_10.1_uv_fluorescence_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument ATS 10.1, therefore assume uses instrument with ultraviolet fluorescence method as stated.',              'measuring_assumption':True}},
            'bendix_8302_flame_photometry_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                                 'instrument':'Bendix 8302'}}, 
            'bubbler_24_h_colorimetric_with_p-rosaniline_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry'}},
            'bubbler_24_h_ion_chromatography_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                                                  'measurement':{'methodology':'ion chromatography'}}, 
            'bubbler_24_h_photometry_not-automatic':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}}, 
            'bubbler_24_h_titration_not-automatic':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                                                                                  'measurement':{'methodology':'titration'}},
            'columbia_scientific_ic_sa700_uv_fluorescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Meloy SA700'}},
            'dani_environment_af_21_m_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF21M'}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_not-automatic':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                                                       'measurement':{'methodology':'ion chromatography'}},
            'dasibi_4108_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Dasibi 4108'}},
            'dasibi_4108_so2_analyser_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Dasibi 4108'}},
            'diffusion_tube_ion_chromatography_not-automatic':                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                    'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                                       'measurement':{'methodology':'ion chromatography'}},
            'doas_ultraviolet_(uv)_photometry_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                                                                                                    'process_details':'Asssume method is DOAS not ultraviolet photometry',                                                                              'measuring_assumption':True}},
            'enc_37_unknown_not-automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}}, 
            'environment_s.a._af_22_uv_fluorescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}},
            'environment_uv_fluorescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_af21m_so2_analyzer_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af22m_so2_analyzer_chemiluminescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af22m_so2_analyzer_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv_fluorescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._sanoa_multigas_longpath_monitoring_system_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'Environnement SA SANOA'}},
            'filter_2pack_ion_chromatography_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'2-Stage Filter Pack'},                                                                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},   
            'filter_3pack_ion_chromatography_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'},                                                                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'filter_ion_chromatography_not-automatic':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography'}},   
            'filter_pack_spectrophotometry_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                                                       'measurement':{'methodology':'spectrophotometry'}},
            'filter_sampling_capillary_electrophoresis_not-automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                            'measurement':{'methodology':'capillary electrophoresis'}},
            'filter_sampling_ion_chromatography_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography'}},   
            'filter_thorin_not-automatic':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                     'preparation_2':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                                              'measurement':{'methodology':'titration'}},
            'filter_unknown_not-automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                              'process_details':'Assume method is ion chromatography after sampling.',                                                                            'measuring_assumption':True}},   
            'gas_meter_unknown_not-automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'horiba_model_apna_360_nox_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA360'}},
            'horiba_model_apoa_350_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apsa_300_so2_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA300'}},
            'horiba_model_apsa_350_so2_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA350'}},
            'horiba_model_apsa_350e_so2_analyser_uv_fluorescence_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA350'}},
            'horiba_model_apsa_360_so2_analyser_unknown_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360_so2_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360a_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360e_so2_analyser_uv_fluorescence_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Horiba APSA370'}},
            'impinger_hydrogen_peroxide_tcm_unknown_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                                                                                                                                                                                        'measurement':{'methodology':'titration'}},
            'impregnated_filter_ion_chromatography_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                                                  'measurement':{'methodology':'ion chromatography'}},
            'lap_885_uv_fluorescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Lap 885, therefore assume uses instrument with ultraviolet fluorescence method as stated.',               'measuring_assumption':True}},
            "lipinski's_aspirator_agt24_colorimetric_with_p-rosaniline_not-automatic":                                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry'}},
            "lipinski's_aspirator_colorimetric_with_p-rosaniline_not-automatic":                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry'}},
            "lipinski's_aspirator_spectrophotometry_not-automatic":                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry'}},    
            "lipinski's_aspirator_unknown_not-automatic":                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Lipinski AGT24'},                                      'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry',                                                                                                                               'process_details':'Asssume method is spectrophotometry',                                                                                            'measuring_assumption':True}},    
            'macleod_collector_sf1_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Macleod collector SF1, therefore assume uses instrument with ultraviolet fluorescence method as stated.', 'measuring_assumption':True}},
            'mcv_64-fuv_uv_fluorescence_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'MCV 64FUV'}},                       
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_uv_fluorescence_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Meloy SA700'}},
            'monitor_labs_model_8810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_8850_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850_so2_analyser_uv_fluorescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850m_so2_analyser_uv_fluorescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML8850M'}},
            'monitor_labs_model_8850s_so2_analyser_uv_fluorescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML8850S'}},
            'monitor_labs_model_9830_co_analyser_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9840_nox_analyser_uv_fluorescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Assume method is ultraviolet fluorescence as stated, and not chemiluminescence as Teledyne ML9840 instrument uses',              'measuring_assumption':True}},         
            'monitor_labs_model_9850_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850b_so2_analyser_uv_fluorescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne ML9850B'}},
            'monitor_labs_undetermined_ultraviolet_(uv)_photometry_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry'}},
            'monitor_labs_undetermined_uv_fluorescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'nilu_filter_pack_ion_chromatography_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                                                       'measurement':{'methodology':'ion chromatography'}},
            'oecd_spectrophotometry_not-automatic':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'orion_chemiluminescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                  'process_details':'Cannot find instrument Orion, therefore assume uses gas chromatography - sulphur chemiluminescence.',                            'measuring_assumption':True}},
            'orion_uv_fluorescence_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet fluorescence method as stated.',                 'measuring_assumption':True}},
            'philips_43a_uv_fluorescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43A'}},
            'philips_coulometry_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry'}},
            'philips_k50033_api_100e_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Philips K50033'}},
            'philips_k50206_00_pulsed_fluorescence_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Philips K50206/00'}},
            'philips_k50206_00_pulsed_fluorescence_so2_analyser_uv_fluorescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Philips K50206/00'}},
            'philips_pw9755_coulometry_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                                  'instrument':'Philips PW9755'}},
            'philips_uv_fluorescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'photocolorimeter_spectrophotometry_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry'}},
            'ptfe_tube_-_gf_filter_unknown_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                                                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                                                              'process_details':'Assume method is ion chromatography after sampling.',                                                                            'measuring_assumption':True}}, 
            's-5001_uv_fluorescence_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'SIR S-5001'}},
            'sequential_air_sampler,_type_ek._koh-impregnated_whatman_40_filter_ion_chromatography_not-automatic':                       {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'NILU EK'},                                             'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                                                                                                  'measurement':{'methodology':'ion chromatography'}},
            'seres_sf2_uv_fluorescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Seres SF2000G',                                                       'process_details':'Cannot find Seres SF2 instrument, assume is Seres SF2000G instrument.',                                                          'measuring_assumption':True}},
            'seres_sf300_uv_fluorescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Cannot find Seres SF300 instrument, assume method is ultraviolet fluorescence as stated',                                        'measuring_assumption':True}},                           
            'sfi_af22m_uv_fluorescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}},
            'teflon_bag+desorption_tube_colorimetric_with_p-rosaniline_not-automatic':                                                   {'primary_sampling':{'sampling_type':'high volume continuous',                                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'bag'},                                                                                        'preparation_2':{'sample_preparation_type':'sorbent trapping'}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'}, 'measurement':{'methodology':'colorimetry'}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100'}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100A'}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100A'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100E'}},
            'teledyne_api_101_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 101'}},
            'teledyne_api_101a_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 101A'}},   
            'teledyne_api_200a_chemiluminescent_nox_analyser_ultraviolet_(uv)_photometry_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200A'}},
            'teledyne_api_400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'thermo_450i_uv_fluorescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 450i'}},
            'thermo_electron_instruments_uv_fluorescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}}, 
            'thermo_environmental_instruments_af22m_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Environnement SA AF22M'}}, 
            'thermo_model_42_no_nox_analyser_uv_fluorescence_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Assume method is ultraviolet fluorescence not chemiluminescence as Thermo 42 instrument uses',                                   'measuring_assumption':True}},  
            'thermo_model_43_so2_analyser_uv_fluorescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43'}},
            'thermo_model_43a_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43A'}},
            'thermo_model_43b_so2_analyser_ultraviolet_(uv)_photometry_automatic':                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43B'}},
            'thermo_model_43b_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43B'}},
            'thermo_model_43c-tl_so2_analyser_uv_fluorescence_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43C-TLE'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43C'}},
            'thermo_model_43h_pulsed_fluorescence_so2_analyser_uv_fluorescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43H'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv_fluorescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43i'}},
            'thermo_model_43s_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43S'}},
            'thermo_model_43w_so2_analyser_uv_fluorescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43W'}},
            'thermo_model_450i_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 450i'}},
            'thermo_model_48_co_analyser_ultraviolet_(uv)_photometry_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48'}},
            'thermo_model_48c_co_analyser_uv_fluorescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Assume method is ultraviolet fluorescence not NDIR as Thermo 48C instrument uses',                                               'measuring_assumption':True}},  
            'thermo_model_49_o3_analyser_uv_fluorescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                        'process_details':'Assume method is ultraviolet fluorescence not ultraviolet photometry as Thermo 49 instrument uses',                              'measuring_assumption':True}},                                                 
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo_so2_43i_chemiluminescence_not-automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43i',                                                          'process_details':'Assume method is ultraviolet fluorescence as Thermo 43i instrument uses not chemiluminescence as stated',                        'measuring_assumption':True}},     
            'too_generic_uv_fluorescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'total_ans_filter_pack_ece-emep_ion_chromatography_not-automatic':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                                                                                                       'measurement':{'methodology':'ion chromatography'}},
            'unknown_chemiluminescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                               'process_details':'Assume internal molybdenum converter',                                                                                           'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - unknown detection',                                                                'sampling_type':'continuous injection'}},
            'unknown_colorimetric_with_p-rosaniline_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry'}},
            'unknown_conductimetry_not-automatic':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True},                                                                                                                                                                                                                                                                                                                  'measurement':{'methodology':'conductimetry',                                                                                         'sampling_type':'injection'}},
            'unknown_coulometry_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_flame_photometry_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                       'sampling_type':'continuous injection'}},
            'unknown_isotope_dilution_not-automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                     'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                 'measuring_assumption':True}},
            'unknown_spectrophotometry_automatic':                                                                                       {                                                                                                                                   'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry',                                                                                     'sampling_type':'continuous injection'}},
            'unknown_spectrophotometry_not-automatic':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'spectrophotometry'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_uv_fluorescence_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_uv_fluorescence_not-automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'waters_chromatography_automatic':                                                                                           {                                                                                                                                   'preparation_1':{'sample_preparation_type':'filter',                                                      'preparation_assumption':True},                                                                                                                                                                        'measurement':{'methodology':'ion chromatography',                                          'instrument':'Waters Alliance',           'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography',                                                                                            'measuring_assumption':True}}, 
            'west-gaeke_reflectometry_not-automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif run_details['parameter'] in ['CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','NMVOC','NMHC','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE']:
            mapping = {
            '':                                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            '8610_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'af_20_m_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                  'measurement':{'methodology':'gas chromatography - mass spectrometry'}},  
            'airmotec_airmoscan2000_chromatography_automatic':                                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'airmotec_btx-1000_flame_photometry_automatic':                                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                    'measuring_assumption':True}},
            'airmotec_gas_chromatography_with_photo_ionization_detector_automatic':                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'airmotec_hc1010_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'airmozone_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'alpha_401_chromatography_automatic':                                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection',   'process_details':'Evidence that method uses photoionisation detection'}},
            'ama_gc5000_btx_type_fid_gas_chromatography_with_photo_ionization_detector_automatic':                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'analysis_automation_limited_526_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'btx_2000_gas_chromatography_with_photo_ionization_detector_automatic':                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'byron_101_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'byron_302_chromatography_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},    
            'carbopack_b_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask'},                                                                                      'preparation_2':{'sample_preparation_type':'preconcentration'},  'measurement':{'methodology':'gas chromatography - mass spectrometry'}}, 
            'carbopack_b_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                          'sampling_type':'continuous injection'}},
            'carbotap_tube_300_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_not-automatic':                                                    {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'},  'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}},
            'chromatosud_hc_-_1000_btx_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'chromosorb_106_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'},  'measurement':{'methodology':'gas chromatography - mass spectrometry'}}, 
            'chrompack_btx_cp7001_monitor_chromatography_automatic':                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'chrompack_btx_cp7001_monitor_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'chrompack_cp9000_flame_ionization_detection_(fid)_not-automatic':                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'},  'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'chrompack_voc-air_chromatography_not-automatic':                                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'},  'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'cpv-cov_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'dani_410_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'dani_btx_unknown_automatic':                                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'danish_lmp_filterpack_spm_collector_gravimetric_analysis_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'dasibi_2108_nox_analyser_chemiluminescence_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                              'instrument':'Dasibi 2108'}},
            'dasibi_7001_chromatography_automatic':                                                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection',   'process_details':'Assume instrument is Chrompack 7001, not Dasabi 7001',                                                         'measuring_assumption':True}},
            'desconocido_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'diffusive_tenaxta_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                             {'primary_sampling':{'sampling_type':'passive',                                                                                                                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'digitel_active_sampler_chromatography_not-automatic':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                  'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'digitel_active_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                  'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'draeger_niosh_(active_sampling)_chromatography_not-automatic':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection'}}, 
            'environment_hc_51m_chromatography_automatic':                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},           
            'environment_sa_collector_fi15_potentiometry_not-automatic':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True},                                                                                                                                                                                                            'measurement':{'methodology':'conductimetry',                                                                                                   'sampling_type':'injection',              'process_details':'Cannot find instrument Environment fi15, assume method is conductimetry',                                      'measuring_assumption':True}},
            'environnement_voc71m_chromatography_automatic':                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID',                                                                                      'measuring_assumption':True}},
            'environnement_voc71m_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'environnement_voc71m_gas_chromatography_with_photo_ionization_detector_automatic':                                                                {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'filter_sampling_gas_chromotography_(gc)_not-automatic':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                      'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'gc_varian_3400_chromatography_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'gc_varian_chromatography_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'hp_5890_2_-_hewlett_packard_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},  
            'hp-6890_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'horiba_model_apha_360_hydrocarbons_analyser_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA360'}},
            'horiba_model_apha_360e_hydrocarbons_analyser_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                             {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA360'}}, 
            "lipinski's_aspirator_chromatography_not-automatic":                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Lipinski AGT24'},                                                                                                                                 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                             'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'markes_international_passive_sampler_automatic':                                                                                                  {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'markes_thermal_desorber_+_agilent_gas_chromatograph_mass_spectrometer_gas_chromatography_+_mass_spectrometry_gc-ms_after_solvent_o_not-automatic':{'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'mlu_950-btx_gas_chromatography_with_photo_ionization_detector_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'monitor_labs_undetermined_chromatography_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},   
            'not-automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500'}},
            'orbo_32_s_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'orion_gas_chromatography_with_photo_ionization_detector_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'orsa_passive_sampler_chromatography_not-automatic':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},  
            'orsa_passive_sampler_gas_chromotography_(gc)_not-automatic':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'passive_sampler_(carbopack_e),perkin-elmer_atd_400,hp_gc_5890,hp_ms_5792_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'passive_sampling_gas_chromotography_(gc)_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'pcf_elettronica_mod._256_chromatography_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'pcf_mod.527_chromatography_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'pcf_mod.527_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'pcf_r526_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'perkin_elmer_turbomatrix_desorber,_perkin_elmer_clarus_500_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'philips_chromatography_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'philips_k50031_chromatographic_analyser_chromatography_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'philips_k50031_chromatographic_analyser_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'pumped_tube_chromatography_not-automatic':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                             'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'radiello_tube_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_not-automatic':                                                        {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}}, 
            'radiello_tube_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                  'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'rancon_btx_chromatography_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}}, 
            'rivm_-_universal_sampler_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_not-automatic':                                             {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Rivm Universal Sampler'},                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}}, 
            'rotork_526_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'shimadzu_gc-14a_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                               {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                               'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'skypostgas_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Terora Skypost Gas'},                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}}, 
            'siemens_u_102_btx_chromatography_automatic':                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},    
            'sintech_spectras_btx_gc_855_serie_600_gas_chromatography_with_photo_ionization_detector_automatic':                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_chromatography_automatic':                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},    
            'sintech_spectras_btx_gc_855_series_undetermined_flame_photometry_automatic':                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection',   'process_details':'Assume method is GC - unknown detection and not flame photometry as stated',                                   'measuring_assumption':True}},        
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_with_photo_ionization_detector_automatic':                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_ultraviolet_(uv)_photometry_automatic':                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection',   'process_details':'Assume method is GC - unknown detection and not ultraviolet photometry as stated',                             'measuring_assumption':True}},
            'sintech_spectras_btx_gc_955_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'spectras_syntec_995_chromatography_automatic':                                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'sri_8610_chromatography_automatic':                                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'sri_8610_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-600_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-600_gas_chromatography_with_photo_ionization_detector_automatic':                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-800_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-800_gas_chromatography_with_photo_ionization_detector_automatic':                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_chromatography_automatic':                                                                                                {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_with_photo_ionization_detector_automatic':                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'tecora_gas_chromotography_(gc)_not-automatic':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Terora Skypost Gas', 'process_details':'Make assumption sampling instrument is TCR Terora Skypost Gas.', 'primary_sampling_assumption':True},                                                                                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection'}}, 
            'teflon_bag+desorption_tube_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_not-automatic':                                           {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                               'preparation_1':{'sample_preparation_type':'bag'},                                                                                         'preparation_2':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}}, 
            'thermo_scientific:_trace_gc_ultra_+_trace_ms_gas_chromatography_+_mass_spectrometry_gc-ms_after_solvent_o_not-automatic':                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping',                                             'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'too_generic_chromatography_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'unknown_atomic_absorption_spectrometry_automatic':                                                                                                {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                            'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                                                          
            'unknown_automated_sampling_and_on_line_analysis,_cryogenic_sample_pr_automatic':                                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection',   'process_details':'Assume method is GC-unknown detection',                                                                        'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},
            'unknown_chromatography_not-automatic':                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_conductimetry_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'conductimetry'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy'}}, 
            'unknown_flame_ionization_detection_(fid)_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                           'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',     'measuring_assumption':True}},
            'unknown_flame_photometry_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'flame photometric detection'}}, 
            'unknown_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_not-automatic':                                                              {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}},
            'unknown_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gas_chromotography_(gc)_not-automatic':                                                                                                   {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromatography_(ref)_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - unknown detection',                                                                          'sampling_type':'continuous injection'}},   
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_-_flame_ionization_not-automatic':                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}}, 
            'unknown_gas_chromatography_-_mass_spectrometry_not-automatic':                                                                                    {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_not-automatic':                                                                                                        {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                        'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_with_photo_ionization_detector_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                  'sampling_type':'continuous injection'}},
            'unknown_passive_sampler_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'varian_chrompack_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                 'sampling_type':'continuous injection'}}}
#------------------------------------------- 
        elif run_details['parameter'] in ['VOC','HC']:
            mapping = {
            'dani_410_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                    {'preparation_1':{'sample_preparation_type':'preconcentration'},                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                    'sampling_type':'continuous injection'}},        
            'horiba_model_apha_360_hydrocarbons_analyser_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic': {                                                                                                'measurement':{'methodology':'selective combustion - flame ionisation detection', 'instrument':'Horiba APHA360'}},
            'horiba_model_apha_360e_hydrocarbons_analyser_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':{                                                                                                'measurement':{'methodology':'selective combustion - flame ionisation detection', 'instrument':'Horiba APHA360'}}, 
            'pcf_r526_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                    {                                                                                                'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                    'sampling_type':'continuous injection'}},   
            'rotork_526_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                  {                                                                                                'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                    'sampling_type':'continuous injection'}},
            'unknown_atomic_absorption_spectrometry_automatic':                                                                   {'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                               'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                                                          
            'unknown_chromatography_automatic':                                                                                   {                                                                                                'measurement':{'methodology':'gas chromatography - unknown detection',                                             'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                     {                                                                                                'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                    'sampling_type':'continuous injection'}},
            'unknown_unknown_automatic':                                                                                          {                                                                                                'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HCHO':
            mapping = {
            '':                                                                                  {                                                                                                                                                                                                  'measurement':{'methodology':'unknown'}},
            'photocolorimeter_spectrophotometry_not-automatic':                                  {'primary_sampling':{'sampling_type':'low volume continuous',  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},        
            'unknown_gas_chromatography_not-automatic':                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',            'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - unknown detection'}},                       
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detect_not-automatic':{'primary_sampling':{'sampling_type':'high volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',            'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'NH3':
            mapping = {
            '':                                                                                           {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'ecn_-_amor_conductimetry_automatic':                                                         {                                                                                                                                                                                                 'measurement':{'methodology':'conductimetry',                                                          'instrument':'ECN AMOR'}},
            'environment_nh3_om_chemiluminescence_automatic':                                             {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Cannot find Environment OM instrument, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).',                  'measuring_assumption':True}},
            'mlu_201_chemiluminescence_automatic':                                                        {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Cannot find Monitor Labs 201 instrument, so assume it uses chemiluminescence (internal molybdenum and stainless steel converters) as 201A instrument uses', 'measuring_assumption':True}}, 
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                                                                                                                                 'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'photocolorimeter_spectrophotometry_not-automatic':                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},           
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':                {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'unknown_chemiluminescence_automatic':                                                        {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).',                'measuring_assumption':True}},
            'unknown_unknown_automatic':                                                                  {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'HNO3':
            mapping = {
            '':{'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'PAN':
            mapping = {
            '':                         {'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':{'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HCl':
            mapping = {
            '':                                                                                           {'measurement':{'methodology':'unknown'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':{'measurement':{'methodology':'differential optical absorption spectroscopy', 'instrument':'OPSIS AR500'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HF':
            mapping = {
            '':                         {'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':{'measurement':{'methodology':'unknown'}}}  
#-------------------------------------------
        elif run_details['parameter'] == 'H2S':
            mapping = {
            '':                                                                                           {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_automatic':                      {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Environnement SA AF21M'}}, 
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence_automatic':                               {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Horiba APSA370'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                                                                                                                                 'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'OPSIS AR500'}},
            'photocolorimeter_spectrophotometry_not-automatic':                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'sintech_spectras_btx_gc_855_series_undetermined_uv_fluorescence_automatic':                  {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                                                                    'process_details':'Assume method is ultraviolet fluorescence and not GC as instrument uses', 'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                     {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Teledyne 100'}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                    {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Teledyne 100A'}},
            'teledyne_api_101a_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                    {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Teledyne 101A'}},
            'thermo_450i_uv_fluorescence_automatic':                                                      {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Thermo 450i'}},
            'thermo_model_42_no_nox_analyser_chemiluminescence_automatic':                                {                                                                                                                                                                                                 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Thermo 42'}},
            'thermo_model_43_so2_analyser_uv_fluorescence_automatic':                                     {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Thermo 43'}},
            'thermo_model_43a_so2_analyser_uv_fluorescence_automatic':                                    {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Thermo 43A'}},
            'thermo_model_450i_uv_fluorescence_automatic':                                                {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence',                          'instrument':'Thermo 450i'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                  {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                              {                                                                                                                                                                                                 'measurement':{'methodology':'unknown'}},
            'unknown_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                 'measurement':{'methodology':'ultraviolet fluorescence'}}} 
#------------------------------------------- 
        elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
            mapping = {
            '':                                                                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}},
            'adm_9000_beta_ray_attenuation_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'ADM 9000'}},
            'airpointer_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Airpointer-NOx'}}, 
            'analysis_automation_beta_ray_attenuation_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with beta-attenuation method as stated.',       'measuring_assumption':True}},       
            'andersen_gmw_model_fh621-n_beta_monitor_beta_ray_attenuation_automatic':                                         {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_tapered_element_oscillating_microbalance_(teom)_automatic':              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'apm-1_elecos_beta_ray_attenuation_automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                              'measurement':{'methodology':'beta-attenuation'}},
            'atmoservice_lvs_3d_gravimetric_analysis_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'atmoservice_pns3d15_gravimetric_analysis_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'atmoservice_pp_-_8k_atomic_absorption_spectrometry_(aas)_not-automatic':                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                             'process_details':'Cannot find sampling instrument Atmoservice PP 8K, assume low volume sampling',    'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                                                                                                                                                                         'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',               'measuring_assumption':True}},
            'atmoservice_pp_-_8k_gravimetric_analysis_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                             'process_details':'Cannot find sampling instrument Atmoservice PP 8K, assume low volume sampling',    'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'beta_tapered_element_oscillating_microbalance_(teom)_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'biolaffite_collector_reflectometry_not-automatic':                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}},  
            'cav_-_a_gravimetry_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'MCV CAV-A/mb',               'process_details':'Make assumption sampling instrument is MCV CAV-A/mb.',                             'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'das_beta_ray_attenuation_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                    'measuring_assumption':True}},
            'das_rancon_adam_c_beta_ray_attenuation_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                    'measuring_assumption':True}},
            'dasibi_7001_beta_ray_attenuation_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Dasibi 7001'}}, 
            'derenda_gravimetric_analysis_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T',      'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.',                    'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'desconocido_gravimetric_analysis_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'dha-80_gravimetry_not-automatic':                                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_gravimetry_automatic':                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_not-automatic':                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_da_80_h_gravimetry_automatic':                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_gravimetry_automatic':                                                                             {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',             'process_details':'Make assumption sampling instrument is Digitel DHA-80.',                           'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_gravimetric_analysis_not-automatic':                                                           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_gravimetric_analysis_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',             'process_details':'Make assumption sampling instrument is Digitel DHA-80.',                           'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'eberline_beta_ray_attenuation_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'environment_beta_ray_attenuation_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_ray_attenuation_automatic':                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp2.5m_undefined_beta_ray_attenuation_automatic':                                       {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                                         'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                      'measuring_assumption':True}},
            'environnement_s.a._model_mpsi_100_beta_ray_attenuation_automatic':                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._mp101m_beta_ray_attenuation_not-automatic':                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'f-701_beta_ray_attenuation_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'f-701_verewa_beta_ray_attenuation_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'filter_gravimetric_analysis_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'gmw_graseby_gravimetric_analysis_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200',  'process_details':'Make assumption sampling instrument is Graseby Andersen GMW 1200',                 'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'gravimetrisk_kleinfilter_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_differential_optical_absorption_spectroscopy_(doas)_automatic':           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                                       'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_laser_spectroscopy_automatic':                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                                       'process_details':'Assume method is not laser spectroscopy as stated but uses optical particle counting as reported instrument uses.',          'measuring_assumption':True}},          
            'grimm_model_edm_180_for_pm10_and_pm2.5_light_scattering_automatic':                                              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                                       'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.', 'measuring_assumption':True}},          
            'grimm_model_edm_180_for_pm10_and_pm2.5_optoelectronic_automatic':                                                {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                                       'process_details':'Assume method is not optoelectronic as stated but uses optical particle counting as reported instrument uses.',              'measuring_assumption':True}},          
            'grimm_model_edm_180_for_pm10_and_pm2.5_scattering_of_laser_light_automatic':                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                                       'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.', 'measuring_assumption':True}},          
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_differential_optical_absorption_spectroscopy_(doas)_automatic':{                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                                       'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},  
            '"grimm_model_edm_365_for_pm10,_pm2.5,_pm1_and_tc"_beta_ray_attenuation_automatic':                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365',                                                                                                                                                                                       'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',            'measuring_assumption':True}}, 
            '"grimm_model_edm_365_for_pm10,_pm2.5,_pm1_and_tc"_unknown_automatic':                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365'}},        
            'horiba_model_apda_350e_tsp_analyser_beta_ray_attenuation_automatic':                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E'}},
            'horiba_model_apda_351_tsp_analyser_beta_ray_attenuation_automatic':                                              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA351'}},
            'horiba_model_apda_351e_tsp_analyser_beta_ray_attenuation_automatic':                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA351E'}},
            'hvs_gravimetric_analysis_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_not-automatic':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_not-automatic':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry'}},  
            'kimoto_168s_beta_ray_attenuation_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto 168S'}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry'}},
            'leckel_kleinfilter_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3',                'process_details':'Make assumption sampling instrument is Leckel LVS3.',                              'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'leckel_sequential_lvs_lvs_-_automatic_filter_change_2,3_m3_h_not-automatic':                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3',                'process_details':'Make assumption sampling instrument is Leckel LVS3.',                              'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'lvs_gravimetric_analysis_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'mcz,_pns-lvs_16_seq_gravimetric_analysis_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'metone_bam-1020_beta_ray_attenuation_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'mlu_beta_ray_attenuation_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Airpointer-PMBA'}},
            'muhlberg_filter_disk_beta_ray_attenuation_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Muhlberg filter disc, therefore just assume uses beta-attenuation method as stated.',                 'measuring_assumption':True}},
            'mvs_gravimetric_analysis_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},                                           
            'only_used_in_one_station_in_rs_beta_ray_attenuation_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'opsis_ar500_open_path_monitor_beta_ray_attenuation_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Assume method is beta-attenuation as stated and not DOAS as reported instrument uses.',                                      'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'differential optical absorption spectroscopy',                                                 'instrument':'OPSIS AR500'}},
            'opsis_sm200_gravimetric_analysis_not-automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                                     'process_details':'Assume method is not gravimetry as stated, but uses beta-attenuation as reported instrument uses.',                          'measuring_assumption':True}},                                              
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation_automatic':                        {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'orion_beta_ray_attenuation_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                            'measuring_assumption':True}},
            'osiris_environmental_particle_monitor_beta_ray_attenuation_automatic':                                           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Turnkey Osiris',                                                                                                                                                                                  'process_details':'Assume method is optical particle counter as instrument uses, and not beta-attenuation as stated',                           'measuring_assumption':True}},
            'partisol_gravimetric_analysis_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000',       'process_details':'Make assumption sampling instrument is Thermo Partisol 2000.',                     'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'philips_mp_101m_beta_ray_attenuation_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'pm2.5_beta_ray_attenuation_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}}, 
            'pns_3d_15_gravimetric_analysis_not-automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2000_gravimetric_analysis_not-automatic':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2000_gravimetry_automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_not-automatic':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'rancon_adam_c_beta_ray_attenuation_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                    'measuring_assumption':True}},
            'sentinel_gravimetric_analysis_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Tecora Sentinel',        'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sequential_air_sampler_-_type_ek_gravimetry_automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'NILU EK'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_gravimetric_analysis_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_lvs3_gravimetric_analysis_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_seq47_50_gravimetric_analysis_not-automatic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'sharp_beta_absorption_not-automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP',                                                                                                                                                                               'process_details':'Assume instrument is Thermo 5030 SHARP',                                                                                     'measuring_assumption':True}},
            'sierra_dichotomous_sampler_gravimetry_automatic':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 244 Dichot', 'process_details':'Make assumption sampling instrument is Sierra Andersen 244 Dichot',                'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},       
            'sm200_(adam)_beta_absorption_not-automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm200_(adam)_beta_ray_attenuation_not-automatic':                                                                {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_(air_dust_automatic_monitor)_beta_ray_attenuation_automatic':                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_beta_ray_attenuation_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'spm-613d_beta_ray_attenuation_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto SPM-613D'}},
            'staplex_gravimetric_analysis_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                             'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_not-automatic':                                                            {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_seq47_50_gravimetric_analysis_not-automatic':                                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'swam_5a_dual_channel_monitor_beta_ray_attenuation_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'tcr_tecora,_model_sentinel_pm_gravimetric_analysis_not-automatic':                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry'}},
            'tcr_tecora,_sentinel_pm_gravimetric_analysis_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',         'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'teom_1400_gravimetry_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)_automatic':                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_volatile_correction_model_(vcm)_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                                     'process_details':'Assume method is TEOM as instrument uses, and not VCM',                                                                      'measuring_assumption':True}},
            'teom_1400a_gravimetry_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_automatic':                                           {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400ab_gravimetry_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)_automatic':                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_automatic':                  {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                           'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_with_filter_automatic':      {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                           'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1405f_with_fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                 {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},
            'thermo_5014i_beta_ray_attenuation_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_gravimetry_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                                        'process_details':'Assume method is beta-attenuation as instrument uses, and not gravimetry as stated',                                         'measuring_assumption':True}},    
            'thermo_andersen_esm_fh_62_i-r_tapered_element_oscillating_microbalance_(teom)_automatic':                        {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                                        'process_details':'Assume method is beta-attenuation as instrument uses, and not TEOM as stated',                                               'measuring_assumption':True}},    
            'thermo_andersen_esm_fh_62_i-r_unknown_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}}, 
            'thermo_andersen_series_fh_62_c14_continuous_pm10_ambient_particulate_monitor_beta_ray_attenuation_automatic':    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'thermo_electron_instruments_beta_ray_attenuation_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'thermo_esm_andersen,_fh_gravimetric_analysis_not-automatic':                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                                        'process_details':'Assume instrument is Thermo Andersen FH62 I-R.',                                                                             'measuring_assumption':True}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation_automatic':              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_raas10-300_gravimetric_analysis_not-automatic':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS10 300'},                                                                                                                                                                                                                                         'measurement':{'methodology':'gravimetry'}}, 
            'umwelttechnik_mcz_gmbh_micro_pns_lvs_16_gravimetric_analysis_not-automatic':                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'umwelttechnik_mcz_gmbh_pns3d15_gravimetric_analysis_not-automatic':                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15',        'process_details':'Make assumption sampling instrument is Atmoservice PNS3D15.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'undefined_gravimetry_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                   'process_details':'Assume method is gravimetry',                                                                                                'measuring_assumption':True}},
            'unitec_beta_ray_attenuation_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                                   'process_details':'Assume instrument is Unitec LSPM10.',                                                                                        'measuring_assumption':True}},
            'unitec_lspm10_beta_ray_attenuation_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                                   'process_details':'Assume method is light scattering photometry, and not beta-attenuation as stated',                                           'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_automatic':                                                               {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                       'sampling_type':'continuous injection',                                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',               'measuring_assumption':True}},                                                          
            'unknown_beta_absorption_not-automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_not-automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},  
            'unknown_chemiluminescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                                                            'process_details':'Make assumption uses internal molybdenum converter',                                                                         'measuring_assumption':True}},
            'unknown_chromatography_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                                     'sampling_type':'continuous injection'}},
            'unknown_conductimetry_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'conductimetry'}},
            'unknown_dif._optical_absorption_spec._not-automatic':                                                            {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'differential optical absorption spectroscopy'}},         
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'differential optical absorption spectroscopy'}},   
            'unknown_gravimetric_analysis_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                   'process_details':'Assume method is gravimetry',                                                                                                'measuring_assumption':True}},
            'unknown_gravimetry_automatic':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                   'process_details':'Assume method is gravimetry',                                                                                                'measuring_assumption':True}},
            'unknown_gravimetry_not-automatic':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                                   'process_details':'Assume method is gravimetry',                                                                                                'measuring_assumption':True}},
            'unknown_light_scattering_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'light scattering photometry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_automatic':                                              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_not-automatic':                                          {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                      'sampling_type':'injection'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_with_filter_automatic':                                  {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'unknown'}},
            'verewa_f_701_beta_ray_attenuation_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_701_gravimetry_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701',                                                                                                                                                                                     'process_details':'Assume method is beta-attenuation as instrument uses, not gravimetry as stated',                                             'measuring_assumption':True}},
            'verewa_f_703_beta_ray_attenuation_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'wedding_&_as._beta_ray_attenuation_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}},
            'wedding_&_as._pm10_gravimetric_analysis_not-automatic':                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},                              
            'wedding_beta_ray_attenuation_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}}, 
            'x_beta_ray_attenuation_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                             'process_details':'Cannot find instrument X, assume beta-attenuation as stated',                                                                'measuring_assumption':True}}}
#------------------------------------------- 
        elif parameter_details['matrix'] == 'pm':   
            mapping = {
            '':                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}},
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_not-automatic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                      'instrument':'GBC 932'}},
            'anderson_a-2000_hvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                        {'primary_sampling':{'sampling_type':'high volume continuous',                                            'process_details':'Cannot find Anderson A-2000 sampling instrument'},                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},        
            'anderson_hvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                               {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'atmoservice_lvs_3d_atomic_absorption_spectrometry_(aas)_not-automatic':                                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Atmoservice PNS3D15'},                                                                                                                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                                                                    'process_details':'Make assumption method is graphite furnace atomic absorption spectroscopy',                                                                     'measuring_assumption':True}},
            'bubbler_24_h_atomic_absorption_spectrometry_(aas)_not-automatic':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},                   
            'collector_atomic_absorption_spectrometry_(aas)_not-automatic':                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'danish_lmp_filterpack_element_collector_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':{'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                             'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'danish_lmp_filterpack_element_collector_x-ray_emission_not-automatic':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                             'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'derenda_lvs_3.1_gravimetric_analysis_not-automatic':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Make assumption method is ion chromatography',                                                                                                  'measuring_assumption':True}},
            'derenda_lvs_3.2_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.2'},                                                                                                                                               'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},       
            'desconocido_atomic_absorption_spectrometry_(aas)_not-automatic':                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_not-automatic':                                              {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_not-automatic':                          {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'digitel_dha80_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_not-automatic':                      {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80'},                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':        {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80',            'process_details':'Make assumption sampling instrument is Digitel DHA-80.',                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Digitel DHA-80',            'process_details':'Make assumption sampling instrument is Digitel DHA-80.',                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'fh62i-n_atomic_absorption_spectrometry_(aas)_not-automatic':                                                    {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation',            'instrument':'Andersen FH621-N'},                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'filter_pack_griess-saltzman_reaction_not-automatic':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',                                                                                              'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction',         'specific_technique':'Griess-Saltzman'},                                                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                                                                  'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                            'measuring_assumption':True}},      
            'filter_pack_spectrophotometry_not-automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                             'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},      
            'filter_3pack_ion_chromatography_not-automatic':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                                                    'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},     
            'filter_sampling_atomic_absorption_spectrometry_(aas)_not-automatic':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'filter_sampling_continous_flow_analysis_(cfa)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis'}},
            'filter_sampling_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'filter_sampling_ion_chromatography_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'filter_atomic_absorption_spectrometry_(aas)_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'filter_ion_chromatography_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'filter_unknown_not-automatic':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                                                         'measuring_assumption':True}},
            'finigan-element2_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Thermo Finnigan ELEMENT2'}},      
            'gmw_high_volume_air_sampler_atomic_absorption_spectrometry_(aas)_not-automatic':                                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Graseby Andersen GMW 1200', 'process_details':'Make assumption sampling instrument is Graseby Andersen GMW 1200',                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'gmw_high_volume_air_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':            {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'Graseby Andersen GMW 1200', 'process_details':'Make assumption sampling instrument is Graseby Andersen GMW 1200',                 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'high_volume_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                    {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'hvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                                        {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'hvs_ion_chromatography_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'icp-oes_as_16_thermo_jarrell_ash_atomic_absorption_spectrometry_(aas)_not-automatic':                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',   'instrument':'Thermo Jarrell Ash Atomscan 16',                                           'process_details':'Assume method is inductively coupled plasma atomic emission spectroscopy as instrument uses and not Atomic Absorption Spectroscopy as stated.', 'measuring_assumption':True}},
            'kleinfilter_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                                                         'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'm-type_sampler_atomic_absorption_spectrometry_(aas)_not-automatic':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'm-type_sampler_spectrophotometry_not-automatic':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},       
            'mcv_cav_-_a_atomic_absorption_spectrometry_(aas)_not-automatic':                                                {'primary_sampling':{'sampling_type':'high volume continuous',  'instrument':'MCV CAV-A/mb',              'process_details':'Make assumption sampling instrument is MCV CAV-A/mb.',                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}}, 
            'mlu_100a_atomic_absorption_spectrometry_(aas)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                            {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation',            'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'opsis_sm200_x-ray_emission_not-automatic':                                                                      {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation',            'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'partisol_2025_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},        
            'perkin_elmer_4100_zl_atomic_absorption_spectrometry_automatic':                                                 {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',           'instrument':'Perkin Elmer 4110 ZL',           'sampling_type':'continuous injection'}},                        
            'perkin_elmer_aanalyst_800_atomic_absorption_spectrometry_automatic':                                            {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                      'instrument':'Perkin Elmer AAnalyst 800',      'sampling_type':'continuous injection'}},         
            'perkin_elmer_aas4100zl_atomic_absorption_spectrometry_(aas)_not-automatic':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',           'instrument':'Perkin Elmer 4110 ZL'}},            
            'perkin_elmer_elan_dr_ii_inductive_coupled_plasma_mass_spectrometry_(icp-ms)_automatic':                         {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Perkin Elmer ELAN DRC II',       'sampling_type':'continuous injection'}},                    
            'perkin_elmer_elan_drc_ii_inductive_coupled_plasma_mass_spectrometry_(icp-ms)_automatic':                        {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Perkin Elmer ELAN DRC II',       'sampling_type':'continuous injection'}},       
            'philips_coulometry_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'coulometry'}},
            'photocolorimeter_spectrophotometry_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},      
            'pns_3d_15_atomic_absorption_spectrometry_(aas)_not-automatic':                                                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Atmoservice PNS3D15'},                                                                                                                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'r&p_partisol_2025_gravimetric_analysis_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2025'},                                                                                                                                                'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                                                         'measuring_assumption':True}},
            'sequential_sampler_gravimetry_automatic':                                                                       {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                       'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography',                                                                                                           'measuring_assumption':True}},
            'sequential_sampler_lvs3_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_not-automatic':            {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel LVS3'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'sm200_(adam)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                           {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation',            'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'spektraa_220ss_varian_atomic_absorption_spectrometry_(aas)_not-automatic':                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                      'instrument':'Varian SpectrAA 220'}},
            'staplex_atomic_absorption_spectrometry_(aas)_not-automatic':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                             'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'sunset_inc_thermal_optical_automatic':                                                                          {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol',               'instrument':'Sunset OCEC Model 5',            'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                                                  'measuring_assumption':True}},
            'sven_leckel_seq47_50_continous_flow_analysis_(cfa)_not-automatic':                                              {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'continuous flow analysis'}},                                                                         
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},          
            'sven_leckel_seq47_50_ion_chromatography_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'ion chromatography'}},         
            'sven_leckel_mvs6_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                       {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel MVS6'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},  
            'tecora_atomic_absorption_spectrometry_(aas)_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'TCR Tecora Sentinel',       'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'tecora_gravimetric_analysis_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'TCR Tecora Sentinel',       'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                                                         'measuring_assumption':True}},
            'tekran_2537_atomic_absorption_spectrometry_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',              'instrument':'Tekran 2537A-1130-1135',                                                   'process_details':'Assume instrument is Tekran 2537A-1130-1135',                                                                                                   'measuring_assumption':True}},  
            'tekran_gas-phase_mercury_vapour_analyser_(model_2537a)_atomic_fluorescence_spectrometry_automatic':             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',              'instrument':'Tekran 2537A-1130-1135'}},
            'termo_esm_andersen,_fh_95_kf_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Andersen FH95-KF'},                                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                                     'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'thermo_atomic_absorption_spectrometry_(aas)_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_automatic':                                                              {                                                                                                                                                                                                                                                    'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True},                                                                                                                                                                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                     'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'unknown_cold_vapour_atomic_absorption_spectrometry_(cv-aas)_not-automatic':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic absorption spectroscopy'}},                                                                   
            'unknown_cold_vapor_atomic_fluorescence_spectrometry_not-automatic':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},                                                                   
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_(cv-afs)_not-automatic':                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},                                                                   
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gravimetric_analysis_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                                 'process_details':'Assume method is ion chromatography after gravimetry.',                                                                                         'measuring_assumption':True}},
            'unknown_icp-oes_not-automatic':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_not-automatic':                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'unknown_mass_espectrometry_(icp-ms)_not-automatic':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_unknown_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}},
            'unknown_x-ray_emission_not-automatic':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'unknown_x-ray_fluorescence_spectroscopy_(xrf)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'ut-3000_mercury_ultra_tracer_atomic_absorption_spectrometry_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'measurement':{'methodology':'cold vapour atomic absorption spectroscopy',                 'instrument':'Envea UT-3000'}},           
            'variaxi_aa_220_atomic_absorption_spectrometry_(aas)_not-automatic':                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                       'instrument':'Varian SpectrAA 220'}},
            'wsl_atomic_absorption_spectrometry_(aas)_not-automatic':                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                               'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                                  'measuring_assumption':True}},
            'xrf_mvs_x-ray_fluorescence_spectroscopy_(xrf)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                                   'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}}} 
#------------------------------------------- 
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
            mapping = {
            '':                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'adsorption_tube_cold_vapour_atomic_fluorescence_spectrometry_(cv-afs)_not-automatic':                   {'primary_sampling':{'sampling_type':'passive',                                                                                                                                   'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',         'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},
            'atmoservice_lvs_3d_atomic_absorption_spectrometry_(aas)_not-automatic':                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                    'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                                                                                                'process_details':'Make assumption method is graphite furnace atomic absorption spectroscopy',                                    'measuring_assumption':True}},
            'derenda_continous_flow_analysis_(cfa)_not-automatic':                                                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis'}},
            'derenda_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'derenda_ion_chromatography_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'derenda_lvs_3.1_gravimetric_analysis_not-automatic':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda LVS 3.1'},                                                                                                                  'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Make assumption method is ion chromatography',                                                                 'measuring_assumption':True}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                    {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',         'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'dionex_2010i_ion_chromatography_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                        'instrument':'Dionex 2010i'}},
            'dionex_3000_ion_chromatography_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                        'instrument':'Thermo ICS-3000'}},
            'dionex_ion_chromatography_not-automatic':                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'filter_3pack_ion_chromatography_not-automatic':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                                            'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},     
            'filter_sampling_atomic_absorption_spectrometry_(aas)_not-automatic':                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'filter_sampling_ion_chromatography_not-automatic':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'high_volume_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':            {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'icp_ms_atomic_absorption_spectrometry_(aas)_not-automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown',                                                                                                                                        'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'leckel_seq47_50_with_cooler_thermal-optical_carbon_analysis_not-automatic':                             {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},   
            'low_volume_sampler_leckel_ion_chromatography_not-automatic':                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'lvs_-_automatic_filter_change_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':  {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'marga_ion_chromatography_not-automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'ion chromatography',                                        'instrument':'Metrohm MARGA'}},
            'micropns_lvs1_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1'},                                                                                                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}, 
            'partisol_+_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_+_unknown_not-automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'partisol_+_zeeman_atomic_absorption_spectrometry_(zeeman_aas)_not-automatic':                           {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy'}},
            'partisol_2300_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                  {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                                                                                                  'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_2300_unknown_not-automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                                                                                                  'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'partisol_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_not-automatic':                       {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'partisol_hydride_generation_electrothermal_-_atomic_absorption_spectr_not-automatic':                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'hydride generation atomic absorption spectroscopy'}},
            'partisol_unknown_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                               'measuring_assumption':True}},
            'perkin-elmer_elan_6000_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',              'instrument':'Perkin Elmer ELAN 6000'}},
            'philips_pw2400_x-ray_fluorescence_spectroscopy_(xrf)_not-automatic':                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy',                           'instrument':'Philips PW2400'}},
            'sm200_(adam)_chromatography_not-automatic':                                                             {                                                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Make assumption method is ion chromatography',                                                                 'measuring_assumption':True}},              
            'sm200_(adam)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                   {                                                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                            'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},              
            'sm200_(adam)_ion_chromatography_not-automatic':                                                         {                                                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography'}},              
            'sunset_inc_thermal_optical_automatic':                                                                  {                                                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol',               'instrument':'Sunset OCEC Model 5',          'sampling_type':'continuous injection', 'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sunset_laboratory_inc_thermal___optical_carbon_analyser_eusaar_2_method_not-automatic':                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                        'instrument':'Sunset OCEC Model 5',                                                  'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},  
            'sunset_laboratory_thermal-optical_carbon_analysis_not-automatic':                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol',               'instrument':'Sunset OCEC Model 5',                                                  'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sven_leckel_seq47_50_atomic_absorption_spectrometry_(aas)_not-automatic':                               {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},            
            'sven_leckel_seq47_50_ion_chromatography_not-automatic':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},        
            'sven_leckel_seq47_50_thermal-optical_carbon_analysis_not-automatic':                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},         
            'teom_accu_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':              {                                                                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'tapered element oscillating microbalance - gravimetry'},                                                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'thermo_icp_iris_intrepid_ii_atomic_absorption_spectrometry_(aas)_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',   'instrument':'Thermo IRIS Intrepid II',                                              'process_details':'Assume method is ICP-AES not AAS',                                                                             'measuring_assumption':True}},  
            'thermo_icp_iris_intrepid_ii_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-a_not-automatic':{'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',   'instrument':'Thermo IRIS Intrepid II'}},       
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_not-automatic':                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                           'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_chromatography_not-automatic':                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Make assumption method is ion chromatography',                                                                 'measuring_assumption':True}},                                                 
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_(cv-afs)_not-automatic':                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},
            'unknown_differential_optical_absorption_spectrometry_not-automatic':                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_gravimetric_analysis_not-automatic':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'unknown_high_vol_sampler_not-automatic':                                                                {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_hvs_-_automatic_filter_change_30_m3.h-1_not-automatic':                                         {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                        'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-a_not-automatic':                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_ion_chromatography_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromotagraphy_not-automatic':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_lvs_-_automatic_filter_change_2,3_m3_h_not-automatic':                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                             'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}}, 
            'unknown_mass_espectrometry_(icp-ms)_not-automatic':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_niosh_not-automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_optical_termical_not-automatic':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                                                                                                    'process_details':'Asssume method is thermal-optical analysis',                                                                   'measuring_assumption':True}},
            'unknown_spectrophotometry_not-automatic':                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},      
            'unknown_unknown_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_x-ray_fluorescence_spectroscopy_(xrf)_not-automatic':                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'varian_spectre_aa_400_graphit_gta-96_atomic_absorption_spectrometry_(aas)_not-automatic':               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',           'instrument':'Varian SpectrAA 400 GTA-96'}},  
            'vgelementalpq2_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                     'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'xrf_mvs_x-ray_fluorescence_spectroscopy_(xrf)_not-automatic':                                           {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                         'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}}} 
           
#--------------------------------------------------------------
#EEA AQ eReporting E1a/E2a / MITECO
#--------------------------------------------------------------
    elif run_details['network'] in ['EEA_AQ_eReporting','MITECO']:
        if run_details['parameter'] == 'O3':
            mapping = {
            '':                                                                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            '342_m_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},                                                                                                    
            'active':                                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'analysis_automation_ultraviolet_(uv)_photometry_automatic':                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet photometry method as stated.',                                'measuring_assumption':True}},
            'analysis_automation_mod._427_ultraviolet_(uv)_photometry_automatic':                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument Analysis Automation, therefore assume uses instrument with ultraviolet photometry method as stated.',                                'measuring_assumption':True}},
            'api100a_uv-fl_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100A'}},
            'api100a_uv_fluorescence_automatic':                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Teledyne 100A'}},
            'api100e_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                       'measuring_assumption':True}},
            'api100e_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                       'measuring_assumption':True}}, 
            'api200_chemi_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200'}},
            'api200_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Teledyne 200 instrument uses',                               'measuring_assumption':True}},
            'api200e_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200E'}},
            'api200e_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Teledyne 200E'}},
            'api400_uv_fluorescence_automatic':                                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400',             'process_details':'Make assumption the method is not ultraviolet fluorescence but is ultraviolet photometry as reported instrument uses',                                      'measuring_assumption':True}},   
            'api400_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'api400_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}}, 
            'api400a_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'api400a_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'api400a_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'api400e_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'api400e_light-abs_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E',            'process_details':'Assume method is not light absorption as stated but is ultraviolet photometry as reported instrument uses.',                                                'measuring_assumption':True}},
            'api400e_light_absorption_automatic':                                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E',            'process_details':'Assume method is not light absorption as stated but is ultraviolet photometry as reported instrument uses.',                                                'measuring_assumption':True}},
            'api400e_ndir_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E',            'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                                            'measuring_assumption':True}},                   
            'api400e_other,_please_specify_automatic':                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'api400e_uv-fl_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E',            'process_details':'Assume method is not ultraviolet fluorescence as stated but is ultraviolet photometry as reported instrument uses.',                                        'measuring_assumption':True}},
            'api400e_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'api400e_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'apit300_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne T300'}},    
            'apit300_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},
            'apit300_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},  
            'apit400_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},   
            'apit400_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400',            'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'apit400_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400',            'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'apit400_other,_please_specify_automatic':                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'apit400_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'apit400_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'automatic':                                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'bam1020_beta_ray_attenuation_automatic':                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'beta-attenuation',                                            'instrument':'Met One BAM1020'}},
            'columbia_scientific_ic_3100_uv-p_automatic':                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'CSI 3100'}},
            'columbia_scientific_ic_3100_ultraviolet_(uv)_photometry_automatic':                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'CSI 3100'}},
            'c.s.i._3.100_ultraviolet_(uv)_photometry_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'CSI 3100'}},
            'dasibi_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'dasibi_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'dasibi1008_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH',           'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                                                    'measuring_assumption':True}},
            'dasibi1008_ultraviolet_(uv)_photometry_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH',           'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                                                    'measuring_assumption':True}},
            'dasibi_1008_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH',           'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                                                    'measuring_assumption':True}},
            'dasibi_1008_o3_analyser_ultraviolet_(uv)_photometry_active':                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH',           'process_details':'Make assumption measuring instrument is Dasibi 1008-AH',                                                                                                    'measuring_assumption':True}},
            'dasibi_1008-ah_o3_analyser_ultraviolet_(uv)_photometry_active':                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH'}},    
            'dasibi1008ah_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH'}},
            'dasibi1008ah_ultraviolet_(uv)_photometry_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-AH'}},
            'dasibi1008rs_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-RS'}},
            'dasibi1008rs_ultraviolet_(uv)_photometry_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1008-RS'}},         
            'dasibi1108_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1108 W/GEN',        'process_details':'Make assumption measuring instrument is Dasibi 1108 W/GEN',                                                                                                 'measuring_assumption':True}},
            'dasibi1108_ultraviolet_(uv)_photometry_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1108 W/GEN',        'process_details':'Make assumption measuring instrument is Dasibi 1108 W/GEN',                                                                                                 'measuring_assumption':True}},
            'dasibi5014_uv-p_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'SIR S-5014',               'process_details':'Make assumption measuring instrument is SIR S-5014',                                                                                                        'measuring_assumption':True}},
            'dasibi5014_ultraviolet_(uv)_photometry_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'SIR S-5014',               'process_details':'Make assumption measuring instrument is SIR S-5014',                                                                                                        'measuring_assumption':True}},
            'environnement_s.a._model_1003-rs_o3_ultraviolet_(uv)_photometry_active':                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Dasibi 1003-RS'}},
            'environnement_s.a._model_o331m_uv_ozone_analyzer_uv-p_automatic':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O331M'}},
            'environnement_s.a._model_o331m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_automatic':       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O331M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_automatic':       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_uv-p_automatic':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_other,_please_specify_automatic':             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o341m_uv_ozone_analyzer_ultraviolet_(uv)_photometry_active':          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_ultraviolet_(uv)_photometry_automatic':        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_uv-p_automatic':                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},
            'environnement_s.a._model_o342m_uv_ozone_analyze_ultraviolet_(uv)_photometry_active':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},
            'enviroo341m_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'enviroo341m_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O341M'}},
            'enviroo342e_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342e'}},
            'enviroo342e_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342e'}},
            'enviroo342e_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342e'}},
            'enviroo342m_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}}, 
            'enviroo342m_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},
            'enviroo342m_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Environnement SA O342M'}},
            'gc855btx-600_ultraviolet_(uv)_photometry_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry',                                                                                                      'measuring_assumption':True}},
            'horiba_model_apma_360e_co_analyser_uv-p_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Horiba APMA360 instrument uses',                                          'measuring_assumption':True}},                                                                                        
            'horibaapma360e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horibaapna360_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminsecence as the reported Horiba APNA360 instrument uses',                             'measuring_assumption':True}},                                                                   
            'horiba_model_apna_360_nox_analyser_chemiluminescence_active':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},
            'horibaapna370_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminsecence as the reported Horiba APNA370 instrument uses',                             'measuring_assumption':True}},                                                                   
            'horiba_model_apoa_300_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA300'}},
            'horiba_model_apoa_300_o3_analyser_ultraviolet_(uv)_photometry_active':                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA300'}},
            'horiba_model_apoa_350_o3_analyser_ultraviolet_(uv)_photometry_active':                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_uv-p_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_ultraviolet_(uv)_photometry_automatic':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_350e_o3_analyser_ultraviolet_(uv)_photometry_active':                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horiba_model_apoa_360_o3_analyser_ultraviolet_(uv)_photometry':                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_360_o3_analyser_ultraviolet_(uv)_photometry_active':                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_360_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_360_o3_analyser_uv-p_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry':                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horiba_model_apoa_370_o3_analyser_uv-p_automatic':                                             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry_automatic':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horiba_model_apoa_370_o3_analyser_ultraviolet_(uv)_photometry_active':                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horibaapoa300_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA300'}},
            'horibaapoa350_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horibaapoa350_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horibaapma360e_ndir_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360'}},
            'horibaapma360e_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Horiba APMA360 instrument uses',                                          'measuring_assumption':True}},                                                                   
            'horibaapna360_chemi_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},
            'horibaapna360_chemiluminescence_automatic':                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Horiba APNA360'}},
            'horibaapoa300_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA300'}},
            'horibaapoa350e_uv-p_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horibaapoa350e_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA350'}},
            'horibaapoa360_other_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horibaapoa360_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horibaapoa360_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horibaapoa360e_uv-p_automatic':                                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},
            'horibaapoa360e_ultraviolet_(uv)_photometry_automatic':                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA360'}},     
            'horibaapoa370_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horibaapoa370_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},
            'horibaapoa370_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Horiba APOA370'}},                                        
            'ion_chromatography':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                     'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'iseo_argopol-sam-xair_uv-p_automatic':                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument ISEO argopol sam-xair, therefore assume uses instrument with ultraviolet photometry method as stated.',                              'measuring_assumption':True}},
            'iseo_argopol-sam-xair_ultraviolet_(uv)_photometry_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument ISEO argopol sam-xair, therefore assume uses instrument with ultraviolet photometry method as stated.',                              'measuring_assumption':True}},
            'mcv48auv_uv-p_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'MCV 48AUV'}},
            'mcv48auv_ultraviolet_(uv)_photometry_automatic':                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'MCV 48AUV'}},   
            'ml2010_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument ML2010, therefore assume uses instrument with ultraviolet photometry method as stated.',                                             'measuring_assumption':True}},
            'ml2010_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument ML2010, therefore assume uses instrument with ultraviolet photometry method as stated.',                                             'measuring_assumption':True}},
            'ml8410_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ethylene chemiluminescence',                                  'instrument':'Teledyne ML8410E',         'process_details':'Assume method is not ultraviolet photometry as stated but is ethylene chemiluminescence as reported instrument uses',                                       'measuring_assumption':True}},
            'ml8810_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'ml8810_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'ml9810_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'ml9810_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'ml9810a_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810A'}},
            'ml9810a_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810A'}},
            'ml9810b_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810B'}},            
            'ml9810b_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810B'}},
            'ml9810b_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810B'}},
            'ml9811_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9811'}},
            'ml9812_uv-p_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9812'}},
            'ml9812_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9812'}},
            'ml_uv-p_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'ml_ultraviolet_(uv)_photometry_automatic':                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'monitor_labs_model_8410_nox_analyser_uv-p_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ethylene chemiluminescence',                                  'instrument':'Teledyne ML8410E',         'process_details':'Assume method is not ultraviolet photometry as stated but is ethylene chemiluminescence as reported instrument uses',                                       'measuring_assumption':True}},              
            'monitor_labs_model_8410_nox_analyser_ultraviolet_(uv)_photometry_active':                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ethylene chemiluminescence',                                  'instrument':'Teledyne ML8410E',         'process_details':'Assume method is not ultraviolet photometry as stated but is ethylene chemiluminescence as reported instrument uses',                                       'measuring_assumption':True}},              
            'monitor_labs_model_8810_o3_analyser_ultraviolet_(uv)_photometry_active':                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_8810_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_8810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML8810'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry_active':                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9810_o3_analyser_ultraviolet_(uv)_photometry_automatic':                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9810'}},
            'monitor_labs_model_9812_o3_analyser_uv-p_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9812'}},
            'monitor_labs_model_9812_o3_analyser_ultraviolet_(uv)_photometry_automatic':                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne ML9812'}},
            'monitor_labs_undetermined_uv-p_automatic':                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},                                                                   
            'monitor_labs_undetermined_ultraviolet_(uv)_photometry_automatic':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},   
            'non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                          'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                                            'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':  {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_with_er_150_differential_optical_absorption_spectroscopy_(doas)':{                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er110_differential_optical_absorption_spectroscopy_(doas)_automatic':               {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er120_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er120_differential_optical_absorption_spectroscopy_(doas)_automatic':               {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_automatic':               {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_remote':                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500_automatic':                                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500_doas_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_automatic':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}}, 
            'opsisar500_doas_remote':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_remote':                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy',                'instrument':'OPSIS AR500'}},
            'orion_uv-p_automatic':                                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet photometry method as stated.',                                              'measuring_assumption':True}},
            'orion_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet photometry method as stated.',                                              'measuring_assumption':True}},
            'other_automatic':                                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_automatic':                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other,_please_specify':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'other_ultraviolet_(uv)_photometry':                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'other_ultraviolet_(uv)_photometry_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'other_uv-fl_automatic':                                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv-p_automatic':                                                                         {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},  
            'passive':                                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'philipsk50094_uv-p_automatic':                                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Philips K50094'}},
            'philipsk50094_ultraviolet_(uv)_photometry_automatic':                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Philips K50094'}},
            'philips_k50094_api_400_ultraviolet_(uv)_photometry_automatic':                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Philips K50094',           'process_details':'Make assumption the instrument is Philips K50094, not Teledyne 400',                                                                                        'measuring_assumption':True}},
            's-5014_ultraviolet_(uv)_photometry_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'SIR S-5014'}},
            'serinus10_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Ecotech Serinus10'}},     
            'serinus10_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Ecotech Serinus10'}},
            'serinus40_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Serinus40 instrument uses',                                  'measuring_assumption':True}},
            'sri8610_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not gas chromatography as the reported SRI 8610 instrument uses',                                  'measuring_assumption':True}},
            'sri8610_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not gas chromatography as the reported SRI 8610 instrument uses',                                  'measuring_assumption':True}},
            'teledyne_api_200_chemiluminescent_nox_analyser_ultraviolet_(uv)_photometry':                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Teledyne 200 instrument uses',                               'measuring_assumption':True}},               
            'teledyne_api_400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_uv-p_automatic':                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_active':               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_other_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400'}},
            'teledyne_api_400_uv_photometric_o3_analyser_uv-fl_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400',             'process_details':'Assume method is not ultraviolet fluorescence as stated but is ultraviolet photometry as reported instrument uses.',                                        'measuring_assumption':True}},
            'teledyne_api_400a_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_active':              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'teledyne_api_400a_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'teledyne_api_400a_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'teledyne_api_400a_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400A'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_other,_please_specify_automatic':                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_active':              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_uv-fl_automatic':                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_400e_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_m400e_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_active':             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne 400E'}},
            'teledyne_api_t300_gas_filter_correlation_co_analyser_ultraviolet_(uv)_photometry':             {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},
            'teledyne_api_t300_gas_filter_correlation_co_analyser_ultraviolet_(uv)_photometry_automatic':   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not NDIR as the reported Teledyne T300 instrument uses',                                           'measuring_assumption':True}},
            'teledyne_api_t400_uv_photometric_o3_analyser_uv-p_automatic':                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'teledyne_api_t400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry':                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'teledyne_api_t400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_automatic':           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'teledyne_api_t400_uv_photometric_o3_analyser_ultraviolet_(uv)_photometry_active':              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Teledyne T400'}},
            'thermo42_chemi_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',           'instrument':'Thermo 42'}},
            'thermo42_ultraviolet_(uv)_photometry_automatic':                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence as the reported Thermo 42 instrument uses',                                  'measuring_assumption':True}},
            'thermo43c_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Thermo 43C instrument uses',                          'measuring_assumption':True}},
            'thermo43c_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not ultraviolet fluorescence as the reported Thermo 43C instrument uses',                          'measuring_assumption':True}},
            'thermo43s_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43S'}},
            'thermo43s_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet fluorescence',                                    'instrument':'Thermo 43S'}},
            'thermo49_automatic':                                                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}},
            'thermo49_uv-p_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}},
            'thermo49_ultraviolet_(uv)_photometry_automatic':                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}},
            'thermo49_ndir_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49',                'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                                            'measuring_assumption':True}},
            'thermo49c_automatic':                                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}}, 
            'thermo49c_chemi_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C',               'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},  
            'thermo49c_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}},                       
            'thermo49c_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}}, 
            'thermo49cps_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49CPS'}},
            'thermo49cps_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49CPS'}},
            'thermo49i_automatic':                                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo49i_chemi_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i',               'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'thermo49i_chemiluminescence_automatic':                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i',               'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet photometry as reported instrument uses.',                                               'measuring_assumption':True}},
            'thermo49i_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo49i_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo49i_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i',               'process_details':'Assume method is not ultraviolet fluorescence as stated but is ultraviolet photometry as reported instrument uses.',                                        'measuring_assumption':True}},
            'thermo49w_uv-p_automatic':                                                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49w'}},
            'thermo49w_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49w'}},
            'thermo_49_cps_ozone_primary_standard_ultraviolet_(uv)_photometry_automatic':                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49CPS'}},
            'thermo_model_42_no_nox_analyser_ultraviolet_(uv)_photometry_automatic':                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption the method is ultraviolet photometry and not chemiluminescence (internal molybdenum converter) as the reported Thermo 42 instrument uses.', 'measuring_assumption':True}},
            'thermo_model_49_o3_analyser_ultraviolet_(uv)_photometry_active':                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}}, 
            'thermo_model_49_o3_analyser_ultraviolet_(uv)_photometry_automatic':                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}}, 
            'thermo_model_49_o3_analyser_uv-p_automatic':                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry_active':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry_automatic':                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}},
            'thermo_model_49c_o3_analyser_ultraviolet_(uv)_photometry':                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}}, 
            'thermo_model_49c_o3_analyser_uv-p_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49C'}}, 
            'thermo_model_49i_o3_analyser_uv-fl_automatic':                                                 {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i',               'process_details':'Assume method is not ultraviolet fluorescnce as stated but is ultraviolet photometry as reported instrument uses.',                                          'measuring_assumption':True}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_active':                              {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry':                                     {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_automatic':                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},
            'thermo_model_49i_o3_analyser_uv-p_automatic':                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                      'instrument':'Thermo 49i'}},                                                                                                                                                                                                                                     
            'too_generic_other_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'too_generic_ultraviolet_(uv)_photometry':                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_ultraviolet_(uv)_photometry_automatic':                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_ultraviolet_(uv)_photometry_not-automatic':                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_ultraviolet_(uv)_photometry_active':                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'too_generic_uv-p_automatic':                                                                   {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'ultraviolet_(uv)_photometry':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'ultraviolet_(uv)_photometry_automatic':                                                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'undetermined_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'undetermined_uv-p_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'undetermined_ultraviolet_(uv)_photometry_automatic':                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'undetermined_unknown_automatic':                                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},         
            'unknown_automatic':                                                                            {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'unknown_chemi_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}},
            'unknown_doas_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                        {                                                                                                                                                                                                                                           'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_light-abs_automatic':                                                                  {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption light absorption refers to ultraviolet photometry.',                                                                                        'measuring_assumption':True}},
            'unknown_light_absorption_automatic':                                                           {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry',                                                                               'process_details':'Make assumption light absorption refers to ultraviolet photometry.',                                                                                        'measuring_assumption':True}},
            'unknown_ultraviolet_(uv)_photometry':                                                          {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                    {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},                                    
            'unknown_other_automatic':                                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_automatic':                                                      {                                                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},      
            'unknown_uv-p_automatic':                                                                       {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'uv-p_automatic':                                                                               {                                                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}}}
#------------------------------------------- 
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {
            'active':                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'airponter_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Airpointer-NOx'}},
            'airpointer_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Airpointer-NOx'}},
            'airpointer_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Airpointer-NOx'}},
            'analysis_automation_mod._447_chemiluminescence_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Cannot find instrument Analysis Automation Mod 447, therefore assumes uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                                             'measuring_assumption':True}},
            'api100e_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                                                           'measuring_assumption':True}},
            'api100e_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses',                                                           'measuring_assumption':True}},
            'api100e_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Teledyne 100E'}},
            'api200_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'api200_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'api200_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'api200a_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'api200a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'api200a_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'api200a_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A',                                                      'process_details':'Assume method is not ultraviolet fluorescence as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                            'measuring_assumption':True}},
            'api200a_gc-fid_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A',                                                      'process_details':'Assume method is not GCFID as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                                               'measuring_assumption':True}},
            'api200au_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200AU'}},
            'api200au_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200AU'}},
            'api200e_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'api_200e_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},    
            'api200e_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'api200e_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'api200e_differential_optical_absorption_spectroscopy_(doas)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E',                                                      'process_details':'Assume method is not DOAS as stated but is chemiluminescence as reported instrument uses.',                                                                                                                                'measuring_assumption':True}},            
            'api200e_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E',                                                      'process_details':'Assume method is not ultraviolet fluorescence as stated but is chemiluminescence (internal molybdenum converter) as reported instrument uses.',                                                                            'measuring_assumption':True}},
            'api200eu-bl_differential_optical_absorption_spectroscopy_(doas)_automatic':                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU',                                                     'process_details':'Assume method is not DOAS as stated but is chemiluminescence as reported instrument uses.',                                                                                                                                'measuring_assumption':True}},
            'api200eu_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'api200eu_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'api200eu_differential_optical_absorption_spectroscopy_(doas)_automatic':                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU',                                                     'process_details':'Assume method is not DOAS as stated but is chemiluminescence as reported instrument uses.',                                                                                                                                'measuring_assumption':True}},
            'apit200u_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200U'}},
            'api200up_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                      'instrument':'Teledyne T200UP',                                                    'process_details':'Assume instrument referenced is Teledyne T200UP',                                                                                                                                                                          'measuring_assumption':True}},
            'api200up_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                      'instrument':'Teledyne T200UP',                                                    'process_details':'Assume instrument referenced is Teledyne T200UP',                                                                                                                                                                          'measuring_assumption':True}},
            'api201a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'api201a_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'apim200e_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'apim200e_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'apit200_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}}, 
            'apit200_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'apit200_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'apit200_spectrophotometry_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200',                                                      'process_details':'Assume method is not spectrophotometry as stated but is chemiluminescence as reported instrument uses.',                                                                                                                   'measuring_assumption':True}},
            'apit200u_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200U'}},
            'aspirator_spectrometry_neda_griess-ylosway_active':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}},  
            'aspirator_spectrophotometry_neda_griess-ylosway_active':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}},
            'automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'bendix_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-B',                                                      'process_details':'Make assumption instrument used is Bendix 8101-B, as only instrument manufactuer is given (Bendix). Methodology given is chemiluminescence, which the instrument uses (make assumption of internal molybdenum converter)', 'measuring_assumption':True}},
            'bendix_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-B',                                                      'process_details':'Make assumption instrument used is Bendix 8101-B, as only instrument manufactuer is given (Bendix). Methodology given is chemiluminescence, which the instrument uses (make assumption of internal molybdenum converter)', 'measuring_assumption':True}},
            'bendix_combustion_engineering_model_8101-c_oxides_of_nitrogen_analyze_chemi_automatic':                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-C'}},
            'bendix_combustion_engineering_model_8101-c_oxides_of_nitrogen_analyze_chemiluminescence_automatic':            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-C'}},
            'bendix_combustion_engineering_model_8101-c_oxides_of_nitrogen_analyze_chemiluminescence_active':               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Bendix 8101-C'}},
            'chemi_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'chemiluminescence_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'chemiluminescence':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'colorimetry_passive':                                                                                          {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'colorimetry'}},
            'csi1600_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'CSI 1600'}},
            'csi1600_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'CSI 1600'}},
            'columbia_scientific_industries_models_1600_chemiluminescence_automatic':                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'CSI 1600'}},
            'dasibi2108_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi2108_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_chemi_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_chemiluminescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_other_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_chemiluminescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Dasibi 2108'}},
            'eco_physics_cld_700_al_chemi_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'ecophysics-cld700al_chemi_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'ecophysics-cld700al_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},                                           
            'ecophysics-cld700al_differential_optical_absorption_spectroscopy_(doas)_automatic':                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL',                                                'process_details':'Make assumption method is Chemiluminescence (internal molybdenum converter) as reported instrument uses, not DOAS',                                                                                                        'measuring_assumption':True}},                                                     
            'eco_physics_cld_700_al_chemiluminescence':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_al_chemiluminescence_active':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_al_chemiluminescence_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'eco_physics_cld_700_chemiluminescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecophysics CLD700AL'}},
            'enrivoac30m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'enrivoac30m_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'enrivoac31m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'enrivoac31m_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'enviroac32e_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32e'}},
            'enviroac32e_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32e'}},  
            'enrivoac32e_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32e'}},         
            'enrivoac32m_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'enrivoac32m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'enrivoac32m_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'enviroas32m_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M',                                             'process_details':'Make assumption method is CAPS as reported instrument uses, not chemiluminescence',                                                                                                                                        'measuring_assumption':True}},
            'enviroas32m_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M',                                             'process_details':'Make assumption method is CAPS as reported instrument uses, not chemiluminescence',                                                                                                                                        'measuring_assumption':True}},
            'enviroas32m_caps_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_other,_please_specify_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac30m_no2_analyzer_chemiluminescence_active':                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC30M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence_active':                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac31m_no2_analyzer_other_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC31M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence_active':                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_ac32m_no2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Environnement SA AC32M'}},
            'environnement_s.a._model_af21m_so2_analyzer_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Environnement SA AF21M instrument uses',                                                  'measuring_assumption':True}},
            'environnement_s.a._model_af21m_so2_analyzer_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Environnement SA AF21M instrument uses',                                                  'measuring_assumption':True}},
            'environnement_s.a._model_af21m_so2_analyzer_uv-fl_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_as32m_nitrogen_dioxide_analyzer_caps':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}}, 
            'environnement_s.a._model_as32m_nitrogen_dioxide_analyzer_caps_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M'}},
            'environnement_s.a._model_as32m_nitrogen_dioxide_analyzer_chemi_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                             'instrument':'Environnement SA AS32M',                                             'process_details':'Make assumption method is CAPS as reported instrument uses, not chemiluminescence',                                                                                                                                        'measuring_assumption':True}},
            'fidas200e_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not optical particle counting as the reported instrument uses',                                                                        'measuring_assumption':True}},
            'filter_pack_spectrophotometry_spectrophotometry_passive':                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry'}},
            'filter_pack_spectrophotometry_spectrophotometry_active':                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry'}},
            'filter_sampling_spectrophotometry_photometry_active':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                     'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}}, 
            'flame_ionization_detection_(fid)_passive':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'flame ionisation detection',                                                                                      'sampling_type':'passive'}},
            'gc855btx-600_chemiluminescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'glass_filter_spectrophotometry_spectrophotometry_active':                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}}, 
            'griess-saltzman_reaction_passive':                                                                             {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}},
            'horibaapha350e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not selective combustion - flame ionisation detection as the reported Horiba APHA350 instrument uses',                                 'measuring_assumption':True}},
            'horibaapha350e_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not selective combustion - flame ionisation detection as the reported Horiba APHA350 instrument uses',                                 'measuring_assumption':True}},    
            'horibaapha360_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not selective combustion - flame ionisation detection as the reported Horiba APHA360 instrument uses',                                 'measuring_assumption':True}},    
            'horibaapma360_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not NDIR as the reported Horiba APMA360 instrument uses',                                                                              'measuring_assumption':True}}, 
            'horibaapna300_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horibaapna300_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horibaapna350_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},       
            'horibaapna350_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna350_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna350e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna350e_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horibaapna360_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360_other,_please_specify_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360_ultraviolet_(uv)_photometry_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360',                                                     'process_details':'Make assumption method is chemiluminescence as reported instrument uses, not ultraviolet photometry',                                                                                                                      'measuring_assumption':True}},
            'horibaapna360e_atomic_absorption_spectrometry_(aas)_chemiluminescence_automatic':                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360e_chemi_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360e_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horibaapna360e_griess-saltzman_reaction_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360',                                                     'process_details':'Make assumption method is chemiluminescence as reported instrument uses, not Colourmetric',                                                                                                                                'measuring_assumption':True}},
            'horibaapna370_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},                          
            'horibaapna370_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_atomic_absorption_spectrometry_(aas)_chemi_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_atomic_absorption_spectrometry_(aas)_chemiluminescence_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_other,_please_specify_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapna370_saltzman_-_colorimetric_microdetermination_of_no2_automatic':                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370',                                                     'process_details':'Make assumption method is chemiluminescence as reported instrument uses, not Colourmetric',                                                                                                                                'measuring_assumption':True}},                 
            'horibaapna370_unknown_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horibaapoa360_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Horiba APOA360 instrument uses',                                                            'measuring_assumption':True}}, 
            'horibaapoa370_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Horiba APOA370 instrument uses',                                                            'measuring_assumption':True}}, 
            'horibaapsa360_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Horiba APSA360 instrument uses',                                                          'measuring_assumption':True}}, 
            'horibaapsa360a_chemiluminescence_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Horiba APSA360 instrument uses',                                                          'measuring_assumption':True}},                     
            'horibaapsa370_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Horiba APSA370 instrument uses',                                                          'measuring_assumption':True}}, 
            'horiba_model_apna_300_nox_analyser_chemiluminescence_active':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horiba_model_apna_300e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horiba_model_apna_300e_nox_analyser_chemiluminescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA300'}},
            'horiba_model_apna_350_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}}, 
            'horiba_model_apna_350_nox_analyser_chemiluminescence_active':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_350e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_350e_nox_analyser_chemiluminescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_350e_nox_analyser_chemiluminescence_active':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA350'}},
            'horiba_model_apna_360_nox_analyser_chemi_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_active':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemi_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemiluminescence_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_ion_chromatography_chemiluminescence_passive':                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_360e_nox_analyser_chemiluminescence_active':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA360'}},
            'horiba_model_apna_370_nox_analyser_chemi_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence_active':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_chemiluminescence_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_other_automatic':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_atomic_absorption_spectrometry_(aas)_chemiluminescence_active':             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_atomic_absorption_spectrometry_(aas)_chemiluminescence_automatic':          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'horiba_model_apna_370_nox_analyser_uv-p_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Horiba APNA370'}},
            'il-punkkinen_for_no2_no_hourly_measurements_saltzman_-_colorimetric_microdetermination_of_no2_active':         {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_griess-saltzman_reaction_active':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                  'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrometry_spectrophotometry_active':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrometry_spectrophotometry_passive':                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_active':                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_not-automatic':                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'impregnated_filter_spectrophotometry_spectrophotometry_active':                                                {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}},  
            'impregnated_filter_spectrophotometry_spectrophotometry_passive':                                               {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}}, 
            'ion_chromatography':                                                                                           {'primary_sampling':{'sampling_type':'passive',                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_passive':                                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'lachat_spectrophotometry_spectrophotometry_active':                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}},
            'mcv30ql_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'MCV 30QL'}},
            'mcv30ql_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'MCV 30QL'}},
            'mcv_30-ql_chemiluminescence':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'MCV 30QL'}},
            'mcv_30-ql_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'MCV 30QL'}},
            'ml2041_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML2041'}},
            'ml2041_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML2041'}},
            'ml8840_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'ml8840_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'ml8841_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},
            'ml8841_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},
            'ml9841_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'ml9841_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'ml9841_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'ml9841a_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841A'}},
            'ml9841a_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841A'}},
            'ml9841b_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'ml9841b_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'ml9841b_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'ml9841t_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'ml9841t_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'monitor_labs_model_8840_nox_analyser_chemi_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'monitor_labs_model_8840_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'monitor_labs_model_8840_nox_analyser_chemiluminescence_active':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8840'}},
            'monitor_labs_model_8841_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},         
            'monitor_labs_model_8841_nox_analyser_chemiluminescence_active':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML8841'}},         
            'monitor_labs_model_9841_nox_analyser_chemiluminescence_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'monitor_labs_model_9841_nox_analyser_chemiluminescence_active':                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841'}},
            'monitor_labs_model_9841a_nox_analyser_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841A'}},
            'monitor_labs_model_9841b_nox_analyser_chemiluminescence_active':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'monitor_labs_model_9841b_nox_analyser_chemi_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'monitor_labs_model_9841b_nox_analyser_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'monitor_labs_model_9841b_nox_analyser_other,_please_specify_automatic':                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841B'}},
            'monitor_labs_model_9841t_nox_analyser_chemiluminescence':                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'monitor_labs_model_9841t_nox_analyser_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'monitor_labs_model_9841t_nox_analyser_chemiluminescence_active':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne ML9841T'}},
            'opsisar500_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500_doas_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500_doas_remote':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_remote':                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er110_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er120_chemiluminescence_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500',                                                        'process_details':'Make assumption method is DOAS as reported instrument uses, not chemiluminescence',                                                                                                                                        'measuring_assumption':True}},
            'opsisar500-er120_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er120_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er120_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_chemiluminescence_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500',                                                        'process_details':'Make assumption method is DOAS as reported instrument uses, not chemiluminescence',                                                                                                                                        'measuring_assumption':True}},
            'opsisar500-er150_doas_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_automatic':                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_remote':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_with_er_150_differential_optical_absorption_spectroscopy_(doas)':                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_active':                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy',                           'instrument':'OPSIS AR500'}},
            'orion_chemi_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Cannot find instrument Orion, therefore assume uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                                                                    'measuring_assumption':True}},
            'orion_chemiluminescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Cannot find instrument Orion, therefore assume uses instrument with chemiluminescence method as stated (assuming uses internal molybdenum converter).',                                                                    'measuring_assumption':True}},
            'other_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'other_chemi_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'other_chemiluminescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'other_chemiluminescence':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'other_photometry_sp_automatic':                                                                                {                                                                                                                               'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'other,_please_specify':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},   
            'other_spectrophotometry_sp_automatic':                                                                         {                                                                                                                               'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'other_spectrophotometry_spectrophotometry_automatic':                                                          {                                                                                                                               'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry',                                                                                               'sampling_type':'continuous injection'}},
            'other_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence'}},
            'palmes_colorimetry_passive_sampler_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'colorimetry'}},
            'passive':                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},              
            'philips_k50034_api_200a_chemiluminescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Philips K50034'}},
            'philips_k50102_no_chemiluminescence_automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Philips K50102'}},
            'philips_k50235_00_no-nox-no2_analyser_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Philips K50235'}},
            'photometry':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'photometry_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'photometry_passive':                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            's-5012_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'SIR S-5012'}},
            'sequential_air_sampler,_type_ss2000._nai-impregnated_glass_sinters_spectrophotometry_spectrophotometry_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU SS2000'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry'}},
            'serinus40_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecotech Serinus40'}},
            'serinus40_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecotech Serinus40'}},
            'serinus40_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecotech Serinus40'}},
            'serinus_40_nox_chemiluminescence':                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Ecotech Serinus40'}},
            'spectrometry_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrometry_passive':                                                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry_passive':                                                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'spectrophotometry'}},
            'teledyne_api_undertermined_chemiluminescence_automatic':                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemi_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200_chemiluminescent_nox_analyser_chemiluminescence_active':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_active':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_other_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_200e_chemiluminescent_nox_analyser_chemiluminescence_active':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_m200e_chemiluminescent_nox_analyser_chemiluminescence_active':                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200E'}},
            'teledyne_api_200eu_chemiluminescent_nox_analyser_-_blue_light_other,_please_specify_active':                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'teledyne_api_200eu_chemiluminescent_nox_analyser_chemiluminescence_active':                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'teledyne_api_200eu_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200EU'}},
            'teledyne_api_400_uv_photometric_o3_analyser_chemi_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Teledyne 400 instrument uses',                                                              'measuring_assumption':True}},
            'teledyne_api_400_uv_photometric_o3_analyser_chemiluminescence_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Teledyne 400 instrument uses',                                                              'measuring_assumption':True}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemiluminescence':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'teledyne_api_t200_chemiluminescent_nox_analyser_chemiluminescence_active':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne T200'}},
            'thermo14b_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B'}},
            'thermo14b_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B'}},
            'thermo14b-e_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo14b-e_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo42_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo42_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo42_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo42c_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo42c_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo42c_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},                 
            'thermo42c_saltzman_-_colorimetric_microdetermination_of_no2_automatic':                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C',                                                         'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not colourmetric',                                                                                                                        'measuring_assumption':True}},
            'thermo42c-tl_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo42c-tl_chemiluminescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo42c-tl_differential_optical_absorption_spectroscopy_(doas)_automatic':                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL',                                                      'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not DOAS',                                                                                                                                'measuring_assumption':True}},
            'thermo42c-tl_other,_please_specify_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo_model_42c-tl_(trace_level_nox)_chemi_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo42i_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo42i_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo42i_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo42i_differential_optical_absorption_spectroscopy_(doas)_automatic':                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i',                                                         'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not DOAS',                                                                                                                                'measuring_assumption':True}},
            'thermo42i_ultraviolet_(uv)_photometry_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i',                                                         'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not ultraviolet photometry',                                                                                                              'measuring_assumption':True}},
            'thermo42i-bl_other_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i',               'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'thermo42i-bl_other,_please_specify_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i',               'overwrite_methodology':'chemiluminescence (internal photolytic converter)'}},
            'thermo42i-tl_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL'}},
            'thermo42i-tl_chemiluminescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL'}},
            'thermo42i-tl_differential_optical_absorption_spectroscopy_(doas)_automatic':                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL',                                                      'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not DOAS',                                                                                                                                'measuring_assumption':True}},
            'thermo42i-tl_saltzman_-_colorimetric_microdetermination_of_no2_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL',                                                      'process_details':'Make assumption the method is chemiluminescence as the reported instrument uses, not colourmetric',                                                                                                                        'measuring_assumption':True}},
            'thermo42s_chemi_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo42s_chemiluminescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo43i-tle_chemi_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Thermo 43i-TLE instrument uses',                                                          'measuring_assumption':True}},
            'thermo43i-tle_chemiluminescence_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet fluorescence as the reported Thermo 43i-TLE instrument uses',                                                          'measuring_assumption':True}},
            'thermo49_chemi_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Thermo 49 instrument uses',                                                                 'measuring_assumption':True}},
            'thermo49_chemiluminescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Make assumption the method is Chemiluminescence (internal molybdenum converter) and not ultraviolet photometry as the reported Thermo 49 instrument uses',                                                                 'measuring_assumption':True}},
            'thermo49_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet photometry',                                                 'instrument':'Thermo 49'}},     
            'thermo450i_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Thermo 450i'}},
            'thermo450i_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'ultraviolet fluorescence',                                               'instrument':'Thermo 450i'}},
            'thermo_model_14b_e_chemiluminescence_no-no2-nox_chemi_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo_model_14b_e_chemiluminescence_no-no2-nox_chemiluminescence_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B/E'}},
            'thermo_model_14b_chemiluminescence_no-no2-nox_chemiluminescence_active':                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 14B'}},
            'thermo_model_42_no_nox_analyser_chemiluminescence_active':                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}}, 
            'thermo_model_42_no_nox_analyser_chemiluminescence':                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}}, 
            'thermo_model_42_no_nox_analyser_chemi_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},                            
            'thermo_model_42_no_nox_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42'}},
            'thermo_model_42c_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_other_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},           
            'thermo_model_42c_no_nox_analyser_other,_please_specify_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c_no_nox_analyser_chemiluminescence_active':                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C'}},
            'thermo_model_42c-tl_(trace_level_nox)_chemiluminescence_active':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo_model_42c-tl_(trace_level_nox)_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42C-TL'}},
            'thermo_model_42i_no_nox_analyser_-_blue_light_other,_please_specify_active':                                   {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence':                                                           {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_active':                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i'}},
            'thermo_model_42i-tl_(trace_level_nox)_chemiluminescence_automatic':                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL'}},
            'thermo_model_42i-tl_(trace_level_nox)_chemiluminescence_active':                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42i-TL'}},
            'thermo_model_42s_no_nox_analyser_chemi_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo_model_42s_no_nox_analyser_chemiluminescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'thermo_model_42s_no_nox_analyser_chemiluminescence_active':                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Thermo 42S'}},
            'too_generic_chemi_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence':                                                                                {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence_active':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_chemiluminescence_not-automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},       
            'too_generic_chemiluminescence_passive':                                                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'too_generic_other_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'undetermined_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'undetermined_caps_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy'}},
            'undetermined_chemi_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'undetermined_chemiluminescence_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'undetermined_unknown_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},              
            'unknown_atomic_absorption_spectrometry_automatic':                                                             {                                                                                                                               'preparation_1':{'sample_preparation_type':'filter',                                                      'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                            'sampling_type':'continuous injection',   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                                                          
            'unknown_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},              
            'unknown_caps_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'cavity attenuated phase shift spectroscopy'}},
            'unknown_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_active':                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_not-automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_passive':                                                                            {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},
            'unknown_chemiluminescence_remote':                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                         'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                                                                                              'measuring_assumption':True}},           
            'unknown_colorimetry_automatic':                                                                                {                                                                                                                               'preparation_1':{'sample_preparation_type':'reagent reaction',                                            'preparation_assumption':True},                                                                                                'measurement':{'methodology':'colorimetry',                                                                                                     'sampling_type':'continuous injection'}},
            'unknown_doas_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_ion_chromatography_chemiluminescence_passive':                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_ion_chromatography_diffusive_sampler_automatic':                                                       {                                                                                                                               'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                                          'measurement':{'methodology':'ion chromatography',                                                                                              'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_diffusive_sampler_passive':                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_unknown_passive':                                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},  
            'unknown_other_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},                                                   
            'unknown_other,_please_specify_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_remote':                                                                         {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_photometry_chemiluminescence_passive':                                                                 {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                   'process_details':'Two different measurement methodologies stated, setting as Unknown.'}}, 
            'unknown_photometry_diffusive_sampler_passive':                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                'measurement':{'methodology':'spectrophotometry'}},                        
            'unknown_unknown_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},
            'unknown_unknown_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}},              
            '':                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'CO':
            mapping = {
            '':                                                                                                            {'measurement':{'methodology':'unknown'}},    
            'api100e_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses', 'measuring_assumption':True}},
            'api100e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Teledyne 100E instrument uses', 'measuring_assumption':True}},                                         
            'api200a_chemi_automatic':                                                                                     {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',             'instrument':'Teledyne 200A'}},
            'api200a_infrared_gas_filter_correlation_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200A instrument uses',        'measuring_assumption':True}},
            'api200a_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200A instrument uses',        'measuring_assumption':True}},
            'api200_ndir_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200 instrument uses',         'measuring_assumption':True}},
            'api200_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200 instrument uses',         'measuring_assumption':True}},
            'api300a_ir-gfc_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'api300a_infrared_gas_filter_correlation_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'api300a_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation) .',        'measuring_assumption':True}},
            'api300a_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation) .',        'measuring_assumption':True}},                 
            'api300e_automatic':                                                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300e_ir-gfc_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300e_infrared_gas_filter_correlation_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300e_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'api300_gc-fid_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300',             'process_details':'Assume method is not GCFID as stated but is NDIR (Gas-Filter Correlation) as reported instrument uses.',                                     'measuring_assumption':True}},                                                  
            'api300_ir-gfc_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_infrared_gas_filter_correlation_automatic':                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_light-abs_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_ndir_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api300_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'api400e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Teledyne 400E instrument does not use NDIR method, make assumption method is NDIR Gas-Filter Correlation',                                   'measuring_assumption':True}},   
            'api_ndir_automatic':                                                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'api_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'apit300_ir-gfc_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'apit300_infrared_gas_filter_correlation_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'apit300_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'apit300_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'apit300_automatic':                                                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'automatic':                                                                                                   {'measurement':{'methodology':'unknown'}},
            'chemi_automatic':                                                                                             {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                'measuring_assumption':True}},
            'chemiluminescence_automatic':                                                                                 {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                'measuring_assumption':True}},
            'chemiluminescence':                                                                                           {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                'measuring_assumption':True}},
            'dasibi3008_ndir_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Dasibi 3008'}},
            'dasibi3008_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Dasibi 3008'}},
            'dasibi_3008_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Dasibi 3008'}},
            'dasibi_3008_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Dasibi 3008'}},
            'enviroco11m_ir-gfc_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},   
            'enviroco11m_infrared_gas_filter_correlation_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'enviroco11m_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'enviroco11m_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'enviroco12e_infrared_gas_filter_correlation_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12e'}}, 
            'enviroco12m_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},        
            'enviroco12m_ir-gfc_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'enviroco12m_infrared_gas_filter_correlation_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'enviroco12m_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'enviroco12m_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},
            'environnement_s.a._model_co10m_co_analyzer_ndir_automatic':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO10M'}},
            'environnement_s.a._model_co10m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO10M'}},
            'environnement_s.a._model_co10m_co_analyzer_other,_please_specify_automatic':                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO10M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_active':               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_ir-gfc_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_infrared_gas_filter_correlation_automatic':                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_ndir_automatic':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)':                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co11m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO11M'}},
            'environnement_s.a._model_co12m_co_analyzer_ndir_automatic':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}},  
            'environnement_s.a._model_co12m_co_analyzer_non-dispersive_infrared_spectroscopy_(ndir)_automatic':            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Environnement SA CO12M'}}, 
            'gc855btx-600_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'horibaapma300_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA300'}},
            'horibaapma300_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA300'}},
            'horibaapma350e_ndir_automatic':                                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horibaapma350e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},                 
            'horibaapma350_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horibaapma350_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horibaapma350e_uv_fluorescence_automatic':                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350',          'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not ultraviolet fluorescence',                                'measuring_assumption':True}},    
            'horibaapna350e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',                                            'process_details':'Make assumption the method is NDIR (cross-flow modulation) and not chemiluminescence as the reported APNA350 instrument uses',                'measuring_assumption':True}},
            'horibaapha360_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',                                            'process_details':'Make assumption the method is NDIR (cross-flow modulation) and not SC-FID as the reported APHA360 instrument uses',                           'measuring_assumption':True}},
            'horibaapma360ce_infrared_gas_filter_correlation_automatic':                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360ce_ndir_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360ce_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                       {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360e_infrared_gas_filter_correlation_automatic':                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360e_ndir_automatic':                                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360e_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360_cavity_ring_down_spectroscopy_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not CRDS',                                                   'measuring_assumption':True}},    
            'horibaapma360_infrared_gas_filter_correlation_automatic':                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapma360_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horibaapsa360_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',                                             'process_details':'Make assumption the method is NDIR (cross-flow modulation) and not ultraviolet fluorescence as the reported APSA360 instrument uses',        'measuring_assumption':True}},
            'horibaapma370_crds_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not CRDS',                                                   'measuring_assumption':True}},
            'horibaapma370_ir-gfc_automatic':                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horibaapma370_infrared_gas_filter_correlation_automatic':                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horibaapma370_light_absorption_automatic':                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not light absorption',                                       'measuring_assumption':True}},
            'horibaapma370_ndir_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horibaapma370_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horibaapma370_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_300_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA300'}},
            'horiba_model_apma_350_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350_co_analyser_ir-gfc_automatic':                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350_co_analyser_infrared_gas_filter_correlation_automatic':                                 {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                       {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_infrared_gas_filter_correlation_automatic':                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_ndir_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_350e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                    {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA350'}},
            'horiba_model_apma_360_co_analyser_cavity_ring_down_spectroscopy_automatic':                                   {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not CRDS',                                                   'measuring_assumption':True}},
            'horiba_model_apma_360_co_analyser_light_absorption_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360',           'process_details':'Assume method is NDIR (cross-flow modulation), as reported instrument uses, and not light absorption',                                       'measuring_assumption':True}},
            'horiba_model_apma_360_co_analyser_infrared_gas_filter_correlation_automatic':                                 {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},    
            'horiba_model_apma_360_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                               {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                              {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_360e_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                       {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA360'}},
            'horiba_model_apma_370_co_analyser_ndir_automatic':                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                               {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                     {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'horiba_model_apma_370_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                        {'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',    'instrument':'Horiba APMA370'}},
            'infrared_gas_filter_correlation':                                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},  
            'ir-gfc_automatic':                                                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'infrared_gas_filter_correlation_automatic':                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'maihak_unor_6n_ir-gfc_automatic':                                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},
            'maihak_unor_6n_infrared_gas_filter_correlation_automatic':                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_non-dispersive_infrared_spectroscopy_(ndir)_automatic':{'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Assume method is NDIR (gas filter correlation), and not UV fluorescence as reported instrument uses',                                        'measuring_assumption':True}}, 
            'ml8830_ndir_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'ml8830_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'ml9830b_automatic':                                                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},         
            'ml9830b_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'ml9830b_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'ml9830_ir-gfc_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'ml9830_infrared_gas_filter_correlation_automatic':                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'ml9830_ndir_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},                                             
            'ml9830_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_8830_co_analyser_ir-gfc_automatic':                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8830_co_analyser_infrared_gas_filter_correlation_automatic':                               {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                      {'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                     'instrument':'Teledyne ML8830'}},
            'monitor_labs_model_8930_co_analyser_infrared_gas_filter_correlation_active':                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML8930'}},                         
            'monitor_labs_model_9830b_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9830b_co_analyser_ndir_automatic':                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9830b_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830B'}},
            'monitor_labs_model_9830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9830_co_analyser_ndir_automatic':                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'monitor_labs_model_9830_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne ML9830'}},
            'ndir_automatic':                                                                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'non-dispersive_infrared_spectroscopy_(ndir)':                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'opsisar500_non-dispersive_infrared_spectroscopy_(ndir)_remote':                                               {'measurement':{'methodology':'differential optical absorption spectroscopy',                  'instrument':'OPSIS AR500',              'process_details':'Assume the method is DOAS as reported instrument used, and not NDIR',                                                                        'measuring_assumption':True}},                                              
            'orion_ndir_automatic':                                                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Orion, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',                 'measuring_assumption':True}},
            'orion_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Orion, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',                 'measuring_assumption':True}},
            'other_beta_automatic':                                                                                        {'measurement':{'methodology':'beta-attenuation'}},
            'other_chemi_automatic':                                                                                       {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                'measuring_assumption':True}},
            'other_crds_automatic':                                                                                        {'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'other_cavity_ring_down_spectroscopy_automatic':                                                               {'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'other_ir-gfc_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'other_infrared_gas_filter_correlation_automatic':                                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'other_light_absorption_automatic':                                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption method light absorption refers to NDIR (Gas-Filter Correlation)',                                                            'measuring_assumption':True}},
            'other_ndir_automatic':                                                                                        {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'other_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'other_automatic':                                                                                             {'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                                       {'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify':                                                                                 {'measurement':{'methodology':'unknown'}},
            'philips_k50109_00_gas_filter_correlation_co_analyser_ndir_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Philips K50109'}},
            'philips_k50109_00_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Philips K50109'}},
            'picarrog2302_cavity_ring_down_spectroscopy_automatic':                                                        {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2302'}},
            'picarrog2302_light_absorption_automatic':                                                                     {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2302',            'process_details':'Assume light absorption refers to CRDS',                                                                                                     'measuring_assumption':True}},
            'picarrog2303_crds_automatic':                                                                                 {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2303'}},
            'picarrog2303_cavity_ring_down_spectroscopy_automatic':                                                        {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2303'}},
            'picarrog2401_crds_automatic':                                                                                 {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2401'}},
            'picarrog2401_cavity_ring_down_spectroscopy_automatic':                                                        {'measurement':{'methodology':'cavity ringdown spectroscopy',                                  'instrument':'Picarro G2401'}},
            's-5006_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'SIR S-5006'}},
            'serinus40_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Serinus40 instrument uses.',           'measuring_assumption':True}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_ndir_automatic':                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200A instrument uses.',       'measuring_assumption':True}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Teledyne 200A instrument uses.',       'measuring_assumption':True}},
            'teledyne_api_200_chemiluminescent_nox_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 200, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',          'measuring_assumption':True}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_ir-gfc_automatic':                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_infrared_gas_filter_correlation_automatic':              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR (Gas-Filter Correlation) method as stated.',                'measuring_assumption':True}},
            'teledyne_api_300a_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Cannot find instrument Teledyne 300a, therefore assume uses instrument with NDIR method as stated (assume Gas-Filter Correlation).',         'measuring_assumption':True}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_ndir_automatic':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_infrared_gas_filter_correlation_automatic':              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300e_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300E'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_ndir_automatic':                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'teledyne_api_300_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_active':      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne 300'}},
            'teledyne_api_t300_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'teledyne_api_t300_gas_filter_correlation_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Teledyne T300'}},
            'teledyne_api_undertermined_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'teledyne_api_undertermined_non-dispersive_infrared_spectroscopy_(ndir)_active':                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'thermo42_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42 instrument uses',            'measuring_assumption':True}},
            'thermo42_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42 instrument uses',            'measuring_assumption':True}},
            'thermo42i_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42i instrument uses',           'measuring_assumption':True}},
            'thermo42i_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not chemiluminescence as the reported Thermo 42i instrument uses',           'measuring_assumption':True}},
            'thermo43i_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Thermo 43i instrument uses',    'measuring_assumption':True}},
            'thermo43i_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Thermo 43i instrument uses',    'measuring_assumption':True}},
            'thermo48c_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48c_infrared_gas_filter_correlation_automatic':                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48c_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48c_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo48i_automatic':                                                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48i_beta_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i',               'process_details':'Make assumption method used is NDIR (Gas-Filter Correlation), as instrument reported uses, and not beta.',                                   'measuring_assumption':True}},
            'thermo48i_ir-gfc_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48i_infrared_gas_filter_correlation_automatic':                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48i_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48i_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo48_ir-gfc_automatic':                                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48_infrared_gas_filter_correlation_automatic':                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo48w_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48W'}},
            'thermo48w_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48W'}},
            'thermo49_ndir_automatic':                                                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49 instrument uses.',      'measuring_assumption':True}},
            'thermo49_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                              {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49 instrument uses.',      'measuring_assumption':True}},
            'thermo49c_ndir_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49C instrument uses.',     'measuring_assumption':True}},
            'thermo49c_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet photometry as the reported Thermo 49C instrument uses.',     'measuring_assumption':True}},
            'thermo_model_43a_so2_analyser_infrared_gas_filter_correlation_automatic':                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption the method is NDIR (Gas-Filter Correlation) and not ultraviolet fluorescence as the reported Thermo 43A instrument uses.',   'measuring_assumption':True}},
            'thermo_model_48c_co_analyser_ir-gfc_automatic':                                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48c_co_analyser_infrared_gas_filter_correlation_automatic':                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48c_co_analyser_ndir_automatic':                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48c_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48C'}},
            'thermo_model_48_co_analyser_ndir_automatic':                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo_model_48_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo_model_48_co_analyser_(gas_filter_correlation)_infrared_gas_filter_correlation_active':                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48'}},
            'thermo_model_48i_co_analyser_infrared_gas_filter_correlation_active':                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_infrared_gas_filter_correlation_automatic':                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_ndir_automatic':                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)':                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'thermo_model_48i_co_analyser_other,_please_specify_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Thermo 48i'}},
            'too_generic_ir-gfc_automatic':                                                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'too_generic_infrared_gas_filter_correlation_automatic':                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'too_generic_ndir_automatic':                                                                                  {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'too_generic_non-dispersive_infrared_spectroscopy_(ndir)':                                                     {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'too_generic_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'undetermined_automatic':                                                                                      {'measurement':{'methodology':'unknown'}},
            'undetermined_ndir_automatic':                                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'undetermined_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_automatic':                                                                                           {'measurement':{'methodology':'unknown'}},
            'unknown_ir-gfc_automatic':                                                                                    {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'unknown_infrared_gas_filter_correlation_automatic':                                                           {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}},
            'unknown_infrared_absorption_not-automatic':                                                                   {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_light-abs_automatic':                                                                                 {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption light absorption refers to NDIR (Gas-Filter Correlation)',                                                                   'measuring_assumption':True}},                                                        
            'unknown_light_absorption_automatic':                                                                          {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Make assumption light absorption refers to NDIR (Gas-Filter Correlation)',                                                                   'measuring_assumption':True}},                                         
            'unknown_ndir_automatic':                                                                                      {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_non-dispersive_infrared_spectroscopy_(ndir)':                                                         {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                            'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation',                                                             'measuring_assumption':True}},
            'unknown_other_automatic':                                                                                     {'measurement':{'methodology':'unknown'}},
            'unknown_vakuum-ultraviolet_(vuv)_spectrometry_automatic':                                                     {'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence'}},
            'unor6n_ndir_automatic':                                                                                       {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},                                           
            'unor6n_infrared_gas_filter_correlation_automatic':                                                            {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},
            'unor6n_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 6N'}},
            'unor610_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                               {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',   'instrument':'Maihak Unor 610'}},
            'vakuum-ultraviolet_(vuv)_spectrometry_automatic':                                                             {'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence'}}}
#------------------------------------------- 
        elif run_details['parameter'] in ['CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','NMVOC','NMHC','ACETYLENE','GLYOXAL','ACETALDEHYDE','ETHANOL','1,3-BUTADIENE','1-BUTENE','ISOBUTANE','N-PENTANE','N-HEXANE','XYLENE','1,2,4-TRIMETHYLBENZENE','BENZO[a]PYRENE']:
            mapping = {
            '':                                                                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},    
            'active':                                                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'af_20_m_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                                                                {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'af_20_m_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'agilent_hp_6890_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'agilenthp6890_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmoozone_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                       {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmoozone_gc-fid_automatic':                                                                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmotec_airmoscan2000_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'airmotec_airmoscan2000_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_passive':                                                                                  {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'airmotec_airmoscan2000_gc-fid_chromatography_active':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'airmotec_btx-1000_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'airmotec_flame_ionization_detection_(fid)_automatic':                                                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmotec_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_with_photo_ionization_detector_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - photoionisation detection'}},
            'airmotec_other,_please_specify_automatic':                                                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'airmovoc_btx1000_automatic':                                                                                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                         'measuring_assumption':True}}, 
            'airmovoc_btx1000_flame_ionization_detection_(fid)_automatic':                                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                         'measuring_assumption':True}},        
            'airmovoc_btx1000_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmovoc_btx1000_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'airmovoc_btx1000_other,_please_specify_automatic':                                                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                         'measuring_assumption':True}},
            'airmovoc_btx_other,_please_specify_automatic':                                                                                                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                         'measuring_assumption':True}},
            'airmovoc_btx_automatic':                                                                                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as airmo BTX 1000 instrument uses',                                                         'measuring_assumption':True}},
            'airmovoc_btx_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'airmovoc_btx_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'airmovoc_btx_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'airtoxica73022_automatic':                                                                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},      
            'airtoxica73022_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'airtoxica73022_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'airtoxica73022_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'alpha400_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'ama_gc5000_btx_type_fid_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_with_photo_ionization_detector_active':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Multiple detectors stated, make assumption detector is FID as is instrument default',                               'measuring_assumption':True}},
            'ama_gc5000_btx_type_fid_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'ama_gc5000_btx_type_fid_other,_please_specify_active':                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},
            'annox_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'automatic':                                                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},      
            'carbopack_b_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry'}}, 
            'carbopack_b_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry'}}, 
            'carbotap_tube_300_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'cavity_ring_down_spectroscopy_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'cavity ringdown spectroscopy'}},   
            'chemiluminescence':                                                                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',       'measuring_assumption':True}},              
            'chemiluminescence_automatic':                                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                      'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',       'measuring_assumption':True}},   
            'chrompack_btx_cp7001_monitor_gas_chromatography_automatic':                                                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'chrompack_btx_cp7001_monitor_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'chrompack_btx_cp7001_monitor_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},                                                                  
            'chrompack_btx_cp7001_monitor_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_passive':                                                                            {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'chrompack_voc-air_other,_please_specify_chromatography_active':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'cp7001-btx_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                       {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'cpv-cov_gc-ms_type_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},
            'dasibi_2108_nox_analyser_gc-fid_chemiluminescence_active':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                              'instrument':'Dasibi 2108'}},
            'dasibi_2108_nox_analyser_gc-fid_chemiluminescence_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                              'instrument':'Dasibi 2108'}},
            'desconocido_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'diffusive_tenaxta_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'diffusive_tenaxta_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'diffusive_tenaxta_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_passive':                                                                                          {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'digitel_active_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'digitel_active_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_passive':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},            
            'draeger_niosh_(active_sampling)_chromatography_not-automatic':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection'}}, 
            'draeger_niosh_(active_sampling)_chromatography_active':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection'}}, 
            'draeger_niosh_(active_sampling)_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_not-automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry'}},                                                                       
            'draeger_niosh_(active_sampling)_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry'}},                                                                       
            'environnement_voc71m_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'environnement_voc71m_gc-fid_gas_chromatography_with_photo_ionization_detector_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection'}},
            'environnement_voc71m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},           
            'environnement_voc72m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)':                                                                                       {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection',   'process_details':'Assume automatic continuous injection',                                                                            'measuring_assumption':True}},
            'environnement_voc72m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - photoionisation detection'}},
            'environnement_voc72m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'flame_ionization_detection_(fid)_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',         'measuring_assumption':True}},
            'flame_ionization_detection_(fid)':                                                                                                                                              {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',         'measuring_assumption':True}},
            'gas_chromatography_automatic':                                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},               
            'gas_chromatography_+_mass_spectrometry_(gc-ms)_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                                                                        {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'gc5000btx-fid_gas_chromatography_automatic':                                                                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc5000btx-fid_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc5000btx-fid_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                          'sampling_type':'continuous injection'}},
            'gc5000btx-fid_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection',                                  'sampling_type':'continuous injection'}},
            'gc5000btx-fid_other,_please_specify_automatic':                                                                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc5000btx-fid_other,_please_specify_chromatography_automatic':                                                                                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc5000btx-pid_automatic':                                                                                                                                                       {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},  
            'gc5000btx-pid_other,_please_specify_automatic':                                                                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},        
            'gc5000btx-pid_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection',                                  'sampling_type':'continuous injection'}},
            'gc5000btx-pid_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc5000btx-pid_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry - photoionisation detection',                                           'sampling_type':'continuous injection'}},
            'gc855btx-600_automatic':                                                                                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},          
            'gc855btx-600_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc855btx-600_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc855btx_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc855btx_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'gc855btx_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc855btx_gc-fid_chromatography_automatic':                                                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'gc955btx_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},           
            'gc955-600_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                                       
            'gc955-600_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'gc955-600_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc955-600_other,_please_specify_automatic':                                                                                                                                     {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gc955-600_automatic':                                                                                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gc955-601_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                                       
            'gc955-800_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                                   
            'gc955-800_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                        {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc955_gas_chromatography_automatic':                                                                                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gc955_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                                       
            'gc955_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'gc955_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc955_gas_chromatography_with_photo_ionization_detector_automatic':                                                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'gc955_gc-fid_type_gas_chromatography_with_photo_ionization_detector_automatic':                                                                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection',                                  'sampling_type':'continuous injection'}},
            'gc955_other,_please_specify_automatic':                                                                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gc955_ultraviolet_(uv)_photometry_automatic':                                                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection',   'process_details':'Ultraviolet Photometry cannot measure VOCs and GC instrument referenced, so assume method is GC-unknown detection', 'measuring_assumption':True}}, 
            'gc955_automatic':                                                                                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'gc-fid':                                                                                                                                                                        {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'gc-fid_active':                                                                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'gc-fid_automatic':                                                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                                       
            'gc-fid_passive':                                                                                                                                                                {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'gc-fid_type_active':                                                                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'gc-fid_type_passive':                                                                                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'gc-ms_type_active':                                                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'gravimetric_analysis_active':                                                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                      'preparation_assumption':True},                                                                 'measurement':{'methodology':'gravimetry'}},
            'horibaapha350e_flame_ionization_detection_(fid)_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA350'}},
            'horibaapha360_flame_ionization_detection_(fid)_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA360'}},
            'horibaapma360_flame_ionization_detection_(fid)_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is not NDIR (cross-flow modulation) as instrument uses, but is GC-FID as stated',                     'measuring_assumption':True}},                                                                                       
            'horibaapha370_flame_ionization_detection_(fid)_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA370'}},                                                                                                               
            'horibaapma370_flame_ionization_detection_(fid)_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is not NDIR (cross-flow modulation) as instrument uses, but is GC-FID as stated',                     'measuring_assumption':True}},                                                                                       
            'horiba_model_apha_360_hydrocarbons_analyser_flame_ionization_detection_(fid)_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'selective combustion - flame ionisation detection',                              'instrument':'Horiba APHA360'}},
            'hp_5890_2_-_hewlett_packard_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}}, 
            'hp-6890_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_not-automatic':                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'hp-6890_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'markes_international_gc-fid_passive_sampler_passive':                                                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'markes_thermal_desorber_+_agilent_gas_chromatograph_mass_spectrometer_gas_chromatography_+_mass_spectrometry_gc-ms_after_solvent_o_active':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'markes_thermal_desorber_+_agilent_gas_chromatograph_mass_spectrometer_gc-ms_type_gas_chromatography_+_mass_spectrometry_gc-ms_after_solvent_o_active':                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},                         
            'opsisar500_chemiluminescence_remote':                                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500',                                             'process_details':'Assume method is not chemiluminescence as stated but DOAS as OPSIS AR500 instrument uses',                          'measuring_assumption':True}},                               
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_remote':                                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500'}}, 
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500'}}, 
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_remote':                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500'}}, 
            'opsis_ar500_open_path_monitor_with_er_150_differential_optical_absorption_spectroscopy_(doas)':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'differential optical absorption spectroscopy',                                   'instrument':'OPSIS AR500'}}, 
            'orbo_32_s_gc-fid_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                                                       {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Conflicting methods stated, assume GC-FID',                                                                         'measuring_assumption':True}},
            'orsa_passive_sampler_chromatography_active':                                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'orsa_passive_sampler_chromatography_not-automatic':                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'orsa_passive_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'orsa_passive_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_passive':                                                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'orsa_passive_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'orsa_passive_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_passive':                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'orsa_passive_sampler_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'other_automatic':                                                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'other_flame_ionization_detection_(fid)_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},   
            'other_gas_chromatography_automatic':                                                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'other_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'other_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'other_gc-fid_automatic':                                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'other_gc-fid_chromatography_remote':                                                                                                                                            {'primary_sampling':{'sampling_type':'passive',                                                                                                                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'other_gc-fid_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'other_gc-fid_passive_sampler_remote':                                                                                                                                           {'primary_sampling':{'sampling_type':'passive',                                                                                                                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'other_gc-fid_remote':                                                                                                                                                           {'primary_sampling':{'sampling_type':'passive',                                                                                                                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'other_other,_please_specify_automatic':                                                                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'other,_please_specify':                                                                                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},           
            'other,_please_specify_automatic':                                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'other,_please_specify_passive':                                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},                
            'passive':                                                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'passive_sampler_(carbopack_e),perkin-elmer_atd_400,hp_gc_5890,hp_ms_5792_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':{'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}}, 
            'passive_sampling_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler',  'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'passive_sampling_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'pcf_r526_flame_ionization_detection_(fid)_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'pcf_mod.527_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'perkinelmerc500_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},                                                                           
            'perkinelmerc500_gc-fid_automatic':                                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'perkinelmerc500_flame_ionization_detection_(fid)_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'perkin_elmer_clarus_500_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_active':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'perkin_elmer_turbomatrix_desorber,_perkin_elmer_clarus_500_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                 {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'philips_k50031_chromatographic_analyser_flame_ionization_detection_(fid)_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'philips_k50031_chromatographic_analyser_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'pumped_tube_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'rivm_-_universal_sampler_gc-fid_type_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_automatic':                                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Rivm Universal Sampler'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection',                                 'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_serie_600_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_chromatography_automatic':                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                  {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gc-fid_chromatography_automatic':                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gc-fid_chromatography_passive':                                                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_other,_please_specify_automatic':                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)':                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume automatic continuous injection',                                                                             'measuring_assumption':True}},
            'sintech_spectras_btx_gc_855_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_active':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},                                                     
            'sintech_spectras_btx_gc_955_series_undetermined_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},          
            'skypostgas_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Terora Skypost Gas'},                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}}, 
            'sri8610_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'sri_8610_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                                               {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'sri_8610_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                         {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},   
            'sri_8610_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                   {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},           
            'sri8610_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                    {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'sri8610_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'swam5adual_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID as reported, not beta-attenuation as reported instrument uses',                             'measuring_assumption':True}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_automatic':                                                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_active':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},            
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                            {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_passive':                                                        {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - photoionisation detection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_followed_by_flame_ionization_detection_(g_passive':               {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},
            'syntech_spectras_gc_955_series_undetermined_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_with_photo_ionization_detector_active':                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - photoionisation detection'}},
            'syntech_spectras_gc_955_series_undetermined_gc-fid_type_gas_chromatography_with_photo_ionization_detector_passive':                                                             {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection'}},
            'syntech_spectras_gc_955_series_undetermined_gc-fid_type_gas_chromatography_with_photo_ionization_detector_automatic':                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection - photoionisation detection',                                  'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-600_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)':                                                                                          {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection',   'process_details':'Assume automatic continuous injection',                                                                             'measuring_assumption':True}},
            'syntech_spectras_gc_955-600_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-800_gas_chromatography_with_photo_ionization_detector_automatic':                                                                                       {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'syntech_spectras_gc_955-800_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                      {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},            
            'tecora_other,_please_specify_gas_chromotography_(gc)_active':                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Terora Skypost Gas', 'process_details':'Make assumption sampling instrument is TCR Terora Skypost Gas.', 'primary_sampling_assumption':True},                                                                                                                                                                                                           'measurement':{'methodology':'gas chromatography - unknown detection'}}, 
            'teflon_bag+desorption_tube_gc-fid_type_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_active':                                                                    {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                               'preparation_1':{'sample_preparation_type':'bag'},                                                                                        'preparation_2':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}}, 
            'thermo_scientific:_trace_gc_ultra_+_trace_ms_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'too_generic_flame_ionization_detection_(fid)_active':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True},                                                                                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},          
            'too_generic_flame_ionization_detection_(fid)_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},
            'too_generic_flame_ionization_detection_(fid)':                                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True},                                                                                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},
            'undetermined_automatic':                                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},         
            'undetermined_flame_ionization_detection_(fid)_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},
            'undetermined_gas_chromatography_(ref)_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},          
            'undetermined_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'undetermined_gc-fid_automatic':                                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'undetermined_gc-fid_chromatography_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},        
            'unknown_atomic_absorption_spectrometry_(aas)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                        'process_details':'Conflicting methods stated, assume GC-FID',                                                                         'measuring_assumption':True}},
            'unknown_automatic':                                                                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},         
            'unknown_chromatography_active':                                                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_chromatography_automatic':                                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'unknown_chromatography_passive':                                                                                                                                                {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},         
            'unknown_flame_ionization_detection_(fid)_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},
            'unknown_gas_chromatography_active':                                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromatography_automatic':                                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'unknown_gas_chromotography_(gc)_not-automatic':                                                                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)':                                                                                                              {'primary_sampling':{'sampling_type':'high volume continuous',                                                                                                                           'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_followed_by_mass_spectrometry_(gc-ms)_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                       'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'unknown_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_chromatography_active':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - photoionisation detection'}},
            'unknown_gas_chromatography_-_mass_spectrometry_active':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_chromatography_passive':                                                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_active':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_followed_by_flame_ionization_detection_(g_passive':                                                   {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                 {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_active':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_passive':                                                                                             {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_passive':                                                                                                                {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_unknown_active':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_unknown_passive':                                                                                                        {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_passive':                                                                                                                                            {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromotography_(gc)_active':                                                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - unknown detection'}},
            'unknown_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gas_chromatography_(ref)_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},   
            'unknown_gc-fid_chromatography_active':                                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_chromatography_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'unknown_gc-fid_chromatography_passive':                                                                                                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_active':                                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_-_flame_ionization_active':                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_followed_by_flame_ionization_detection_(g_passive':                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}},
            'unknown_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(fid)_+_ecd_(gc-f_passive':                                                                                           {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection - electron capture detection'}},
            'unknown_gc-fid_gas_chromatography_(gc)_+_flame_ionisation_(gc-fid)_active':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromatography_(ref)_active':                                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_gas_chromotography_(gc)_active':                                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_passive_sampler_passive':                                                                                                                                        {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_type_chromatography_active':                                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_type_chromatography_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'unknown_gc-fid_type_chromatography_passive':                                                                                                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_unknown_active':                                                                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-fid_unknown_passive':                                                                                                                                                {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                              'preparation_1':{'sample_preparation_type':'flask',                                                       'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'unknown_gc-ms_type_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection'}},                                                                                  
            'unknown_gc-ms_type_gas_chromatography_-_mass_spectrometry_active':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gc-ms_type_gas_chromotography_(gc)_active':                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gc-ms_type_unknown_active':                                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                        {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown',                                                                                                                                                'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},        
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_gas_chromatography_followed_by_flame_ionization_detection_(g_active':                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown',                                                                                                                                                'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gas_chromotography_(gc)_active':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                            'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',                                            'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_other_automatic':                                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_remote':                                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'unknown_remote':                                                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},         
            'unknown_ultraviolet_(uv)_photometry_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'ultraviolet photometry'}},
            'unknown_unknown_automatic':                                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'varian_chrompack_gc-fid_automatic':                                                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'varian_chrompack_other,_please_specify_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                                       'sampling_type':'continuous injection'}},
            'varian3800_gc-fid_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                       'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'voc71m_automatic':                                                                                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID',                                                                                           'measuring_assumption':True}},
            'voc71m_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'voc71m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}},
            'voc72m_automatic':                                                                                                                                                              {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID',                                                                                           'measuring_assumption':True}},
            'voc72m_flame_ionization_detection_(fid)_automatic':                                                                                                                             {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Method stated purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',          'measuring_assumption':True}},       
            'voc72m_gas_chromatography_automatic':                                                                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection',   'process_details':'Assume method is GC-FID',                                                                                           'measuring_assumption':True}},
            'voc72m_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                              'sampling_type':'continuous injection'}},
            'voc72m_gas_chromatography_followed_by_photo_ionization_detection_(gc-pid)_automatic':                                                                                           {                                                                                                                                                                                                                             'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                                                                                           'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                               'sampling_type':'continuous injection'}}}
#------------------------------------------- 
        elif run_details['parameter'] in ['VOC','HC']:
            mapping = {
            'horibaapha350e_flame_ionization_detection_(fid)_automatic':                         {'measurement':{'methodology':'selective combustion - flame ionisation detection', 'instrument':'Horiba APHA350'}},    
            'horibaapha360_flame_ionization_detection_(fid)_automatic':                          {'measurement':{'methodology':'selective combustion - flame ionisation detection', 'instrument':'Horiba APHA360'}},          
            'other_gas_chromatography_followed_by_flame_ionization_detection_(gc-fid)_automatic':{'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                    'sampling_type':'continuous injection'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HCHO':
            mapping = {
            'airmoozone_spectrophotometry_automatic':                                                                                                          {                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                       'sampling_type':'continuous injection', 'process_details':'Assume method is GC-FID not spectrophotometry as airmoozone system uses GC-FID', 'measuring_assumption':True}},
            'gc-fid_passive':                                                                                                                                  {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask',             'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},       
            'high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_passive':                                                              {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask',             'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},
            'spectrophotometry_active':                                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',  'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_active':                                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',  'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_high_performance_liquid_chromatography_-_mass_spectrometry_(hplc-ms)_high_performance_liquid_chromatography_-_fluorescence_detect_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping',  'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - mass spectrometry - fluorescence detection'}}} 
#------------------------------------------- 
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '':                                                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'airponter_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Airpointer-SO2'}},
            'airpointer_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Airpointer-SO2'}},
            'api100_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'api100_coulometry_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'api100_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'api100_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},         
            'api100a_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},   
            'api100a_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'api100a_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'api100e_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'api100e_ndir_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E',                                                      'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'api100e_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'api100e_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'api101_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101'}},
            'api101_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101'}},
            'api101a_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101A'}},
            'api101a_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101A'}},
            'api101a_ultraviolet_(uv)_photometry_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 101A',                                                      'process_details':'Assume method is not ultraviolet photometry as stated but is ultraviolet fluorescence as reported instrument uses.',                     'measuring_assumption':True}},
            'apit100_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100'}},
            'apit100_cv-afs_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100',                                                      'process_details':'Assume method is not CV-AFS as stated but is ultraviolet fluorescence as reported instrument uses.',                                     'measuring_assumption':True}},
            'apit100_light_scattering_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100',                                                      'process_details':'Assume method is not light scattering as stated but is ultraviolet fluorescence as reported instrument uses.',                           'measuring_assumption':True}},
            'apit100_ion_chromatography_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100',                                                      'process_details':'Assume method is not ion chromatography as stated but is ultraviolet fluorescence as reported instrument uses.',                         'measuring_assumption':True}},          
            'apit100_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100'}},
            'apit100_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne T100'}},
            'automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'bubbler_24_h_titration_titration_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction'},                                                                                                                                                      'measurement':{'methodology':'titration'}}, 
            'chemi_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence',           'measuring_assumption':True}},
            'chemiluminescence_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence',           'measuring_assumption':True}},
            'chemiluminescence':                                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence',           'measuring_assumption':True}},
            'columbia_scientific_ic_sa700_other_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'columbia_scientific_ic_sa700_other,_please_specify_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'columbia_scientific_ic_sa700_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'columbia_scientific_ic_sa700_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'dani_environment_af_21_m_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_active':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'danish_lmp_filterpack_poll._collector_ion_chromatography_ion_chromatography_active':                              {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'dasibi4108_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'dasibi4108_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'dasibi4108_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'dasibi_4108_so2_analyser_uv-fl_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'dasibi_4108_so2_analyser_uv_fluorescence_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Dasibi 4108'}},
            'diffusion_tube_ion_chromatography_passive':                                                                       {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'diffusion_tube_ion_chromatography_ion_chromatography_passive':                                                    {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'sorbent trapping',  'specific_technique':'Diffusive Sampler'},                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'enc_37_other,_please_specify_unknown_active':                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'enveaaf22e_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Envea AF22e'}},                                                                                 
            'enviroaf20m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF20M'}}, 
            'enviroaf20m_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF20M'}}, 
            'enviroaf21m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'enviroaf21m_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'enviroaf21m_ultraviolet_(uv)_photometry_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M',                                             'process_details':'Assume method is ultraviolet fluorescence as reported instrument uses, not ultraviolet photometry',                                      'measuring_assumption':True}},                                    
            'enviroaf22m_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},          
            'enviroaf22m_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'enviroaf22m_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af20m_so2_analyzer_uv_fluorescence_active':                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF20M'}},
            'environnement_s.a._model_af21m_so2_analyzer_other_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_other,_please_specify_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_active':                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af21m_so2_analyzer_uv_fluorescence_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF21M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv_fluorescence_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'environnement_s.a._model_af22m_so2_analyzer_uv_fluorescence_active':                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Environnement SA AF22M'}},
            'filter_pack_ion_chromatography_spectrophotometry_active':                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography',                                                                                                                            'process_details':'Both ion chromatography and spectrophotometry are given as methods, assume method is ion chromatography.',                               'measuring_assumption':True}},
            'filter_sampling_capillary_electrophoresis_(ce)_capillary_electrophoresis_active':                                 {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                'measurement':{'methodology':'capillary electrophoresis'}},
            'filter_sampling_spectrophotometry_ion_chromatography_active':                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                            'process_details':'Both ion chromatography and spectrophotometry are given as methods, assume method is ion chromatography.',                               'measuring_assumption':True}},
            'filter_spectrophotometry_thorin_active':                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                      'preparation_2':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'}, 'measurement':{'methodology':'spectrophotometry'}},
            'filter_unknown_active':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'gas_meter_other,_please_specify_unknown_active':                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'gc855btx-600_uv_fluorescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence',                                                                                 'measuring_assumption':True}},     
            'grimm-edm180c_light_scattering_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'optical particle counter',                                   'instrument':'Grimm 180'}},
            'horibaapha360_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not SC-FID as the reported Horiba APHA360 instrument uses.',                  'measuring_assumption':True}},
            'horibaapma360_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not NDIR as the reported Horiba APMA360 instrument uses.',                    'measuring_assumption':True}},
            'horiba_model_apna_360_nox_analyser_chemiluminescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}}, 
            'horiba_model_apna_360_nox_analyser_chemiluminescence_active':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}}, 
            'horibaapna370_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Horiba APNA370 instrument uses.',       'measuring_assumption':True}},
            'horiba_model_apsa_350_so2_analyser_uv_fluorescence_active':                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horiba_model_apsa_350e_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horiba_model_apsa_360_so2_analyser_uv-fl_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360_so2_analyser_uv_fluorescence_active':                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360e_so2_analyser_uv_fluorescence_active':                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_360e_so2_analyser_uv_fluorescence_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horiba_model_apsa_370_so2_analyser_uv-fl_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horiba_model_apsa_370_so2_analyser_uv_fluorescence_active':                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horibaapda351e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not beta-attenuation as the reported Horiba APDA351 instrument uses.',        'measuring_assumption':True}},
            'horibaapda351e_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not beta-attenuation as the reported Horiba APDA351 instrument uses.',        'measuring_assumption':True}},
            'horibaapma370_ndir_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)', 'instrument':'Horiba APMA370'}},
            'horibaapma370_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)', 'instrument':'Horiba APMA370'}},
            'horibaapna360_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}},
            'horibaapna360_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Horiba APNA360'}},
            'horibaapna360_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported instrument uses.',                      'measuring_assumption':True}},                                                                      
            'horibaapsa350_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},    
            'horibaapsa350_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horibaapsa350e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horibaapsa350e_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA350'}},
            'horibaapsa360_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360a_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},          
            'horibaapsa360a_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360a_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360e_uv-fl_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360e_uv_fluorescence_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360'}},
            'horibaapsa360e_ion_chromatography_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA360',                                                     'process_details':'Assume method is not ion chromatography as stated but is ultraviolet fluorescence as reported instrument uses.',                         'measuring_assumption':True}},
            'horibaapsa370_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horibaapsa370_chemi_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370',                                                     'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'horibaapsa370_chemiluminescence_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370',                                                     'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'horibaapsa370_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'horibaapsa370_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Horiba APSA370'}},
            'ion_chromatography':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}}, 
            'ion_chromatography_passive':                                                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'light-scat_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'light scattering photometry'}},
            'light_scattering_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'light scattering photometry'}},
            'lvs_spectrophotometry_not-automatic':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous'},                                                              'preparation_1':{'sample_preparation_type':'reagent reaction',  'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'spectrophotometry'}},
            'macleod_collector_sf1_uv_fluorescence_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Macleod collector SF1, therefore assume uses instrument with ultraviolet fluorescence method as stated.',         'measuring_assumption':True}},        
            'mcv64fuv_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'MCV 64FUV'}},
            'mcv_64-fuv_uv_fluorescence':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'MCV 64FUV'}},
            'mcv_64-fuv_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'MCV 64FUV'}},
            'mcv64fuv_uv_fluorescence_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'MCV 64FUV'}},
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_uv-fl_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'meloy_model_sa_700_fluorescence_sulfur_dioxide_analyze_uv_fluorescence_automatic':                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'meloysa700_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'meloysa700_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Meloy SA700'}},
            'ml2040_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Teledyne ML2040, therefore assume uses instrument with ultraviolet fluorescence method as stated.',               'measuring_assumption':True}},
            'ml2050_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Teledyne ML2050, therefore assume uses instrument with ultraviolet fluorescence method as stated.',               'measuring_assumption':True}},
            'ml2050_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Teledyne ML2050, therefore assume uses instrument with ultraviolet fluorescence method as stated.',               'measuring_assumption':True}},
            'ml8850_coulometry_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850',                                                    'process_details':'Assume method is not coulometry as stated but is ultraviolet fluorescence as reported instrument uses.',                                 'measuring_assumption':True}},
            'ml8850_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'ml8850_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'ml8850s_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850S'}},
            'ml8850s_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850S'}},
            'ml9810_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Teledyne ML9810 instrument uses.', 'measuring_assumption':True}},                                                                                                                                                                   
            'ml9810_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Teledyne ML9810 instrument uses.', 'measuring_assumption':True}},
            'ml9850_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850',                                                    'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'ml9850_uv-fl_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'ml9850_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'ml9850b_ndir_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B',                                                   'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'ml9850b_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B',                                                   'process_details':'Assume method is not NDIR as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'ml9850b_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B'}},
            'ml9850b_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B'}},
            'ml_uv-fl_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'ml_uv_fluorescence_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'monitor_labs_model_8850_so2_analyser_ion_chromatography_uv_fluorescence_passive':                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850',                                                    'process_details':'Assume method is continuous ultraviolet fluorescence, as Teledyne ML8850 instrument uses.',                                              'measuring_assumption':True}},
            'monitor_labs_model_8850_so2_analyser_uv-fl_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850_so2_analyser_uv_fluorescence_active':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850_so2_analyser_uv_fluorescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850_so2_analyser_uv_fluorescence_passive':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850'}},
            'monitor_labs_model_8850s_so2_analyser_uv_fluorescence_active':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850S'}},
            'monitor_labs_model_8850s_so2_analyser_uv_fluorescence_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML8850S'}},
            'monitor_labs_model_9810_o3_analyser_uv_fluorescence':                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Teledyne ML9810 instrument uses.', 'measuring_assumption':True}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence_active':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv_fluorescence_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850_so2_analyser_uv-fl_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850'}},
            'monitor_labs_model_9850b_so2_analyser_uv_fluorescence_active':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne ML9850B'}},
            'monitor_labs_undetermined_uv-fl_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'monitor_labs_undetermined_uv_fluorescence_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'nilu_filter_pack_ion_chromatography_active':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}}, 
            'opsis_ar500_open_path_monitor_with_er_150_differential_optical_absorption_spectroscopy_(doas)':                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_active':                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er110_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er110_differential_optical_absorption_spectroscopy_(doas)_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er120_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er120_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er120_differential_optical_absorption_spectroscopy_(doas)_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_doas_remote':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500-er150_differential_optical_absorption_spectroscopy_(doas)_remote':                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_doas_remote':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'opsisar500_differential_optical_absorption_spectroscopy_(doas)_remote':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy',               'instrument':'OPSIS AR500'}},
            'orion_chemi_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Cannot find instrument Orion, therefore assume uses gas chromatography - sulphur chemiluminescence.',                                    'measuring_assumption':True}},
            'orion_chemiluminescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Cannot find instrument Orion, therefore assume uses gas chromatography - sulphur chemiluminescence.',                                    'measuring_assumption':True}},
            'orion_uv-fl_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet fluorescence method as stated.',                         'measuring_assumption':True}},
            'orion_uv_fluorescence_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Cannot find instrument Orion, therefore assume uses instrument with ultraviolet fluorescence method as stated.',                         'measuring_assumption':True}},
            'other_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'other_ion_chromatography_other_automatic':                                                                        {                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'other_ion_chromatography_unknown_automatic':                                                                      {                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                                  'sampling_type':'continuous injection'}},
            'other_photometry_other_automatic':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},            'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},            'measurement':{'methodology':'spectrophotometry'}},  
            'other_photometry_unknown_automatic':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},            'preparation_2':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},            'measurement':{'methodology':'spectrophotometry'}},
            'other,_please_specify_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'other_uv-fl_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv_fluorescence_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'other_uv_fluorescence':                                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'passive':                                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'philips_k50033_api_100e_uv_fluorescence_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Philips K50033'}},                                                      
            'philips_pw9755_00_so2_analyser_coulometry_active':                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'coulometry',                                                 'instrument':'Philips PW9755'}}, 
            's-5001_uv_fluorescence_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'SIR S-5001'}},
            'sequential_air_sampler,_type_ek._koh-impregnated_whatman_40_filter_ion_chromatography_ion_chromatography_active': {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'NILU EK'},                                      'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'seres_sf2_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Seres SF2000G',                                                      'process_details':'Cannot find Seres SF2 instrument, assume is Seres SF2000G instrument.',                                                                  'measuring_assumption':True}},
            'serinus10_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not ultraviolet photometry as the reported Serinus 10 instrument uses.',      'measuring_assumption':True}},
            'serinus50_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Ecotech Serinus50'}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence_active':                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A',                                                      'process_details':'Cannot find Teledyne 100 instrument, assume is Teledyne 100A instrument.',                                                               'measuring_assumption':True}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'teledyne_api_100a_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100A'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv-fl_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'teledyne_api_100e_uv_fluorescent_so2_analyser_uv_fluorescence_active':                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Teledyne 100E'}},
            'thermo42i_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Thermo 42i'}},
            'thermo42i_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',          'instrument':'Thermo 42i'}},
            'thermo42i_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Thermo 42i instrument uses.',           'measuring_assumption':True}},
            'thermo42i_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Thermo 42i instrument uses.',           'measuring_assumption':True}},
            'thermo43_uv-fl_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo43_uv_fluorescence_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo43a_coulometry_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A',                                                         'process_details':'Assume method is not coulometry as stated but is ultraviolet fluorescence as reported instrument uses.',                                 'measuring_assumption':True}},
            'thermo43a_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo43a_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo43a_uv-p_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A',                                                         'process_details':'Assume method is not ultraviolet photometry as stated but is ultraviolet fluorescence as reported instrument uses.',                     'measuring_assumption':True}},
            'thermo43b_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43B'}},
            'thermo43b_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43B'}},
            'thermo43c_coulometry_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C',                                                         'process_details':'Assume method is not coulometry as stated but is ultraviolet fluorescence as reported instrument uses.',                                 'measuring_assumption':True}},
            'thermo43c_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo43c_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo43ctl_differential_optical_absorption_spectroscopy_(doas)_automatic':                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE',                                                     'process_details':'Assume method is not DOAS as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},                                            
            'thermo43ctl_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE'}},
            'thermo43ctl_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE'}},
            'thermo43i-tle_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},      
            'thermo43i-tle_differential_optical_absorption_spectroscopy_(doas)_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE',                                                     'process_details':'Assume method is not DOAS as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},                                            
            'thermo43i-tle_uv-fl_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo43i-tle_uv_fluorescence_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo43i_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo43i_chemi_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo43i_chemiluminescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},   
            'thermo43i_cv-afs_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not CV-AFS as stated but is ultraviolet fluorescence as reported instrument uses.',                                     'measuring_assumption':True}},
            'thermo43i_differential_optical_absorption_spectroscopy_(doas)_automatic':                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not DOAS as stated but is ultraviolet fluorescence as reported instrument uses.',                                       'measuring_assumption':True}},
            'thermo43i_ion_chromatography_unknown_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not ion chromatography as stated but is ultraviolet fluorescence as reported instrument uses.',                         'measuring_assumption':True}},
            'thermo43i_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo43i_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo43s_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43S'}},
            'thermo43s_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43S'}},
            'thermo43w_other,_please_specify_uv-fl_automatic':                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43W'}},
            'thermo43w_uv-fl_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43W'}},
            'thermo43w_uv_fluorescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43W'}},
            'thermo450i_chemi_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i',                                                        'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo450i_uv-fl_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i'}},
            'thermo48_ndir_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)','instrument':'Thermo 48'}},
            'thermo48_uv_fluorescence_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Assume method is ultraviolet fluorescence as stated, not NDIR as reported instrument uses.',                                             'measuring_assumption':True}},
            'thermo49_ndir_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49',                                                          'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                         'measuring_assumption':True}},
            'thermo49_non-dispersive_infrared_spectroscopy_(ndir)_automatic':                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49',                                                          'process_details':'Assume method is not NDIR as stated but is ultraviolet photometry as reported instrument uses.',                                         'measuring_assumption':True}},
            'thermo49_uv_fluorescence_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Assume method is ultraviolet fluorescence as stated, not ultraviolet photometry as reported instrument uses.',                           'measuring_assumption':True}},
            'thermo_450i_chemiluminescence':                                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i',                                                        'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo450i_chemiluminescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i',                                                        'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},
            'thermo_450i_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i'}},
            'thermo450i_uv_fluorescence_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 450i'}},
            'thermo_model_42_no_nox_analyser_uv_fluorescence_automatic':                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Thermo 42 instrument uses.',            'measuring_assumption':True}},
            'thermo_model_42i_no_nox_analyser_uv_fluorescence':                                                                {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                                                                                                      'process_details':'Make assumption the method is ultraviolet fluorescence and not chemiluminescence as the reported Thermo 42i instrument uses.',           'measuring_assumption':True}},
            'thermo_model_43_so2_analyser_uv-fl_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43_so2_analyser_uv_fluorescence_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43_so2_analyser_uv_fluorescence_active':                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43'}},
            'thermo_model_43a_so2_analyser_other_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_other,_please_specify_automatic':                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},
            'thermo_model_43a_so2_analyser_uv_fluorescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43A'}},                                                
            'thermo_model_43b_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43B'}},
            'thermo_model_43c_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c_so2_analyser_uv_fluorescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C'}},
            'thermo_model_43c-tl_so2_analyser_uv_fluorescence_active':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE'}},
            'thermo_model_43c-tl_so2_analyser_uv_fluorescence_automatic':                                                      {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43C-TLE'}},
            'thermo_model_43h_pulsed_fluorescence_so2_analyser_uv-fl_automatic':                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43H'}},
            'thermo_model_43h_pulsed_fluorescence_so2_analyser_uv_fluorescence_automatic':                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43H'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv-fl_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv_fluorescence_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv_fluorescence_not-automatic':                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i-tle_(enhanced_trace_level)_uv_fluorescence_active':                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i-TLE'}},
            'thermo_model_43i_so2_analyser_uv-fl_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_uv_fluorescence':                                                                   {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_uv_fluorescence_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43i_so2_analyser_chemiluminescence_active':                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i',                                                         'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                          'measuring_assumption':True}},                                                   
            'thermo_model_43i_so2_analyser_uv_fluorescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43i'}},
            'thermo_model_43s_so2_analyser_uv_fluorescence_active':                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence',                                   'instrument':'Thermo 43S'}},
            'thermo_model_49i_o3_analyser_ultraviolet_(uv)_photometry_automatic':                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet photometry',                                     'instrument':'Thermo 49i'}},
            'too_generic_other_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'too_generic_uv-fl_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence':                                                                                     {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence_active':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'too_generic_uv_fluorescence_passive':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'total_ans_filter_pack_ece-emep_ion_chromatography_ion_chromatography_active':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},      
            'total_ans_filter_pack_ece-emep_ion_chromatography_ion_chromatography_passive':                                    {'primary_sampling':{'sampling_type':'passive'},                                                                            'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                                                           'measurement':{'methodology':'ion chromatography'}},
            'undetermined_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'undetermined_uv-fl_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'undetermined_uv_fluorescence_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},     
            'unknown_chemi_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence',           'measuring_assumption':True}},
            'unknown_chemiluminescence_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                                                                                'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence',           'measuring_assumption':True}},
            'unknown_doas_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                           {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'differential optical absorption spectroscopy'}},
            'unknown_gas_chromatography_followed_by_flame_ionization_detection_(g_automatic':                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                     'sampling_type':'continuous injection'}},
            'unknown_ion_chromatography_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',            'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_uv_fluorescence_passive':                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown',                                                                                                                                       'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_light_scattering_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'light scattering photometry'}},
            'unknown_other_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'unknown_other,_please_specify_remote':                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'unknown_other_remote':                                                                                            {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}}, 
            'unknown_spectrophotometry_not-automatic':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction',  'preparation_assumption':True},                                                                                                                      'measurement':{'methodology':'spectrophotometry'}},
            'unknown_unknown_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'unknown'}},
            'unknown_uv-fl_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_uv_fluorescence_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'unknown_uv_fluorescence_remote':                                                                                  {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},         
            'uv-fl_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv_fluorescence_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}},
            'uv_fluorescence':                                                                                                 {                                                                                                                                                                                                                                                                                                                                                'measurement':{'methodology':'ultraviolet fluorescence'}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'NH3':
            mapping = {
            'api201a_chemi_automatic':                                                    {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'api201a_chemiluminescence_automatic':                                        {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)', 'instrument':'Teledyne 201A'}},
            'chemi_automatic':                                                            {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'chemiluminescence_automatic':                                                {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'ion_chromatography_active':                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',           'preparation_assumption':True},            'measurement':{'methodology':'ion chromatography'}}, 
            'other_chemi_automatic':                                                      {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'other_chemiluminescence_automatic':                                          {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},
            'other_other_automatic':                                                      {                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_automatic':                                      {                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other_potentiometry_automatic':                                              {                                                                                                                                                                                                            'measurement':{'methodology':'conductimetry'}},
            'passive':                                                                    {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'spectrophotometry',                                                                                      'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                                                         'measuring_assumption':True}}, 
            'spectrophotometry':                                                          {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},            'measurement':{'methodology':'spectrophotometry'}},
            'spectrophotometry_passive':                                                  {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True},            'measurement':{'methodology':'spectrophotometry'}},
            'teledyne_api_200a_chemiluminescent_nox_analyser_chemiluminescence_automatic':{                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                      'instrument':'Teledyne 200A'}}, 
            'undetermined_potentiometry_automatic':                                       {                                                                                                                                                                                                            'measurement':{'methodology':'conductimetry'}},
            'unknown_automatic':                                                          {                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_chemi_automatic':                                                    {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}},                                                                                                                                                                                                          
            'unknown_chemiluminescence_automatic':                                        {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and stainless steel converters)',                                 'process_details':'Only chemiluminescence stated as method, make assumption method is Chemiluminescence (internal molybdenum and stainless steel converters).', 'measuring_assumption':True}}}  
#------------------------------------------- 
        elif run_details['parameter'] == 'HNO3':
            mapping = { 
            'active':                   {                                                                                                                                                                                       'measurement':{'methodology':'unknown'}},
            'ion_chromatography_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}}} 
#-------------------------------------------
        elif run_details['parameter'] == 'PAN':
            mapping = {}
#-------------------------------------------
        elif run_details['parameter'] == 'HCl':
            mapping = {
            'ion_chromatography_active':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}}} 
#-------------------------------------------
        elif run_details['parameter'] == 'HF':
            mapping = {}   
#-------------------------------------------
        elif run_details['parameter'] == 'H2S':
            mapping = {
            'airpointer_uv_fluorescence_automatic':                                     {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Airpointer-SO2'}}, 
            'api101_uv_fluorescence_automatic':                                         {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 101'}},
            'apit100_uv_fluorescence_automatic':                                        {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne T100'}},
            'enviroaf21m_uv_fluorescence_automatic':                                    {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF21M'}},
            'enviroaf22m_uv_fluorescence_automatic':                                    {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF22M'}},   
            'horibaapsa370_automatic':                                                  {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA370'}},
            'horibaapsa370_chemiluminescence_automatic':                                {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA370',           'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                'measuring_assumption':True}},
            'horibaapsa370_uv_fluorescence_automatic':                                  {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA370'}},
            'other_chemiluminescence_automatic':                                        {'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                          'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses gas chromatography - sulphur chemiluminescence', 'measuring_assumption':True}},
            'other_uv_fluorescence_automatic':                                          {'measurement':{'methodology':'ultraviolet fluorescence'}},
            'sintech_spectras_btx_gc_855_series_undetermined_uv_fluorescence_automatic':{'measurement':{'methodology':'ultraviolet fluorescence',                                                                'process_details':'Make assumption the method is ultraviolet fluorescence and not GC as the reported sintech instrument uses.',                   'measuring_assumption':True}},
            'thermo43i_uv_fluorescence_automatic':                                      {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43i'}},
            'thermo450i_chemiluminescence_automatic':                                   {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 450i',              'process_details':'Assume method is not chemiluminescence as stated but is ultraviolet fluorescence as reported instrument uses.',                'measuring_assumption':True}},
            'thermo450i_uv_fluorescence_automatic':                                     {'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 450i'}},
            'unknown_uv_fluorescence_automatic':                                        {'measurement':{'methodology':'ultraviolet fluorescence'}}}
#-------------------------------------------     
        elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
            mapping = {
            '':                                                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'active':                                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'automatic':                                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'andersenfh621n_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh621n_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh621n_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh621n_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh62c14_beta_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'andersenfh62c14_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'andersenfh62c14_light_scattering_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},              
            'andersenfh62in_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh62in_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersenfh62ir_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_gravimetric_analysis_beta_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}}, 
            'andersenfh62ir_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'andersenfh62ir_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_tapered_element_oscillating_microbalance_(teom)_with_filter_automatic':            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},                                                                                                
            'andersenfh62ir_light_scattering_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},                                               
            'andersenfh62ir_nephelometry_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not lnephelometry as stated but uses beta-attenuation as reported instrument uses.',                                               'measuring_assumption':True}},                                               
            'andersenfh62ir_nephelometry_+_beta_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not beta-attenuation - nephelometry as stated but uses beta-attenuation as reported instrument uses.',                             'measuring_assumption':True}},
            'andersenfh62ir_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle cunting as stated but uses beta-attenuation as reported instrument uses.',                                    'measuring_assumption':True}},
            'andersenfh62ir_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'andersenfh62ir_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'andersenfh621n_gravimetric_analysis_beta_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},            
            'andersen_gmw_model_fh621-n_beta_monitor_beta_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_beta_ray_attenuation_active':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},  
            'andersen_gmw_model_fh621-n_beta_monitor_beta_ray_attenuation_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_gravimetric_analysis_beta_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_gravimetric_analysis_beta_ray_attenuation_active':                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_gravimetric_analysis_beta_ray_attenuation_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_light-scat_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_light_scattering_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_nephelometry_beta_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_nephelometry_+_beta_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_opc-cmc_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_other,_please_specify_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersen_gmw_model_fh621-n_beta_monitor_tapered_element_oscillating_microbalance_(teom)_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Andersen FH621-N'}},
            'andersengps-1-modif_beta_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Andersen GPS-1 Modified, therefore assume uses Beta-Attenuation as stated.',                                                 'measuring_assumption':True}},
            'andersengps-1-modif_beta_ray_attenuation_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Andersen GPS-1 Modified, therefore assume uses Beta-Attenuation as stated.',                                                 'measuring_assumption':True}},
            'api200_chemi_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Teledyne 200'}},
            'api200e_chemi_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Teledyne 200E'}},
            'apit100_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                     'instrument':'Teledyne T100'}}, 
            'apit100_beta_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'apit100_beta_ray_attenuation_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'apit100_opc-cmc_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'apit100_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'apit100_optoelectronic_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                                     'instrument':'Teledyne T100',                                                                                                                                                                        'process_details':'Assume method is ultraviolet fluorescence as reported instrument uses, and not optoelectronic.',                                                    'measuring_assumption':True}}, 
            'apm-1_elecos_beta_automatic':                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'apm-1_elecos_beta_ray_attenuation_automatic':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation'}},
            'apm-1_elecos_nephelometry_automatic':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Elecos APM-1'},                                                                                                                                                                                                                                         'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not nephelometry as stated but uses beta-attenuation.',                                                                            'measuring_assumption':True}},
            'aqd_-_analytical_techniques*_active':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},       
            'atmoservice_lvs_3d_gravimetric_analysis_gravimetric_analysis_active':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}},
            'atmoservice_pns3d15_gravimetric_analysis_active':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}},
            'atmoservice_pns3d15_gravimetric_analysis_gravimetric_analysis_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}}, 
            'bam1020_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_beta_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_beta_ray_attenuation_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_gravimetric_analysis_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},           
            'bam1020_gravimetric_analysis_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_nephelometry_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not nephelometry as stated but uses beta-attenuation as reported instrument uses.',                                                'measuring_assumption':True}},
            'bam1020heated_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},                                                                                             
            'bam1020heated_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},                                                                                             
            'bam1020heated_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020heated_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'bam1020heated_teom-fdms_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'bam1020heated_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020heated_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'bam1020heated_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020heated_unknown_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_light_scattering_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},                
            'bam1020_nephelometry_+_beta_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation - nephelometry as stated but uses beta-attenuation as reported instrument uses.',                             'measuring_assumption':True}},
            'bam1020_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'bam1020_other,_please_specify_beta_ray_attenuation_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_other,_please_specify_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'bam1020_teom_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'bam1020unheated_beta_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020unheated_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'bam1020_uv_fluorescence_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not ultraviolet fluorescence as stated but uses beta-attenuation as reported instrument uses.',                                    'measuring_assumption':True}},
            'beta_automatic':                                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_beta_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_ray_attenuation':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_ray_attenuation_active':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'beta_ray_attenuation_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'cav_-_a_gravimetric_analysis_gravimetry_active':                                                                                                           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'MCV CAV-A/mb',          'process_details':'Make assumption sampling instrument is MCV CAV-A/mb.',          'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'chemiluminescence':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                                                 'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                       'measuring_assumption':True}},
            'chemiluminescence_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                                                 'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',                                       'measuring_assumption':True}},
            'das_rancon_adam_c_beta_ray_attenuation_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                                           'measuring_assumption':True}},        
            'dasibi7001_beta_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Dasibi 7001'}},
            'dasibi7001_beta_ray_attenuation_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Dasibi 7001'}}, 
            'dasibi_7001_beta_ray_attenuation_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Dasibi 7001'}}, 
            'derenda_gravimetric_analysis_active':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'derenda_gravimetric_analysis_gravimetric_analysis_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'derenda_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'derenda_gravimetric_analysis_not-automatic':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Comde Derenda PNS 16T', 'process_details':'Make assumption sampling instrument is Comde Derenda PNS 16T.', 'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'desconocido_gravimetric_analysis_gravimetric_analysis_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'desconocido_gravimetric_analysis_not-automatic':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'dha-80_gravimetry_active':                                                                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'dha-80_gravimetric_analysis_gravimetry_active':                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_active':                                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_gravimetric_analysis_active':                                                                                           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha80_gravimetric_analysis_not-automatic':                                                                                                         {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_gravimetric_analysis_gravimetry_active':                                                                                                    {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_active':                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_automatic':                                                             {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_dha-80_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                                  {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_dha80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                         {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_dha-80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hivol_gravimetry_active':                                                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_gravimetry_automatic':                                                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_active':                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hivol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_automatic':                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},                                           
            'digitel_hivol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                                   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hivol_other_automatic':                                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hivol_da_80_h_gravimetric_analysis_gravimetry_automatic':                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_gravimetric_analysis_active':                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},                 
            'digitel_hvs_dha80_gravimetric_analysis_gravimetric_analysis_active':                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_gravimetric_analysis_not-automatic':                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_dha80_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                                                                                                                                       'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is Gravimetry, as is almost aways used after use of low volume samplers.',                                                            'measuring_assumption':True}},
            'digitel_hvs_gravimetric_analysis_active':                                                                                                                  {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'digitel_hvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                             {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',        'process_details':'Make assumption sampling instrument is Digitel DHA-80.',        'primary_sampling_assumption':True},                                                                                                         'measurement':{'methodology':'gravimetry'}},
            'eberlinefh62-1_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'eberlinefh62-1_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'eberlinefh62-1_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'eberlinefh62-1_teom_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'eberline_model_fh_62-1_beta_ray_attenuation_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Eberline FH62-1'}},
            'enviromp101m_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_gravimetric_analysis_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_gravimetric_analysis_beta_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp101m_light_scattering_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not light-scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},
            'enviromp101m_nephelometry_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_nephelometry_+_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'enviromp101m_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'enviromp101m_teom_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'enviromp101m_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'enviromp101m_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'enviromp101m_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM-FDMS as stated but uses beta-attenuation as reported instrument uses.',                                                   'measuring_assumption':True}},
            'enviromp101m_unknown_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}}, 
            'enviromp2.5m_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                                             'measuring_assumption':True}},
            'enviromp2.5m_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                                             'measuring_assumption':True}},
            'enviromp2.5m_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                                             'measuring_assumption':True}},
            'envirompsi100_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'envirompsi100_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'enviroo331m_light_scattering_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                                                                                                                                                                                                                       'process_details':'Assume method is light scattering photometry and not ultraviolet photometry as reported instrument uses',                                           'measuring_assumption':True}},
            'enviroo331m_light_scattering_not-automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                                                                                                                                                                                                                       'process_details':'Assume method is light scattering photometry and not ultraviolet photometry as reported instrument uses',                                           'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation':                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation_active':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_ray_attenuation_not-automatic':                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_beta_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_other_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_(may_be_used_for_pm10_and_pm2.5)_teom_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_ray_attenuation_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_beta_ray_attenuation_active':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_gravimetric_analysis_beta_ray_attenuation_active':                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mp101m_pm10_monitor_nephelometry_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not nephelometry as stated but uses beta-attenuation as reported instrument uses.',                                                'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_opc-cmc_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'environnement_s.a._model_mp101m_pm10_monitor_other,_please_specify_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'environnement_s.a._model_mpsi_100_beta_ray_attenuation_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._model_mpsi_100_gravimetric_analysis_beta_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._model_mpsi_100_gravimetric_analysis_beta_ray_attenuation_active':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._model_mpsi_100_nephelometry_+_beta_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MPSI100'}},
            'environnement_s.a._model_mp2.5m_undefined_beta_ray_attenuation_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M',                                                                                                                                                              'process_details':'Cannot find instrument Environnement SA MP2.5M, assume it is Environnement SA MP101M.',                                                             'measuring_assumption':True}},
            'f-701-20_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_beta_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_beta_ray_attenuation_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_verewa___durag_beta_ray_attenuation':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'f-701-20_verewa___durag_beta_ray_attenuation_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701-20'}},
            'fidas200_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_beta_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200_beta_ray_attenuation_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200_gravimetric_analysis_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_opc-cmc_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not gravimetry as stated but uses optical particle counting as reported instrument uses.',                                         'measuring_assumption':True}},
            'fidas200_light-scat_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'fidas200_light_scattering_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'fidas200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not ICP-MS as stated but uses optical particle counting as reported instrument uses.',                                             'measuring_assumption':True}},
            'fidas200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_ray_attenuation_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200_nephelometry_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'fidas200_opc-cmc_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_other,_please_specify_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},        
            'fidas200_scanning_mobilty_particle_sizer_(smps)_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
            'fidas200_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'fidas200_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'fidas200_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'fidas200e_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E'}},
            'fidas200e_beta_ray_attenuation_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E',                                                                                                                                                                     'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200e_gravimetric_analysis_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E'}},
            'fidas200e_nephelometry_+_beta_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E',                                                                                                                                                                     'process_details':'Assume method is not nephelometry+beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                      'measuring_assumption':True}},
            'fidas200e_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E'}},
            'fidas200e_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E',                                                                                                                                                                     'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},       
            'fidas200e_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E',                                                                                                                                                                     'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'fidas200e_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E',                                                                                                                                                                     'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'fidas200s_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S'}},
            'fidas200s_beta_ray_attenuation_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S',                                                                                                                                                                     'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'fidas200s_light-scat_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S',                                                                                                                                                                     'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'fidas200s_light_scattering_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S',                                                                                                                                                                     'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'fidas200s_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S'}},
            'fidas200s_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200S',                                                                                                                                                                     'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'gravimetric_analysis':                                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'gravimetric_analysis_active':                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'gravimetric_analysis_automatic':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                           'primary_sampling_assumption':True},                    'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},  
            'gravimetric_analysis_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                                                                                                                                                                                         'measuring_assumption':True}},
            'grimm-edm180_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180mc_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180c_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180d_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},     
            'grimm_edm_180_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm+edm+180_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180_gravimetric_analysis_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180mc_gravimetric_analysis_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm180_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_gravimetric_analysis_light-scat_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180_gravimetric_analysis_light_scattering_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_edm_180_gravimetric_analysis_opc-cmc_automatic':                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_gravimetric_analysis_opc-cmc_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_gravimetric_analysis_scattering_of_laser_light_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180_light_absorption_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180_light-scat_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180_light_scattering_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_edm_180_light_scattering_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180c_light_scattering_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180mc_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180mc_gravimetric_analysis_opc-cmc_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Cannot find instrument Grimm 180-MC, therefore assume instrument is Grimm 180.',                                                                    'measuring_assumption':True}},                             
            'grimm-edm180mc_opc-cmc_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Cannot find instrument Grimm 180-MC, therefore assume instrument is Grimm 180.',                                                                    'measuring_assumption':True}},                                                                                                            
            'grimm-edm180mc_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_nephelometry_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},                                              
            'grimm+edm+180_nephelometry_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'grimm_edm_180_opc-cmc_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_opc-cmc_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_edm_180_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_optoelectronic_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_other,_please_specify_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm-edm180_scattering_of_laser_light_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm180c_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},      
            'grimm-edm180d_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},   
            'grimm_edm_180_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm-edm180_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm+edm+180_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm_edm_180_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'grimm-edm180_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'grimm-edm180_teom_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm-edm365_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm365_light-scat_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm365_light_scattering_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 365',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm107_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm107_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm-edm107_light-scat_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm-edm107_light_scattering_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_differential_optical_absorption_spectroscopy_(doas)_active':                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},             
            '"grimm_model_edm_107_for_pm10,_pm2.5,_pm1_and_tc"_differential_optical_absorption_spectroscopy_(doas)_automatic':                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 107',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},             
            'grimm_model_edm_180_for_pm10_and_pm2.5_beta_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not DOAS as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_gravimetric_analysis_light_scattering_active':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_gravimetric_analysis_scattering_of_laser_light_active':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_laser_spectroscopy_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not laser spectroscopy as stated but uses optical particle counting as reported instrument uses.',                                 'measuring_assumption':True}},          
            'grimm_model_edm_180_for_pm10_and_pm2.5_light_scattering_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_light-scat_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_nephelometry_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_nephelometry_active':                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not nephelometry as stated but uses optical particle counting as reported instrument uses.',                                       'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_opc-cmc_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optical_particle_counter_+_conversion_to_mass_concentration':                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_optoelectronic_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not optoelectronic as stated but uses optical particle counting as reported instrument uses.',                                     'measuring_assumption':True}},                   
            'grimm_model_edm_180_for_pm10_and_pm2.5_other_automatic':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_other,_please_specify_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_scattering_of_laser_light_active':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180_for_pm10_and_pm2.5_scattering_of_laser_light_automatic':                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180',                                                                                                                                                                            'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'grimm_model_edm_180-mc_for_pm10,_pm2.5,_pm1,_tc_and_31_size_channels_optical_particle_counter_+_conversion_to_mass_concentration_active':                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Grimm 180'}}, 
            'high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_active':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},                                      
            'horibaapda350_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E',                                                                                                                                                                      'process_details':'Cannot find Horiba APDA350 instrument, assume it is Horiba APDA350E.',                                                                              'measuring_assumption':True}},
            'horibaapda350_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E',                                                                                                                                                                      'process_details':'Cannot find Horiba APDA350 instrument, assume it is Horiba APDA350E.',                                                                              'measuring_assumption':True}},
            'horibaapda351e_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA351E'}},
            'horibaapda351e_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA351E'}},
            'horibaapda371_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371'}},
            'horibaapda371_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371'}},
            'horibaapda371_light-scat_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371',                                                                                                                                                                       'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},
            'horibaapda371_light_scattering_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371',                                                                                                                                                                       'process_details':'Assume method is not light scattering photometry as stated but uses beta-attenuation as reported instrument uses.',                                 'measuring_assumption':True}},
            'horibaapda372_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Horiba APDA372'}},
            'horibaapda372_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Horiba APDA372',                                                                                                                                                                       'process_details':'Assume method is not beta-attenuation as stated but uses light scattering photometry as reported instrument uses.',                                 'measuring_assumption':True}},            
            'horibaapda372_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Horiba APDA372',                                                                                                                                                                       'process_details':'Assume method is not TEOM as stated but uses light scattering photometry as reported instrument uses.',                                             'measuring_assumption':True}},        
            'horibaapna370_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not chemiluminescence (internal molybdenum converter) as reported instrument uses.',                'measuring_assumption':True}},
            'horibaapna370_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not chemiluminescence (internal molybdenum converter) as reported instrument uses.',                'measuring_assumption':True}},
            'horibafh621r_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba FH621-R'}},
            'horibafh621r_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba FH621-R'}},
            'horiba_model_apda_350e_tsp_analyser_gravimetric_analysis_beta_ray_attenuation_active':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA350E'}},
            'horiba_model_apda_371_beta_ray_attenuation':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Horiba APDA371'}},
            'hvs_gravimetric_analysis_active':                                                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},               
            'hvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                                                                                                                 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_active':                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_active':                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_gravimetric_analysis_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_2.5_gravimetric_analysis_not-automatic':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'ivl_pmodell_s_10_gravimetric_analysis_not-automatic':                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IVL P-Model'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'kimoto_168s_beta_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto 168S'}},
            'kimoto_168s_beta_ray_attenuation_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto 168S'}},
            'leckel_-_lvs_3_gravimetric_analysis_active':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_-_lvs_3_gravimetric_analysis_gravimetric_analysis_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'leckel_sequential_lvs_lvs_-_automatic_filter_change_2,3_m3_h_active':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3',           'process_details':'Make assumption sampling instrument is Leckel LVS3.',                              'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'light_scattering':                                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'light-scat_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'light_scattering_active':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},   
            'light_scattering_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'lvs_gravimetric_analysis_active':                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'lvs_gravimetric_analysis_gravimetric_analysis_active':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'lvs_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'mcz,_pns-lvs_16_seq_gravimetric_analysis_active':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'mcz,_pns-lvs_16_seq_gravimetric_analysis_gravimetric_analysis_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry'}},
            'metone_bam-1020_beta_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},        
            'metone_bam-1020_beta_ray_attenuation':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_beta_ray_attenuation_active':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},  
            'metone_bam-1020_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_gravimetric_analysis_beta_ray_attenuation_active':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_ray_attenuation':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_beta_ray_attenuation_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_other_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_heated_teom_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'metone_bam-1020_other_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_other,_please_specify_beta_ray_attenuation_active':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},
            'metone_bam-1020_teom_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},
            'ml9841b_nephelometry_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                                                                                                                                                                                                                                   'process_details':'Assume method is beta-attenuation - nephelometry as stated and not chemiluminescence (internal molybdenum converter) as reported instrument uses.', 'measuring_assumption':True}},
            'muhlberg_filter_disk_beta_ray_attenuation_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Muhlberg filter disc, therefore just assume uses beta-attenuation method as stated.',                                        'measuring_assumption':True}},
            'nephelometry_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'nephelometry'}},
            'nephelometry_+_beta':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'nephelometry_+_beta_active':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'nephelometry_+_beta_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'nephelometry_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'opc-cmc_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'opsis_ar500_open_path_monitor_beta_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not DOAS as reported instrument uses.',                                                             'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_beta_ray_attenuation_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation as stated and not DOAS as reported instrument uses.',                                                             'measuring_assumption':True}},
            'opsis_ar500_open_path_monitor_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                                                 'instrument':'OPSIS AR500'}},
            'opsis_ar500_open_path_monitor_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',                                                 'instrument':'OPSIS AR500'}},
            'opsissm200_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_beta_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_beta_ray_attenuation_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_gravimetric_analysis_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not gravimetry as stated, but uses beta-attenuation as reported instrument uses.',                                                 'measuring_assumption':True}},
            'opsis_sm200_gravimetric_analysis_active':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not gravimetry as stated, but uses beta-attenuation as reported instrument uses.',                                                 'measuring_assumption':True}},
            'opsissm200_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_other_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_other,_please_specify_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsissm200_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not TEOM as stated, but uses beta-attenuation as reported instrument uses.',                                                       'measuring_assumption':True}},
            'opsissm200_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not TEOM-FDMS as stated, but uses beta-attenuation as reported instrument uses.',                                                  'measuring_assumption':True}},
            'opsis_sm200_gravimetric_analysis_gravimetric_analysis_active':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200',                                                                                                                                                                          'process_details':'Assume method is not gravimetry as stated, but uses beta-attenuation as reported instrument uses.',                                                 'measuring_assumption':True}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_automatic':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation_active':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'opsis_sm200_particulate_analyser_with_pm2.5_&_pm10_heads_beta_ray_attenuation_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'optical_particle_counter_+_conversion_to_mass_concentration':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'orion_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                                                   'measuring_assumption':True}},
            'orion_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Orion, therefore assume uses Beta-Attenuation as stated.',                                                                   'measuring_assumption':True}},
            'osiris_nephelometry_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Turnkey Osiris'}},
            'other_automatic':                                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_beta_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},                                                                                     
            'other_beta_ray_attenuation_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'other_beta_ray_attenuation_active':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'other_gravimetric_analysis_automatic':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'other_gravimetric_analysis_beta_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation',                                                                                                                 'measuring_assumption':True}},
            'other_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation',                                                                                                                 'measuring_assumption':True}},
            'other_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'other_gravimetric_analysis_gravimetry_automatic':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'other_gravimetric_analysis_other_automatic':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'other_gravimetric_analysis_teom_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'other_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_automatic':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                       'process_details':'Assume method is gravimetry, not ICP-MS',                                                                                                            'measuring_assumption':True}},                                                                                                                                                                                                                   
            'other_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_opc-cmc_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                         'process_details':'Assume method is optical particle counting, not ICP-MS',                                                                                             'measuring_assumption':True}},                          
            'other_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                                           {                                                                                                                                                                                                                                               'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                                                               'sampling_type':'continuous injection'}},
            'other_light-scat_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'other_light_scattering_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'other_opc-cmc_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},                                                                                                                                                                                                                                                                                                                                                                                                                                                 
            'other_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'other_other_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify_not-automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},      
            'other_other,_please_specify_active':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other,_please_specify':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other,_please_specify_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'other_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'other_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_teom_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'other_teom-fdms_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'other_uv-fl_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence'}},
            'palas_model_fidas_200_beta_ray_attenuation_automatic':                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not beta-attenuation as stated but uses optical particle counting as reported instrument uses.',                                   'measuring_assumption':True}},
            'palas_model_fidas_200_light_scattering':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},                                                                                             
            'palas_model_fidas_200_light_scattering_automatic':                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not light scattering photometry as stated but uses optical particle counting as reported instrument uses.',                        'measuring_assumption':True}},
            'palas_model_fidas_200_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}}, 
            'palas_model_fidas_200_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}}, 
            'palas_model_fidas_200_optical_particle_counter_+_conversion_to_mass_concentration':                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}}, 
            'palas_model_fidas_200_other_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},                                                                                                                                                          
            'palas_model_fidas_200_teom_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'palas_model_fidas_200_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM as stated but uses optical particle counting as reported instrument uses.',                                               'measuring_assumption':True}},
            'palas_model_fidas_200_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200',                                                                                                                                                                      'process_details':'Assume method is not TEOM-FDMS as stated but uses optical particle counting as reported instrument uses.',                                          'measuring_assumption':True}},
            'palas_model_fidas_200e_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E'}}, 
            'palas_model_fidas_200e_optical_particle_counter_+_conversion_to_mass_concentration':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200E'}}, 
            'partisol_gravimetric_analysis_active':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000',  'process_details':'Make assumption sampling instrument is Thermo Partisol 2000.',                     'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'partisol_gravimetric_analysis_gravimetric_analysis_active':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000',  'process_details':'Make assumption sampling instrument is Thermo Partisol 2000.',                     'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'philips_mp_101m_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Environnement SA MP101M'}},
            'pm2.5_beta_ray_attenuation_automatic':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}}, 
            'pns_3d_15_gravimetric_analysis_gravimetric_analysis_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Atmoservice PNS3D15'},                                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry'}},
            'rancon_adam_c_beta_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                                           'measuring_assumption':True}},
            'rancon_adam_c_beta_ray_attenuation_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Cannot find instrument Rancon Adam C, therefore assume uses Beta-Attenuation as stated.',                                                           'measuring_assumption':True}},
            'r&p_partisol_2000_gravimetric_analysis_gravimetric_analysis_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2000_gravimetric_analysis_gravimetry_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_active':                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_gravimetric_analysis_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'r&p_partisol_2025_gravimetric_analysis_not-automatic':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry'}},
            'sentinel_gravimetric_analysis_gravimetric_analysis_active':                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Tecora Sentinel',   'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sentinel_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'TCR Tecora Sentinel',   'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_gravimetric_analysis_gravimetric_analysis_active':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_lvs3_gravimetric_analysis_gravimetric_analysis_active':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                                                                                                                                          'measurement':{'methodology':'gravimetry'}},
            'sequential_sampler_seq47_50_gravimetric_analysis_gravimetric_analysis_active':                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry'}},
            'sm200_(adam)_beta_ray_attenuation_active':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm200_(adam)_beta_ray_attenuation_not-automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm200_(adam)_gravimetric_analysis_beta_ray_attenuation_active':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},      
            'sm_200_beta_ray_attenuation_active':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_(air_dust_automatic_monitor)_gravimetric_analysis_beta_ray_attenuation_active':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},                             
            'sm_200_(air_dust_automatic_monitor)_beta_ray_attenuation_active':                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'sm_200_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'OPSIS SM200'}},
            'spm-613d_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Kimoto SPM-613D'}},
            'staplex_beta_automatic':                                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                        'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is gravimetry, not beta-attenuation, as used with primary sampling instrument',                                                       'measuring_assumption':True}},
            'staplex_beta_ray_attenuation_automatic':                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                        'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is gravimetry, not beta-attenuation, as used with primary sampling instrument',                                                       'measuring_assumption':True}},
            'staplex_gravimetric_analysis_gravimetric_analysis_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                        'process_details':'There are many Staplex samplers that exist, assume this is low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_mvs6_gravimetric_analysis_not-automatic':                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel MVS6'},                                                                                                                                                                                                                                           'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_seq47_50_gravimetric_analysis_active':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'sven_leckel_seq47_50_gravimetric_analysis_gravimetric_analysis_active':                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry'}},
            'swam5adual_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam5adual_beta_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam5adual_beta_ray_attenuation_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam5adual_caps_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a',                                                                                                                                                                          'process_details':'Assume method is not CAPS as stated but uses beta-attenuation as reported instrument uses.',                                                        'measuring_assumption':True}},                                                                                                                                                              
            'swam_5a_dual_channel_monitor_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam_5a_dual_channel_monitor_beta_ray_attenuation':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'swam5asingle_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'FAI Swam 5a'}},
            'tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'tcr_tecora,_model_sentinel_pm_gravimetric_analysis_gravimetric_analysis_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry'}},
            'tcr_tecora,_sentinel_pm_gravimetric_analysis_gravimetric_analysis_active':                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel'},                                                                                                                                                                                                                                   'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_active':                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_not-automatic':                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_gravimetric_analysis_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'tecora_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'TCR Tecora Sentinel',    'process_details':'Make assumption sampling instrument is TCR Tecora Sentinel.',                      'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry'}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_beta_ray_attenuation_active':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_beta_ray_attenuation_automatic':                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but beta-attenuation.',                                                  'measuring_assumption':True}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_optical_particle_counter_+_conversion_to_mass_concentration_active':                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'teledyne_api_t100_uv_fluorescent_so2_analyser_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter',                                                                                                                                                                                                                                                                          'process_details':'Assume method is not ultraviolet fluorescence as reported instrument uses, but optical particle counting.',                                         'measuring_assumption':True}},
            'teom1400_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},                                                                                                                                                                              
            'teom1400_beta_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},    
            'teom1400a_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_beta_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400a_beta_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom_1400a_beta_ray_attenuation_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400a_beta_ray_attenuation_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400a_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},   
            'teom1400a_gravimetric_analysis_teom-fdms_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400a_other,_please_specify_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom1400ab_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}}, 
            'teom1400ab_beta_ray_attenuation_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400ab-fdms_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400ab-fdms_beta_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'process_details':'Assume method is not beta-attenuation as stated but uses TEOM-FDMS as reported instrument uses.',                                                   'measuring_assumption':True}},
            'teom1400ab-fdms_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'process_details':'Assume method is not beta-attenuation as stated but uses TEOM-FDMS as reported instrument uses.',                                                   'measuring_assumption':True}},
            'teom1400ab-fdms_teom_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_teom-fdms_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_teom-fdms_automatic':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_tapered_element_oscillating_microbalance_(teom)_with_filter_automatic':           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400ab-fdms_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'process_details':'Assume method is not optical particle counting as stated but uses TEOM-FDMS as reported instrument uses.',                                          'measuring_assumption':True}},
            'teom1400ab-fdms_other,_please_specify_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400ab-fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400ab-fdms_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400ab_nephelometry_+_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation - nephelometry as stated but uses TEOM as reported instrument uses.',                                         'measuring_assumption':True}},
            'teom1400ab_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                                                                                                                                        'process_details':'Assume method is not optical particle counting as stated but uses TEOM as reported instrument uses.',                                               'measuring_assumption':True}},
            'teom1400ab_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},     
            'teom_1400ab_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}}, 
            'teom1400ab_teom_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab'}},                                                                                                    
            'teom1400ab_teom-fdms_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                                                               
            'teom_1400ab_with_fdms_module_(8500)_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_opc-cmc_automatic':                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_other_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_other,_please_specify_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_active':                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_automatic':                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400ab_with_fdms_module_(8500)_teom_automatic':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400ab_with_fdms_module_(8500)_teom-fdms_automatic':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400ab',                                                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400a_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom1400a_nephelometry_+_beta_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation - nephelometry as stated but uses TEOM as reported instrument uses.',                                         'measuring_assumption':True}},
            'teom1400a_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                                                                                                                                         'process_details':'Assume method is not optical particle counting as stated but uses TEOM as reported instrument uses.',                                               'measuring_assumption':True}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)':                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_active':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom1400a_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400a_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400a_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400a_teom_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom_1400a_teom-fdms_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                             
            'teom1400a_teom_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a'}},
            'teom1400a_teom-fdms_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400a',                                                 'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom1400_beta_ray_attenuation_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1400fdms-8500b-cb_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400fdms-8500b-cb_beta_ray_attenuation_automatic':                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'process_details':'Assume method is not beta-attenuation as stated but uses TEOM-FDMS as reported instrument uses.',                                                   'measuring_assumption':True}},      
            'teom1400fdms-8500b-cb_unknown_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},  
            'teom1400fdms-8500b-cb_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'process_details':'Assume method is not optical particle counting as stated but uses TEOM-FDMS as reported instrument uses.',                                          'measuring_assumption':True}},
            'teom1400fdms-8500b-cb_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400fdms-8500b-cb_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400fdms-8500b-cb_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1400fdms-8500b-cb_teom-fdms_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'teom_1400_fdms_+_8500_b_or_cb_dryer_other_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_fdms_+_8500_b_or_cb_dryer_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_fdms_+_8500_b_or_cb_dryer_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_fdms_+_8500_b_or_cb_dryer_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom_1400_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_light_scattering_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not light scattering photometry as stated but uses TEOM as reported instrument uses.',                                             'measuring_assumption':True}},
            'teom_1400_opc-cmc_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not optical particle counting as stated but uses TEOM as reported instrument uses.',                                               'measuring_assumption':True}},
            'teom_1400_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not optical particle counting as stated but uses TEOM as reported instrument uses.',                                               'measuring_assumption':True}},
            'teom1400_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                                                                                                                                          'process_details':'Assume method is not optical particle counting as stated but uses TEOM as reported instrument uses.',                                               'measuring_assumption':True}},
            'teom_1400_other_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_other,_please_specify_automatic':                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)_active':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom_1400_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_teom_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
            'teom1400_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1405df-fdms_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_gravimetric_analysis_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},     
            'teom1405df-fdms_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},          
            'teom1405df-fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_teom_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_teom-fdms_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405df-fdms_unknown_automatic':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},        
            'teom_1405df_with_fdms_dichotomous_monitor_for_pm2.5_&_pm10_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':{                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405-DF'}},
            'teom1405d_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405D'}},
            'teom1405d_teom_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405D'}},
            'teom1405f-fdms_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},       
            'teom1405f-fdms_beta_ray_attenuation_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F',                                                                                                                                                                         'process_details':'Assume method is not beta-attenuation as stated but uses TEOM-FDMS as reported instrument uses.',                                                   'measuring_assumption':True}},                                                                        
            'teom1405f-fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},
            'teom1405f-fdms_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},
            'teom1405f-fdms_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},       
            'teom1405f-fdms_teom_automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},
            'teom1405f-fdms_teom-fdms_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                             
            'teom_1405f_with_fdms_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                                                           
            'teom_1405f_with_fdms_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                                                           
            'teom_1405f_with_fdms_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',   'instrument':'Thermo 1405F'}},                                                           
            'teom1405_beta_ray_attenuation_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                                                                                                                                          'process_details':'Assume method is not beta-attenuation as stated but uses TEOM as reported instrument uses.',                                                        'measuring_assumption':True}},
            'teom1405_nephelometry_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                                                                                                                                          'process_details':'Assume method is not nephelometry as stated but uses TEOM as reported instrument uses.',                                                            'measuring_assumption':True}},                                                                                        
            'teom1405_nephelometry_+_beta_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                                                                                                                                          'process_details':'Assume method is not beta-attenuation - nephelometry as stated but uses TEOM as reported instrument uses.',                                         'measuring_assumption':True}},                                                                                        
            'teom_1405_tapered_element_oscillating_microbalance_(teom)_active':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},
            'teom1405_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},
            'teom1405_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'teom1405_teom_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405'}},
            'teom1405_teom-fdms_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1405',                                                  'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},                                                                           
            'teom_automatic':                                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'teom-fdms_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'thermo42i_automatic':                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo42i_chemi_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo42i_chemiluminescence_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo5014i_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo5014i_beta_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo5014i_beta_ray_attenuation_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo5014i_gravimetry_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},              
            'thermo_5014i_beta_continuous_ambient_particulate_monitor_beta_ray_attenuation_active':                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo_5014i_beta_continuous_ambient_particulate_monitor_beta_ray_attenuation_automatic':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},
            'thermo5014i_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume method is not DOAS as stated but uses beta-attenuation as reported instrument uses',                                                         'measuring_assumption':True}},            
            'thermo5014i_gravimetric_analysis_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i'}},         
            'thermo5014i_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume method is not TEOM as stated but uses beta-attenuation as reported instrument uses',                                                         'measuring_assumption':True}},
            'thermo5030sharp_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_beta_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_beta_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_gravimetric_analysis_nephelometry_beta_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_nephelometry_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_nephelometry_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_nephelometry_+_beta_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_other_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_scattering_of_laser_light_automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo5030sharp_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP',                                                                                                                                                                    'process_details':'Assume method is not TEOM as stated but uses beta-attenuation - nephelometry as reported instrument uses.',                                         'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_beta_automatic':                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},                                                                                             
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation':                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_active':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}}, 
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_beta_ray_attenuation_not-automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},      
            'thermo_andersen_esm_fh_62_i-r_gravimetric_analysis_other_automatic':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_gravimetric_analysis_beta_ray_attenuation_active':                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_nephelometry_active':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_nephelometry_beta_automatic':                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_nephelometry_+_beta_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_opc-cmc_automatic':                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_optical_particle_counter_+_conversion_to_mass_concentration_active':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume method is not optical particle counting as stated but uses beta-attenuation as reported instrument uses.',                                   'measuring_assumption':True}},
            'thermo_andersen_esm_fh_62_i-r_other_automatic':                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_esm_fh_62_i-r_other,_please_specify_automatic':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R'}},
            'thermo_andersen_series_fh_62_c14_continuous_pm10_ambient_particulate_monitor_beta_ray_attenuation_automatic':                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo FH62 C14'}},
            'thermo_esm_andersen,_fh_gravimetric_analysis_gravimetric_analysis_active':                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume instrument is Thermo Andersen FH62 I-R.',                                                                                                    'measuring_assumption':True}},
            'thermo_esm_andersen,_fh_gravimetric_analysis_active':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo Andersen FH62 I-R',                                                                                                                                                             'process_details':'Assume instrument is Thermo Andersen FH62 I-R.',                                                                                                    'measuring_assumption':True}},
            'thermoi_beta_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume instrument is Thermo 5014i.',                                                                                                                'measuring_assumption':True}},
            'thermoi_beta_ray_attenuation_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume instrument is Thermo 5014i.',                                                                                                                'measuring_assumption':True}},
            'thermo_i-series_beta_ray_attenuation_automatic':                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Thermo 5014i',                                                                                                                                                                         'process_details':'Assume instrument is Thermo 5014i.',                                                                                                                'measuring_assumption':True}},
            'thermo_model_42i_no_nox_analyser_chemiluminescence_automatic':                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                            'instrument':'Thermo 42i'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation_automatic':                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_beta_ray_attenuation_active':                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_gravimetric_analysis_beta_ray_attenuation_active':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_nephelometry_+_beta_active':                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'thermo_model_5030_sharp_particular_monitor_with_pm10_&_pm2.5_heads_nephelometry_+_beta_automatic':                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry',                                                              'instrument':'Thermo 5030 SHARP'}},
            'too_generic_beta_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'too_generic_beta_ray_attenuation_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'too_generic_beta_ray_attenuation':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'too_generic_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'too_generic_optical_particle_counter_+_conversion_to_mass_concentration':                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'too_generic_other_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'too_generic_other,_please_specify':                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction':                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)':                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_active':                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'too_generic_teom_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'too_generic_teom-fdms_automatic':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'umwelttechnik_mcz_gmbh_micro_pns_lvs_16_gravimetric_analysis_active':                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry'}},
            'umwelttechnik_mcz_gmbh_micro_pns_lvs_16_gravimetric_analysis_gravimetric_analysis_active':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'MCZ LVS16'},                                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry'}},
            'undefined_gravimetric_analysis_gravimetry_active':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'undetermined_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'undetermined_beta_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'undetermined_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'undetermined_opc-cmc_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}}, 
            'undetermined_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'undetermined_other_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'undetermined_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'undetermined_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'undetermined_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'undetermined_teom_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'undetermined_teom-fdms_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'undetermined_unknown_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unitec_beta_ray_attenuation_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume instrument is Unitec LSPM10.',                                                                                                               'measuring_assumption':True}},
            'unitec_lspm10_gravimetric_analysis_beta_ray_attenuation_active':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation as stated but uses Light Scattering Photometry as reported instrument uses.',                                 'measuring_assumption':True}},
            'unitec_lspm10_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry',                                                                  'instrument':'Unitec LSPM10',                                                                                                                                                                        'process_details':'Assume method is not beta-attenuation as stated but uses Light Scattering Photometry as reported instrument uses.',                                 'measuring_assumption':True}},
            'unknown_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_beta_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_active':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},
            'unknown_beta_ray_attenuation_not-automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation'}},                                                                               
            'unknown_differential_optical_absorption_spectroscopy_(doas)_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},    
            'unknown_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},            
            'unknown_dif._optical_absorption_spec._not-automatic':                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},   
            'unknown_gravimetric_analysis_active':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is gravimetry',                                                                                                                       'measuring_assumption':True}},
            'unknown_gravimetric_analysis_automatic':                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                        'process_details':'Assume method is gravimetry',                                                                                                                       'measuring_assumption':True}},
            'unknown_gravimetric_analysis_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation',                                                                                                                 'measuring_assumption':True}},                     
            'unknown_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation',                                                                                                                 'measuring_assumption':True}},
            'unknown_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                  'process_details':'Assume method is beta-attenuation',                                                                                                                 'measuring_assumption':True}},
            'unknown_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_active':                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},                                                                                                                                                                                                                                                                                      
            'unknown_gravimetric_analysis_differential_optical_absorption_spectroscopy_(doas)_automatic':                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy'}},                                                                                                                                                                                                                                                                                      
            'unknown_gravimetric_analysis_gravimetric_analysis_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_gravimetric_analysis_automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},                                                                                       
            'unknown_gravimetric_analysis_gravimetric_analysis_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},                                                                                                                                                                                                                                                                                                                                                                                                                                                               
            'unknown_gravimetric_analysis_gravimetry_active':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_gravimetry_automatic':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry'}},
            'unknown_gravimetric_analysis_light-scat_automatic':                                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_gravimetric_analysis_light_scattering_active':                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_gravimetric_analysis_nephelometry_beta_automatic':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_active':                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_gravimetric_analysis_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_gravimetric_analysis_teom_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_gravimetric_analysis_unknown_active':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry',                                                                                                                      'measuring_assumption':True}},
            'unknown_gravimetric_analysis_unknown_automatic':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                              'process_details':'Assume method is TEOM',                                                                                                                            'measuring_assumption':True}},
            'unknown_gravimetric_analysis_not-automatic':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry',                                                                                                                      'measuring_assumption':True}},
            'unknown_gravimetry_active':                                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry',                                                                                                                      'measuring_assumption':True}},
            'unknown_gravimetry_automatic':                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry',                                                                                                                      'measuring_assumption':True}},
            'unknown_gravimetry_not-automatic':                                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry',                                                                                                                      'measuring_assumption':True}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_ray_attenuation_active':                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                   'process_details':'Assume method is beta-attenuation, not ICP-MS',                                                                                                    'measuring_assumption':True}},                        
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_ray_attenuation_automatic':                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                   'process_details':'Assume method is beta-attenuation, not ICP-MS',                                                                                                    'measuring_assumption':True}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_beta_automatic':                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                   'process_details':'Assume method is beta-attenuation, not ICP-MS',                                                                                                    'measuring_assumption':True}},                                                             
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetry_automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                              'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                                                         'process_details':'Assume method is gravimetry, not ICP-MS',                                                                                                          'measuring_assumption':True}},      
            'unknown_light_absorption_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                                                                                                                                                                                                                                   'process_details':'Assume light absorption refers to beta-attenuation',                                                                                               'measuring_assumption':True}},
            'unknown_light-scat_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering_active':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_light_scattering_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'light scattering photometry'}},
            'unknown_nephelometry':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'nephelometry'}},
            'unknown_nephelometry_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'nephelometry'}},
            'unknown_nephelometry_beta_automatic':                                                                                                                      {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_nephelometry_+_beta_active':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_nephelometry_+_beta_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'unknown_opc-cmc_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},                                                                                                                                                                                                                                                                                                                                                                                                                                            
            'unknown_optical_particle_counter_+_conversion_to_mass_concentration_automatic':                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'optical particle counter'}},
            'unknown_other_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_automatic':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_active':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},   
            'unknown_other,_please_specify_remote':                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)':                                                                                                  {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_active':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_automatic':                                                                                        {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_not-automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                      'sampling_type':'injection'}},  
            'unknown_tapered_element_oscillating_microbalance_(teom)_remote':                                                                                           {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},       
            'unknown_tapered_element_oscillating_microbalance_(teom)_+_correction_with_estimated_volatile_fraction_automatic':                                          {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_tapered_element_oscillating_microbalance_(teom)_with_filter_dynamics_measurement_system_(fdms)_automatic':                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
            'unknown_teom_automatic':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'unknown_teom-fdms_automatic':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}}, 
            'unknown_unknown_active':                                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown_not-automatic':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown'}},
            'verewa_f_701_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_701_nephelometry_beta_automatic':                                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_701_nephelometry_+_beta_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F701'}},
            'verewa_f_703_beta_ray_attenuation_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'verewa_f_703_gravimetric_analysis_beta_ray_attenuation_active':                                                                                            {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'verewa_f_703_gravimetric_analysis_beta_ray_attenuation_automatic':                                                                                         {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Verewa F703'}},
            'wedding_gravimetric_analysis_beta_automatic':                                                                                                              {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}},  
            'wedding_beta_ray_attenuation_automatic':                                                                                                                   {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}},  
            'wedding_gravimetric_analysis_beta_ray_attenuation_active':                                                                                                 {                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Wedding and Associates Beta'}}}  
#------------------------------------------- 
        elif parameter_details['matrix'] == 'pm':   
            mapping = {
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_active':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                              'instrument':'GBC 932'}},
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                              'instrument':'GBC 932'}},
            'aa_gbc_932_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                              'instrument':'GBC 932'}},
            'active':                                                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},
            'ae22-pm2.5_light-abs_automatic':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE22'}},
            'ae22-pm2.5_light_absorption_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE22'}},
            'ae31_light-abs_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE31'}},
            'ae31_light_absorption_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE31'}},
            'ae33_aethalometer_light_absorption':                                                                                                                              {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE33'}},
            'ae33_light-abs_automatic':                                                                                                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE33'}}, 
            'ae33_light_absorption_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE33'}},
            'ae33-pm2.5_light_absorption_automatic':                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE33'}}, 
            'ae33_automatic':                                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer',                                                      'instrument':'Magee AE33'}}, 
            'atomic_absorption_spectrometry_(aas)':                                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'atomic_absorption_spectrometry_(aas)_active':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'automatic':                                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},        
            'chemiluminescence_automatic':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                                               'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter',  'measuring_assumption':True}},          
            'cold_vapour_atomic_fluorescence_spectrometry_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},                                                                   
            'danish_lmp_filterpack_element_collector_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':{'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack'},                                                                                                                  'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_active':                                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                                  {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'digitel_dha80_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                                                                                   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'digitel_dha80_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_not-automatic':                                                                            {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'digitel_dha80_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_active':               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'digitel_dha80_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_active':                                                                               {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'digitel_dha80_inductive_coupled_plasma_optical_emission_spectrometry_(icp-_not-automatic':                                                                        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80'},                                                                                                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                 {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_hvs_150mm_qurtz_filters_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active': {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'digitel_hvs_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'digitel_hvs_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                              {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Digitel DHA-80',          'process_details':'Make assumption sampling instrument is Digitel DHA-80.', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement_automatic':                          {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                       'instrument':'Sunset OCEC Model 5', 'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'fh62i-n_atomic_absorption_spectrometry_(aas)_active':                                                                                                             {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'Andersen FH621-N'},                                                                                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'fh62i-n_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                      {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'Andersen FH621-N'},                                                                                                                                                                     'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},      
            'filter_sampling_atomic_absorption_spectrometry_(aas)_active':                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                       'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'filter_sampling_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                       'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_active':                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},                                       
            'leckel_-_lvs_3_gravimetric_analysis_active':                                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                              'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'leckel_-_lvs_3_gravimetric_analysis_not-automatic':                                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                                                                                                              'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'leckel_-_lvs_3_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_gravimetric_analysis_active':                                                             {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Leckel LVS3'},                                                                                                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                            'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'lvs_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'lvs_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_passive':                                                                                         {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'sorbent trapping',         'specific_technique':'Diffusive Sampler',                            'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'ion_chromatography':                                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_passive':                                                                                                                                      {'primary_sampling':{'sampling_type':'passive'},                                                                                                                                                                        'preparation_1':{'sample_preparation_type':'sorbent trapping',         'specific_technique':'Diffusive Sampler',                            'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                     {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                              {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'opsis_sm200_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                            {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'beta-attenuation', 'instrument':'OPSIS SM200'},                                                                                                                                                                          'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'other_light-abs_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer'}},
            'other_light_absorption_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer'}},
            'other_light_scattering_automatic':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer'}},
            'other_other_automatic':                                                                                                                                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},
            'other_other,_please_specify':                                                                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},
            'other,_please_specify_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},   
            'other_thermo-optical_ec_oc_measurement_automatic':                                                                                                                {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                                                           'sampling_type':'continuous injection'}},
            'sequential_sampler_gravimetry_active':                                                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                          'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                       'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                'measurement':{'methodology':'ion chromatography',                                                                                                                              'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'sequential_sampler_gravimetry_automatic':                                                                                                                         {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                                                                                                                                                          'measurement':{'methodology':'ion chromatography',                                                                                    'sampling_type':'continuous injection',   'process_details':'Assume method is ion chromatography',                                                                          'measuring_assumption':True}},
            'sunsettoca_thermo-optical_ec_oc_measurement_automatic':                                                                                                           {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',  'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sunsettoca_to-ecoc_automatic':                                                                                                                                    {                                                                                                                                                                                                                       'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',  'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'sven_leckel,seq_47_50_atomic_absorption_spectrometry_(aas)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                  {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':           {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel,seq_47_50_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'sven_leckel_seq47_50_continous_flow_analysis_(cfa)_continous_flow_analysis_(cfa)_active':                                                                         {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'continuous flow analysis'}},
            'sven_leckel_seq47_50_ion_chromatography_ion_chromatography_active':                                                                                               {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Leckel SEQ 47-50'},                                                                                                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'ion chromatography'}},
            'tekran1130_cold_vapour_atomic_fluorescence_spectrometry_automatic':                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                     'instrument':'Tekran 2537X-1130-1135'}},   
            'termo_esm_andersen,_fh_95_kf_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Andersen FH95-KF'},                                                                                                                  'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'termo_esm_andersen,_fh_95_kf_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                             {'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Andersen FH95-KF'},                                                                                                                  'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                          'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}, 
            'thermal-optical_carbon_analysis_active':                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},        
            'thermo_atomic_absorption_spectrometry_(aas)_active':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'thermo_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'thermo5012maap_light_absorption_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'multi angle absorption photometer',                                 'instrument':'Thermo 5012'}},
            'undetermined_light_scattering_automatic':                                                                                                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'aethalometer'}},
            'unknown_atomic_absorption_spectrometry_(aas)_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_automatic':                                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_not-automatic':                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown',                                                                                                                                         'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':                         {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_ion_chromatography_unknown_active':                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                      'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                        'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography'}},                         
            'unknown_other,_please_specify_remote':                                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'measurement':{'methodology':'unknown'}}}
#------------------------------------------- 
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
            mapping = {
            '':                                                                                                                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'active':                                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'atmoservice_lvs_3d_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_atomic_absorption_spectrometry_(aas)_active':                               {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Atmoservice PNS3D15'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'atomic_absorption_spectrometry_(aas)':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'atomic_absorption_spectrometry_(aas)_active':                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'cold_vapour_atomic_absorption_spectrometry_(cv-aas)_active':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic absorption spectroscopy'}},
            'cold_vapour_atomic_fluorescence_spectrometry':                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},
            'cold_vapour_atomic_fluorescence_spectrometry_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy'}},
            'continous_flow_analysis_(cfa)_active':                                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis'}},
            'coulometry_active':                                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'coulometry',                                                                                                    'sampling_type':'injection'}},
            'derenda_continous_flow_analysis_(cfa)_continous_flow_analysis_(cfa)_active':                                                                            {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'continuous flow analysis'}},
            'derenda_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'derenda_ion_chromatography_ion_chromatography_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'derenda_lvs_3.1_ion_chromatography_gravimetric_analysis_active':                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'ion chromatography'}},
            'derenda_lvs_3.1_thermal-optical_carbon_analysis_gravimetric_analysis_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Comde Derenda LVS 3.1'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}}, 
            'dionex_ion_chromatography_ion_chromatography_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'dionex_2010i_ion_chromatography_ion_chromatography_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                               'instrument':'Dionex 2010i'}},
            'dionex_3000_ion_chromatography_ion_chromatography_active':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                               'instrument':'Thermo ICS-3000'}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement':                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',                                                    'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'ec_oc_measurements_sunset_laboratory_inc_thermal___optical_carbon_analyser_(near-real_time)_thermo-optical_ec_oc_measurement_automatic':                {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',          'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'filter_sampling_ion_chromatography_ion_chromatography_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'filter_3pack_ion_chromatography_ion_chromatography_active':                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',              'specific_technique':'3-Stage Filter Pack'},                                                            'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'flame_atomic_absorption_spectrometry_(f-aas)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy'}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)':                                                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},           
            'gc-ms_type_active':                                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'graphite_furnace_atomic_absorption_spectrometry_(gf-aas)':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},  
            'gravimetric_analysis_active':                                                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                               'measurement':{'methodology':'ion chromatography',                                                                                                                                      'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            'high_performance_liquid_chromatography_hplc_with_uv_detection_(hplc-uv)_active':                                                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - ultraviolet detection'}}, 
            'high_performance_liquid_chromatography_-_mass_spectrometry_(hplc-ms)_active':                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - mass spectrometry'}},
            'high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_active':                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},
            'hydride_generation_electrothermal_-_atomic_absorption_spectrometry_(hg-et-aas)_active':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'hydride generation atomic absorption spectroscopy'}}, 
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)':                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_automatic':                                                                                    {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                                  'sampling_type':'continuous injection'}}, 
            'inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_other_automatic':                                                                              {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                                  'sampling_type':'continuous injection'}},
            'ion_chromatography':                                                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'ion_chromatography_active':                                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'leckel_seq47_50_with_cooler_gravimetric_analysis_thermal-optical_carbon_analysis_active':                                                               {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                                                                                                                                              'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'leckel_seq47_50_with_cooler_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'low_volume_sampler_leckel_ion_chromatography_ion_chromatography_active':                                                                                {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'lvs_-_automatic_filter_change_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':{'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'marga_ic_automatic':                                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                               'instrument':'Metrohm MARGA'}},
            'marga_ion_chromatography_automatic':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'ion chromatography',                                               'instrument':'Metrohm MARGA'}},
            'micropns_lvs1_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'MCZ LVS1'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}}, 
            'other_automatic':                                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other_other_automatic':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},    
            'other,_please_specify':                                                                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other,_please_specify_active':                                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other,_please_specify_automatic':                                                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'other_thermo-optical_ec_oc_measurement_automatic':                                                                                                      {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                                                                   'sampling_type':'continuous injection'}},
            'partisol_2300_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_2300_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_2300_other,_please_specify_unknown_active':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                                        'measuring_assumption':True}},
            'partisol_2300_other,_please_specify_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Thermo Partisol 2300'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_active':                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_hydride_generation_electrothermal_-_atomic_absorption_spectr_active':                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'partisol_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_+_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_+_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                   {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_+_other,_please_specify_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                                                      {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'partisol_+_other,_please_specify_unknown_active':                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                                                                                                                                                        'measuring_assumption':True}},
            'perkin-elmer_elan_6000_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                     'instrument':'Perkin Elmer ELAN 6000'}},
            'perkin-elmer_elan_6000_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_not-automatic':{'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                     'instrument':'Perkin Elmer ELAN 6000'}},
            'spectrometry':                                                                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                        'measuring_assumption':True}},
            'spectrometry_active':                                                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.',                        'measuring_assumption':True}},
            'spectrophotometry_active':                                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'sunset_laboratory_inc_thermal___optical_carbon_analyser_thermal-optical_carbon_analysis_eusaar_2_method_active':                                        {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - EUSAAR2',                               'instrument':'Sunset OCEC Model 5',                                                    'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sunset_laboratory_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',                                                    'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sunsettoca_to-ecoc_automatic':                                                                                                                          {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',          'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sunsettoca_thermo-optical_ec_oc_measurement_automatic':                                                                                                 {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol',                      'instrument':'Sunset OCEC Model 5',          'sampling_type':'continuous injection',   'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.',                                                 'measuring_assumption':True}},
            'sven_leckel_seq47_50_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                 {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'sven_leckel_seq47_50_ion_chromatography_ion_chromatography_active':                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'ion chromatography'}},
            'sven_leckel_seq47_50_thermal-optical_carbon_analysis_thermal-optical_carbon_analysis_active':                                                           {'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Leckel SEQ 47-50'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',                                             'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'tekran2537x_cold_vapour_atomic_fluorescence_spectrometry_automatic':                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy',                     'instrument':'Tekran 2537X-1130-1135'}}, 
            'teom_accu_sampler_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'teom_accu_sampler_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':            {                                                                                                                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'tapered element oscillating microbalance - gravimetry'},                                                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'thermal-optical_carbon_analysis':                                                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'thermal-optical_carbon_analysis_active':                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'thermo_icp_iris_intrepid_ii_other_automatic':                                                                                                           {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',          'instrument':'Thermo IRIS Intrepid II',      'sampling_type':'continuous injection'}}, 
            'thermo_icp_iris_intrepid_ii_other,_please_specify_automatic':                                                                                           {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',          'instrument':'Thermo IRIS Intrepid II',      'sampling_type':'continuous injection'}},
            'unknown_atomic_absorption_spectrometry_(aas)_active':                                                                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_mass_espectrometry_(icp-ms)_active':                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_active':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_atomic_absorption_spectrometry_(aas)_active':                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_ion_chromatography_active':                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_not-automatic':                                                                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_mass_espectrometry_(icp-ms)_not-automatic':                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_spectrophotometry_active':                                                                                 {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_atomic_absorption_spectrometry_(aas)_unknown_active':                                                                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                                                                    'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},
            'unknown_atomic_absorption_spectrometry_(aas)_gas_chromatography_-_mass_spectrometry_active':                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},     
            'unknown_cold_vapour_atomic_absorption_spectrometry_(cv-aas)_atomic_absorption_spectrometry_(aas)_active':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'cold vapour atomic absorption spectroscopy'}},
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_mass_espectrometry_(icp-ms)_active':                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_cold_vapour_atomic_fluorescence_spectrometry_mass_espectrometry_(icp-ms)_not-automatic':                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_continous_flow_analysis_(cfa)_ion_chromatography_active':                                                                                       {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_mass_espectrometry_(icp-ms)_not-automatic':                                                      {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_mass_espectrometry_(icp-ms)_active':                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_+_mass_spectrometry_(gc-ms)_active':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},       
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromatography_active':                                                                      {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},      
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_active':                                                                 {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_gas_chromotography_(gc)_not-automatic':                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},     
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_high_performance_liquid_chromatography_-_fluorescence_detect_active':                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_gas_chromatography_+_mass_spectrometry_(gc-ms)_not-automatic':                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - mass spectrometry'}},
            'unknown_gas_chromotography_(gc)_active':                                                                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - unknown detection'}},          
            'unknown_gas_chromotography_(gc)_not-automatic':                                                                                                         {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'gas chromatography - unknown detection'}},  
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detect_not-automatic':                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},    
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_conductimetry_active':                                               {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},           
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_gas_chromotography_(gc)_active':                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - fluorescence detection'}},    
            'unknown_high_performance_liquid_chromatography_-_mass_spectrometry_(hplc-ms)_gas_chromotography_(gc)_active':                                           {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - mass spectrometry'}},       
            'unknown_high_performance_liquid_chromatography_-_mass_spectrometry_(hplc-ms)_high_performance_liquid_chromatography_-_fluorescence_detect_active':      {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - mass spectrometry - fluorescence detection'}},                  
            'unknown_high_performance_liquid_chromatography_-_mass_spectrometry_(hplc-ms)_high_performance_liquid_chromatography_active':                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'high performance liquid chromatography - mass spectrometry'}},
            'unknown_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_atomic_absorption_spectrometry_(aas)_active':                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'unknown_graphite_furnace_atomic_absorption_spectrometry_(gf-aas)_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy'}},
            'unknown_high_performance_liquid_chromatography_-_fluorescence_detection_(hplc-fld)_atomic_absorption_spectrometry_(aas)_active':                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_atomic_absorption_spectrometry_active':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_atomic_absorption_spectrometry_(aas)_active':                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_differential_optical_absorption_spectrometry_active':                           {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_ion_chromatography_active':                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_lvs_-_automatic_filter_change_2,3_m3_h_active':                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}}, 
            'unknown_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-aes)_unknown_active':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_chromatography_active':                                                                {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_high_vol_sampler_active':                                                              {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_hvs_-_automatic_filter_change_30_m3.h-1_active':                                       {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                             'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_atomic_emission_spectrometry_(icp-a_active':                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_active':                      {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_lvs_-_automatic_filter_change_2,3_m3_h_active':                                        {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_mass_espectrometry_(icp-ms)_active':                                                   {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_mass_espectrometry_(icp-ms)_not-automatic':                                            {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_unknown_active':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'inductively coupled plasma mass spectrometry'}},
            'unknown_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_x-ray_fluorescence_spectroscopy_(xrf)_active':                                         {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_ion_chromatography_active':                                                                                                                     {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_unknown_active':                                                                                                             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_ion_chromotagraphy_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_ion_chromatography_active':                                                                                                  {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_lvs_-_automatic_filter_change_2,3_m3_h_active':                                                                              {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_lvs_-_automatic_filter_change_2,3_m3_h_not-automatic':                                                                       {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_not-automatic':                                                                                                              {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography'}},
            'unknown_ion_chromatography_high_performance_liquid_chromatography_-_fluorescence_detect_active':                                                        {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},      
            'unknown_lvs_-_automatic_filter_change_2,3_m3_h_active':                                                                                                 {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_lvs_-_automatic_filter_change_2,3_m3_h_not-automatic':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                              'preparation_1':{'sample_preparation_type':'filter'},                                                                                                                          'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                                                                      'process_details':'Assume method is ion chromatography after sampling.',                                                          'measuring_assumption':True}},
            'unknown_other_automatic':                                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_automatic':                                                                                                               {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_remote':                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_niosh_active':                                                                                                            {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_other,_please_specify_unknown_active':                                                                                                          {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_other_remote':                                                                                                                                  {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_reflectometry_niosh_active':                                                                                                                    {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},                                                                                      
            'unknown_spectrophotometry_ion_chromatography_active':                                                                                                   {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_spectrophotometry_spectrophotometry_active':                                                                                                    {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry'}},
            'unknown_thermal-optical_carbon_analysis_ion_chromatography_active':                                                                                     {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown',                                                                                                                                                 'process_details':'Two different measurement methodologies stated, setting as Unknown.'}},
            'unknown_thermal-optical_carbon_analysis_optical_termical_active':                                                                                       {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'thermal-optical analysis - unknown protocol'}},
            'unknown_unknown_active':                                                                                                                                {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_unknown_automatic':                                                                                                                             {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'unknown'}},
            'unknown_x-ray_fluorescence_spectroscopy_(xrf)_unknown_active':                                                                                          {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'X-ray fluorescence spectroscopy'}},
            'varian_spectre_aa_400_graphit_gta-96_inductive_coupled_plasma_mass_spectrometryscopy_(icp-ms)_atomic_absorption_spectrometry_(aas)_active':             {'primary_sampling':{'sampling_type':'low volume continuous',                                          'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',                  'instrument':'Varian SpectrAA 400 GTA-96'}}, 
            'xrf_mvs_other_automatic':                                                                                                                               {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                               'sampling_type':'continuous injection'}}, 
            'xrf_mvs_other,_please_specify_automatic':                                                                                                               {                                                                                                                                           'preparation_1':{'sample_preparation_type':'filter',                                                                                           'preparation_assumption':True},                                                                                                                                                                                                                                   'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                               'sampling_type':'continuous injection'}}}

#--------------------------------------------------------------
#JAPAN_NIES
#--------------------------------------------------------------
    elif run_details['network'] == 'JAPAN_NIES':
        if run_details['parameter'] == 'O3':
            mapping = {
            '':                 {                                                                                                                                                                                                           'measurement':{'methodology':'unknown'}},  
            'absorptiometry':   {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Potassium Iodide'}, 'measurement':{'methodology':'spectrophotometry'}}, 
            'coulometry':       {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True},                                                                                                          'measurement':{'methodology':'coulometry',                   'sampling_type':'injection'}},
            'uv_absorption':    {                                                                                                                                                                                                           'measurement':{'methodology':'ultraviolet photometry'}},
            'chemiluminescence':{                                                                                                                                                                                                           'measurement':{'methodology':'ethylene chemiluminescence'}}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            mapping = {
            '':                 {                                                                                                                                                                                                          'measurement':{'methodology':'unknown'}},
            'absorptiometry':   {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'}, 'measurement':{'methodology':'spectrophotometry'}}, 
            'chemiluminescence':{                                                                                                                                                                                                          'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'process_details':'Only Chemiluminescence stated as methodology, make assumption the method uses internal molybdenum converter', 'measuring_assumption':True}}}
        elif run_details['parameter'] == 'CO':
            mapping = {
            '':                                  {'measurement':{'methodology':'unknown'}},
            'non-dispersive_infrared_absorption':{'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation', 'measuring_assumption':True}}}        
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '':                                      {                                                                                                  'measurement':{'methodology':'unknown'}},
            'conventional_solution_conductivity':    {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'measurement':{'methodology':'conductimetry',              'sampling_type':'injection'}},
            'high_sensitivity_solution_conductivity':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'measurement':{'methodology':'conductimetry',              'sampling_type':'injection'}},
            'uv_fluorescence':                       {                                                                                                  'measurement':{'methodology':'ultraviolet fluorescence'}}}
        elif run_details['parameter'] in ['HCl','HF','H2S']:
            mapping = {
            '':{'measurement':{'methodology':'unknown'}}} 
        elif run_details['parameter'] in ['CH4','NMHC']:
            mapping = {
            '':                                                                      {'measurement':{'methodology':'unknown'}},
            'direct':                                                                {'measurement':{'methodology':'gas chromatography - flame ionisation detection', 'sampling_type':'continuous injection', 'process_details':'Assume continuous injection', 'measuring_assumption':True}},
            'difference_with_propane_conversion':                                    {'measurement':{'methodology':'gas chromatography - flame ionisation detection', 'sampling_type':'continuous injection', 'process_details':'Assume continuous injection', 'measuring_assumption':True}},
            'difference_with_no_propane_conversion':                                 {'measurement':{'methodology':'gas chromatography - flame ionisation detection', 'sampling_type':'continuous injection', 'process_details':'Assume continuous injection', 'measuring_assumption':True}},
            'total_hydrocarbon_automatic_measuring_machine_with_methane_conversion': {'measurement':{'methodology':'flame ionisation detection'}},
            'total_hydrocarbon_automatic_measuring_machine_with_propane_conversion': {'measurement':{'methodology':'flame ionisation detection'}},
            'total_hydrocarbon_automatic_measuring_machine_with_ethylene_conversion':{'measurement':{'methodology':'flame ionisation detection'}}}
        elif run_details['parameter'] == 'HC':
            mapping = {
            '':                                                                      {'measurement':{'methodology':'unknown'}},
            'direct':                                                                {'measurement':{'methodology':'gas chromatography - flame ionisation detection', 'sampling_type':'continuous injection', 'process_details':'Assume continuous injection', 'measuring_assumption':True}},
            'difference_with_propane_conversion':                                    {'measurement':{'methodology':'flame ionisation detection'}},
            'difference_with_no_propane_conversion':                                 {'measurement':{'methodology':'flame ionisation detection'}},
            'total_hydrocarbon_automatic_measuring_machine_with_methane_conversion': {'measurement':{'methodology':'flame ionisation detection'}},
            'total_hydrocarbon_automatic_measuring_machine_with_propane_conversion': {'measurement':{'methodology':'flame ionisation detection'}},
            'total_hydrocarbon_automatic_measuring_machine_with_ethylene_conversion':{'measurement':{'methodology':'flame ionisation detection'}}}
        elif run_details['parameter'] == 'PM10':
            mapping = {
            '':                            {'measurement':{'methodology':'unknown'}},
            'light_scattering':            {'measurement':{'methodology':'nephelometry'}},
            'piezoelectric_balance_method':{'measurement':{'methodology':'quartz crystal microbalance - gravimetry'}},
            'beta-ray_absorption':         {'measurement':{'methodology':'beta-attenuation'}}}
        elif run_details['parameter'] == 'PM2.5':
            mapping = {   
            '':                                                                 {'measurement':{'methodology':'unknown'}},
            'beta-ray_absorption':                                              {'measurement':{'methodology':'beta-attenuation'}},
            'beta-ray_absorption_other':                                        {'measurement':{'methodology':'beta-attenuation'}},
            'beta-ray_absorption_frm-compliant_sampler':                        {'measurement':{'methodology':'beta-attenuation'}},
            'beta-ray_absorption_kimoto_pm-712':                                {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Kimoto PM-712'}},
            'beta-ray_absorption_kimoto_pm-717':                                {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Kimoto PM-717'}},
            'beta-ray_absorption_toa_dkk_fpm-377':                              {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'DKK-TOA FPM-377'}},
            'beta-ray_absorption_horiba_apda-3750a':                            {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Horiba APDA3750A'}},
            'beta-ray_absorption_thermo_fh62c14':                               {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo FH62 C14'}},
            'beta-ray_absorption_environnement_sa_mp101m':                      {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Environnement SA MP101M'}}, 
            'beta-ray_absorption_thermo_5014i':                                 {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo 5014i'}},
            'beta-ray_absorption_thermo_sharp_5030':                            {'measurement':{'methodology':'beta-attenuation - nephelometry',                         'instrument':'Thermo 5030 SHARP'}},
            'beta-ray_absorption_light_scattering_hybrid':                      {'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'beta-ray_absorption_light_scattering_hybrid_frm-compliant_sampler':{'measurement':{'methodology':'beta-attenuation - nephelometry'}},
            'beta-ray_absorption_light_scattering_hybrid_thermo_sharp_5030':    {'measurement':{'methodology':'beta-attenuation - nephelometry',                         'instrument':'Thermo 5030 SHARP'}},
            'beta-ray_absorption_light_scattering_hybrid_kimoto_pm-712':        {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Kimoto PM-712'}},
            'beta-ray_absorption_light_scattering_hybrid_thermo_fh62c14':       {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo FH62 C14'}},
            'filter_vibration':                                                 {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'filter_vibration_other':                                           {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},            
            'filter_vibration_frm-compliant_sampler':                           {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry'}},
            'filter_vibration_kimoto_pm-712':                                   {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Kimoto PM-712',             'process_details':'Assume method is beta-attenuation as reported instrument uses and not TEOM as stated',             'measuring_assumption':True}},           
            'filter_vibration_thermo_fh62c14':                                  {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'Thermo FH62 C14',           'process_details':'Assume method is beta-attenuation as reported instrument uses and not TEOM as stated',             'measuring_assumption':True}},      
            'filter_vibration_toa_dkk_fpm-377':                                 {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'DKK-TOA FPM-377',           'process_details':'Assume method is beta-attenuation as reported instrument uses and not TEOM as stated',             'measuring_assumption':True}},                          
            'light_scattering':                                                 {'measurement':{'methodology':'nephelometry'}},
            'light_scattering_frm-compliant_sampler':                           {'measurement':{'methodology':'nephelometry'}},
            'light_scattering_toa_dkk_fpm-377':                                 {'measurement':{'methodology':'beta-attenuation',                                        'instrument':'DKK-TOA FPM-377',           'process_details':'Assume method is beta-attenuation as reported instrument uses and not light scattering as stated', 'measuring_assumption':True}}}                          
                        

#--------------------------------------------------------------
#MEXICO_CDMX
#--------------------------------------------------------------
    elif run_details['network'] == 'MEXICO_CDMX':
        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information  = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)'}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}
        elif run_details['parameter'] == 'SO2':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence'}}
        elif run_details['parameter'] in ['PM10','PM2.5']:
            standard_measurement_information = {'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry', 'process_details':'Two different methods stated as possible TEOM and beta-attenuation. Assume TEOM.', 'measuring_assumption':True}}

#--------------------------------------------------------------
#NOAA ISD
#--------------------------------------------------------------
    elif run_details['network'] == 'NOAA_ISD':
        standard_measurement_information = {'measurement':{'methodology':'unknown'}}

#--------------------------------------------------------------
#SEARCH
#--------------------------------------------------------------
    elif run_details['network'] == 'SEARCH':
        if run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information  = {'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}

#--------------------------------------------------------------
#UK AIR
#--------------------------------------------------------------
    elif run_details['network'] == 'UK_AIR':
        if raw_measurement_info != '':
            if run_details['parameter'] == 'NH3':
                mapping = {        
                'nh3active':       {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'},                                                     'measurement':{'methodology':'ion chromatography', 'process_details':'Assume method is ion chromatography after sampling.',                'measuring_assumption':True}},
                'nh3diffusiontube':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'spectrophotometry',  'process_details':'Assume measurement methodology is spectrophotometry after sampling', 'measuring_assumption':True}},
                'nh3passive':      {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'spectrophotometry',  'process_details':'Assume measurement methodology is spectrophotometry after sampling', 'measuring_assumption':True}}}
            elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
                mapping = {
                'bam':        {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                             'instrument':'Met One BAM1020'}},             
                'fidas':      {                                                                                                                                                        'measurement':{'methodology':'optical particle counter',                                                                     'instrument':'Palas Fidas 200'}},
                'grav_emfab': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'}, 'measurement':{'methodology':'gravimetry'}},
                'grav_eq':    {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                                                               'process_details':'Make assumption that the gravimetry equivalent methodology is beta-attenuation as TEOM is otherwise stated', 'measuring_assumption':True}},
                'grav_quartz':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'}, 'measurement':{'methodology':'gravimetry'}},
                'grav_teflon':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter'}, 'measurement':{'methodology':'gravimetry'}},
                'indic.grav': {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                                                               'process_details':'Make assumption that the gravimetry equivalent methodology is beta-attenuation as TEOM is otherwise stated', 'measuring_assumption':True}},        
                'ref.eq':     {                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                                                               'process_details':'Make assumption that the gravimetry equivalent methodology is beta-attenuation as TEOM is otherwise stated', 'measuring_assumption':True}},
                'teom':       {                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}},
                'teom_fdms':  {                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry'}},
                'vcm':        {                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                        'instrument':'Thermo 1400'}}}
            elif parameter_details['matrix'] == 'pm':
                mapping = {
                'improve-a': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}},
                'niosh-5040':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040'}}}
            elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']:
                mapping = {
                'improve-a': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A'}},
                'niosh-5040':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040'}}}
        elif run_details['parameter'] == 'O3':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2'):
            standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'process_details':'No instrument information other than method is chemiluminescence, assume chemiluminescence (internal molybdenum converter)', 'measuring_assumption':True}}
        elif run_details['parameter'] == 'CO':
            standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}
        elif run_details['parameter'] == 'SO2':
            standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence'}}
        elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
            standard_measurement_information = {'measurement':{'methodology':'unknown'}}
        else:
            standard_measurement_information = {'measurement':{'methodology':'unknown'}}

#--------------------------------------------------------------
#US EPA AirNow DOS
#--------------------------------------------------------------
    elif run_details['network'] == 'US_EPA_AirNow_DOS':
        standard_measurement_information = {'measurement':{'methodology':'unknown'}}

#--------------------------------------------------------------
#US EPA AQS
#--------------------------------------------------------------
    elif run_details['network'] == 'US_EPA_AQS':
        if run_details['parameter'] == 'O3':
            mapping = {
            '3':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Meloy 0A325-2R',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '4':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Meloy 0A350-2R',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '7':  {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Bendix 8002',                'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '11': {'measurement':{'methodology':'ethylene chemiluminescence',                                                                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}, 'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume ethylene chemiluminescence', 'measuring_assumption':True}},
            '14': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-1',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '15': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-2',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '16': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'McMillan 1100-3',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}}, 
            '17': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Teledyne ML8410E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '19': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1003-AH',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '20': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'Beckman 950A',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '23': {'measurement':{'methodology':'rhodamine B chemiluminescence',                     'instrument':'Philips PW9771',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '36': {'measurement':{'methodology':'ethylene chemiluminescence',                        'instrument':'CSI 2000',                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '47': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Thermo 49',                  'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '53': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne ML8810',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '55': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'PCI LC-12',                  'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '56': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Dasibi 1008-AH',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '78': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environics 300',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '87': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne 400',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '91': {'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne ML9810',            'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '103':{'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'OPSIS AR500',                'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '105':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O341M',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '112':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA360',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '134':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'DKK-TOA GUX-113E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '137':{'measurement':{'methodology':'differential optical absorption spectroscopy',      'instrument':'Environnement SA SANOA',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '148':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O342M',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '160':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Horiba APOA370',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '161':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Seres OZ 2000G',             'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '164':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'SIR S-5014',                 'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '165':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Tanabyte 722',               'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '169':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'DKK-TOA GUX-313E',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '187':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Ecotech Serinus10',          'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '190':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 202',                     'reported_lower_limit_of_detection':{'O3':'3.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '199':{'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'instrument':'Teledyne 265E',              'reported_lower_limit_of_detection':{'O3':'0.6'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '214':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Teledyne T204-O3',           'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '215':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 211',                     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '218':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 106-L',                   'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '222':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Sabio 6030',                 'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '225':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'Environnement SA O342e',     'reported_lower_limit_of_detection':{'O3':'5.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '227':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B POM',                     'reported_lower_limit_of_detection':{'O3':'4.0'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}},
            '901':{'measurement':{'methodology':'ultraviolet photometry',                            'instrument':'2B 202',                     'reported_lower_limit_of_detection':{'O3':'1.5'}, 'reported_upper_limit_of_detection':{'O3':'500.0'}}}}
#-------------------------------------------
        elif (run_details['parameter'] == 'NO') or (run_details['parameter'] == 'NO2') or (run_details['parameter'] == 'NOx') or (run_details['parameter'] == 'NOy'):
            mapping = {
            '10': {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'},                                                                                                                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                                              'reported_lower_limit_of_detection':{'NO2':'0.1'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '11': {                                                                                                  'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Lyshkow'},                                                                                                                                                                                                                                                           'measurement':{'methodology':'colorimetry',                                                                                                                     'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'10.0','NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'},                  'sampling_type':'continuous injection'}},
            '12': {                                                                                                  'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'colorimetry',                                                                                                                     'reported_lower_limit_of_detection':{'NO2':'10.0'},                                        'reported_upper_limit_of_detection':{'NO2':'400.0'},                                              'sampling_type':'continuous injection'}}, 
            '13': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'coulometry',                                                                                                                      'reported_lower_limit_of_detection':{'NO2':'10.0','NOx':'10.0'},                           'reported_upper_limit_of_detection':{'NO2':'400.0','NOx':'1200.0'}}},
            '14': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                               'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'},                                                            'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                      'measuring_assumption':True}},
            '16': {                                                                                                  'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman',    'process_details':'Preparation method is stated as Poland Method. Believe this is another name for the Griess-Saltzman method, see paper: https://www.sciencedirect.com/science/article/pii/S0269749197000389', 'preparation_assumption':True}, 'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'10.0'},                                        'reported_upper_limit_of_detection':{'NO2':'400.0'},                                              'sampling_type':'continuous injection'}},
            '21': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8440E',       'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '22': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Bendix 8101-C',          'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '25': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'CSI 1600',               'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '31': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Meloy NA530R',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '34': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Beckman 952A',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '35': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 14B/E',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '37': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 14D/E',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '38': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Bendix 8101-B',          'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '40': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Philips PW9762/02',      'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '42': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8840',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'400.0','NOx':'1200.0'}}},
            '51': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Thermo 17C',             'reported_lower_limit_of_detection':{'NOx':'1.0'},                                         'reported_upper_limit_of_detection':{'NOx':'1200.0'}}},
            '61': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0'},                                         'reported_upper_limit_of_detection':{'NO':'500.0'}}},
            '71': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0'},                                         'reported_upper_limit_of_detection':{'NO':'500.0'}}},
            '74': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42',              'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'1.0', 'NOx':'10.0', 'NOy':'0.5'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '75': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42S',             'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05', 'NOy':'0.05'}, 'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '82': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200',           'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0', 'NOy':'0.5'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '83': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML8841',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0'},               'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '89': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Dasibi 2108',            'reported_lower_limit_of_detection':{'NO':'2.7', 'NO2':'5.0', 'NOx':'2.7'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '90': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne ML9841',        'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'5.0', 'NOx':'10.0', 'NOy':'5.0'},  'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOy':'500.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Griess-Saltzman'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO':'10.0','NO2':'2.7', 'NOx':'2.7'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '94': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Sodium Arsenite'},                                                                                                                                                                                                                                                   'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '95': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TEA'},                                                                                                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '96': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TGS-ANSA'},                                                                                                                                                                                                                                                          'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '97': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TEA'},                                                                                                                                                                                                                                                               'measurement':{'methodology':'spectrophotometry',                                                                                                               'reported_lower_limit_of_detection':{'NO2':'2.7'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '99': {                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200E',          'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'2.7', 'NOx':'5.0', 'NOy':'0.5'},   'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0','NOx':'500.0'}}},
            '100':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                                                               'reported_lower_limit_of_detection':{'NO':'1000.0'},                                       'reported_upper_limit_of_detection':{'NO':'1200.0'},                                                                                        'process_details':'No instrument information other than method is chemiluminescence, and instrument sampling is continuous, assume chemiluminescence (internal molybdenum converter)',                                                                                                                                                                                                                      'measuring_assumption':True}},
            '102':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'differential optical absorption spectroscopy',                                             'instrument':'OPSIS AR500',            'reported_lower_limit_of_detection':{'NO2':'1.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '104':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Environnement SA AC31M', 'reported_lower_limit_of_detection':{'NO2':'5.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '111':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Horiba APNA360',         'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '121':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-114E',       'reported_lower_limit_of_detection':{'NO2':'5.0','NOx':'5.0'},                             'reported_upper_limit_of_detection':{'NO2':'400.0','NOx':'1200.0'}}},
            '139':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'differential optical absorption spectroscopy',                                             'instrument':'Environnement SA SANOA', 'reported_lower_limit_of_detection':{'NO2':'1.0'},                                         'reported_upper_limit_of_detection':{'NO2':'400.0'}}},
            '146':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Environnement SA AC32M', 'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'1.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '152':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'SIR S-5012',             'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'400.0','NO2':'400.0','NOx':'400.0'}}},
            '157':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Horiba APNA370',         'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '163':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Seres 2000G',            'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'400.0','NOx':'1200.0'}}},
            '171':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-314E',       'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'1200.0','NO2':'1200.0','NOx':'1200.0'}}},
            '186':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech Serinus40',      'reported_lower_limit_of_detection':{'NO':'5.0', 'NO2':'5.0', 'NOx':'5.0'},                'reported_upper_limit_of_detection':{'NO':'500.0','NO2':'500.0','NOx':'500.0'}}},
            '200':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                                        'instrument':'Teledyne T200UP',        'reported_lower_limit_of_detection':{'NO':'0.1', 'NO2':'0.1', 'NOx':'0.1'},                'reported_upper_limit_of_detection':{'NO':'2000.0','NO2':'2000.0','NOx':'2000.0'}}},
            '210':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                                               'instrument':'Environnement SA AS32M', 'reported_lower_limit_of_detection':{'NO2':'0.04'},                                        'reported_upper_limit_of_detection':{'NO2':'1000.0'}}},
            '212':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                                               'instrument':'Teledyne T500U',         'reported_lower_limit_of_detection':{'NO2':'0.04'},                                        'reported_upper_limit_of_detection':{'NO2':'1000.0'}}},
            '241':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal photolytic converter)',                                        'instrument':'Teledyne T200P',         'reported_lower_limit_of_detection':{'NO':'0.4', 'NO2':'0.4', 'NOx':'0.4'},                'reported_upper_limit_of_detection':{'NO':'4000.0','NO2':'4000.0','NOx':'4000.0'}}},
            '256':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'cavity attenuated phase shift spectroscopy',                                               'instrument':'Teledyne N500',          'reported_lower_limit_of_detection':{'NO':'0.1', 'NO2':'0.1', 'NOx':'0.1'},                'reported_upper_limit_of_detection':{'NO':'1000.0','NO2':'500.0','NOx':'1000.0'}}},
            '574':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Thermo 42C-TL',          'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'1000.0','NO2':'1000.0','NOx':'1000.0'}}},
            '590':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech EC9841T',        'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'2000.0','NO2':'2000.0','NOx':'2000.0'}}},
            '591':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Ecotech EC9843',         'reported_lower_limit_of_detection':{'NO':'0.05','NOx':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOx':'2000.0'}}},
            '599':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Teledyne 200EU',         'reported_lower_limit_of_detection':{'NO':'0.05','NO2':'0.05','NOx':'0.05'},               'reported_upper_limit_of_detection':{'NO':'1000.0','NO2':'200.0','NOx':'2000.0'},                                                           'process_details':'In methods_all.csv the instrument is stated as Teledyne 200EU-NOy, but this only measures NO/NOy. It gives an equivalent method of Teledyne 200A/Teledyne 200E, however these have detection limits of 0.4ppbv. In both methods_all.csv and seen in the datafiles, the applicable detection limit is many times 0.05ppbv, this suggests the instrument in fact used is Teledyne 200EU.', 'measuring_assumption':True}},
            '674':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Thermo 42C-NOy',         'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'1000.0','NOy':'1000.0'}}},
            '690':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'Ecotech EC9841T',        'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOy':'500.0'}}},
            '691':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Ecotech EC9843',         'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'2000.0','NOy':'500.0'}}},
            '699':{                                                                                                                                                                                                                                                                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (external molybdenum converter)',                                        'instrument':'Teledyne 200EU-NOy',     'reported_lower_limit_of_detection':{'NO':'0.05','NOy':'0.05'},                            'reported_upper_limit_of_detection':{'NO':'1000.0','NOy':'520.0'}}}}
#------------------------------------------- 
        elif run_details['parameter'] == 'CO':
            mapping = {
            '8':  {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Bendix 8501-5CA',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '11': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                                                          'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'},                                             'process_details':'Only NDIR given as method. As it is an old AQS code and other methods around it use Non-Dispersive Infrared Absorption (Luft) method, assume this.', 'measuring_assumption':True}},
            '12': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Beckman 866',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '13': {'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Detection Tube'}, 'measurement':{'methodology':'colorimetry',                                                                                        'reported_lower_limit_of_detection':{'CO':'5000.0'}, 'reported_upper_limit_of_detection':{'CO':'50000.0'},   'sampling_type':'continuous injection'}},
            '14': {                                                                                                       'measurement':{'methodology':'dual isotope fluorescence',                                                                          'reported_lower_limit_of_detection':{'CO':'400.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '18': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'MSA 202S',               'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '21': {                                                                                                       'measurement':{'methodology':'gas chromatography - unknown detection',                                                             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'},   'sampling_type':'continuous injection'}}, 
            '33': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Horiba AQM10',           'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '41': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8310',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '48': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Horiba 300E',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '50': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Massachusetts CO1',      'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '51': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Dasibi 3003',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '54': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C',             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '55': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48C-TL',          'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '66': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (luft)',                   'instrument':'Teledyne ML8830',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '67': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Dasibi 3008',            'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '88': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne ML9830',        'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '93': {                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300',           'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '106':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA360',         'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '108':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO11M', 'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '130':{                                                                                                       'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',              'instrument':'Los Gatos CO',           'reported_lower_limit_of_detection':{'CO':'0.05'}}},
            '147':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Environnement SA CO12M', 'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '158':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',  'instrument':'Horiba APMA370',         'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '167':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'DKK-TOA GFC-311E',       'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '172':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'SIR S-5006',             'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '174':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Ecotech Serinus30',      'reported_lower_limit_of_detection':{'CO':'500.0'},  'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '554':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Thermo 48i-TLE',         'reported_lower_limit_of_detection':{'CO':'40.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '562':{                                                                                                       'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',              'instrument':'Los Gatos CO',           'reported_lower_limit_of_detection':{'CO':'0.05'}}},
            '588':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Ecotech EC9830T',        'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '593':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300EU',         'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}},
            '594':{                                                                                                       'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)', 'instrument':'Teledyne 300EU',         'reported_lower_limit_of_detection':{'CO':'20.0'},   'reported_upper_limit_of_detection':{'CO':'50000.0'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'SO2':
            mapping = {
            '1':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'unknown',                                                                               'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '5':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'second derivative spectrophotometry',            'instrument':'Ecotech SM1000',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '6':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Meloy SA185-2A',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '9':  {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43',              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '10': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                     'instrument':'Philips PW9755',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '11': {                                                                                                                                        'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'colorimetry',                                                                           'reported_lower_limit_of_detection':{'SO2':'10.0'},  'reported_upper_limit_of_detection':{'SO2':'1500.0'},   'sampling_type':'continuous injection'}}, 
            '13': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                                                         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '14': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                                                            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '16': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                                                           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '18': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '19': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                                                           'reported_lower_limit_of_detection':{'SO2':'1.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '20': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '21': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'second derivative spectrophotometry',                                                   'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '22': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                  'instrument':'Asarco 500',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '23': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '24': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                  'instrument':'Asarco 500',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '29': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Beckman 953',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '30': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Bendix 8303',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '31': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'gas chromatography - sulphur chemiluminescence',                                        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'Chemiluminescence only stated as method. Assume gas chromatography - sulphur chemiluminescence is method.',                                            'measuring_assumption':True}},
            '32': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Meloy SA285E',           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '33': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'conductimetry',                                                                         'reported_lower_limit_of_detection':{'SO2':'10.0'},  'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '39': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML8850',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'10000.0'}}},
            '46': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Meloy SA700',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '49': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech AM2020',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'AQS states instrument as using Second Derivative Spectrophotometry method, whereas reference suggests Ultraviolet Fluorescence, side with reference.', 'measuring_assumption':True}},
            '50': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                      'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '59': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                                                              'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '60': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43A',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'10000.0'}}},
            '61': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Dasibi 4108',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '75': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML8850S',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '77': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100',           'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '84': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF21M', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '92': {                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne ML9850',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '94': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '95': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                                        'preparation_assumption':True}, 'measurement':{'methodology':'polarography',                                                                          'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '96': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Hydrogen Peroxide'},                                   'measurement':{'methodology':'titration',                                                                             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '98': {'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'TGS-ANSA'},                                            'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '100':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100A',          'reported_lower_limit_of_detection':{'SO2':'0.4'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '101':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',   'instrument':'OPSIS AR500',            'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '114':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA360',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '115':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-32',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '133':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-112E',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '138':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'differential optical absorption spectroscopy',   'instrument':'Environnement SA SANOA', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '149':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Environnement SA AF22M', 'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '159':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Horiba APSA370',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '166':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'SIR S-5001',             'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '168':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'DKK-TOA GFS-312E',       'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '188':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech Serinus50',      'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '194':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Seres SF2000G',          'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '237':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Sabio 6020',             'reported_lower_limit_of_detection':{'SO2':'0.3'},   'reported_upper_limit_of_detection':{'SO2':'500.0'}}},                                   
            '511':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'coulometry',                                     'instrument':'Philips PW9700',         'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '513':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'flame photometric detection',                    'instrument':'Teledyne ML8450',        'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '560':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Thermo 43C-TLE',         'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '592':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Ecotech EC9850T',        'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '600':{                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'ultraviolet fluorescence',                       'instrument':'Teledyne 100EU',         'reported_lower_limit_of_detection':{'SO2':'0.2'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '801':{'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Pararosaniline'},                                      'measurement':{'methodology':'spectrophotometry',                                                                     'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '811':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'IMPROVE Module D'},                                                                                                                                                                                      'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'}}},
            '850':{'primary_sampling':{'sampling_type':'low volume continuous', 'instrument':'Thermo Partisol 2300'},                                                                                                                                                                                  'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'2.0'},   'reported_upper_limit_of_detection':{'SO2':'1500.0'},                                              'process_details':'Only sequential sampler given as method, make assumption concentrations are determined using Ion Chromatography',                                      'measuring_assumption':True}},
            '900':{'primary_sampling':{'sampling_type':'low volume continuous',                                       'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                                 'measurement':{'methodology':'ion chromatography',                                                                    'reported_lower_limit_of_detection':{'SO2':'0.026'}, 'reported_upper_limit_of_detection':{'SO2':'1500.0'}}}}                                                                                                                                                                                                                               
#-------------------------------------------
        elif run_details['parameter'] in ['CH4','ETHENE','ETHANE','PROPENE','PROPANE','ISOPRENE','BENZENE','TOLUENE','MONOTERPENES','NMHC','NMVOC']:
            mapping = {
            '11': {                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','CH4':'0.1','NMVOC':'1.0'},                                                                                                         'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'},                                                                                              'sampling_type':'continuous injection',   'process_details':'AQS states purely as FID, but FID cannot solely distinguish individual compounds, therefore assume uses GC beforehand.', 'measuring_assumption':True}},       
            '12': {'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'NMVOC':'0.2'},                                                                                                                                   'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                                                                        'process_details':'AQS states purely as FID, but FID cannot solely distinguish individual compounds, therefore assume uses GC beforehand.', 'measuring_assumption':True}},                       
            '51': {'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'6.0','TOLUENE':'19.0'}}}, 
            '91': {                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','CH4':'0.1','NMVOC':'0.1'},          'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'},                                                                                              'sampling_type':'continuous injection',   'process_details':'AQS states purely as FID, but FID cannot solely distinguish VOC, therefore assume uses GC beforehand.',                  'measuring_assumption':True}},
            '92': {                                                                                                  'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                                                                                            'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'},                                                                                                                                                                                                                                                                                      'sampling_type':'continuous injection'}},          
            '101':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                                                                                                       'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                                                                        'process_details':'AQS states method as multi detector GC, assume Gas Chromatography - Mass Spectrometry - Flame Ionisation Detection',     'measuring_assumption':True}},
            '102':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                                                                                            'reported_lower_limit_of_detection':{'BENZENE':'0.31'}}},
            '104':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'PROPANE':'0.33','BENZENE':'0.1','TOLUENE':'0.1'}}},
            '108':{'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'bag'},                                                                              'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                                                                                            'reported_lower_limit_of_detection':{'BENZENE':'0.5','TOLUENE':'0.1'}}},
            '109':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.1','TOLUENE':'0.1'}}},
            '110':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'0.033','PROPANE':'0.033','BENZENE':'0.0166','TOLUENE':'0.01429'}}},
            '113':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'0.01','BENZENE':'0.0033','TOLUENE':'0.002857'}}},
            '115':{'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.1','TOLUENE':'0.1'}}},
            '116':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'9.8','PROPANE':'0.1066','BENZENE':'0.2','TOLUENE':'0.1'}}},
            '119':{'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.01','TOLUENE':'0.01'}}},
            '120':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '122':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '123':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.005','ETHANE':'0.005','PROPENE':'0.0033','PROPANE':'0.0033','ISOPRENE':'0.002','BENZENE':'0.00166','TOLUENE':'0.001429','CH4':'0.01','NMVOC':'0.01'}, 'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'}}},
            '125':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.068','ETHANE':'0.068','PROPENE':'0.0453','PROPANE':'0.0453','ISOPRENE':'0.0272','BENZENE':'0.02266','TOLUENE':'0.019429','NMVOC':'0.136'},            'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection'}},
            '126':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','CH4':'0.1','NMVOC':'0.01'},         'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '127':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','CH4':'0.1','NMVOC':'0.1'},          'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'}}},
            '128':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.005','ETHANE':'0.005','PROPENE':'0.0033','PROPANE':'0.0033','ISOPRENE':'0.002','BENZENE':'0.00166','TOLUENE':'0.001429','NMVOC':'0.1'},               'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection'}},
            '129':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '134':{'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                     'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                                                                                            'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'},                                                                                                                                                                                                                                                                                      'sampling_type':'continuous injection'}},         
            '135':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - direct temperature resolved mass spectrometry',                                                                                                                        'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection'}},                            
            '136':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                                                                                                       'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.4'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection'}},
            '137':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.05','TOLUENE':'0.05','NMVOC':'0.1'},                                                                                                                                                                                                 'sampling_type':'continuous injection'}},
            '142':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.05829','ETHANE':'0.15903','PROPENE':'0.1313','PROPANE':'0.1586466','ISOPRENE':'0.1141','BENZENE':'0.11198','TOLUENE':'0.12165','NMVOC':'0.1'},                                                                                                                                                                              'sampling_type':'continuous injection',   'process_details':'AQS states method as dual column GC, assume Gas Chromatography - Dual Flame Ionisation Detection',                       'measuring_assumption':True}},
            '144':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'}}},            
            '145':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'}}},
            '146':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','CH4':'0.1','NMVOC':'0.1'},          'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '147':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'PROPENE':'0.4466','PROPANE':'0.26','ISOPRENE':'0.148','BENZENE':'0.08833','TOLUENE':'0.06714'}}},
            '148':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - fourier transform infrared spectroscopy - mass spectrometry',                                                                                                          'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429'}}},
            '149':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                                                                                                       'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'}}},
            '150':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '151':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - electron capture detection - photoionisation detection',                                                                                                               'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'}}},
            '152':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},
            '153':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                                                                                                       'reported_lower_limit_of_detection':{'PROPENE':'0.1','BENZENE':'0.05','TOLUENE':'0.05'},                                                                                                                                                                                                                                                                                                                        'process_details':'AQS states method as multi detector GC, assume Gas Chromatography - Mass Spectrometry - Flame Ionisation Detection',     'measuring_assumption':True}},
            '155':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - photoionisation detection',                                                                                                                                            'reported_lower_limit_of_detection':{'BENZENE':'0.1'}}},
            '157':{                                                                                                                                                                                                                                  'measurement':{'methodology':'differential optical absorption spectroscopy',                                     'instrument':'OPSIS AR500',                                                                              'reported_lower_limit_of_detection':{'BENZENE':'1.0','TOLUENE':'2.0'}}},
            '160':{                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'CH4':'0.1','NMVOC':'0.1'},                                                                                                                       'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'},                                                                                              'sampling_type':'continuous injection',   'process_details':'Make assumption the measurement method is GC-FID',                                                                       'measuring_assumption':True}},  
            '161':{                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'CH4':'0.1','NMVOC':'0.1'},                                                                                                                       'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection',   'process_details':'Make assumption the measurement method is GC-FID',                                                                       'measuring_assumption':True}},  
            '162':{                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'CH4':'0.1','NMVOC':'0.1'},                                                                                                                       'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'},                                                                                              'sampling_type':'continuous injection',   'process_details':'Make assumption the measurement method is GC-FID',                                                                       'measuring_assumption':True}}, 
            '164':{                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                                                       'instrument':'Thermo 55C',    'overwrite_methodology':'gas chromatography - flame ionisation detection', 'reported_lower_limit_of_detection':{'CH4':'0.1','NMVOC':'0.1'},                                                                                                                       'reported_upper_limit_of_detection':{'CH4':'70000.0','NMVOC':'20000.0'},                                                                                              'sampling_type':'continuous injection',   'process_details':'Instrument stated as Thermo 55, assume it is Thermo 55C',                                                                'measuring_assumption':True}},
            '168':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - electron capture detection - photoionisation detection',                                                                                                               'reported_lower_limit_of_detection':{'BENZENE':'0.5','TOLUENE':'0.2'}}},       
            '169':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'bag'},                                                                              'measurement':{'methodology':'gas chromatography - electron capture detection - photoionisation detection',                                                                                                               'reported_lower_limit_of_detection':{'BENZENE':'0.5','TOLUENE':'0.2'}}},
            '171':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.0166','TOLUENE':'0.01429'}}},      
            '172':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'0.033','PROPANE':'0.033','BENZENE':'0.0166','TOLUENE':'0.01429'}}},          
            '173':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'NMVOC':'0.1'},                                                                                                                                   'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection',   'process_details':'AQS states purely as FID, but FID cannot solely distinguish individual compounds, therefore assume uses GC beforehand.', 'measuring_assumption':True}},                                                 
            '174':{                                                                                                  'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033'},                                                                                                                                                                                                                                                             'sampling_type':'continuous injection'}},
            '175':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429'}}},                                                           
            '176':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'0.059','PROPANE':'0.02','BENZENE':'0.04','TOLUENE':'0.04'}}},                 
            '177':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                                                                                                                       'reported_lower_limit_of_detection':{'ETHENE':'0.05','ETHANE':'0.05','PROPENE':'0.033','PROPANE':'0.033','ISOPRENE':'0.02','BENZENE':'0.0166','TOLUENE':'0.01429','NMVOC':'0.1'},                      'reported_upper_limit_of_detection':{'NMVOC':'20000.0'}}},                                                               
            '200':{                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                                                                                    'reported_lower_limit_of_detection':{'NMVOC':'0.1'},                                                                                                                                   'reported_upper_limit_of_detection':{'NMVOC':'30000.0'},                                                                                                              'sampling_type':'continuous injection',   'process_details':'Assume GC as not explicitely stated',                                                                                    'measuring_assumption':True}},                                   
            '202':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - ultraviolet detection',                                                                                                                            'reported_lower_limit_of_detection':{'CH4':'0.1'}}},
            '206':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'CH4':'66.0'}}},                       
            '210':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'ISOPRENE':'0.02','BENZENE':'0.02','TOLUENE':'0.02'}}},
            '211':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'PROPENE':'0.13','BENZENE':'0.1','TOLUENE':'0.12'}}},                         
            '221':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '222':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '223':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '224':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '225':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '226':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '227':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '228':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '229':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '230':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '231':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '232':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                                                                                                                      'reported_lower_limit_of_detection':{'ETHENE':'0.25','ETHANE':'0.25','PROPENE':'0.166','PROPANE':'0.166','ISOPRENE':'0.1','BENZENE':'0.0833','TOLUENE':'0.071429','NMVOC':'0.5'},                      'reported_upper_limit_of_detection':{'ETHENE':'500.0','ETHANE':'500.0','PROPENE':'333.33','PROPANE':'333.33','ISOPRENE':'200.0','BENZENE':'166.66','TOLUENE':'142.86','NMVOC':'0.5'}, 'sampling_type':'continuous injection'}},
            '300':{                                                                                                                                                                                                                                  'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                                                                                    'reported_lower_limit_of_detection':{'NMVOC':'0.1'},                                                                                                                                   'reported_upper_limit_of_detection':{'NMVOC':'20000.0'},                                                                                                              'sampling_type':'continuous injection',   'process_details':'Assume GC as not explicitely stated',                                                                                    'measuring_assumption':True}},                                   
            '561':{                                                                                                                                                                                                                                  'measurement':{'methodology':'cavity ringdown spectroscopy',                                                     'instrument':'Picarro G1301',                                                                            'reported_lower_limit_of_detection':{'CH4':'0.21'},                                                                                                                                    'reported_upper_limit_of_detection':{'CH4':'70000.0'}}},
            '903':{'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                                                                                                           'reported_lower_limit_of_detection':{'BENZENE':'0.8066','TOLUENE':'0.671429'}}}, 
            '908':{'primary_sampling':{'sampling_type':'low volume continuous'},                                     'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'gas chromatography - mass spectrometry',                                                                                                                                                    'reported_lower_limit_of_detection':{'BENZENE':'0.1966','TOLUENE':'0.132857'}}}} 
#-------------------------------------------
        elif run_details['parameter'] in ['VOC','HC']:
            mapping = {
            '11': {                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                                                       'reported_lower_limit_of_detection':{'HC':'200.0'}, 'reported_upper_limit_of_detection':{'HC':'70000.0'}}},        
            '104':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'}}},    
            '120':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'}}},
            '122':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                             'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'}}},         
            '126':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'HC':'0.01'},  'reported_upper_limit_of_detection':{'HC':'70000.0'}}},       
            '152':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                  'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'}}},
            '160':{                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                                                       'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'},   'process_details':'Make assumption the measurement method is FID',           'measuring_assumption':True}},    
            '161':{                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                                                       'reported_lower_limit_of_detection':{'HC':'0.1'},                                                           'process_details':'Make assumption the measurement method is FID',           'measuring_assumption':True}},          
            '162':{                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                                                       'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'},   'process_details':'Make assumption the measurement method is FID',           'measuring_assumption':True}}, 
            '164':{                                                                                                                                                                                                                                  'measurement':{'methodology':'flame ionisation detection',                           'instrument':'Thermo 55C',  'reported_lower_limit_of_detection':{'HC':'0.1'},   'reported_upper_limit_of_detection':{'HC':'70000.0'},   'process_details':'Instrument stated as Thermo 55, assume it is Thermo 55C', 'measuring_assumption':True}}}
#-------------------------------------------           
        elif run_details['parameter'] == 'HCHO':
            mapping = {
            '101':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - mass spectrometry - flame ionisation detection',                               'reported_lower_limit_of_detection':{'HCHO':'0.1'},                                               'process_details':'AQS states method as multi detector GC, assume Gas Chromatography - Mass Spectrometry - Flame Ionisation Detection', 'measuring_assumption':True}},          
            '102':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - ultraviolet detection',                                    'reported_lower_limit_of_detection':{'HCHO':'0.1'}}}, 
            '114':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - photodiode array detection',                               'reported_lower_limit_of_detection':{'HCHO':'0.085'}}},           
            '120':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'}, 'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                   'reported_lower_limit_of_detection':{'HCHO':'0.1'}}},          
            '122':{'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'flask'},                                                                            'measurement':{'methodology':'gas chromatography - dual flame ionisation detection',                                              'reported_lower_limit_of_detection':{'HCHO':'0.1'}}},                    
            '125':{                                                                                                  'preparation_1':{'sample_preparation_type':'preconcentration'},                                                                 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                   'reported_lower_limit_of_detection':{'HCHO':'0.136'},   'sampling_type':'continuous injection'}},      
            '126':{'primary_sampling':{'sampling_type':'high volume continuous'},                                    'preparation_1':{'sample_preparation_type':'flask'},            'preparation_2':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                   'reported_lower_limit_of_detection':{'HCHO':'0.1'}}},          
            '156':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},          'preparation_2':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'high performance liquid chromatography - photodiode array detection',                               'reported_lower_limit_of_detection':{'HCHO':'0.175'}}},                   
            '157':{                                                                                                                                                                                                                                  'measurement':{'methodology':'differential optical absorption spectroscopy',                        'instrument':'OPSIS AR500',   'reported_lower_limit_of_detection':{'HCHO':'1.0'}}},            
            '165':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - photodiode array detection',                               'reported_lower_limit_of_detection':{'HCHO':'0.03'}}},                     
            '166':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'ion chromatography',                                                  'instrument':'Dionex DX-300', 'reported_lower_limit_of_detection':{'HCHO':'0.03'}}},                       
            '179':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - photodiode array detection',                               'reported_lower_limit_of_detection':{'HCHO':'0.06'}}},                     
            '202':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},          'preparation_2':{'sample_preparation_type':'sorbent trapping'}, 'measurement':{'methodology':'high performance liquid chromatography - ultraviolet detection',                                    'reported_lower_limit_of_detection':{'HCHO':'0.1'}}}, 
            '203':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping'},                                                                 'measurement':{'methodology':'high performance liquid chromatography - ultraviolet detection',                                    'reported_lower_limit_of_detection':{'HCHO':'0.1'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'NH3':
            mapping = {
            '10': {'primary_sampling':{'sampling_type':'passive'},                                                   'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'ion chromatography',                                                                                                               'reported_lower_limit_of_detection':{'NH3':'0.1'}}},
            '50': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                    'measurement':{'methodology':'ion chromatography',                                                                                                               'reported_lower_limit_of_detection':{'NH3':'0.03'}}}, 
            '51': {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Thermo 17C',              'reported_lower_limit_of_detection':{'NH3':'1.0'}}},
            '82': {                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'instrument':'Teledyne 201E',           'reported_lower_limit_of_detection':{'NH3':'1.0'}}},
            '91': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Nessler'},           'measurement':{'methodology':'spectrophotometry',                                                                                                                'reported_lower_limit_of_detection':{'NH3':'7.2'}}},
            '92': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'reagent reaction', 'specific_technique':'Sodium Phenolate'},  'measurement':{'methodology':'colorimetry',                                                                                                                      'reported_lower_limit_of_detection':{'NH3':'7.2'}}},
            '93': {                                                                                                                                                                                                            'measurement':{'methodology':'electrochemical membrane diffusion',                                                                                               'reported_lower_limit_of_detection':{'NH3':'1000.0'}}}, 
            '101':{                                                                                                                                                                                                            'measurement':{'methodology':'photoacoustic spectroscopy',                                                               'instrument':'Pranalytica Nitrolux200', 'reported_lower_limit_of_detection':{'NH3':'0.2'}}},
            '121':{                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)',                                        'instrument':'DKK-TOA GLN-114E',        'reported_lower_limit_of_detection':{'NH3':'5.0'}}},
            '400':{                                                                                                                                                                                                            'measurement':{'methodology':'photoacoustic spectroscopy',                                                               'instrument':'Pranalytica Nitrolux200', 'reported_lower_limit_of_detection':{'NH3':'0.2'}}},
            '592':{                                                                                                                                                                                                            'measurement':{'methodology':'chemiluminescence (internal molybdenum and quartz converters)',                            'instrument':'Ecotech EC9842T',         'reported_lower_limit_of_detection':{'NH3':'0.2'}}}}
#-------------------------------------------
        elif run_details['parameter'] == 'HNO3':
            mapping = {
            '50': {'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'denuder'},                                                                                   'measurement':{'methodology':'ion chromatography', 'reported_lower_limit_of_detection':{'HNO3':'0.00777'}}},      
            '101':{'primary_sampling':{'sampling_type':'passive',               'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography', 'reported_lower_limit_of_detection':{'HNO3':'0.77656'}}},
            '900':{'primary_sampling':{'sampling_type':'low volume continuous', 'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter pack',      'specific_technique':'3-Stage Filter Pack'},                              'measurement':{'methodology':'ion chromatography', 'reported_lower_limit_of_detection':{'HNO3':'0.00272'}}}}                                                                                                                             
#-------------------------------------------
        elif run_details['parameter'] == 'PAN':
            mapping = {
            '11':{'measurement':{'methodology':'gas chromatography - electron capture detection', 'reported_lower_limit_of_detection':{'PAN':'0.1'}, 'sampling_type':'continuous injection'}}} 
#-------------------------------------------
        elif run_details['parameter'] in ['PM10','PM2.5','PM1']:
            mapping = {
            '1':            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 244 Dichot'},                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'1.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},
            '2':            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'General Metal Works 9200'},                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'1.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                    
            '3':            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Wedding and Associates Dichot'},                                                                                                                                                                                                            'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'1.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                  
            '4':            {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 246B Dichot'},                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'1.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                              
            '25':           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 254'},                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                        
            '40':           {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Wedding and Associates Beta',                                                                                                                       'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}}, 
            '41':           {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Sibata BAM102',                                                                                                                                     'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},       
            '51':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Sierra Andersen 321A',           'process_details':'Make assumption sampling instrument is Sierra Andersen 321A', 'primary_sampling_assumption':True},                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},
            '52':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Sierra Andersen 321A'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},            
            '53':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'General Metal Works 9000'},                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                        
            '54':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                  
            '62':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                          
            '63':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                          
            '64':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                          
            '65':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-C'},                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                          
            '71':           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Oregon DEQ'},                                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                  
            '73':           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 246B Dichot'},                                                                                                                                                                                                              'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                      
            '76':           {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Andersen FH621-N',                                                                                                                                  'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},           
            '79':           {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                      'instrument':'Thermo 1400',                                                                                                                                                                                                          'reported_upper_limit_of_detection':{'PM10':'13000.0'}}},
            '81':           {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Wedding and Associates Beta',                                                                                                                                                                                          'reported_upper_limit_of_detection':{'PM10':'5000.0'}}}, 
            '98':           {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                   
            '116':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                       
            '117':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                     
            '118':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                        
            '119':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 100'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                      
            '120':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 300'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                        
            '122':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1020',                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                  
            '124':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ100'},                                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                              
            '125':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                              
            '126':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                   
            '127':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'17000.0'}}},                                                                              
            '130':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS10 100'},                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                                          
            '132':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS10 300'},                                                                                                                                                                                                                      'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                                      
            '141':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Tisch Environmental 6070'},                                                                                                                                                                                                                 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                     
            '142':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                                                                                                                                                                                                'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                         
            '143':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                               
            '144':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                            
            '145':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                                    
            '150':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Thermo FH62 C14',                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},         
            '153':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 100'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                                  
            '155':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 300'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                                  
            '162':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Ecotech 3000'},                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                               
            '170':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1020',                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'5.0'},               'reported_upper_limit_of_detection':{'PM2.5':'975.0'}}},                  
            '171':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1022',                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'5.0'},               'reported_upper_limit_of_detection':{'PM2.5':'975.0'}}},                         
            '177':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                                        
            '178':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                      'instrument':'Thermo 1405',                                                                                                                                       'reported_lower_limit_of_detection':{'PM2.5':'2.0'}}},        
            '179':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0', 'PM2.5':'2.0'}, 'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                     
            '181':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                      'instrument':'Thermo 1400a',                'overwrite_methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},
            '182':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'instrument':'Thermo 1405-DF',                                                                                                                                    'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},        
            '183':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Thermo 5014i',                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                         
            '183_pm2.5_alt':{                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'instrument':'Thermo 1405F',                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'2.0'}}},                
            '184':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation - nephelometry',                                                            'instrument':'Thermo 5030 SHARP',                                                                                                                                 'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                    
            '195':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'optical particle counter',                                                                   'instrument':'Grimm 180',                                                                                                                                         'reported_lower_limit_of_detection':{'PM10':'0.1','PM2.5':'0.1'},  'reported_upper_limit_of_detection':{'PM10':'1500.0','PM2.5':'1500.0'}}},                                               
            '197':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                      
            '198':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                              
            '204':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Teledyne 602 BetaPlus',                                                                                                                             'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'1500.0'}}},                                    
            '205':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Teledyne 602 BetaPlus',                                                                                                                             'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'1500.0'}}},                    
            '208':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'instrument':'Thermo 1405-DF',                                                                                                                                    'reported_lower_limit_of_detection':{'PM10':'2.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},  
            '209':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1022',                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'5.0'},               'reported_upper_limit_of_detection':{'PM2.5':'975.0'}}},                        
            '216':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Tisch Environmental Wilbur'},                                                                                                                                                                                                               'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                                     
            '221':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One E-FRM-DC'},                                                                                                                                                                                                                         'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},  
            '226':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One E-BAM Plus',                                                                                                                                'reported_lower_limit_of_detection':{'PM10':'10.0'},               'reported_upper_limit_of_detection':{'PM10':'10000.0'}}},                  
            '236':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'optical particle counter',                                                                   'instrument':'Teledyne 640',                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'0.1'},               'reported_upper_limit_of_detection':{'PM2.5':'10000.0'}}},                          
            '238':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'optical particle counter',                                                                   'instrument':'Teledyne 640x',                                                                                                                                     'reported_lower_limit_of_detection':{'PM2.5':'0.1'},               'reported_upper_limit_of_detection':{'PM2.5':'10000.0'}}},                              
            '239':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'optical particle counter',                                                                   'instrument':'Teledyne 640x',                                                                                                                                     'reported_lower_limit_of_detection':{'PM10':'0.1'},                'reported_upper_limit_of_detection':{'PM10':'10000.0'}}},                   
            '246':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One E-SEQ-FRM'},                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},                                                                                                
            '521':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One E-FRM-DC'},                                                                                                                                                                                                                         'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},             
            '545':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One E-SEQ-FRM'},                                                                                                                                                                                                                        'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},         
            '581':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'instrument':'Thermo 1405F',                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},               
            '701':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},               
            '702':          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},            
            '702_pm2.5_alt':{                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                          
            '703':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                 
            '704':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                 
            '707':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module A'},                                                                                                                                                                                                                         'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'}}},                     
            '711':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                 
            '712':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                  
            '713':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                  
            '715':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},               
            '716':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                
            '717':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                
            '718':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                 
            '722':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                  
            '723':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                    
            '724':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                   
            '731':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1020',                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'5.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                       
            '733':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One BAM1020',                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'5.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                      
            '734':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Met One E-BAM',                                                                                                                                     'reported_lower_limit_of_detection':{'PM10':'10.0','PM2.5':'6.0'}, 'reported_upper_limit_of_detection':{'PM10':'10000.0','PM2.5':'5000.0'}}},                            
            '740':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'pressure drop tape sampling',                                                                'instrument':'CAMMS',                                                                                                                                             'reported_lower_limit_of_detection':{'PM2.5':'3.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}}, 
            '741':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'pressure drop tape sampling',                                                                'instrument':'CAMMS',                                                                                                                                             'reported_lower_limit_of_detection':{'PM2.5':'3.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},      
            '750':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Andersen FH621-N',                                                                                                                                                                                                     'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},
            '753':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'beta-attenuation',                                                                           'instrument':'Andersen FH621-N',                                                                                                                                                                                                     'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                           
            '760':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.06'},              'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},
            '761':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry',                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.06'},              'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},
            '771':          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},
            '771_pm2.5_alt':{                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'nephelometry',                                                                               'instrument':'Radiance Research 903',                                                                                                                             'reported_lower_limit_of_detection':{'PM2.5':'0.4'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}}, 
            '773':          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0'},                'reported_upper_limit_of_detection':{'PM10':'5000.0'}}},
            '781':          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                 
            '785':          {'primary_sampling':{'sampling_type':'low volume continuous',                                                                                                                                  'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM10':'4.0','PM2.5':'2.0'},  'reported_upper_limit_of_detection':{'PM10':'5000.0','PM2.5':'5000.0'}}},           
            '789':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                    
            '790':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - filter dynamics measurement system - gravimetry', 'instrument':'Thermo 1405-DF',                                                                                                                                                                                                       'reported_upper_limit_of_detection':{'PM10':'5000.0','PM2.5':'5000.0'}}},  
            '791':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'unknown',                                                                                                                                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},    
            '792':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                      'instrument':'Thermo 1405D'}},           
            '793':          {'primary_sampling':{'sampling_type':'low volume continuous'},                                                                                                                                                                      'preparation_1':{'sample_preparation_type':'filter', 'preparation_assumption':True}, 'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'2.0'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},   
            '802':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'nephelometry',                                                                               'instrument':'Ambilabs 2-WIN',                                                                                                                                    'reported_lower_limit_of_detection':{'PM2.5':'0.7'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},            
            '810':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.104'},             'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                                
            '812':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'nephelometry',                                                                               'instrument':'Ecotech 9003',                                                                                                                                      'reported_lower_limit_of_detection':{'PM2.5':'0.4'},               'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}}, 
            '820':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Teflon'},                                                                                                                                                                                                                  'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.04'},              'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                            
            '830':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 400'},                                                                                                                                                                                                                             'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.04'},              'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                                    
            '850':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                                                                                                                                                                                                     'measurement':{'methodology':'gravimetry',                                                                                                                                                                                                                                                   'reported_lower_limit_of_detection':{'PM2.5':'0.04'},              'reported_upper_limit_of_detection':{'PM2.5':'5000.0'}}},                                                                         
            '879':          {                                                                                                                                                                                                                                                                                                                        'measurement':{'methodology':'tapered element oscillating microbalance - gravimetry',                                                                                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM10':'5000.0'}}}}
#------------------------------------------- 
        elif parameter_details['matrix'] == 'pm':
            mapping = {
            '43': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                    'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'},   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},          
            '44': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',         'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                      
            '45': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                      
            '57': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                              
            '59': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',         'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                               
            '68': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                         
            '69': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                        
            '70': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                       
            '74': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                    'reported_lower_limit_of_detection':{'Pb':'0.002'},    'reported_upper_limit_of_detection':{'Pb':'100.0'},   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                     
            '85': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                 
            '86': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                 
            '89': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0547'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                               
            '90': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.003'},    'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                        
            '91': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                             
            '92': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                    'reported_lower_limit_of_detection':{'Pb':'0.002'},    'reported_upper_limit_of_detection':{'Pb':'100.0'},   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                     
            '94': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                    
            '95': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                              
            '96': {'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.002'},    'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                               
            '101':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                    'reported_lower_limit_of_detection':{'Pb':'0.001'},    'reported_upper_limit_of_detection':{'Pb':'100.0'},   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                  
            '107':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'graphite furnace atomic absorption spectroscopy',         'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                     
            '108':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.00001'},  'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                        
            '109':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                                          
            '110':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},          
            '113':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                                       
            '114':{'primary_sampling':{'sampling_type':'low volume continuous'},                               'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                         'reported_lower_limit_of_detection':{'Pb':'0.00001'},  'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                                                 
            '115':{'primary_sampling':{'sampling_type':'low volume continuous'},                               'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.00001'},  'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                                                 
            '140':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                                                                       
            '189':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                  
            '191':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.00002'},  'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                         
            '192':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                            
            '193':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                           
            '196':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.0075'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                           
            '304':{'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Xonteck 924'}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                                                                    'measurement':{'methodology':'X-ray fluorescence spectroscopy',                         'reported_lower_limit_of_detection':{'Pb':'0.004'},    'reported_upper_limit_of_detection':{'Pb':'100.0'}}}, 
            '305':{'primary_sampling':{'sampling_type':'low volume continuous',   'instrument':'Xonteck 924'}, 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.003'},    'reported_upper_limit_of_detection':{'Pb':'100.0'}}},         
            '803':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'flame atomic absorption spectroscopy',                    'reported_lower_limit_of_detection':{'Pb':'0.01'},     'reported_upper_limit_of_detection':{'Pb':'100.0'},   'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},            
            '813':{'primary_sampling':{'sampling_type':'high volume continuous'},                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_3':{'sample_preparation_type':'reagent reaction'}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',            'reported_lower_limit_of_detection':{'Pb':'0.0002'},   'reported_upper_limit_of_detection':{'Pb':'100.0'}}},                                       
            '905':{'primary_sampling':{'sampling_type':'low volume continuous'},                               'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'reagent reaction'},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.35'}}},                                                 
            '910':{'primary_sampling':{'sampling_type':'low volume continuous'},                               'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'reagent reaction'},                                                                                                                                    'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'reported_lower_limit_of_detection':{'Pb':'0.0486'}}}}                                                          
#------------------------------------------- 
        elif run_details['parameter'] in ['PM10_C','PM10_EC','PM10_OC','PM2.5_C','PM2.5_EC','PM2.5_OC','PM1_C','PM1_EC','PM1_OC']:
            mapping = {
            '62':         {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal analysis',                                                                                                            'reported_lower_limit_of_detection':{'PM10_C':'4.0'},                                            'process_details':'Make assumption method is thermal analysis',                   'measuring_assumption':True}},
            '102':        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal analysis',                                                                                                            'reported_lower_limit_of_detection':{'PM10_C':'1.0'},                                            'process_details':'Make assumption method is thermal analysis',                   'measuring_assumption':True}},      
            '108':        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',  'instrument':'DRI Model 2001',                                                        'reported_lower_limit_of_detection':{'PM10_C':'0.2','PM10_EC':'1.0','PM10_OC':'0.2'}}}, 
            '111':        {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040', 'instrument':'DRI Model 2001',                                                        'reported_lower_limit_of_detection':{'PM10_EC':'1.0'}}},         
            '805':        {'primary_sampling':{'sampling_type':'low volume continuous',                                             'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.02','PM2.5_OC':'0.087'}}},                                     
            '809':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module C'},                                               'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},          
            '813':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_C':'0.245','PM2.5_EC':'0.245','PM2.5_OC':'0.245'}}},             
            '815_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.233','PM2.5_OC':'1.005'}}},          
            '815_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},            
            '816_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},          
            '816_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},                    
            '823':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Quartz'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.224','PM2.5_OC':'0.224'}}},     
            '825':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Quartz'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.213','PM2.5_OC':'0.918'}}},                   
            '831':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG 3000N'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002'}}},                              
            '833':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 450'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.098','PM2.5_OC':'0.098'}}},           
            '837':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 450'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.025','PM2.5_OC':'0.025'}}},         
            '838_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG 3000N'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},          
            '838_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG 3000N'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},                  
            '839':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG 3000N'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_OC':'0.002'}}},             
            '840':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG 3000N'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002'}}},              
            '843_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},
            '843_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                   'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},                       
            '843_tot_alt':{'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.098','PM2.5_OC':'0.098'}}},
            '847':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.098','PM2.5_OC':'0.098'}}},        
            '853':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.163','PM2.5_OC':'0.163'}}},      
            '855':        {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.155','PM2.5_OC':'0.67'}}},    
            '860':        {                                                                                                                                                                                                                                                                                                                                                                                                                    'measurement':{'methodology':'thermal analysis',                      'instrument':'Rupprecht and Patashnick 5400',                                         'reported_lower_limit_of_detection':{'PM2.5_C':'0.2','PM2.5_EC':'0.2','PM2.5_OC':'0.2'}}},
            '867':        {                                                                                                                                              'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040', 'instrument':'Sunset OCEC Model 5',           'sampling_type':'continuous injection', 'reported_lower_limit_of_detection':{'PM2.5_C':'0.2','PM2.5_EC':'0.2','PM2.5_OC':'0.2'},         'process_details':'Make assumption measuring instrument is Sunset OCEC Model 5.', 'measuring_assumption':True}},
            '883_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},                  
            '883_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},                  
            '888_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM10_EC':'0.002','PM10_OC':'0.002'}}},                          
            '888_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM10_EC':'0.002','PM10_OC':'0.002'}}},                          
            '892_tor':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                                        'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}},          
            '892_tot':    {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                    'measurement':{'methodology':'thermal-optical analysis - NIOSH 5040',                                                                                       'reported_lower_limit_of_detection':{'PM2.5_EC':'0.002','PM2.5_OC':'0.002'}}}}          
#------------------------------------------- 
        elif parameter_details['matrix'] in ['pm10','pm2.5','pm1']: 
            mapping = {
            '62':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry'},                                                                                                                                                                                'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_Cl-':'4.0','PM10_K+':'4.0','PM10_NH4+':'4.0','PM10_NO3-':'4.0','PM10_SO4--':'4.0'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           'process_details':'Assume method is ion chromatography after gravimetry.',                                                        'measuring_assumption':True}},
            '85':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_Al':'0.01','PM10_Ca++':'0.0072','PM10_K+':'0.0017','PM10_Mg++':'0.0128','PM10_Na+':'0.0213','PM10_Se':'0.0128'}}},           
            '91':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction',                                                  'preparation_assumption':True}, 'measurement':{'methodology':'colorimetry',                                                                                            'reported_lower_limit_of_detection':{'PM10_SO4--':'0.05'}}},   
            '92':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                   'reported_lower_limit_of_detection':{'PM10_As':'0.02','PM10_Cd':'0.001','PM10_Co':'0.004','PM10_Cr':'0.002','PM10_Cu':'0.001','PM10_Fe':'0.01','PM10_K+':'0.03','PM10_Mg++':'0.001','PM10_Mn':'0.001','PM10_Ni':'0.004','PM10_Pb':'0.002','PM10_Zn':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                       
            '92_cl_alt':    {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_Cl-':'0.11'}}},                   
            '92_no3_alt':   {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction',         'specific_technique':'Griess-Saltzman'},                                 'measurement':{'methodology':'spectrophotometry',                                                                                      'reported_lower_limit_of_detection':{'PM10_NO3-':'0.1'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                 'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                           'measuring_assumption':True}},               
            '93':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_NO3-':'0.05','PM10_SO4--':'0.5'}}},    
            '94':           {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction',         'specific_technique':'Griess-Saltzman'},                                 'measurement':{'methodology':'spectrophotometry',                                                                                      'reported_lower_limit_of_detection':{'PM10_NO3-':'0.05'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                'process_details':'Assume measurement methodology is spectrophotometry after sampling',                                           'measuring_assumption':True}},                
            '97':           {'primary_sampling':{'sampling_type':'low volume continuous',                                               'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_NH4+':'0.5'}}},                               
            '102':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                   'reported_lower_limit_of_detection':{'PM10_Al':'0.01','PM10_As':'0.02','PM10_Cd':'0.0002','PM10_Co':'0.004','PM10_Cr':'0.002','PM10_Cu':'0.001','PM10_Fe':'0.01','PM10_Mn':'0.001','PM10_Ni':'0.004','PM10_Pb':'0.08','PM10_Se':'0.01','PM10_Zn':'0.0002'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},                      
            '102_alt':      {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_Cl-':'0.11','PM10_K+':'0.03','PM10_NH4+':'0.1','PM10_NO3-':'0.05','PM10_SO4--':'0.5'}}}, 
            '103':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                   'reported_lower_limit_of_detection':{'PM10_Al':'0.01','PM10_As':'0.02','PM10_Cd':'0.0002','PM10_Cr':'0.002','PM10_Cu':'0.001','PM10_Pb':'0.08','PM10_Se':'0.01','PM10_Zn':'0.0002'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},      
            '103_alt':      {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_NO3-':'0.05','PM10_SO4--':'0.5'}}},    
            '105':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_As':'0.01','PM10_Cd':'0.001','PM10_Co':'0.004','PM10_Cr':'0.002','PM10_Cu':'0.001','PM10_Fe':'0.01','PM10_Hg':'0.025','PM10_Mn':'0.001','PM10_Ni':'0.004','PM10_Pb':'0.01','PM10_Se':'0.03','PM10_V':'0.001','PM10_Zn':'0.0133'}}},                             
            '106':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_Al':'0.027','PM10_As':'0.01','PM10_Ca++':'0.22','PM10_Cd':'0.001','PM10_Co':'0.02','PM10_Cr':'0.01','PM10_Cu':'0.001','PM10_Fe':'0.01','PM10_K+':'0.15','PM10_Mg++':'0.017','PM10_Mn':'0.001','PM10_Na+':'1.75','PM10_Ni':'0.004','PM10_Pb':'0.08','PM10_Se':'0.014','PM10_V':'0.001','PM10_Zn':'0.0133'}}},                            
            '107':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_Al':'0.014','PM10_As':'0.022','PM10_Ca++':'0.276','PM10_Cd':'0.001','PM10_Co':'0.004','PM10_Cr':'0.002','PM10_Cu':'0.001','PM10_Fe':'0.01','PM10_K+':'0.205','PM10_Mg++':'0.028','PM10_Mn':'0.001','PM10_Na+':'1.769','PM10_Ni':'0.004','PM10_Pb':'0.01','PM10_Se':'0.017','PM10_V':'0.001','PM10_Zn':'0.0133'}}},              
            '109':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.0003','PM10_Cd':'0.0003','PM10_Co':'0.0003','PM10_Cr':'0.0003','PM10_Cu':'0.0003','PM10_Mn':'0.0003','PM10_Ni':'0.0003','PM10_Pb':'0.0003','PM10_V':'0.0003','PM10_Zn':'0.0003'}}},                
            '110':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00013','PM10_Cd':'0.00001','PM10_Co':'0.0000492','PM10_Cr':'0.0004','PM10_Cu':'0.03','PM10_Fe':'0.00004','PM10_Hg':'0.00003','PM10_Mn':'0.00006','PM10_Ni':'0.00007','PM10_Pb':'0.0025','PM10_Se':'0.00004','PM10_V':'0.00004','PM10_Zn':'0.00004'}}},   
            '112':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00001'}}},                                  
            '113':          {'primary_sampling':{'sampling_type':'high volume continuous'},                                                                                  'preparation_1':{'sample_preparation_type':'filter'},                                                                              'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True}, 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00009','PM10_Cd':'0.00002','PM10_Co':'0.0001','PM10_Cr':'0.0006','PM10_Hg':'0.00003','PM10_Mn':'0.00013','PM10_Ni':'0.0005','PM10_Pb':'0.00009','PM10_Se':'0.00003'}}}, 
            '118':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM2.5_As':'0.0001','PM2.5_Cd':'0.00002','PM2.5_Co':'0.00005','PM2.5_Cr':'0.00012','PM2.5_Mn':'0.00022','PM2.5_Ni':'0.00009','PM2.5_Pb':'0.00016','PM2.5_Se':'0.00042'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   'reported_upper_limit_of_detection':{'PM2.5_As':'5.0','PM2.5_Cd':'5.0','PM2.5_Co':'5.0','PM2.5_Cr':'5.0','PM2.5_Mn':'5.0','PM2.5_Ni':'5.0','PM2.5_Pb':'5.0','PM2.5_Se':'5.0'}}},              
            '141':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Tisch Environmental 6070'},                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_As':'0.0003','PM10_Cr':'0.005','PM10_Ni':'0.0012'}}},       
            '201':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00002','PM10_Cd':'0.00002','PM10_Co':'0.00002','PM10_Cr':'0.00002','PM10_Mn':'0.00002','PM10_Ni':'0.00002','PM10_Pb':'0.00002','PM10_Se':'0.00002','PM10_V':'0.00002'}}},               
            '202':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00002','PM10_Cd':'0.00002','PM10_Co':'0.00002','PM10_Cr':'0.00002','PM10_Hg':'0.00002','PM10_Mn':'0.00002','PM10_Ni':'0.00002','PM10_Pb':'0.00002','PM10_Se':'0.00002'}}},      
            '203':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.0043','PM10_Cd':'0.0007','PM10_Co':'0.0012','PM10_Cr':'0.0036','PM10_Hg':'0.0000464','PM10_Mn':'0.0006','PM10_Ni':'0.0012','PM10_Pb':'0.0007','PM10_Se':'0.0012'}}},      
            '204':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00022','PM10_Cd':'0.00018','PM10_Co':'0.00022','PM10_Cr':'0.00496','PM10_Hg':'0.00207','PM10_Mn':'0.00122','PM10_Ni':'0.00179','PM10_Pb':'0.00066','PM10_Se':'0.00027'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                      'reported_upper_limit_of_detection':{'PM10_As':'10.0','PM10_Cd':'10.0','PM10_Co':'10.0','PM10_Cr':'10.0','PM10_Hg':'10.0','PM10_Mn':'10.0','PM10_Ni':'10.0','PM10_Pb':'10.0','PM10_Se':'10.0'}}},   
            '303':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Sierra Andersen 246B Dichot'},                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Al':'0.004','PM2.5_Al':'0.004','PM10_As':'0.002','PM2.5_As':'0.002','PM10_Ca+':'0.003','PM2.5_Ca+':'0.003','PM10_Cd':'0.002','PM2.5_Cd':'0.002','PM10_Cl-':'0.005','PM2.5_Cl-':'0.005','PM10_Co':'0.006','PM2.5_Co':'0.006','PM10_Cr':'0.001','PM2.5_Cr':'0.001','PM10_Cu':'0.001','PM2.5_Cu':'0.001','PM10_Fe':'0.001','PM2.5_Fe':'0.001','PM10_Hg':'0.002','PM2.5_Hg':'0.002','PM10_K+':'0.003','PM2.5_K+':'0.003','PM10_Mn':'0.001','PM2.5_Mn':'0.001','PM10_Ni':'0.001','PM2.5_Ni':'0.001','PM10_Pb':'0.002','PM2.5_Pb':'0.002','PM10_Se':'0.001','PM2.5_Se':'0.001','PM10_V':'0.004','PM2.5_V':'0.004','PM10_Zn':'0.001','PM2.5_Zn':'0.001'}}},                                                                                                                                            
            '304':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 1200'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_As':'0.005','PM10_Al':'0.005','PM10_Ca+':'0.002','PM10_Cd':'0.016','PM10_Cl-':'0.002','PM10_Co':'0.00003','PM10_Cr':'0.002','PM10_Cu':'0.0001','PM10_Fe':'0.00004','PM10_Hg':'0.057','PM10_K+':'0.029','PM10_Mn':'0.00005','PM10_Ni':'0.002','PM10_Pb':'0.0001','PM10_Se':'0.005','PM10_V':'0.00009','PM10_Zn':'0.0006'}}},                                                                                                                            
            '305':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Graseby Andersen GMW 321-B'},                                       'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_As':'0.005','PM10_Ca+':'0.002','PM10_Cd':'0.016','PM10_Cl-':'0.002','PM10_Co':'0.00003','PM10_Cr':'0.002','PM10_Cu':'0.0001','PM10_Fe':'0.00004','PM10_Hg':'0.057','PM10_K+':'0.029','PM10_Mn':'0.00005','PM10_Ni':'0.002','PM10_Pb':'0.0001','PM10_Se':'0.005','PM10_V':'0.00009','PM10_Zn':'0.0006'}}},                                                                            
            '306':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Wedding and Associates PM'},                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_As':'0.005','PM10_Ca+':'0.002','PM10_Cd':'0.016','PM10_Cl-':'0.002','PM10_Co':'0.00003','PM10_Cr':'0.002','PM10_Cu':'0.0001','PM10_Fe':'0.00004','PM10_Hg':'0.057','PM10_K+':'0.029','PM10_Mn':'0.00005','PM10_Ni':'0.002','PM10_Pb':'0.0001','PM10_Se':'0.005','PM10_V':'0.00009','PM10_Zn':'0.0006'}}},                                                                                                                                        
            '501':          {'primary_sampling':{'sampling_type':'high volume continuous', 'instrument':'Tisch Environmental 6000'},                                         'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.0344','PM10_Cd':'0.00001','PM10_Co':'0.00001','PM10_Cr':'0.344','PM10_Hg':'0.00002','PM10_Mn':'0.00002','PM10_Ni':'0.00011','PM10_Pb':'0.00011','PM10_Se':'0.00002'}}}, 
            '800':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module A'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.002','PM2.5_As':'0.002','PM2.5_Ca+':'0.002','PM2.5_Cl-':'0.002','PM2.5_Cr':'0.002','PM2.5_Cu':'0.002','PM2.5_Fe':'0.002','PM2.5_K+':'0.002','PM2.5_Mg++':'0.002','PM2.5_Mn':'0.002','PM2.5_Na+':'0.002','PM2.5_Ni':'0.0002','PM2.5_Pb':'0.002','PM2.5_Se':'0.002','PM2.5_V':'0.002','PM2.5_Zn':'0.002'}}},                                                                                                                            
            '801':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module A'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.002','PM2.5_As':'0.002','PM2.5_Ca+':'0.002','PM2.5_Cl-':'0.002','PM2.5_Cr':'0.002','PM2.5_Cu':'0.002','PM2.5_Fe':'0.002','PM2.5_K+':'0.002','PM2.5_Mg++':'0.002','PM2.5_Mn':'0.002','PM2.5_Na+':'0.002','PM2.5_Ni':'0.0002','PM2.5_Pb':'0.002','PM2.5_Se':'0.002','PM2.5_V':'0.002','PM2.5_Zn':'0.002'}}},                                                                                                                       
            '802':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module A'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'particle induced X-ray emission',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.002','PM2.5_As':'0.002','PM2.5_Ca+':'0.002','PM2.5_Cl-':'0.002','PM2.5_Cr':'0.002','PM2.5_Cu':'0.002','PM2.5_Fe':'0.002','PM2.5_K+':'0.002','PM2.5_Mg++':'0.002','PM2.5_Mn':'0.002','PM2.5_Na+':'0.002','PM2.5_Ni':'0.0002','PM2.5_Pb':'0.002','PM2.5_Se':'0.002','PM2.5_V':'0.002','PM2.5_Zn':'0.002'}}},                                                                                                               
            '803':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module B'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.002','PM2.5_NH4+':'0.002','PM2.5_NO3-':'0.002','PM2.5_SO4--':'0.002'}}},            
            '805':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module B'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.002','PM2.5_NH4NO3':'0.002','PM2.5_NO3-':'0.002','PM2.5_SO4--':'0.002'}}},                   
            '805_pb_alt':   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS10 200'},                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},                               
            '806':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module B'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.002','PM2.5_NO3-':'0.002','PM2.5_SO4--':'0.002'}}},            
            '806_pb_alt':   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS10 300'},                                              'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},                              
            '807':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'IMPROVE Module B'},                                                 'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.002','PM2.5_NO3-':'0.002','PM2.5_SO4--':'0.002'}}},                      
            '807_pb_alt':   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ100'},                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},                            
            '808':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},                            
            '809':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},              
            '811':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.01088','PM2.5_As':'0.00247','PM2.5_Ca+':'0.00347','PM2.5_Cd':'0.0105','PM2.5_Cl-':'0.00578','PM2.5_Co':'0.00141','PM2.5_Cr':'0.00159','PM2.5_Cu':'0.00135','PM2.5_Fe':'0.00196','PM2.5_Hg':'0.00437','PM2.5_K+':'0.00341','PM2.5_Mg++':'0.01841','PM2.5_Mn':'0.00231','PM2.5_Na+':'0.05107','PM2.5_Ni':'0.00125','PM2.5_Pb':'0.00549','PM2.5_Se':'0.00212','PM2.5_V':'0.0015','PM2.5_Zn':'0.00145'}}},
            '811_pb_alt':   {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Pb':'0.001'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        'reported_upper_limit_of_detection':{'PM10_Pb':'100.0'}}},                            
            '812':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.008','PM2.5_K+':'0.014','PM2.5_Na+':'0.03','PM2.5_NH4+':'0.017','PM2.5_NO3-':'0.008','PM2.5_SO4--':'0.012'}}},
            '813':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Met One SASS'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Na+':'0.03'}}},
            '817':          {'primary_sampling':{'sampling_type':'low volume continuous',                                               'primary_sampling_assumption':True}, 'preparation_1':{'sample_preparation_type':'filter',                                               'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry',              'preparation_assumption':True}, 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Al':'0.004','PM10_As':'0.002','PM10_Ca++':'0.003','PM10_Cl-':'0.005','PM10_Co':'0.006','PM10_Cr':'0.001','PM10_Cu':'0.001','PM10_Fe':'0.001','PM10_Hg':'0.002','PM10_K+':'0.003','PM10_Mn':'0.001','PM10_Ni':'0.001','PM10_Pb':'0.002','PM10_Se':'0.001','PM10_V':'0.004','PM10_Zn':'0.001'}}},             
            '820':          {                                                                                                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'X-ray fluorescence spectroscopy',                         'instrument':'Pall Xact 620',                  'reported_lower_limit_of_detection':{'PM10_As':'0.000027','PM10_Ca++':'0.000321','PM10_Cd':'0.001353','PM10_Co':'0.000047','PM10_Cr':'0.000109','PM10_Cu':'0.000064','PM10_Fe':'0.00008','PM10_Hg':'0.000043','PM10_K+':'0.000811','PM10_Mn':'0.000067','PM10_Ni':'0.000035','PM10_Pb':'0.053','PM10_Se':'0.000032','PM10_V':'0.000137','PM10_Zn':'0.000043'}}},
            '820_pm2.5_alt':{                                                                                                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'X-ray fluorescence spectroscopy',                         'instrument':'Pall Xact 625',                  'reported_lower_limit_of_detection':{'PM2.5_As':'0.114','PM2.5_Ca+':'0.902','PM2.5_Cd':'5.748','PM2.5_Co':'0.317','PM2.5_Cr':'0.288','PM2.5_Cu':'0.267','PM2.5_Fe':'0.759','PM2.5_Hg':'0.189','PM2.5_K+':'2.366','PM2.5_Mn':'0.283','PM2.5_Ni':'0.226','PM2.5_Pb':'0.218','PM2.5_Se':'0.141','PM2.5_V':'0.29','PM2.5_Zn':'0.231'}}},
            '821':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Teflon'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cr':'0.00063','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_Hg':'0.00175','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00085','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},
            '822':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Nylon'},                                           'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.013','PM2.5_Na+':'0.028','PM2.5_NH4+':'0.015','PM2.5_NO3-':'0.008','PM2.5_SO4--':'0.011'}}},
            '831':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 400'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_Hg':'0.00175','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00251','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},  
            '834':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 400'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.006','PM2.5_Na+':'0.012','PM2.5_NH4+':'0.007','PM2.5_SO4--':'0.005'}}},
            '835':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 400'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                'reported_lower_limit_of_detection':{'PM2.5_NO3-':'0.003'}}},          
            '841':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cr':'0.00063','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_Hg':'0.00175','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00085','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},  
            '844':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.012','PM2.5_Na+':'0.024','PM2.5_NH4+':'0.014','PM2.5_NO3-':'0.006','PM2.5_SO4--':'0.01'}}},
            '845':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.012','PM2.5_Na+':'0.024','PM2.5_NH4+':'0.014','PM2.5_SO4--':'0.01'}}},
            '846':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cr':'0.00063','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_Hg':'0.00175','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00085','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},    
            '849':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.012','PM2.5_Na+':'0.024','PM2.5_NH4+':'0.014','PM2.5_SO4--':'0.01'}}},
            '851':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cr':'0.00063','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_Hg':'0.00175','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00085','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},    
            '852':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.00949','PM2.5_Na+':'0.02044','PM2.5_NH4+':'0.01095','PM2.5_NO3-':'0.00584','PM2.5_SO4--':'0.00803'}}}, 
            '853':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'thermal-optical analysis - IMPROVE-A',                                                                   'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.002','PM2.5_NH4+':'0.002','PM2.5_NO3-':'0.002','PM2.5_SO4--':'0.002'}}},            
            '861':          {                                                                                                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal reduction - chemiluminescence',                   'instrument':'Rupprecht and Patashnick 8400N', 'reported_lower_limit_of_detection':{'PM2.5_NO3-':'0.2'}}},            
            '862':          {                                                                                                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal reduction - ultraviolet fluorescence',            'instrument':'Rupprecht and Patashnick 8400S', 'reported_lower_limit_of_detection':{'PM2.5_SO4--':'0.2'}}},            
            '875':          {                                                                                                                                                                                                                                                                                                                                                                                                                                   'measurement':{'methodology':'thermal reduction - ultraviolet fluorescence',            'instrument':'Thermo 5020i',                   'reported_lower_limit_of_detection':{'PM2.5_SO4--':'0.5'}}},
            '875_no3_alt':  {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'URG MASS 400'},                                                     'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                'reported_lower_limit_of_detection':{'PM2.5_NO3-':'0.003'}}},          
            '876':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2300'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_SO4--':'0.01'}}},
            '878':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Quartz'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_Cl-':'0.0625','PM2.5_SO4--':'0.0625'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              'reported_upper_limit_of_detection':{'PM2.5_Cl-':'50.0','PM2.5_SO4--':'50.0'}}},  
            '879':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Quartz'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'preparation_2':{'sample_preparation_type':'reagent reaction',                                                  'preparation_assumption':True}, 'measurement':{'methodology':'colorimetry',                                                                                            'reported_lower_limit_of_detection':{'PM2.5_NH4+':'0.0625'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'reported_upper_limit_of_detection':{'PM2.5_NH4+':'50.0'}}},                
            '880':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Andersen RAAS2.5 Quartz'},                                          'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'flame atomic absorption spectroscopy',                                                                   'reported_lower_limit_of_detection':{'PM2.5_K+':'0.03','PM2.5_Na+':'0.007'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    'reported_upper_limit_of_detection':{'PM2.5_K+':'50.0','PM2.5_Na+':'50.0'},                                                                                                                                                                                                              'process_details':'Only atomic absorption spectrometry stated as method, assume method is Flame Atomic Absorption Spectroscopy.', 'measuring_assumption':True}},  
            '881':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                'reported_lower_limit_of_detection':{'PM2.5_NO3-':'0.003'}}},                     
            '884':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM2.5_Al':'0.00436','PM2.5_As':'0.00099','PM2.5_Ca+':'0.00139','PM2.5_Cd':'0.00421','PM2.5_Cl-':'0.00232','PM2.5_Co':'0.00056','PM2.5_Cr':'0.00063','PM2.5_Cu':'0.00054','PM2.5_Fe':'0.00079','PM2.5_K+':'0.00137','PM2.5_Mg++':'0.00738','PM2.5_Mn':'0.00092','PM2.5_Na+':'0.0205','PM2.5_Ni':'0.0005','PM2.5_Pb':'0.0022','PM2.5_Se':'0.00085','PM2.5_V':'0.0006','PM2.5_Zn':'0.00058'}}},              
            '885':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM2.5_K+':'0.012','PM2.5_Na+':'0.024','PM2.5_NH4+':'0.014','PM2.5_SO4--':'0.01'}}},
            '886':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                'reported_lower_limit_of_detection':{'PM10_NO3-':'0.003'}}},                
            '889':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'X-ray fluorescence spectroscopy',                                                                        'reported_lower_limit_of_detection':{'PM10_Al':'0.00436','PM10_As':'0.00099','PM10_Ca++':'0.00139','PM10_Cd':'0.00421','PM10_Cl-':'0.00232','PM10_Co':'0.00056','PM10_Cr':'0.00063','PM10_Cu':'0.00054','PM10_Fe':'0.00079','PM10_K+':'0.00137','PM10_Mg++':'0.00738','PM10_Mn':'0.00092','PM10_Na+':'0.0205','PM10_Ni':'0.0005','PM10_Pb':'0.0022','PM10_Se':'0.00085','PM10_V':'0.0006','PM10_Zn':'0.00058'}}},                                                                                                             
            '890':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'ion chromatography',                                                                                     'reported_lower_limit_of_detection':{'PM10_K+':'0.012','PM10_Na+':'0.024','PM10_NH4+':'0.014','PM10_SO4--':'0.01'}}},
            '891':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'unknown',                                                                                                'reported_lower_limit_of_detection':{'PM2.5_NO3-':'0.003'}}},                     
            '901':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI frmOmni'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.0021','PM2.5_As':'0.0021','PM10_Cd':'0.0003','PM2.5_Cd':'0.0003','PM10_Co':'0.0007','PM2.5_Co':'0.0007','PM10_Cr':'0.0014','PM2.5_Cr':'0.0014','PM10_Hg':'0.0021','PM2.5_Hg':'0.0021','PM10_K+':'0.0139','PM2.5_K+':'0.0139','PM10_Mn':'0.0007','PM2.5_Mn':'0.0007','PM10_Na+':'0.0069','PM2.5_Na+':'0.0069','PM10_Ni':'0.0014','PM2.5_Ni':'0.0014','PM10_Pb':'0.0014','PM2.5_Pb':'0.0014','PM10_Se':'0.0014','PM2.5_Se':'0.0014'},                                                                                                                                                                                                           'reported_upper_limit_of_detection':{'PM2.5_As':'5000.0','PM2.5_Cd':'5000.0','PM2.5_Co':'5000.0','PM2.5_Cr':'5000.0','PM2.5_Hg':'5000.0','PM2.5_K+':'5000.0','PM2.5_Mn':'5000.0','PM2.5_Na+':'5000.0','PM2.5_Ni':'5000.0','PM10_Pb':'5000.0','PM2.5_Pb':'5000.0','PM2.5_Se':'5000.0'}}},                                                                                                             
            '902':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI frmOmni'},                                                      'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_K+':'0.0347','PM2.5_K+':'0.0347','PM10_Na+':'0.0694','PM2.5_Na+':'0.0694'},                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          'reported_upper_limit_of_detection':{'PM2.5_K+':'5000.0','PM2.5_Na+':'5000.0'}}},                                                                                                              
            '903':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_Al':'0.064','PM10_As':'0.0006','PM2.5_As':'0.0006','PM10_Ca++':'0.0908','PM10_Cd':'0.0001','PM2.5_Cd':'0.0001','PM10_Cl-':'0.0002','PM10_Co':'0.0002','PM2.5_Co':'0.0002','PM10_Cr':'0.0004','PM2.5_Cr':'0.0004','PM10_Cu':'0.00072','PM10_Hg':'0.0006','PM2.5_Hg':'0.0006','PM10_K+':'0.0042','PM2.5_K+':'0.0042','PM10_Mg++':'0.00932','PM10_Mn':'0.0002','PM2.5_Mn':'0.0002','PM10_Na+':'0.0021','PM2.5_Na+':'0.0021','PM10_Ni':'0.0004','PM2.5_Ni':'0.0004','PM10_Pb':'0.0004','PM2.5_Pb':'0.0004','PM10_Se':'0.0004','PM2.5_Se':'0.0004','PM10_Zn':'0.00491'}}},                                                                                                             
            '904':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'BGI PQ200'},                                                        'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_K+':'0.0104','PM2.5_K+':'0.0104','PM10_Na+':'0.0208','PM2.5_Na+':'0.0208'}}},                                                                                                                     
            '905':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM2.5_Al':'0.0006','PM10_As':'0.0006','PM2.5_As':'0.0006','PM10_Cd':'0.0001','PM2.5_Cd':'0.0001','PM10_Co':'0.0002','PM2.5_Co':'0.0002','PM10_Cr':'0.0004','PM2.5_Cr':'0.0004','PM10_Hg':'0.0006','PM2.5_Hg':'0.0006','PM10_K+':'0.0042','PM2.5_K+':'0.0042','PM10_Mn':'0.0002','PM2.5_Mn':'0.0002','PM10_Na+':'0.0021','PM2.5_Na+':'0.0021','PM10_Ni':'0.0004','PM2.5_Ni':'0.0004','PM10_Pb':'0.0004','PM2.5_Pb':'0.0004','PM10_Se':'0.0004','PM2.5_Se':'0.0004'}}},                                                                                                                         
            '906':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2000'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy',                                                'reported_lower_limit_of_detection':{'PM10_K+':'0.0104','PM2.5_K+':'0.0104','PM10_Na+':'0.0208','PM2.5_Na+':'0.0208'}}},                                                                                                             
            '907':          {'primary_sampling':{'sampling_type':'low volume continuous',  'instrument':'Thermo Partisol 2025'},                                             'preparation_1':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True},                                                                                                                                                 'measurement':{'methodology':'inductively coupled plasma mass spectrometry',                                                           'reported_lower_limit_of_detection':{'PM10_As':'0.00052','PM10_Cd':'0.00026','PM10_Co':'0.00026','PM10_Cr':'0.00026','PM10_Cr':'0.0004','PM10_Hg':'0.00026','PM10_Mn':'0.00026','PM10_Ni':'0.00026','PM10_Pb':'0.0008','PM10_Se':'0.00026','PM10_V':'0.00026'}}}}                                                                                                                         
                  
#--------------------------------------------------------------
#US EPA CASTNET
#--------------------------------------------------------------
    elif run_details['network'] == 'US_EPA_CASTNET':
        if raw_measurement_info == 'o3':
            if run_details['parameter'] == 'O3':
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet photometry'}}
        #-------------------------------------------
        if raw_measurement_info == 'hourly_gas':
            if run_details['parameter'] == 'NO':
                standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter)', 'process_details':'Assume method is chemiluminescence (internal molybdenum converter)', 'measuring_assumption':True}}
            elif run_details['parameter'] == 'NO2':
                standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal photolytic converter)'}}
            elif run_details['parameter'] == 'NH3':
                standard_measurement_information = {'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'process_details':'Assume method is chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'measuring_assumption':True}}
            elif run_details['parameter'] == 'HNO3':
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'denuder'}, 'measurement':{'methodology':'chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'process_details':'Assume method is chemiluminescence (internal molybdenum converter and external stainless steel converter)', 'measuring_assumption':True}}
            elif run_details['parameter'] == 'CO':
                standard_measurement_information = {'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)'}}
            elif run_details['parameter'] == 'SO2':
                standard_measurement_information = {'measurement':{'methodology':'ultraviolet fluorescence'}}
            elif run_details['parameter'] == 'TEMP_2M':
                standard_measurement_information = {'measurement':{'methodology':'unknown'}}
        #-------------------------------------------
        elif raw_measurement_info == 'drychem':
            if run_details['parameter'] in ['SO2','HNO3','Cl-','NO3-','SO4--']:
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'}, 'measurement':{'methodology':'ion chromatography'}}   
            elif run_details['parameter'] == 'NH4+':
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'}, 'measurement':{'methodology':'continuous flow analysis', 'instrument':'Bran+Luebbe Autoanalyzer 3'}} 
            elif run_details['parameter'] in ['Ca++','K+','Mg++','Na+']:
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'3-Stage Filter Pack'}, 'measurement':{'methodology':'inductively coupled plasma atomic emission spectroscopy', 'instrument':'Perkin Elmer Optima 7300'}} 
        #-------------------------------------------
        elif raw_measurement_info == 'vischem':
            if run_details['parameter'] == 'PM2.5':
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'}, 'measurement':{'methodology':'gravimetry', 'process_details':'Assume method is gravimetry', 'measuring_assumption':True}}
            elif run_details['parameter'] in['PM2.5_Al','PM2.5_As','PM2.5_Ca++','PM2.5_Cd','PM2.5_Cl-','PM2.5_Co','PM2.5_Cr','PM2.5_Cu','PM2.5_Fe','PM2.5_Hg','PM2.5_K+','PM2.5_Mg++','PM2.5_Mn','PM2.5_Na+','PM2.5_Ni','PM2.5_NO3-','PM2.5_Pb','PM2.5_Se','PM2.5_SO4--','PM2.5_V','PM2.5_Zn']:
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'ion chromatography', 'process_details':'Assume method is ion chromatography', 'measuring_assumption':True}} 
            elif run_details['parameter'] in ['PM2.5_C','PM2.5_EC','PM2.5_OC']:
                standard_measurement_information = {'primary_sampling':{'sampling_type':'low volume continuous'}, 'preparation_1':{'sample_preparation_type':'filter pack', 'specific_technique':'1-Stage Filter Pack'}, 'preparation_2':{'sample_preparation_type':'intermediate measurement', 'methodology':'gravimetry', 'preparation_assumption':True}, 'measurement':{'methodology':'thermal-optical analysis - unknown protocol', 'process_details':'Assume method is thermal-optical analysis', 'measuring_assumption':True}}

#--------------------------------------------------------------
#US NADP AMNet
#--------------------------------------------------------------
    elif run_details['network'] == 'US_NADP_AMNet':
        if run_details['parameter'] == 'PM2.5_Hg':
            standard_measurement_information = {'measurement':{'methodology':'cold vapour atomic fluorescence spectroscopy', 'instrument':'Tekran 2537A-1130-1135'}}
#--------------------------------------------------------------
#US NADP AMoN
#--------------------------------------------------------------
    elif run_details['network'] == 'US_NADP_AMoN':
        if run_details['parameter'] == 'NH3':
            standard_measurement_information = {'primary_sampling':{'sampling_type':'passive'}, 'preparation_1':{'sample_preparation_type':'sorbent trapping', 'specific_technique':'Diffusive Sampler'}, 'measurement':{'methodology':'continuous flow analysis'}}

#--------------------------------------------------------------
#WMO_WDCGG
#--------------------------------------------------------------
    elif run_details['network'] == 'WMO_WDCGG':
        if run_details['parameter'] == 'CO':
            mapping = {                                                                                                                                                                                                            
            'aero_laser:al5001_vurf':                                                                                                                                                                                     {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5001'}},   
            'aero_laser_al_5001_vurf':                                                                                                                                                                                    {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5001'}},   
            'aerolaser_5001_vurf':                                                                                                                                                                                        {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5001'}},                        
            'agilent_gc6890_gas_chromatography_(fid)':                                                                                                                                                                    {                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                        'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},         
            'agilent_model_5890_gas_chromatography_(fid)':                                                                                                                                                                {                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                        'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'agilent_model_6890n_gas_chromatography_(fid)':                                                                                                                                                               {                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                        'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},      
            'al_5001_#142_;_aero_laser_gmbh,_germany._vurf':                                                                                                                                                              {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5001'}},   
            'apma_360;_serial_nr.:205003,_id_7062;_horiba_ndir':                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}}, 
            'apma_360_(horiba_ltd.)_ndir':                                                                                                                                                                                {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}}, 
            'apma_370_(horiba_ltd.)_ndir':                                                                                                                                                                                {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA370'}}, 
            'co2_ch4_(co)_h2o_picarro_analyzer_g2301_(g2401)_crds':                                                                                                                                                       {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401',                                              'process_details':'Assume instrument is Picarro G2401, as Picarro G2301 is stated also',            'measuring_assumption':True}},
            'co2_ch4_co_h2o_picarro_analyzer_g2401_crds':                                                                                                                                                                 {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'co_n2o_h2o_lgr_analyzer_(30-ep,_913-0015-1000)_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                    {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},
            'co_n2o_losgatos_analyzer_model:913-005_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                            {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},    
            'flask_fast_carbon_monoxide_analyzer_al5002_aero-laser_gmbh_vurf':                                                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5002',   'sampling_type':'injection'}},
            'flask_hewlett_packard_5890a_gas_chromatography_(cvafs)':                                                                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - cold vapour atomic fluorescence spectroscopy'}},
            'flask_hewlett_packard_5890a_gas_chromatography_(rgd)':                                                                                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection'}},
            'flask_shimadzu_gc-9a_gas_chromatography_(fid)':                                                                                                                                                              {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'flask_trace_analytical_gas_chromatograph_gas_chromatography_(rgd)':                                                                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection'}},
            'flask_unknown':                                                                                                                                                                                              {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                                         'process_details':'Assume method is GC- unknown detection',                                         'measuring_assumption':True}},
            'flask_unknown_gas_chromatography_(rgd)':                                                                                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection'}},
            'flask_unknown_tildas':                                                                                                                                                                                       {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'tunable diode laser absorption spectroscopy',                                                            'sampling_type':'injection'}},      
            'flask_unknown_vurf':                                                                                                                                                                                         {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'}, 'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                                                              'sampling_type':'injection'}},           
            'g2401_co2_ch4_co_h2o_analyzer_from_picarro,_inc._crds':                                                                                                                                                      {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'gc-hgo_reduction_uv_absorption;instrument_type:online_gc;instrument_manufacturer:aglient_peaklaboratories;instrument_model:gc-uecd_6890a_and_pp1;instrument_name:agilent_gc-uecd_6890_pp1_md_tac_gas_chr':   {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},    
            'gc-md_unknown':                                                                                                                                                                                              {                                                                                                                    'measurement':{'methodology':'gas chromatography - unknown detection',                                                                 'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},   
            'gc-rga3,_trace_analytical_gas_chromatography_(rgd)':                                                                                                                                                         {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},
            'horiba,_ltd._ga-360s_ndir':                                                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba GA-360S'}},
            'horiba_apma-360_ce#303003_ndir':                                                                                                                                                                             {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}}, 
            'horiba_apma360,_s_n_890617034_ndir':                                                                                                                                                                         {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}}, 
            'horiba_apma_360ce_ndir':                                                                                                                                                                                     {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}},  
            'horiba_ltd._apma_350e_ndir':                                                                                                                                                                                 {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA350'}},  
            'horiba_ltd._apma_360_ndir':                                                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}},        
            'horiba_ltd._apma_370_ndir':                                                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA370'}},  
            'horiba_model_apma-360._(#712020)_ndir':                                                                                                                                                                      {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',          'instrument':'Horiba APMA360'}}, 
            'lgr_co_n2o_h2o_analyzer_913-0015_ep_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                               {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},        
            'lgr_n2o_co_913-0014-1000_crds':                                                                                                                                                                              {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},                         
            'lgr_n2o_analyser_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},     
            'los_gatos,_ep-30_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},
            'losgatos_501_ep_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                                   {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},
            'losgatosresearch_qcl_n2o_co_30r_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                   {                                                                                                                    'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',                      'instrument':'Los Gatos N2O/CO'}},
            'manufacturer_of_gc_system:trace_analytical;model_and_s_n:rga3,_rga3-101590-007;detector_type(s)_and_manufacturer:gc-hgo_reduction__uv_absorption;temperature:268\xb0_c;carrier_gas:n2_gc50._manufacturer:ge':{                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},          
            'model_48;gfc_ambient_co_analyzer;thermo_environmental_instruments_inc;n_serial_#48-47169-278_ir-gfc':                                                                                                        {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48'}},                                                 
            'pcarro_co_co2_ch4_h2o_g2401_crds':                                                                                                                                                                           {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},     
            'picarro_co_co2_ch4_h2o_g2401_crds':                                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'picarro_co_co2_h2o_g2302_crds':                                                                                                                                                                              {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2302'}},
            'picarro_crds':                                                                                                                                                                                               {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy'}}, 
            'picarro_g2302_crds':                                                                                                                                                                                         {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2302'}},
            'picarro_g2401;#2352-cfkads2196_crds':                                                                                                                                                                        {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'picarro_g2401_(s_n:cfkads2152-20141128)_crds':                                                                                                                                                               {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'picarro_g2401_crds':                                                                                                                                                                                         {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'piccaro,_g2401_crds':                                                                                                                                                                                        {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},      
            'picarro_inc.,_g2401,_s_n_cfkads2028_crds':                                                                                                                                                                   {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'picarro_inc.,_g2401,_s_n_cfkads2031_crds':                                                                                                                                                                   {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},
            'picarro_inc.,_g2401,_s_n_cfkads2158_crds':                                                                                                                                                                   {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},        
            'picarro,_inc.,_g2401_(sn:cfkads2320)_crd':                                                                                                                                                                   {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},       
            'picarro,_inc.,_g2401_(sn:cfkads2320)_crds':                                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy',                                        'instrument':'Picarro G2401'}},       
            'rga3_co_detector_gas_chromatography_(rgd)':                                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},                                                                                                                              
            'rga3_trace_analytical_gas_chromatography_(rgd)':                                                                                                                                                             {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},                                                                                                                              
            'rgd2-trace_analytical_analyzer_gas_chromatography_(rgd)':                                                                                                                                                    {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},                                                                                                                              
            'round_science_inc._tra-1_gas_chromatography_(rgd)':                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'gas chromatography - mercuric oxide reduction detection',                                                'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},                                                                                                                              
            'te_48s_#48s-50873-286_thermo_environmental_instruments_inc._ndir':                                                                                                                                           {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48s'}}, 
            'teco;teco_48s_ndir':                                                                                                                                                                                         {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48s'}}, 
            'tei48c_tl_ndir':                                                                                                                                                                                             {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48C-TL'}}, 
            'tei_48_gfc_ir-gfc':                                                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48'}},                                                 
            'tei_48c_tl_(sn:66838-352)_ndir':                                                                                                                                                                             {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48C-TL'}}, 
            'thermo_48i_ndir':                                                                                                                                                                                            {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48i'}}, 
            'thermo_electron_instruments_48_w_light_absorption_analysis_(ir,_except_ndir)':                                                                                                                               {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48W'}}, 
            'thermo_environmental_instruments_inc._aerolaser_5001_vurf':                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5001'}},
            'thermo_environmental_instruments_inc._aerolaser_5002_vurf':                                                                                                                                                  {                                                                                                                    'measurement':{'methodology':'vacuum ultraviolet resonance fluorescence',                           'instrument':'Aerolaser AL5002'}},
            'thermo_environmental_instruments_inc._te_48s_ndir':                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48s'}}, 
            'thermo_scientific_tei_48c-tl_ndir':                                                                                                                                                                          {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',         'instrument':'Thermo 48C-TL'}}, 
            'university_of_wollongong_(australia)_in-situ_mid-infrared_fourier_transform_spectrometer_mid-ftir':                                                                                                          {                                                                                                                    'measurement':{'methodology':'fourier transform infrared spectroscopy',                             'instrument':'Ecotech Spectronus'}},                                                                                                                
            'unknown_crds':                                                                                                                                                                                               {                                                                                                                    'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'unknown_gas_chromatography_(fid)':                                                                                                                                                                           {                                                                                                                    'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                        'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},
            'unknown_light_absorption_analysis_(ir,_except_ndir)':                                                                                                                                                        {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation', 'measuring_assumption':True}},
            'unknown_ndir':                                                                                                                                                                                               {                                                                                                                    'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                    'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation', 'measuring_assumption':True}},
            'unknown':                                                                                                                                                                                                    {                                                                                                                    'measurement':{'methodology':'unknown'}},
            'unknown_unknown':                                                                                                                                                                                            {                                                                                                                    'measurement':{'methodology':'unknown'}}}
#-------------------------------------------
        elif run_details['parameter'] == 'CH4':
            mapping = {
            'agilent_6890n_gas_chromatography_(fid)':                                                                                                                                                                            {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'agilent_gc6890_gas_chromatography_(fid)':                                                                                                                                                                           {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},             
            'agilent_model_5890_6890n_gas_chromatography_(fid)':                                                                                                                                                                 {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},               
            'agilent_model_5890_gas_chromatography_(fid)':                                                                                                                                                                       {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},              
            'agilent_model_6890n_gas_chromatography_(fid)':                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'aglient_6890_gc_gas_chromatography_(fid)':                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'carlo_erba_gas_chromatography_(fid)':                                                                                                                                                                               {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'chromo-pack_cp_9000_gas_chromatography_(fid)':                                                                                                                                                                      {                                                               'preparation_1':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'chrompack_cp_9000_gas_chromatography_(fid)':                                                                                                                                                                        {                                                               'preparation_1':{'sample_preparation_type':'preconcentration'}, 'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},                 
            'co2_ch4_(co)_h2o_picarro_analyzer_g2301_(g2401)_crds':                                                                                                                                                              {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401',                                                       'process_details':'Assume instrument is Picarro G2401, as Picarro G2301 is stated also',            'measuring_assumption':True}},
            'co2_ch4_co_h2o_picarro_analyzer_g2401_crds':                                                                                                                                                                        {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},
            'co2_ch4_h2o_picarro_analyzer_g2301_crds':                                                                                                                                                                           {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},         
            'co2_ch4_h2o_picarro_analyzer_g230_crds':                                                                                                                                                                            {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301',                                                       'process_details':'Assume instrument is Picarro G2301 as only Picarro G230 is stated',              'measuring_assumption':True}},                  
            'dlt100_(los_gatos_research)_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                              {                                                                                                                               'measurement':{'methodology':'off-axis integrated cavity output spectroscopy'}},      
            'flask_agilent_model_6890n_gas_chromatography_(fid)':                                                                                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},          
            'flask_carle_gas_chromatograph_gas_chromatography_(fid)':                                                                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},         
            'flask_gas_chromatograph_hp_5890_series_2_gas_chromatography_(fid)':                                                                                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},         
            'flask_gas_chromatograph_hp_6890_gas_chromatography_(fid)':                                                                                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},         
            'flask_hewlett_packard;5890_series_ii_gas_chromatography_(fid)':                                                                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},           
            'flask_hewlett_packard_hp5890_series_ii_gas_chromatography_(fid)':                                                                                                                                                   {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},        
            'flask_picarro_g2401_crds':                                                                                                                                                                                          {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401',               'sampling_type':'injection'}},
            'flask_shimadzu_gc-8a_gas_chromatography_(fid)':                                                                                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},   
            'flask_tzvet-500m,_dzerjinsk,_russia_no48_gas_chromatography_(fid)':                                                                                                                                                 {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},             
            'flask_unknown':                                                                                                                                                                                                     {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - unknown detection',                                                                                                           'process_details':'Assume method is GC- unknown detection',                                         'measuring_assumption':True}},
            'flask_unknown_crds':                                                                                                                                                                                                {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'cavity ringdown spectroscopy',                                                                             'sampling_type':'injection'}},        
            'flask_unknown_gas_chromatography_(fid)':                                                                                                                                                                            {'primary_sampling':{'sampling_type':'high volume continuous'}, 'preparation_1':{'sample_preparation_type':'flask'},            'measurement':{'methodology':'gas chromatography - flame ionisation detection'}},
            'g1301_(picarro)_crds':                                                                                                                                                                                              {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G1301'}},                 
            'g1301_co2_ch4_h2o_analyzer_from_picarro,_inc._crds':                                                                                                                                                                {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G1301'}},                
            'g2301_picarro,_inc._crds':                                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},         
            'g2401_(picarro)_crds':                                                                                                                                                                                              {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},          
            'g2401_co2_ch4_co_h2o_analyzer_from_picarro,_inc._crds':                                                                                                                                                             {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},      
            'gc-3800,_dani_gas_chromatography_(fid)':                                                                                                                                                                            {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'gc-3800,_varian_gas_chromatography_(fid)':                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'gc-md_gas_chromatography_(fid)':                                                                                                                                                                                    {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'gc-md_unknown':                                                                                                                                                                                                     {                                                                                                                               'measurement':{'methodology':'gas chromatography - unknown detection',                                                                   'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},    
            'gc_3800,_varian;gc:gc_3800;detector:fid;detector_temp.:250_celsius;oven_temp.:65_celsius_(15_min);flow_rate:60_ml_min;column_type:porapak_q_80_100_mesh_(1_m\u00D7_1_8"\u00D7_2_mm_ss);injection_volume:2_ml;valve':{                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'hewlett_packard;5890_ii_gas_chromatography_(fid)':                                                                                                                                                                  {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'hewlett_packard_modified_hp_6890_chem._station,_#_us00030144_gas_chromatography_(fid)':                                                                                                                             {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},              
            'hewlett_packard_modified_hp_6890_gc_020394_gas_chromatography_(fid)':                                                                                                                                               {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'horiba,_ltd._ga-360s_ndir':                                                                                                                                                                                         {                                                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (cross-flow modulation)',   'instrument':'Horiba GA-360S'}},   
            'hp_5890_gc_gas_chromatography_(fid)':                                                                                                                                                                               {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'infrared_spectrophotometer_based_on_mdr-41_monochromator_manufactured_by_lomo,_st-petersburg,_russia_light_absorption_analysis_(ir,_except_ndir)':                                                                  {                                                                                                                               'measurement':{'methodology':'non-dispersive infrared absorption (gas-filter correlation)',                                                                                      'process_details':'Only NDIR stated as method, make assumption method uses Gas-Filter Correlation', 'measuring_assumption':True}},
            'instrument_type:tracegas_monitor;instrument_manufacturer:picarro;instrument_model:g2301;instrument_name:picarro_crds_g2301_rgl_crds':                                                                               {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},         
            'instrument_type:tracegas_monitor;instrument_manufacturer:picarro;instrument_model:g2301;instrument_name:picarro_crds_g2301_tac_crds':                                                                               {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},                   
            'lgr_gga_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                                                  {                                                                                                                               'measurement':{'methodology':'off-axis integrated cavity output spectroscopy',               'instrument':'Los Gatos GGA'}},   
            'met._cons._gc_sichromat_1-4_gas_chromatography_(fid)':                                                                                                                                                              {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},               
            'nira_venus_301_gas_chromatography_(fid)':                                                                                                                                                                           {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',              'instrument':'Nira Venus 301'}},               
            'picarro,_2301_crds':                                                                                                                                                                                                {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},         
            'picarro_crds':                                                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy'}}, 
            'picarro_envirosense_3000i_crds':                                                                                                                                                                                    {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro Envirosense 3000i'}},                             
            'picarro_g1301_crds':                                                                                                                                                                                                {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G1301'}},                    
            'picarro_g2301_analyzer_for_co2,_ch4_and_h2o_(serial_number_2579-cfads2443)_crds':                                                                                                                                   {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},             
            'picarro_g2301_crds':                                                                                                                                                                                                {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2301'}},         
            'picarro_g2302_cfads2132_crds':                                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2132-i',                                                     'process_details':'Assume instrument is Picarro G2132-i, as Picarro G2302 is stated also',          'measuring_assumption':True}},
            'picarro_g2401;#2352-cfkads2196_crds':                                                                                                                                                                               {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},      
            'picarro_g2401_(s_n:cfkads2152-20141128)_crds':                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},    
            'picarro_g2401_cfads2209_crds':                                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},   
            'picarro_g2401_crds':                                                                                                                                                                                                {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},
            'picarro_inc.,_g2401,_s_n_cfkads2028_crds':                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},
            'picarro_inc.,_g2401,_s_n_cfkads2031_crds':                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},
            'picarro_inc.,_g2401,_s_n_cfkads2158_crds':                                                                                                                                                                          {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},
            'picarro,_inc.,_g2401_(sn:cfkads2320)_crds':                                                                                                                                                                         {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy',                                 'instrument':'Picarro G2401'}},    
            'round_science_inc._rgc-1_gas_chromatography_(fid)':                                                                                                                                                                 {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'shimadzu;gc-14a_equipped_with_an_fid_and_a_data_processor_shimadzu_c-r4a_gas_chromatography_(fid)':                                                                                                                 {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'shimadzu_gc-8a_gas_chromatography_(fid)':                                                                                                                                                                           {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'siemens-meteo_consult_met._cons._gc_sichromat_1-4_gas_chromatography_(fid)':                                                                                                                                        {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},           
            'university_of_wollongong_(australia)_in-situ_mid-infrared_fourier_transform_spectrometer_mid-ftir':                                                                                                                 {                                                                                                                               'measurement':{'methodology':'fourier transform infrared spectroscopy',                      'instrument':'Ecotech Spectronus'}},     
            'unknown_crds':                                                                                                                                                                                                      {                                                                                                                               'measurement':{'methodology':'cavity ringdown spectroscopy'}},
            'unknown_gas_chromatography_(fid)':                                                                                                                                                                                  {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}},  
            'unknown_off-axis_integrated_cavity_output_spectroscopy_(oa-icos)':                                                                                                                                                  {                                                                                                                               'measurement':{'methodology':'off-axis integrated cavity output spectroscopy'}},
            'unknown':                                                                                                                                                                                                           {                                                                                                                               'measurement':{'methodology':'unknown'}},   
            'unknown_unknown':                                                                                                                                                                                                   {                                                                                                                               'measurement':{'methodology':'unknown'}},
            'varian_5890_gas_chromatography_(fid)':                                                                                                                                                                              {                                                                                                                               'measurement':{'methodology':'gas chromatography - flame ionisation detection',                                                          'sampling_type':'continuous injection', 'process_details':'Assume continuous injection',                                                    'measuring_assumption':True}}}           
#--------------------------------------------------------------
    
    #need to map raw measurement information to standard information? 
    if 'mapping' in locals():

        #check if have mapping information for raw network provided information
        if raw_measurement_info not in list(mapping.keys()):
            #if do not, return 'non-mapped' string
            return 'non-mapped'
        
        #map raw measurement information to standard information
        standard_measurement_information = mapping[raw_measurement_info]

    return standard_measurement_information

###--------------------------------------------------------------------------------------------------###
#DEFINE DICTIONARIES WHICH MAP MEASUREMENT STAGE SPECIFIC MEASUREMENT INFORMATION TO STANDARD METADATA VARIABLES 
###--------------------------------------------------------------------------------------------------###

#syntax --> measurement stage specific variable name : standard metadata name

primary_sampling_metadata_mapping = {'sampling_type':             'primary_sampling_type',
                                     'instrument':                'primary_sampling_instrument_name',
                                     'documented_flow_rate':      'primary_sampling_instrument_documented_flow_rate',
                                     'reported_flow_rate':        'primary_sampling_instrument_reported_flow_rate',
                                     'process_details':           'primary_sampling_process_details',
                                     'instrument_manual_name':    'primary_sampling_instrument_manual_name',
                                     'instrument_further_details':'primary_sampling_further_details'
                                    }

sample_preparation_metadata_mapping = {'sample_preparation_types':          'sample_preparation_types',
                                       'sample_preparation_techniques':     'sample_preparation_techniques',
                                       'sample_preparation_process_details':'sample_preparation_process_details',
                                       'sample_preparation_further_details':'sample_preparation_further_details'
                                      }

measurement_metadata_mapping = {'methodology':                        'measurement_methodology',
                                'instrument':                         'measuring_instrument_name',
                                'sampling_type':                      'measuring_instrument_sampling_type',
                                'documented_flow_rate':               'measuring_instrument_documented_flow_rate',
                                'reported_flow_rate':                 'measuring_instrument_reported_flow_rate',
                                'process_details':                    'measuring_instrument_process_details',
                                'instrument_manual_name':             'measuring_instrument_manual_name',
                                'instrument_further_details':         'measuring_instrument_further_details',
                                'documented_lower_limit_of_detection':'measuring_instrument_documented_lower_limit_of_detection',
                                'reported_lower_limit_of_detection':  'measuring_instrument_reported_lower_limit_of_detection',
                                'documented_upper_limit_of_detection':'measuring_instrument_documented_upper_limit_of_detection',
                                'reported_upper_limit_of_detection':  'measuring_instrument_reported_upper_limit_of_detection',
                                'documented_uncertainty':             'measuring_instrument_documented_uncertainty',
                                'reported_uncertainty':               'measuring_instrument_reported_uncertainty',
                                'documented_accuracy':                'measuring_instrument_documented_accuracy',
                                'reported_accuracy':                  'measuring_instrument_reported_accuracy',
                                'documented_precision':               'measuring_instrument_documented_precision',
                                'reported_precision':                 'measuring_instrument_reported_precision',
                                'documented_zero_drift':              'measuring_instrument_documented_zero_drift',
                                'reported_zero_drift':                'measuring_instrument_reported_zero_drift',
                                'documented_span_drift':              'measuring_instrument_documented_span_drift',
                                'reported_span_drift':                'measuring_instrument_reported_span_drift', 
                                'documented_zonal_drift':             'measuring_instrument_documented_zonal_drift',
                                'reported_zonal_drift':               'measuring_instrument_reported_zonal_drift',
                                'documented_resolution':              'measuring_instrument_documented_measurement_resolution',
                                'reported_resolution':                'measuring_instrument_reported_measurement_resolution',
                                'documented_absorption_cross_section':'measuring_instrument_documented_absorption_cross_section',
                                'reported_absorption_cross_section':  'measuring_instrument_reported_absorption_cross_section'
                               }

###--------------------------------------------------------------------------------------------------###
#DEFINE FUNCTION TO MAP NETWORK PROVIDED MEASUREMENT INFORMATION TO STANDARD MEASUREMENT INFORMATION 
###--------------------------------------------------------------------------------------------------###

def map_network_measurement_information_to_standard_measurement_information(raw_provided_measurement_information, site_metadata, run_details, parameter_details, write_to_metadata_array=True):
    
    '''Function which maps network provided measurement information to standard measurement information.
       This takes information defined in standardised mapping dictionaries to accomplish this.
    '''

    import copy
    from GHOST_standards import standard_sampling_types, standard_sampling_preparation_types, standard_measurement_methodologies
    import glob
    import numpy as np
    import os
    import pandas as pd
    import unit_converter

    #get list of af all currently unmapped raw network provided measurement information
    unmapped_provided_measurement_information = sorted(glob.glob(run_details['process_root']+'/run/unmapped_provided_measurement_information/*'))
    unmapped_provided_measurement_information = [info.split('unmapped_provided_measurement_information/')[-1] for info in unmapped_provided_measurement_information]
        
    #make sure raw_measurement_info is lower-case, and replace empty single spaces and forward slashes with underscores
    raw_provided_measurement_information = raw_provided_measurement_information.replace(" ", "_").replace("/", "_").lower()
        
    #get standard measurement information specific for network/parameter being processed 
    #on some occasions the measurements are standard by network therefore the information is pre-standardised
    #on other occasions the networks provide non-standard measurement information that need mapping to standard information
    if len(site_metadata['meta_update_stamp']['value']) > 0:
        standard_provided_measurement_information = network_measurement_information_to_standard_measurement_information_mapping(raw_provided_measurement_information, run_details, parameter_details, metadata_stamp_dt=site_metadata['meta_update_stamp']['value'][-1])
    else:
        standard_provided_measurement_information = network_measurement_information_to_standard_measurement_information_mapping(raw_provided_measurement_information, run_details, parameter_details)

    #iterate through standard sampling types/sample preparation types/measurement methodologies dictionaries
    #taking information specific to the standardised reported measurement information, overwriting the stored dictionary information where necessary  

    #initialise dictionary for storing QA flags related to the measurement information (all flags set as False initially)
    measurement_qa_flags = {'Methodology Not Mapped': False, 
                            'Assumed Primary Sampling' : False,
                            'Assumed Sample Preparation': False,
                            'Assumed Measurement Methodology': False,
                            'Unknown Primary Sampling Type': False,
                            'Unknown Primary Sampling Instrument': False,
                            'Unknown Sample Preparation Type': False,
                            'Unknown Sample Preparation Technique': False,
                            'Unknown Measurement Method': False,
                            'Unknown Measuring Instrument': False,
                            'Erroneous Primary Sampling': False,
                            'Erroneous Sample Preparation': False,
                            'Erroneous Measurement Methodology': False,
                            'Invalid QA Measurement Methodology': False
                           }

    #initialise dictionary for storing sample preparation concatenated string information (can be multiple sample preparation information) 
    specific_sample_preparation_strings = {'sample_preparation_types' : '',
                                           'sample_preparation_techniques' : '',
                                           'sample_preparation_process_details' : '',
                                           'sample_preparation_further_details' : ''
                                          }

    #check if the raw measurement information has not been able to be mapped to standard information
    if standard_provided_measurement_information == 'non-mapped':

        #if not, set QA flag informing of this
        measurement_qa_flags['Methodology Not Mapped'] = True
        
        #check if have previously logged that the raw measurement information is unmapped
        if raw_provided_measurement_information not in unmapped_provided_measurement_information:
        
            #if not, create network:species:raw_measurement_info file for writing out a log that measurement information is unmapped 
            unmapped_provided_measurement_information_filename = run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information)
            
            f= open(unmapped_provided_measurement_information_filename,"w+")
            f.close() 
            #give 770 permissions to file
            os.chmod(unmapped_provided_measurement_information_filename, 0o770)

        if write_to_metadata_array:
            return site_metadata, measurement_qa_flags, 'UNK'
        else:
            return {}, measurement_qa_flags, 'UNK'

    #--------------------------------------------------------------------------------#
    #is there primary sampling information provided?
    specific_primary_sampling_information = {}
    if 'primary_sampling' in list(standard_provided_measurement_information.keys()):
        
        #get standard provided primary sampling information dictionary
        standard_provided_primary_sampling_information = copy.deepcopy(standard_provided_measurement_information['primary_sampling'])
        
        #get information specific to provided primary sampling type in standard sampling types dictionary     
        specific_primary_sampling_information = copy.deepcopy(standard_sampling_types[standard_provided_primary_sampling_information['sampling_type']])
        
        #update specific primary sampling information, updating by "overwrite" variables
        #get list of all variables containing "overwrite" string
        overwrite_vars = [var for var in list(specific_primary_sampling_information.keys()) if 'overwrite' in var]
        for overwrite_var in overwrite_vars:
            original_var = overwrite_var.split('overwrite_')[0]
            specific_primary_sampling_information[original_var] = specific_primary_sampling_information[overwrite_var]

        #if have defined primary sampling instrument name, get dictionary with information specific to that instrument  
        if 'instrument' in list(standard_provided_primary_sampling_information.keys()):
            specific_primary_sampling_instrument_information = standard_sampling_types[standard_provided_primary_sampling_information['sampling_type']]['instruments'][standard_provided_primary_sampling_information['instrument']]   
            #merge specific primary sampling information and specific primary sampling instrument information
            #handle matching variables by keeping instrument specific information by priority
            specific_primary_sampling_information = {**specific_primary_sampling_information, **specific_primary_sampling_instrument_information}

        #merge specific primary sampling information and standard provided primary sampling information
        #handle matching variables by keeping standard provided primary sampling information by priority
        specific_primary_sampling_information  = {**specific_primary_sampling_information, **standard_provided_primary_sampling_information} 
        #get list of variables in generated primary sampling dictionary
        specific_primary_sampling_information_vars = list(specific_primary_sampling_information.keys())        

        #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
        for var in specific_primary_sampling_information_vars:   
            if type(specific_primary_sampling_information[var]) == dict:
                if run_details['parameter'] in list(specific_primary_sampling_information[var].keys()):
                    specific_primary_sampling_information[var] = specific_primary_sampling_information[var][run_details['parameter']]
                else:
                    del specific_primary_sampling_information[var]
                    specific_primary_sampling_information_vars = list(specific_primary_sampling_information.keys()) 

        #if do not have primary sampling type defined (i.e. == unknown), set quality assurance flag informing of this
        if specific_primary_sampling_information['sampling_type'] == 'unknown':
            measurement_qa_flags['Unknown Primary Sampling Type'] = True     

        #if do not have primary sampling instrument defined, set quality assurance flag informing of this
        #this is False for "passive", "remote" and "manual" sampling types as there is no physically real sampling instrument in these cases
        if 'instrument' not in specific_primary_sampling_information_vars:
            if specific_primary_sampling_information['sampling_type'] not in ['manual', 'passive', 'remote']:
                measurement_qa_flags['Unknown Primary Sampling Instrument'] = True     

        #determine if current parameter processing is validly isolated by primary sampling type --> flag if not
        #assume all components with 'gas/'pm'/'precip'/'wetdep'/'meteo' matrices can be sampled by primary sampling instrument (i.e. only check isolates PM sizes)
        if ('isolated_parameters' in specific_primary_sampling_information_vars) & (specific_primary_sampling_information['sampling_type'] != 'unknown'):
            if (run_details['parameter'] not in specific_primary_sampling_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_primary_sampling_information['isolated_parameters']) & (parameter_details['matrix'] not in ['gas','pm','precip','wetdep','meteo']):
                measurement_qa_flags['Erroneous Primary Sampling'] = True

        #if "primary_sampling_assumption" flag is in specific sample preparation dictionary, update primary sampling assumption flag
        if 'primary_sampling_assumption' in specific_primary_sampling_information_vars:
                measurement_qa_flags['Assumed Primary Sampling'] = True

        #if have 'additional_sample_preparation_type' in specific primary sampling information dictionary, append to sampling_preparation_type variable (if exists)
        if 'additional_sample_preparation_type' in specific_primary_sampling_information_vars:
            if 'sample_preparation_type' in specific_primary_sampling_information_vars:
                specific_primary_sampling_information['sample_preparation_type'] = '{};{}'.format(specific_primary_sampling_information['sample_preparation_type'], specific_primary_sampling_information['additional_sample_preparation_type'])
            else:
                specific_primary_sampling_information['sample_preparation_type'] = specific_primary_sampling_information['additional_sample_preparation_type'] 

        #if have sample preparation information provided for the primary sampling instrument, process it
        if 'sample_preparation_type' in specific_primary_sampling_information_vars:
            #get sample preparation variable
            sample_preparation_type = specific_primary_sampling_information['sample_preparation_type']

            #gather all variable names associated with sample preparation in dictionary (i.e. with 'preparation' prefix)
            sample_preparation_vars = [var for var in specific_primary_sampling_information_vars if ('preparation' in var) & (var != 'sample_preparation_type')]
            #create dictionary of all all sample preparation variables names associated their values (dropping 'preparation' prefix')
            sample_preparation_dict = {}
            for var in sample_preparation_vars:   
                if var not in ['preparation_further_details', 'preparation_manual_name', 'preparation_assumption']:          
                    sample_preparation_dict[var.split('preparation_')[-1]] = specific_primary_sampling_information[var] 
                else:
                    sample_preparation_dict[var] = specific_primary_sampling_information[var] 
    
            #now iterate through all distinct provided sample preparation information
            sample_preparation_type_split = sample_preparation_type.split(';')
            sample_preparation_dict_split = {key: value.split(';') for key, value in list(sample_preparation_dict.items())}            
            for preparation_N, sample_preparation_type_iteration in enumerate(sample_preparation_type_split):

                #create specific distinct sample preparation information dictionary
                standard_provided_sample_preparation_information = {key: value[preparation_N] for key, value in list(sample_preparation_dict.items())}
                standard_provided_sample_preparation_information['sample_preparation_type'] = sample_preparation_type_iteration

                #get information specific to sample preparation type in standard sampling preparation types dictionary
                specific_sample_preparation_information = standard_sampling_preparation_types[sample_preparation_type_iteration]
                
                #if have defined sample preparation technique name, get dictionary with information specific to that technique  
                if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_technique_information = standard_sampling_preparation_types[sample_preparation_type_iteration]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                    #merge specific sample preparation type information and specific sample preparation technique information
                    #handle matching variables by keeping technique specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
                #else, if have a intermediate measurement defined as the sample preparation type, get dictionary with information specific to that method
                elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                    #if have defined measuring instrument name, get information specific to that instrument
                    if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                        specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                        #merge specific sample preparation methodology information and specific sample preparation instrument information
                        #handle matching variables by keeping instrument specific information by priority
                        specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                    #merge specific sample preparation type information and specific sample preparation methodology information
                    #handle matching variables by keeping methodology specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

                #merge specific sample preparation information and standard provided sampling preparation information
                #handle matching variables by keeping standard provided sampling preparation information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
                #get list of variables in generated sample preparation dictionary
                specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())  

                #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
                for var in specific_sample_preparation_information_vars:   
                    if type(specific_sample_preparation_information[var]) == dict:
                        if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                            specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']]  
                        else:
                            del specific_sample_preparation_information[var]
                            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())

                #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
                if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                    measurement_qa_flags['Unknown Sample Preparation Type'] = True  

                #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
                #this is False for "flask", "bag", "preconcentration", "filter" preparation types as there are no specific techniques for these types
                if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                    if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','bag','preconcentration','filter']:
                            measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

                #determine if current parameter processing is validly isolated by sample preparation
                if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

               #determine if sample preparation measurement method actually measures current parameter processing 
                if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

                #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True                  

                #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
                if 'preparation_assumption' in specific_sample_preparation_information_vars:
                    measurement_qa_flags['Assumed Sample Preparation'] = True

                #concatenate sample preparation strings 
                if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                    specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])
  
    #--------------------------------------------------------------------------------#
    #is there sample preparation information provided?
    #check for all different sample preparation information (i.e. preparation_1, preparation_2, preparation_3..)
    have_preparation = True
    preparation_n = 1
    while have_preparation:
        if 'preparation_{}'.format(preparation_n) in list(standard_provided_measurement_information.keys()):
            
            #get standard provided sample preparation information dictionary
            standard_provided_sample_preparation_information = copy.deepcopy(standard_provided_measurement_information['preparation_{}'.format(preparation_n)])

            #get information specific to provided sampling preparation type in standard sampling preparation types dictionary     
            specific_sample_preparation_information = copy.deepcopy(standard_sampling_preparation_types[standard_provided_sample_preparation_information['sample_preparation_type']])
            
            #update specific sample preparation information, updating by "overwrite" variables
            #get list of all variables containing "overwrite" string
            overwrite_vars = [var for var in list(specific_sample_preparation_information.keys()) if 'overwrite' in var]
            for overwrite_var in overwrite_vars:
                original_var = overwrite_var.split('overwrite_')[0]
                specific_sample_preparation_information[original_var] = specific_sample_preparation_information[overwrite_var]

            #if have defined sample preparation technique name, get dictionary with information specific to that technique  
            if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                specific_sample_preparation_technique_information = standard_sampling_preparation_types[standard_provided_sample_preparation_information['sample_preparation_type']]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                #merge specific sample preparation type information and specific sample preparation technique information
                #handle matching variables by keeping technique specific information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
            #else if, if have a intermediate measurement defined as the sample preparation type, get dictionary with information specific to that method
            elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                #if have defined measuring instrument name, get information specific to that instrument
                if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                    #merge specific sample preparation methodology information and specific sample preparation instrument information
                    #handle matching variables by keeping instrument specific information by priority
                    specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                #merge specific sample preparation type information and specific sample preparation methodology information
                #handle matching variables by keeping methodology specific information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

            #merge specific sample preparation information and standard provided sampling preparation information
            #handle matching variables by keeping standard provided sampling preparation information by priority
            specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
            #get list of variables in generated sample preparation dictionary
            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())  

            #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
            for var in specific_sample_preparation_information_vars:   
                if type(specific_sample_preparation_information[var]) == dict:
                    if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                        specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']] 
                    else:
                        del specific_sample_preparation_information[var] 
                        specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys()) 

            #if do not have primary sampling determined yet and sample preparation type is an 'intermediate measurement'
            #make a copy of the specific sampling preparation information dictionary to record the primary sampling information
            if len(specific_primary_sampling_information) == 0:
                specific_primary_sampling_information = copy.deepcopy(specific_sample_preparation_information)

            #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
            if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                measurement_qa_flags['Unknown Sample Preparation Type'] = True  

            #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
            #this is False for "flask", "bag", "preconcentration", "filter" preparation types as there are no real specific techniques for these types, rather they are uniform
            if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','bag','preconcentration','filter']:
                        measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

            #determine if current parameter processing is validly isolated by sample preparation
            if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True

           #determine if sample preparation measurement method actually measures current parameter processing 
            if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True

            #determine if sample preparation measurement method is QA accepted for current parameter processing 
            if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                    measurement_qa_flags['Erroneous Sample Preparation'] = True                 

            #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
            if ('preparation_assumption' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                measurement_qa_flags['Assumed Sample Preparation'] = True

            #concatenate sample preparation strings 
            if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                if 'specific_technique' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                elif 'methodology' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                else:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                if 'process_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                if 'preparation_further_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
            else:
                specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                if 'specific_technique' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                elif 'methodology' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                if 'process_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                if 'preparation_further_details' in specific_sample_preparation_information_vars:
                    specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])

            #update preparation_n       
            preparation_n += 1
        else:
            have_preparation = False
        
    #--------------------------------------------------------------------------------#
    #is there measurement information? 
    specific_measurement_method_information = {}
    if 'measurement' in list(standard_provided_measurement_information.keys()):
    
        #get standard provided measurement method information dictionary
        standard_provided_measurement_method_information = copy.deepcopy(standard_provided_measurement_information['measurement'])

        #get information specific to provided measurement method in standard measurement dictionary     
        specific_measurement_method_information = copy.deepcopy(standard_measurement_methodologies[standard_provided_measurement_method_information['methodology']])
        
        #update specific measurement information, updating by "overwrite" variables
        #get list of all variables containing "overwrite" string
        overwrite_vars = [var for var in list(specific_measurement_method_information.keys()) if 'overwrite' in var]
        for overwrite_var in overwrite_vars:
            original_var = overwrite_var.split('overwrite_')[0]
            specific_measurement_method_information[original_var] = specific_measurement_method_information[overwrite_var]

        #if have defined measuring instrument name, get information specific to that instrument
        if 'instrument' in list(standard_provided_measurement_method_information.keys()):
            specific_measurement_instrument_information = standard_measurement_methodologies[standard_provided_measurement_method_information['methodology']]['instruments'][standard_provided_measurement_method_information['instrument']]   
            #merge specific measurement method information and specific measurement instrument information
            #handle matching variables by keeping instrument specific information by priority
            specific_measurement_method_information = {**specific_measurement_method_information, **specific_measurement_instrument_information}

        #merge specific measurement method information and standard provided primary sampling information
        #handle matching variables by keeping standard provided primary sampling information by priority
        specific_measurement_method_information = {**specific_measurement_method_information, **standard_provided_measurement_method_information} 
        #get list of variables in generated primary sampling dictionary
        specific_measurement_method_information_vars = list(specific_measurement_method_information.keys())        
    
        #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
        for var in specific_measurement_method_information_vars:   
            if type(specific_measurement_method_information[var]) == dict:
                if run_details['parameter'] in list(specific_measurement_method_information[var].keys()):
                    specific_measurement_method_information[var] = specific_measurement_method_information[var][run_details['parameter']] 
                else:  
                    del specific_measurement_method_information[var] 
                    specific_measurement_method_information_vars = list(specific_measurement_method_information.keys()) 

        #if do not have measurement method defined (i.e. == unknown), set quality assurance flag informing of this
        if specific_measurement_method_information['methodology'] == 'unknown':
            measurement_qa_flags['Unknown Measurement Method'] = True  

        #if do not have specific measuring instrument defined, set quality assurance flag informing of this
        #if measuring sampling type is "manual" then this should be False, as there is no physically real measuring instrument in this case
        if 'instrument' not in specific_measurement_method_information_vars:
            if specific_measurement_method_information['methodology'] == 'unknown':
                measurement_qa_flags['Unknown Measuring Instrument'] = True
            else:
                if specific_measurement_method_information['sampling_type'] not in ['manual']:
                    measurement_qa_flags['Unknown Measuring Instrument'] = True

       #determine if measurement method actually measures the current parameter processing 
        if ('measured_parameters' in specific_measurement_method_information_vars) & (specific_measurement_method_information['methodology'] != 'unknown'):
            if (run_details['parameter'] not in specific_measurement_method_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_measurement_method_information['measured_parameters']):
                measurement_qa_flags['Erroneous Measurement Methodology'] = True 

        #determine if measurement method is QA accepted for current parameter processing 
        if ('qa_accepted_parameters' in specific_measurement_method_information_vars) & (specific_measurement_method_information['methodology'] != 'unknown'):
            if (run_details['parameter'] not in specific_measurement_method_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_measurement_method_information['qa_accepted_parameters']):
                measurement_qa_flags['Invalid QA Measurement Methodology'] = True    

        #if "measuring_assumption" flag is in specific measurement method dictionary, update measurement method assumption flag
        if 'measuring_assumption' in specific_measurement_method_information_vars:
                measurement_qa_flags['Assumed Measurement Methodology'] = True     

        #if have 'additional_sample_preparation_type' in measurement method dictionary, append to sampling_preparation_type variable (if exists)
        if 'additional_sample_preparation_type' in specific_measurement_method_information_vars:
            if 'sample_preparation_type' in specific_measurement_method_information_vars:
                specific_measurement_method_information['sample_preparation_type'] = '{};{}'.format(specific_measurement_method_information['sample_preparation_type'], specific_measurement_method_information['additional_sample_preparation_type'])
            else:
                specific_measurement_method_information['sample_preparation_type'] = specific_measurement_method_information['additional_sample_preparation_type']                 
      
        #if have sample preparation information provided for the measuring instrument, process it
        if 'sample_preparation_type' in specific_measurement_method_information_vars:
            #get sample preparation variable
            sample_preparation_type = specific_measurement_method_information['sample_preparation_type']

            #gather all variable names associated with sample preparation in dictionary (i.e. with 'preparation' prefix)
            sample_preparation_vars = [var for var in specific_measurement_method_information_vars if ('preparation' in var) & (var != 'sample_preparation_type')]
            #create dictionary of all all sample preparation variables names associated their values (dropping 'preparation' prefix'
            sample_preparation_dict = {}
            for var in sample_preparation_vars:   
                if var not in ['preparation_further_details', 'preparation_manual_name', 'preparation_assumption']:          
                    sample_preparation_dict[var.split('preparation_')[-1]] = specific_measurement_method_information[var] 
                else:
                    sample_preparation_dict[var] = specific_measurement_method_information[var] 

            #now iterate through all distinct provided sample preparation information
            sample_preparation_type_split = sample_preparation_type.split(';')
            sample_preparation_dict_split = {key: value.split(';') for key, value in list(sample_preparation_dict.items())}            
            for preparation_N, sample_preparation_type_iteration in enumerate(sample_preparation_type_split):

                #create specific distinct sample preparation information dictionary
                standard_provided_sample_preparation_information = {key: value[preparation_N] for key, value in list(sample_preparation_dict.items())}
                standard_provided_sample_preparation_information['sample_preparation_type'] = sample_preparation_type_iteration

                #get information specific to sample preparation type/specific technique in standard sampling preparation types dictionary
                specific_sample_preparation_information = standard_sampling_preparation_types[sample_preparation_type_iteration]
                if 'specific_technique' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_technique_information = standard_sampling_preparation_types[sample_preparation_type_iteration]['specific_techniques'][standard_provided_sample_preparation_information['specific_technique']]                
                    #merge specific sample preparation type information and specific sample preparation technique information
                    #handle matching variables by keeping technique specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_technique_information}
                elif 'methodology' in list(standard_provided_sample_preparation_information.keys()):
                    specific_sample_preparation_methodology_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]
                    if 'instrument' in list(standard_provided_sample_preparation_information.keys()):
                        specific_sample_preparation_instrument_information = standard_measurement_methodologies[standard_provided_sample_preparation_information['methodology']]['instruments'][standard_provided_sample_preparation_information['instrument']]
                        #merge specific sample preparation methodology information and specific sample preparation instrument information
                        #handle matching variables by keeping instrument specific information by priority
                        specific_sample_preparation_methodology_information = {**specific_sample_preparation_methodology_information, **specific_sample_preparation_instrument_information}
                    #merge specific sample preparation type information and specific sample preparation methodology information
                    #handle matching variables by keeping methodology specific information by priority
                    specific_sample_preparation_information = {**specific_sample_preparation_information, **specific_sample_preparation_methodology_information}

                #merge specific sample preparation information and standard provided sampling preparation information
                #handle matching variables by keeping standard provided sampling preparation information by priority
                specific_sample_preparation_information = {**specific_sample_preparation_information, **standard_provided_sample_preparation_information}
                #get list of variables in generated sample preparation dictionary
                specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys()) 

                #if have dictionary of values per parameter, get parameter being processed specific values (if parameter exists in dictionary as key)
                for var in specific_sample_preparation_information_vars:   
                    if type(specific_sample_preparation_information[var]) == dict:
                        if run_details['parameter'] in list(specific_sample_preparation_information[var].keys()):
                            specific_sample_preparation_information[var] = specific_sample_preparation_information[var][run_details['parameter']]  
                        else:
                            del specific_sample_preparation_information[var]
                            specific_sample_preparation_information_vars = list(specific_sample_preparation_information.keys())

                #if do not have sample preparation type defined (i.e. == unknown), set quality assurance flag informing of this
                if specific_sample_preparation_information['sample_preparation_type'] == 'unknown':
                    measurement_qa_flags['Unknown Sample Preparation Type'] = True  

                #if do not have sample preparation technique/method defined, set quality assurance flag informing of this
                #this is False for "flask", "bag", "preconcentration", "filter" preparation types as there are no specific techniques for these types
                if ('specific_technique' not in specific_sample_preparation_information_vars) & ('methodology' not in specific_sample_preparation_information_vars):
                    if specific_sample_preparation_information['sample_preparation_type'] not in ['flask','bag','preconcentration','filter']:
                            measurement_qa_flags['Unknown Sample Preparation Technique'] = True     

                #determine if current parameter processing is validly isolated by sample preparation
                if ('isolated_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['isolated_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['isolated_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

               #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('measured_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['measured_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['measured_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True

                #determine if sample preparation measurement method is QA accepted for current parameter processing 
                if ('qa_accepted_parameters' in specific_sample_preparation_information_vars) & (specific_sample_preparation_information['sample_preparation_type'] != 'unknown'):
                    if (run_details['parameter'] not in specific_sample_preparation_information['qa_accepted_parameters']) & (run_details['parameter'].split('_')[0] not in specific_sample_preparation_information['qa_accepted_parameters']):
                        measurement_qa_flags['Erroneous Sample Preparation'] = True                     

                #if "preparation_assumption" flag is in specific sample preparation dictionary, update sample preparation assumption flag
                if ('preparation_assumption' in specific_sample_preparation_information_vars):
                    measurement_qa_flags['Assumed Sample Preparation'] = True

                #concatenate sample preparation strings 
                if len(specific_sample_preparation_strings['sample_preparation_types']) > 0: 
                    specific_sample_preparation_strings['sample_preparation_types'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_types'],specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_techniques'],specific_sample_preparation_information['methodology'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_techniques'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_process_details'],specific_sample_preparation_information['process_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};{}'.format(specific_sample_preparation_strings['sample_preparation_further_details'],specific_sample_preparation_information['preparation_further_details'])
                    else:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = '{};'.format(specific_sample_preparation_strings['sample_preparation_further_details'])
                else:
                    specific_sample_preparation_strings['sample_preparation_types'] = copy.deepcopy(specific_sample_preparation_information['sample_preparation_type'])
                    if 'specific_technique' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['specific_technique'])
                    elif 'methodology' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_techniques'] = copy.deepcopy(specific_sample_preparation_information['methodology'])
                    if 'process_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_process_details'] = copy.deepcopy(specific_sample_preparation_information['process_details'])
                    if 'preparation_further_details' in specific_sample_preparation_information_vars:
                        specific_sample_preparation_strings['sample_preparation_further_details'] = copy.deepcopy(specific_sample_preparation_information['preparation_further_details'])

    #--------------------------------------------------------------------------------#

    #if raw provided measurement information has been successfully mapped, but a file is still contained within unmapped_provided_measurement_information directory, remove file from directory
    if raw_provided_measurement_information in unmapped_provided_measurement_information:
        #check if measurement information still exists before removing it (for parallel cases)
        #also put in try/except just in case for parallel cases where file is deleted before trying to remove it (if statement, should catch these cases but have had failures before)
        if os.path.isfile(run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information)):
            try:
                os.remove(run_details['process_root']+'/unmapped_provided_measurement_information/{}:{}:{}'.format(run_details['network'],run_details['parameter'],raw_provided_measurement_information))
            except:
                pass
    
    #--------------------------------------------------------------------------------#

    #now add measurement stage specific dictionaries to standard site metadata variables
    #convert variable concentration specific units to standard units per variable, e.g. limits of detections etc.
    #all aerosol matrix units in standard methods/instrument dicts in ug/m3, for gases in ppbV, for precip in ug l-1, for wetdep in ug m-2 min-1, for vconc and size in um3 um-2
    #do not convert for meteo. variables
    if parameter_details['matrix'] == 'gas':
        original_units = 'ppbv'
    elif '_COEFF_' in run_details['parameter']:
        original_units = 'Mm-1'
    elif (parameter_details['matrix'] == 'pm') or (parameter_details['matrix'] == 'pm10') or (parameter_details['matrix'] == 'pm2.5') or (parameter_details['matrix'] == 'pm1'):
        original_units = 'ug m-3'
    elif parameter_details['matrix'] == 'precip':
        original_units = 'ug l-1'
    elif parameter_details['matrix'] == 'wetdep':
        original_units = 'ug m-2 min-1'
    elif (parameter_details['matrix'] == 'aod') or (parameter_details['matrix'] == 'extaod') or (parameter_details['matrix'] == 'absaod') or (parameter_details['matrix'] == 'absaod-bc') or (parameter_details['matrix'] == 'absaod-brc') or (parameter_details['matrix'] == 'scataod')  or (parameter_details['matrix'] == 'ssa') or (parameter_details['matrix'] == 'asy') or (parameter_details['matrix'] == 'rin'):
        original_units = 'unitless'
    elif (parameter_details['matrix'] == 'vconc') or (parameter_details['matrix'] == 'size'):
        original_units = 'um3 um-2'
    else:
        original_units = ''
    #set variables to convert concentration units for
    vars_to_convert = ['documented_lower_limit_of_detection','reported_lower_limit_of_detection','documented_upper_limit_of_detection','reported_upper_limit_of_detection','documented_uncertainty','reported_uncertainty',
                       'documented_accuracy','reported_accuracy','documented_precision','reported_accuracy','documented_zero_drift','reported_zero_drift','documented_span_drift','reported_span_drift','documented_zonal_drift',
                       'reported_zonal_drift','documented_resolution','reported_resolution']

    #define metadata dictionary to write
    standard_measurement_metadata = {}    

    #primary sampling metadata
    valid_primary_sampling_vars = list(primary_sampling_metadata_mapping.keys())
    for var in list(specific_primary_sampling_information.keys()):
        if var in valid_primary_sampling_vars:
            standard_measurement_metadata[primary_sampling_metadata_mapping[var]] = specific_primary_sampling_information[var]

    #sample preparation metadata
    valid_sample_preparation_vars = list(sample_preparation_metadata_mapping.keys())
    for var in list(specific_sample_preparation_strings.keys()):
        if var in valid_sample_preparation_vars:
            sample_preparation_string = specific_sample_preparation_strings[var]
            if sample_preparation_string != '':
                standard_measurement_metadata[sample_preparation_metadata_mapping[var]] = sample_preparation_string

    #measurement metadata
    valid_measurement_vars = list(measurement_metadata_mapping.keys())
    for var in list(specific_measurement_method_information.keys()):
        if var in valid_measurement_vars:

            #need to convert units of variable?
            if (var in vars_to_convert) & (original_units != ''):
                #variable value is non-NaN?
                if not pd.isnull(specific_measurement_method_information[var]):
                    #convert all concentration units in strings (taking care with %'s and unit times, spaces, and /)
 
                    #values split by '< >='
                    if ' ' in specific_measurement_method_information[var]:
                        split_space = specific_measurement_method_information[var].split(' ')   

                        less_than_split = split_space[0].split('<')
                        for less_than_index, less_than_var in enumerate(less_than_split):
                            #have '/' in string?
                            if '/' in less_than_var:
                                slash_split = less_than_var.split('/')
                                for slash_index, slash_var in enumerate(slash_split):
                                    if ('%' not in slash_var) & (slash_var not in ['hour','day','7days','week','month','year']):
                                        conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(slash_var), measured_species = parameter_details['chemical_formula'])
                                        slash_split[slash_index] = str(conv_obj.converted_value)   
                                less_than_split[less_than_index] = '/'.join(slash_split)                              
                            #do not have '/' in string?
                            else:
                                if ('%' not in less_than_var) & (less_than_var not in ['hour','day','7days','week','month','year']):
                                    conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(less_than_var), measured_species = parameter_details['chemical_formula'])
                                    less_than_split[less_than_index] = str(conv_obj.converted_value)  
                        #rejoin less than split
                        split_space[0] = '<'.join(less_than_split)

                        greater_than_split = split_space[1].split('>=')
                        for greater_than_index, greater_than_var in enumerate(greater_than_split):
                            #have '/' in string?
                            if '/' in greater_than_var:
                                slash_split = greater_than_var.split('/')
                                for slash_index, slash_var in enumerate(slash_split):
                                    if ('%' not in slash_var) & (slash_var not in ['hour','day','7days','week','month','year']):
                                        conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(slash_var), measured_species = parameter_details['chemical_formula'])
                                        slash_split[slash_index] = str(conv_obj.converted_value)   
                                greater_than_split[greater_than_index] = '/'.join(slash_split)                              
                            #do not have '/' in string?
                            else:
                                if ('%' not in greater_than_var) & (greater_than_var not in ['hour','day','7days','week','month','year']):
                                    conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(greater_than_var), measured_species = parameter_details['chemical_formula'])
                                    greater_than_split[greater_than_index] = str(conv_obj.converted_value)  
                        #rejoin greater than split
                        split_space[1] = '>='.join(greater_than_split)

                        #rejoin everything together
                        specific_measurement_method_information[var] = ' '.join(split_space)

                    #values split by '<'
                    elif '<' in specific_measurement_method_information[var]:
                        less_than_split = specific_measurement_method_information[var].split('<')
                        for less_than_index, less_than_var in enumerate(less_than_split):
                            #have '/' in string?
                            if '/' in less_than_var:
                                slash_split = less_than_var.split('/')
                                for slash_index, slash_var in enumerate(slash_split):
                                    if ('%' not in slash_var) & (slash_var not in ['hour','day','7days','week','month','year']):
                                        conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(slash_var), measured_species = parameter_details['chemical_formula'])
                                        slash_split[slash_index] = str(conv_obj.converted_value)   
                                less_than_split[less_than_index] = '/'.join(slash_split)                              
                            #do not have '/' in string?
                            else:
                                if ('%' not in less_than_var) & (less_than_var not in ['hour','day','7days','week','month','year']):
                                    conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(less_than_var), measured_species = parameter_details['chemical_formula'])
                                    less_than_split[less_than_index] = str(conv_obj.converted_value)  
                        #rejoin less than split
                        specific_measurement_method_information[var] = '<'.join(less_than_split)

                    #values split by '>='
                    elif '>=' in specific_measurement_method_information[var]:
                        greater_than_split = specific_measurement_method_information[var].split('>=')
                        for greater_than_index, greater_than_var in enumerate(greater_than_split):
                            #have '/' in string?
                            if '/' in greater_than_var:
                                slash_split = greater_than_var.split('/')
                                for slash_index, slash_var in enumerate(slash_split):
                                    if ('%' not in slash_var) & (slash_var not in ['hour','day','7days','week','month','year']):
                                        conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(slash_var), measured_species = parameter_details['chemical_formula'])
                                        slash_split[slash_index] = str(conv_obj.converted_value)   
                                greater_than_split[greater_than_index] = '/'.join(slash_split)                              
                            #do not have '/' in string?
                            else:
                                if ('%' not in greater_than_var) & (greater_than_var not in ['hour','day','7days','week','month','year']):
                                    conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(greater_than_var), measured_species = parameter_details['chemical_formula'])
                                    greater_than_split[greater_than_index] = str(conv_obj.converted_value)  
                        #rejoin greater than split
                        specific_measurement_method_information[var] = '>='.join(greater_than_split)

                    #values split by '/'
                    elif '/' in specific_measurement_method_information[var]:
                        slash_split = specific_measurement_method_information[var].split('/')
                        for slash_index, slash_var in enumerate(slash_split):
                            if ('%' not in slash_var) & (slash_var not in ['hour','day','7days','week','month','year']):
                                conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(slash_var), measured_species = parameter_details['chemical_formula'])
                                slash_split[slash_index] = str(conv_obj.converted_value)
                        #rejoin slash split
                        specific_measurement_method_information[var] = '/'.join(slash_split)

                    #just numeric value
                    elif ('%' not in specific_measurement_method_information[var]) & (specific_measurement_method_information[var] not in ['hour','day','7days','week','month','year']):
                        conv_obj = unit_converter.convert_units(original_units,parameter_details['standard_units'],np.float32(specific_measurement_method_information[var]), measured_species = parameter_details['chemical_formula'])
                        specific_measurement_method_information[var] = str(conv_obj.converted_value)
            
            standard_measurement_metadata[measurement_metadata_mapping[var]] = specific_measurement_method_information[var] 

    #write measurement metadata directly to station metadata array if desired (write_to_metadata_array == True)
    if write_to_metadata_array:
        for key, value in standard_measurement_metadata.items():
            site_metadata[key]['value'].append(value)
        return site_metadata, measurement_qa_flags, specific_measurement_method_information['abbreviation']
    else:
        return standard_measurement_metadata, measurement_qa_flags, specific_measurement_method_information['abbreviation']

###--------------------------------------------------------------------------------------------------###
#DEFINE EXTERNAL METADATA MAPPING INFORMATION
###--------------------------------------------------------------------------------------------------###

#define dictionaries which map external metadata file variable names to standardised metadata variable names
#syntax = 'external_metadata_file_variable_name' : 'standard_variable_name' 

#if the word 'manual' appears in the dictionary variable names, then the external metadata field has eben manually compiled

#for mapping variable data contained in 'EBAS/metadata/processed/EBAS_META.npy' to standardised names
EBAS_manual_external_meta_map = {
'other_network_references':'associated_networks',
'setting':                 'area_classification',      
'land_use':                'land_use'
}
    
#for mapping variable data contained in 'EEA_AIRBASE/metadata/processed/EEA_AIRBASE_META.npy' to standardised names
EEA_AIRBASE_external_meta_map = {
'station_european_code':                     'station_reference',                    
'station_latitude_decimal_degrees':          'latitude',
'station_longitude_decimal_degrees':         'longitude', 
'station_altitude':                          'altitude',
'height_sampling_point':                     'sampling_height',                      
'station_name':                              'station_name',                         
'station_type_of_area':                      'area_classification',                  
'type_of_station':                           'station_classification',               
'main_emission_source':                      'main_emission_source',                 
'station_characteristic_of_zone':            'land_use',
'station_subcategory_rural_background':      'area_classification_rural_subdivision',
'station_represented_radius':                'representative_radius',                
'station_city':                              'city',
'station_distance_to_junction':              'distance_to_junction',                 
'station_distance_to_kerb':                  'distance_to_kerb',                     
'station_distance_to_source':                'distance_to_source',                   
'station_street_width':                      'street_width',
'street_type':                               'street_type',
'number_of_vehicles':                        'daily_passing_vehicles',               
'station_traffic_speed':                     'daytime_traffic_speed',                        
'population':                                'population',
'station_description':                       'station_description',
'measurement_equipment':                     'measurement_equipment',
'measurement_technique_principle':           'measurement_method',                   
'measurement_automatic':                     'measurement_automatic',
'measurement_unit':                          'measurement_unit',
'calibration_method':                        'calibration_method',
'calibration_frequency':                     'calibration_frequency',
'calibration_description':                   'calibration_description',
'length_sampling_line':                      'length_sampling_line',
'location_sampling_point':                   'sample_location_miscellaneous',
'EMEP_station':                              'emep_station',                         
'component_code':                            'pollutant_code',                       
'component_name':                            'pollutant',                            
'network_code':                              'local_network_code',                   
'network_short_name':                        'local_network_name',           
'station_local_code':                        'local_reference',               
'network_data_supplier_person_title':        'principal_investigator_title',         
'network_data_supplier_person_first_name':   'principal_investigator_first_name',    
'network_data_supplier_person_last_name':    'principal_investigator_last_name',     
'network_data_supplier_organization_name':   'principal_investigator_institution',     
'network_data_supplier_person_email_address':'principal_investigator_email_address', 
'network_manager_person_title':              'contact_title',                        
'network_manager_person_first_name':         'contact_first_name',                   
'network_manager_person_last_name':          'contact_last_name',                    
'network_manager_organization_name':         'contact_institution',                  
'network_manager_person_email_address':      'contact_email_address',
'network_time_reference_basis':              'network_time_reference_basis'
}       

#for mapping variable data contained in 'EEA_AQ_eReporting/metadata/network_provided/EEA_AQ_eReporting_META.csv' to standardised names
EEA_AQ_eReporting_external_meta_map = {
'AirQualityStationEoICode':'station_reference',     
'Latitude':                'latitude',             
'Longitude':               'longitude',             
'Altitude':                'altitude',
'InletHeight':             'sampling_height',       
'BuildingDistance':        'distance_to_building',  
'KerbDistance':            'distance_to_kerb',      
'AirQualityStationArea':   'area_classification',   
'AirQualityStationType':   'station_classification',
'SamplingPoint':           'sampling_point',        
'SamplingProcess':         'sampling_process',      
'Projection':              'spatial_reference_system',            
'AirQualityNetwork':       'local_network_code',    
'AirQualityStationNatCode':'local_reference',       
'MeasurementEquipment':    'measurement_equipment', 
'MeasurementType':         'measurement_type',
'MeasurementMethod':       'measurement_method',    
'AirPollutantCode':        'pollutant_code',
'AirPollutant':            'pollutant'
}
                                     
#for mapping variable data contained in 'EEA_AQ_eReporting/metadata/network_provided/EEA_AQ_eReporting_AIDE_D.csv' to standardised names
EEA_AQ_eReporting_AIDE_D_external_meta_map = {
'EoICode':                 'station_reference',                 
'Latitude':                'latitude',                          
'Longitude':               'longitude',                         
'Altitude':                'altitude',                          
'StationName':             'station_name',                      
'StationArea':             'area_classification',               
'StationType':             'station_classification',            
'ResponsibleParty':        'principal_investigator_institution',
'SamplingPointId':         'sampling_point',                    
'SamplingProcess':         'sampling_process',                  
'Network':                 'local_network_name',
'MeasurementEquipment':    'measurement_equipment',
'MeasurementMethod':       'measurement_method',
'SamplingEquipment':       'sampling_equipment',
'AnalyticalTechnique':     'analytical_technique',              
'Pollutant':               'pollutant'
}

#for mapping variable data containing in 'MITECO/metadata/network_provided/MITECO_META_YYYY.csv' to standardised names
MITECO_external_meta_map = {
'COD_LOCAL':               'station_reference_copy',                 
'PROVINCIA':               'province_code',
'MUNICIPIO':               'municipality_code',
'ESTACION':                'local_station_id',
'COD_ESTACION_DEM':        'eeacode',
'N_RED':                   'local_network_name',
'NOMBRE':                  'station_name',
'LATITUD_G':               'latitude',                          
'LONGITUD_G':              'longitude',                         
'ALTITUD':                 'altitude',                                                
'N_CCAA':                  'autonomous_community',
'N_PROVINCIA':             'province',
'N_MUNICIPIO':             'municipality',
'TIPO_AREA':               'area_classification',               
'TIPO_ESTACION':           'station_classification', 
'TIPO_SUBAREA_RURAL':      'area_classification_rural_subdivision',
'ZONA':                    'land_use',
'DIRECCION':               'station_address'   
}